/***********************************************************************
 Copyright (C) 1993 CAD Framework Initiative, Inc.
 All Rights Reserved Worldwide.

This program and its accompanying documentation are supplied pursuant 
 to a license agreement from CAD Framework Initiative, Inc.
 The use, disclosure, and copying of this 
 software and documentation are all governed by that license agreement, 
 and any other use, disclosure, or copying of the software or 
 documentation is expressly prohibited.

************************************************************************/

#ifndef CFIDR_H
#define CFIDR_H

#ifdef CFI_ARGS
#undef CFI_ARGS
#endif
#define CFI_ARGS(args)	()

#include "cfi.h"

#define CFIDR_MIN_FLOAT32  -1E+37
#define CFIDR_MAX_FLOAT32   1E+37
#define CFIDR_MIN_INT32    -2147483647
#define CFIDR_MAX_INT32     2147483647

#define CFIDR_TRUE  CFI_TRUE
#define CFIDR_FALSE CFI_FALSE

#ifdef __cpluscplus
extern "C" {
#endif

typedef cfiBooleanT cfidrBooleanT;

typedef cfiLongT    cfidrInt32T;
typedef cfiFloatT   cfidrFloat32T;
typedef cfiStringT  cfidrStringT;
typedef cfiVoidT    cfidrVoidT;

typedef cfiIteratorIdT	cfidrIterIdT;
typedef cfiObjectIdT	cfidrObjectIdT;

typedef cfidrObjectIdT cfidrNamedObjectIdT;
typedef cfidrObjectIdT cfidrLibIdT;
typedef cfidrObjectIdT cfidrCellIdT;
typedef cfidrObjectIdT cfidrPortIdT;
typedef cfidrObjectIdT cfidrPortBundleIdT;
typedef cfidrObjectIdT cfidrPortBusIdT;
typedef cfidrObjectIdT cfidrPortScalarIdT;
typedef cfidrObjectIdT cfidrInstIdT;
typedef cfidrObjectIdT cfidrPortInstIdT;
typedef cfidrObjectIdT cfidrPortInstBundleIdT;
typedef cfidrObjectIdT cfidrPortInstBusIdT;
typedef cfidrObjectIdT cfidrPortInstScalarIdT;
typedef cfidrObjectIdT cfidrNetIdT;
typedef cfidrObjectIdT cfidrNetBundleIdT;
typedef cfidrObjectIdT cfidrNetBusIdT;
typedef cfidrObjectIdT cfidrNetScalarIdT;
typedef cfidrObjectIdT cfidrPropIdT;
typedef cfidrObjectIdT cfidrViewIdT;
typedef cfidrObjectIdT cfidrEncapsulatedViewIdT;
typedef cfidrObjectIdT cfidrNetlistViewIdT;

typedef cfidrIterIdT cfidrCellsIdT;
typedef cfidrIterIdT cfidrInstsIdT;
typedef cfidrIterIdT cfidrLibsIdT;
typedef cfidrIterIdT cfidrNetsIdT;
typedef cfidrIterIdT cfidrNetBundlesIdT;
typedef cfidrIterIdT cfidrPortsIdT;
typedef cfidrIterIdT cfidrPortInstsIdT;
typedef cfidrIterIdT cfidrPropsIdT;
typedef cfidrIterIdT cfidrViewsIdT;

typedef cfidrObjectIdT cfidrSelectorSetIdT;
typedef cfidrObjectIdT cfidrViewSelectorIdT;
typedef cfidrObjectIdT cfidrViewNameSelectorIdT;
typedef cfidrObjectIdT cfidrViewTypeSelectorIdT;
typedef cfidrObjectIdT cfidrLibrarySelectorIdT;
 
typedef cfidrIterIdT cfidrViewSelectorsIdT;
typedef cfidrIterIdT cfidrLibrarySelectorsIdT;

typedef enum {
    CFIDR_UNDEFINED_OBJECTTYPE = 0,
    CFIDR_LIB                 = 1,
    CFIDR_CELL                = 2,
    CFIDR_PORTBUNDLE          = 3,
    CFIDR_PORTBUS             = 4,
    CFIDR_PORTSCALAR          = 5,
    CFIDR_INST                = 6,
    CFIDR_PORTINSTBUNDLE      = 7,
    CFIDR_PORTINSTBUS         = 8,
    CFIDR_PORTINSTSCALAR      = 9,
    CFIDR_NETBUNDLE           = 10,
    CFIDR_NETBUS              = 11,
    CFIDR_NETSCALAR           = 12,
    CFIDR_PROP                = 13,
    CFIDR_NETLISTVIEW         = 14,
    CFIDR_ENCAPSULATEDVIEW    = 15,
    CFIDR_SELECTORSET         = 16,
    CFIDR_LIBSELECTOR         = 17,
    CFIDR_VIEWNAMESELECTOR    = 18,
    CFIDR_VIEWTYPESELECTOR    = 19,
    CFIDR_MAX_OBJECTTYPE     = 20
    } cfidrObjectTypeT;

typedef enum {
    CFIDR_UNDEFINED_VALUETYPE = CFI_UNDEFINED_VALUETYPE,
    CFIDR_INT32               = CFI_LONG,
    CFIDR_STRING              = CFI_STRING,
    CFIDR_FLOAT32             = CFI_FLOAT,
    CFIDR_BOOLEAN             = CFI_BOOLEAN,
    CFIDR_MAX_VALUETYPE       = CFI_MAX_VALUETYPE
    } cfidrValueTypeT;

typedef enum {
    CFIDR_UNDEFINED_PORTDIRECTION = 0,
    CFIDR_INPUT                   = 1,
    CFIDR_OUTPUT                  = 2,
    CFIDR_IO                      = 3,
    CFIDR_MAX_PORTDIRECTION
    } cfidrPortDirectionT;

typedef enum {
  CFIDR_UNDEFINED_ACCESSMODE = 0,
  CFIDR_READ                 = 1,
  CFIDR_UPDATE               = 2,
  CFIDR_MAX_ACCESSMODE     
  } cfidrAccessModeT;

typedef enum {
    CFIDR_UNDEFINED_ATTACHMODE = 0,
    CFIDR_ATTACH_BY_ORDER       = 1,
    CFIDR_ATTACH_BY_NAME        = 2,
    CFIDR_MAX_ATTACHMODE
    } cfidrAttachModeT;

typedef enum {
    CFIDR_UNDEFINED_ITERMODE   = 0,
    CFIDR_ITER_SCALARS          = 1,
    CFIDR_ITER_BUNDLES          = 2,
    CFIDR_ITER_TOP              = 3,
    CFIDR_ITER_ALL              = 4,
    CFIDR_MAX_ITERMODE
    } cfidrIterModeT;

typedef enum {
    CFIDR_NO_ERROR                    =  0,
    CFIDR_DESCRIBER_PORT_NOT_FOUND,
    CFIDR_DESCRIBER_RECURSION,
    CFIDR_DESCRIBER_VIEW_NOT_FOUND,
    CFIDR_INTERNAL_SYSTEM_ERROR,
    CFIDR_INVALID_ATTACH_MODE,
    CFIDR_UNUSABLE_DESCRIBER_OID,
    CFIDR_INVALID_DESCRIBER_TYPE,
    CFIDR_INVALID_DIRECTION,
    CFIDR_INVALID_ERROR_CODE,
    CFIDR_INVALID_VALUE,
    CFIDR_INVALID_ITER,
    CFIDR_INVALID_ITER_TYPE,
    CFIDR_INVALID_MEMBERS,
    CFIDR_INVALID_NAME,
    CFIDR_INVALID_OBJECTTYPE,
    CFIDR_UNUSABLE_OID,
    CFIDR_UNUSABLE_PORT_OID,
    CFIDR_INVALID_PORT_TYPE,
    CFIDR_UNUSABLE_PORTINST_OID,
    CFIDR_INVALID_PORTINST_TYPE,
    CFIDR_INVALID_VIEWTYPE,
    CFIDR_LIB_ALREADY_OPEN,
    CFIDR_NEW_DESCRIBER_PORT_FOUND,
    CFIDR_NO_NAME_MATCHES,
    CFIDR_NAME_IN_USE,
    CFIDR_OBJECT_ALREADY_EXISTS,
    CFIDR_OBJECT_NOT_FOUND,
    CFIDR_PORTINST_ALREADY_ATTACHED,
    CFIDR_PORTINST_NOT_ATTACHED,
    CFIDR_PORT_ALREADY_ATTACHED,
    CFIDR_PORT_MEMBERS_DIFFER,
    CFIDR_PORT_NOT_ATTACHED,
    CFIDR_VIEW_ALREADY_OPEN,
    CFIDR_OPEN_FOR_UPDATE_FAILED,
    CFIDR_INVALID_ITERMODE,
    CFIDR_ILLEGAL_OPERATION,
    CFIDR_CROSS_VIEW_OPERATION,
    CFIDR_SCALAR_BUNDLE_ATTACHMENT,
    CFIDR_LIB_OPEN_FAILED,
    CFIDR_VIEW_OPEN_FAILED,
    CFIDR_INVALID_ACCESSMODE,
    CFIDR_READ_ONLY,
    CFIDR_INVALID_KEY,
    CFIDR_INVALID_NET_TYPE,
    CFIDR_INVALID_PORTBUNDLE_TYPE,
    CFIDR_INVALID_POSITION,
    CFIDR_MEMBER_RECURSION,
    CFIDR_MULTIPLE_VIEWS_MATCH,
    CFIDR_NO_VIEW_MATCHES,
    CFIDR_UNUSABLE_NET_OID,
    CFIDR_UNUSABLE_PORTBUNDLE_OID,
    CFIDR_MAX_ERROR
    } cfidrErrorT;


/* START OF CFI '92 DR PI Public Functions */
cfidrEncapsulatedViewIdT cfidrCellCreateEncapsulatedView CFI_ARGS((
                         cfidrCellIdT owner,
 		         cfidrStringT name,
			 cfidrStringT type,
			 cfidrStringT key,
			 cfidrErrorT *error));

cfidrNetlistViewIdT cfidrCellCreateNetlistView CFI_ARGS((
                         cfidrCellIdT owner,
			 cfidrStringT name,
			 cfidrErrorT *error));

cfidrLibIdT         cfidrCellGetOwner CFI_ARGS((
                         cfidrCellIdT object, 
			 cfidrErrorT *error));

cfidrViewsIdT       cfidrCellGetViews CFI_ARGS((
			 cfidrCellIdT object, 
			 cfidrErrorT *error));

cfidrBooleanT       cfidrInstCheckDescriber CFI_ARGS((
                         cfidrInstIdT object, 
		         cfidrErrorT *error));

cfidrNetlistViewIdT cfidrInstGetDescriber CFI_ARGS((
                         cfidrInstIdT object, 
			 cfidrErrorT *error));

cfidrNetlistViewIdT cfidrInstGetOwner CFI_ARGS((
                         cfidrInstIdT object, 
			 cfidrErrorT *error));

cfidrPortInstsIdT   cfidrInstGetPortInsts CFI_ARGS((
                         cfidrInstIdT object,
			 cfidrIterModeT mode, 
			 cfidrErrorT *error));

cfidrCellIdT        cfidrIterNextCell CFI_ARGS((
                         cfidrCellsIdT iter, 
			 cfidrErrorT *error));

cfidrInstIdT        cfidrIterNextInst CFI_ARGS((
                         cfidrInstsIdT iter, 
			 cfidrErrorT *error));

cfidrLibIdT         cfidrIterNextLib CFI_ARGS((
                         cfidrLibsIdT iter, 
			 cfidrErrorT *error));

cfidrNetIdT         cfidrIterNextNet CFI_ARGS((
                         cfidrNetsIdT iter, 
			 cfidrErrorT *error));

cfidrNetBundleIdT   cfidrIterNextNetBundle CFI_ARGS((
                         cfidrNetBundlesIdT iter, 
			 cfidrErrorT *error));

cfidrPortIdT        cfidrIterNextPort CFI_ARGS((
                         cfidrPortsIdT iter, 
			 cfidrErrorT *error));

cfidrPortInstIdT    cfidrIterNextPortInst CFI_ARGS((
                         cfidrPortInstsIdT iter, 
			 cfidrErrorT *error));

cfidrPropIdT        cfidrIterNextProp CFI_ARGS((
                         cfidrPropsIdT iter, 
			 cfidrErrorT *error));

cfidrViewIdT        cfidrIterNextView CFI_ARGS((
                         cfidrViewsIdT iter, 
			 cfidrErrorT *error));

cfidrVoidT          cfidrIterQuit CFI_ARGS((
                         cfidrIterIdT iter, 
			 cfidrErrorT *error));

cfidrVoidT          cfidrIterReset CFI_ARGS((
                         cfidrIterIdT iter, 
                         cfidrErrorT *error));

cfidrCellIdT        cfidrLibCreateCell CFI_ARGS((
                         cfidrLibIdT owner,
                         cfidrStringT name,
                         cfidrErrorT *error));

cfidrCellsIdT       cfidrLibGetCells CFI_ARGS((
                         cfidrLibIdT object, 
                         cfidrErrorT *error));

cfidrVoidT          cfidrLibPurge CFI_ARGS((
                         cfidrLibIdT object, 
                         cfidrErrorT *error));

cfidrVoidT          cfidrLibSave CFI_ARGS((
                         cfidrLibIdT object, 
                         cfidrErrorT *error));

cfidrStringT        cfidrNamedObjectGetName CFI_ARGS((
                         cfidrNamedObjectIdT object,
                         cfidrErrorT *error));

cfidrVoidT          cfidrNamedObjectSetName CFI_ARGS((
                         cfidrNamedObjectIdT object,
                         cfidrStringT name, 
                         cfidrErrorT *error));

cfidrVoidT          cfidrNetAttachPort CFI_ARGS((
                         cfidrNetIdT object,
			 cfidrPortIdT port,
			 cfidrAttachModeT mode, 
                         cfidrErrorT *error));

cfidrVoidT          cfidrNetAttachPortInst CFI_ARGS((
                         cfidrNetIdT object,
			 cfidrPortInstIdT portInst,
			 cfidrAttachModeT mode, 
                         cfidrErrorT *error));

cfidrVoidT          cfidrNetDetachPort CFI_ARGS((
                         cfidrNetIdT object, 
		         cfidrPortIdT port, 
                         cfidrErrorT *error));

cfidrVoidT          cfidrNetDetachPortInst CFI_ARGS((
                         cfidrNetIdT object,
			 cfidrPortInstIdT portInst, 
                         cfidrErrorT *error));

cfidrNetBundlesIdT  cfidrNetGetNetBundles CFI_ARGS((
                         cfidrNetIdT id, 
                         cfidrErrorT *error));

cfidrNetlistViewIdT cfidrNetGetOwner CFI_ARGS((
                         cfidrNetIdT object, 
                         cfidrErrorT *error));

cfidrPortsIdT       cfidrNetGetPorts CFI_ARGS((
                         cfidrNetIdT object,
                         cfidrErrorT *error));


cfidrPortInstsIdT   cfidrNetGetPortInsts CFI_ARGS((
                         cfidrNetIdT object,
                         cfidrErrorT *error));

cfidrNetIdT         cfidrNetBundleFindNetByPosition CFI_ARGS((
                         cfidrNetBundleIdT netBundle, 
                         cfidrInt32T position, 
                         cfidrErrorT *error));


cfidrNetsIdT        cfidrNetBundleGetNets CFI_ARGS((
                         cfidrNetBundleIdT netB,
			 cfidrIterModeT mode, 
                         cfidrErrorT *error));

cfidrInt32T         cfidrNetBundleGetSize CFI_ARGS((
                         cfidrNetBundleIdT netBundle,
                         cfidrErrorT *error));

cfidrVoidT          cfidrNetBundleInsertNet CFI_ARGS((
                         cfidrNetBundleIdT owner,
			 cfidrNetIdT net,
			 cfidrInt32T position,
                         cfidrErrorT *error));

cfidrVoidT          cfidrNetBundleRemoveNet CFI_ARGS((
                         cfidrNetBundleIdT owner,
			 cfidrInt32T position,
                         cfidrErrorT *error));

cfidrInt32T         cfidrNetBusGetStart CFI_ARGS((
                         cfidrNetBusIdT netBus,
                         cfidrErrorT *error));

cfidrInt32T         cfidrNetBusGetStep CFI_ARGS((
                         cfidrNetBusIdT netBus,
                         cfidrErrorT *error));

cfidrVoidT          cfidrNetBusSetStart CFI_ARGS((
                         cfidrNetBusIdT netBus,
		         cfidrInt32T start,
                         cfidrErrorT *error));

cfidrVoidT          cfidrNetBusSetStep CFI_ARGS((
                         cfidrNetBusIdT netBus,
                         cfidrInt32T step,
                         cfidrErrorT *error));

cfidrBooleanT       cfidrNetScalarGetIsGlobal CFI_ARGS((
                         cfidrNetScalarIdT netScalar,
                         cfidrErrorT* error));

cfidrVoidT          cfidrNetScalarSetIsGlobal CFI_ARGS((
                         cfidrNetScalarIdT netScalar,
                         cfidrBooleanT value,
                         cfidrErrorT* error));

cfidrPropIdT        cfidrObjectCreateProp CFI_ARGS((
                         cfidrObjectIdT owner,
                         cfidrStringT name,
                         cfidrErrorT *error));

cfidrVoidT          cfidrObjectDestroy CFI_ARGS((
                         cfidrObjectIdT object, 
                         cfidrErrorT *error));

cfidrObjectTypeT    cfidrObjectGetObjectType CFI_ARGS((
                         cfidrObjectIdT object, 
                         cfidrErrorT *error));

cfidrPropsIdT       cfidrObjectGetProps CFI_ARGS((
                         cfidrObjectIdT object, 
                         cfidrErrorT *error));

cfidrBooleanT       cfidrObjectIsNull CFI_ARGS((
                         cfidrObjectIdT object,
                         cfidrErrorT *error));

cfidrBooleanT       cfidrObjectIsSame CFI_ARGS((
                         cfidrObjectIdT object1,
                         cfidrObjectIdT object2, 
                         cfidrErrorT *error));

cfidrBooleanT       cfidrObjectIsUsable CFI_ARGS((
                         cfidrObjectIdT object,
                         cfidrErrorT *error));

cfidrLibIdT         cfidrPICreateLib CFI_ARGS((
                         cfidrStringT name,
                         cfidrErrorT *error));

cfidrStringT        cfidrPIGetErrorText CFI_ARGS((
                         cfidrErrorT errorForText, 
                         cfidrErrorT *error));

cfidrLibsIdT        cfidrPIGetLibs CFI_ARGS((
                         cfidrErrorT *error));

cfidrObjectIdT      cfidrPIGetNullId CFI_ARGS((
                         cfidrErrorT *error));

cfidrVoidT          cfidrPIInit CFI_ARGS((
                         cfidrErrorT *error));

cfidrLibIdT         cfidrPIOpenLib CFI_ARGS((
                         cfidrStringT name,
                         cfidrAccessModeT mode,
                         cfidrErrorT *error));

cfidrViewIdT        cfidrPIOpenView CFI_ARGS((
                         cfidrStringT libname,
                         cfidrStringT cellname,
                         cfidrStringT viewname,
                         cfidrAccessModeT mode, 
                         cfidrErrorT *error));

cfidrVoidT          cfidrPIQuit CFI_ARGS((
                         cfidrErrorT *error));

cfidrPortDirectionT cfidrPortScalarGetDirection CFI_ARGS((
                         cfidrPortScalarIdT object, 
                         cfidrErrorT *error));

cfidrObjectIdT      cfidrPortGetOwner CFI_ARGS((
                         cfidrPortIdT object, 
                         cfidrErrorT *error));

cfidrInt32T         cfidrPortGetPosition CFI_ARGS((
                         cfidrPortIdT object, 
                         cfidrErrorT *error));

cfidrVoidT          cfidrPortSetOwnerPortBundle CFI_ARGS((
                         cfidrPortIdT port,
                         cfidrPortBundleIdT owner,
                         cfidrInt32T position,
                         cfidrErrorT *error));

cfidrVoidT          cfidrPortSetOwnerView CFI_ARGS((
                         cfidrPortIdT port,
                         cfidrInt32T position,
                         cfidrErrorT *error));

cfidrPortBundleIdT  cfidrPortBundleCreatePortBundle CFI_ARGS((
                         cfidrPortBundleIdT owner,
                         cfidrStringT name,
                         cfidrInt32T position,
                         cfidrErrorT *error));

cfidrPortBusIdT     cfidrPortBundleCreatePortBus CFI_ARGS((
                         cfidrPortBundleIdT owner,
                         cfidrStringT name,
                         cfidrInt32T position,
                         cfidrInt32T start,
                         cfidrInt32T step,
                         cfidrErrorT *error));

cfidrPortScalarIdT  cfidrPortBundleCreatePortScalar CFI_ARGS((
                         cfidrPortBundleIdT owner,
                         cfidrStringT name,
                         cfidrInt32T position,
                         cfidrPortDirectionT direction,
                         cfidrErrorT *error));

cfidrPortIdT        cfidrPortBundleFindPortByPosition CFI_ARGS((
                         cfidrPortBundleIdT owner,
                         cfidrInt32T position,
                         cfidrErrorT *error));

cfidrNetBundlesIdT  cfidrPortBundleGetNetBundles CFI_ARGS((
                         cfidrPortBundleIdT obj,
                         cfidrErrorT *error));

cfidrPortsIdT       cfidrPortBundleGetPorts CFI_ARGS((
                         cfidrPortBundleIdT object,
                         cfidrIterModeT mode, 
                         cfidrErrorT *error));

cfidrInt32T         cfidrPortBundleGetSize CFI_ARGS((
                         cfidrPortBundleIdT object,
                         cfidrErrorT *error));

cfidrInt32T         cfidrPortBusGetStart CFI_ARGS((
                         cfidrPortBusIdT portBus,
                         cfidrErrorT *error));

cfidrInt32T         cfidrPortBusGetStep CFI_ARGS((
                         cfidrPortBusIdT portBus,
                         cfidrErrorT *error));

cfidrVoidT          cfidrPortBusSetStart CFI_ARGS((
                         cfidrPortBusIdT portBus,
                         cfidrInt32T start,
                         cfidrErrorT *error));

cfidrVoidT          cfidrPortBusSetStep CFI_ARGS((
                         cfidrPortBusIdT portBus,
                         cfidrInt32T step,
                         cfidrErrorT *error));

cfidrNetScalarIdT   cfidrPortScalarGetNetScalar CFI_ARGS((
                         cfidrPortScalarIdT object, 
                         cfidrErrorT *error));

cfidrBooleanT       cfidrPortInstCheckDescriber CFI_ARGS((
                         cfidrPortInstIdT object, 
                         cfidrErrorT *error));

cfidrPortIdT        cfidrPortInstGetDescriber CFI_ARGS((
                         cfidrPortInstIdT object, 
                         cfidrErrorT *error));

cfidrPortDirectionT cfidrPortInstScalarGetDescriberDirection CFI_ARGS((
                         cfidrPortInstScalarIdT object, 
                         cfidrErrorT *error));

cfidrStringT        cfidrPortInstGetDescriberName CFI_ARGS((
                         cfidrPortInstIdT object, 
                         cfidrErrorT *error));

cfidrInstIdT        cfidrPortInstGetOwner CFI_ARGS((
                         cfidrPortInstIdT object, 
                         cfidrErrorT *error));

cfidrNetBundlesIdT  cfidrPortInstBundleGetNetBundles CFI_ARGS((
                         cfidrPortInstBundleIdT object,
                         cfidrErrorT *error));

cfidrPortInstsIdT   cfidrPortInstBundleGetPortInsts CFI_ARGS((
                         cfidrPortInstBundleIdT object,
                         cfidrIterModeT mode, 
                         cfidrErrorT *error));

cfidrInt32T         cfidrPortInstBundleGetSize CFI_ARGS((
                         cfidrPortInstBundleIdT portInstBundle,
                         cfidrErrorT *error));

cfidrInt32T         cfidrPortInstBusGetStart CFI_ARGS((
                         cfidrPortInstBusIdT portInstBus, 
                         cfidrErrorT *error));

cfidrInt32T         cfidrPortInstBusGetStep CFI_ARGS((
                         cfidrPortInstBusIdT portInstBus, 
                         cfidrErrorT *error));

cfidrNetScalarIdT   cfidrPortInstScalarGetNetScalar CFI_ARGS((
                         cfidrPortInstScalarIdT object, 
                         cfidrErrorT *error));

cfidrBooleanT       cfidrPropGetBooleanValue CFI_ARGS((
                         cfidrPropIdT object, 
                         cfidrErrorT *error));

cfidrFloat32T       cfidrPropGetFloat32Value CFI_ARGS((
                         cfidrPropIdT object, 
                         cfidrErrorT *error));

cfidrInt32T         cfidrPropGetInt32Value CFI_ARGS((
                         cfidrPropIdT object, 
                         cfidrErrorT *error));

cfidrObjectIdT      cfidrPropGetOwner CFI_ARGS((
                         cfidrPropIdT object, 
                         cfidrErrorT *error));

cfidrStringT        cfidrPropGetStringValue CFI_ARGS((
                         cfidrPropIdT object, 
                         cfidrErrorT *error));

cfidrValueTypeT     cfidrPropGetValueType CFI_ARGS((
                         cfidrPropIdT object, 
                         cfidrErrorT *error));

cfidrVoidT          cfidrPropSetBooleanValue CFI_ARGS((
                         cfidrPropIdT object,
                         cfidrBooleanT value, 
                         cfidrErrorT *error));

cfidrVoidT          cfidrPropSetFloat32Value CFI_ARGS((
                         cfidrPropIdT object,
                         cfidrFloat32T value, 
                         cfidrErrorT *error));

cfidrVoidT          cfidrPropSetInt32Value CFI_ARGS((
                         cfidrPropIdT object,
                         cfidrInt32T value, 
                         cfidrErrorT *error));

cfidrVoidT          cfidrPropSetStringValue CFI_ARGS((
                         cfidrPropIdT object,
                         cfidrStringT value, 
                         cfidrErrorT *error));

cfidrStringT        cfidrViewGetViewType CFI_ARGS((
                         cfidrViewIdT view,
                         cfidrErrorT *error));

cfidrStringT        cfidrEncapsulatedViewGetKey CFI_ARGS((
                         cfidrEncapsulatedViewIdT owner,
                         cfidrErrorT *error));

cfidrInstIdT        cfidrNetlistViewCreateInst CFI_ARGS((
                         cfidrNetlistViewIdT owner,
                         cfidrNetlistViewIdT desc,
                         cfidrStringT name,
                         cfidrErrorT *error));

cfidrNetBundleIdT   cfidrNetlistViewCreateNetBundle CFI_ARGS((
                         cfidrNetlistViewIdT owner, 
                         cfidrStringT name, 
                         cfidrErrorT *error));

cfidrNetBusIdT      cfidrNetlistViewCreateNetBus CFI_ARGS((
                         cfidrNetlistViewIdT owner,
                         cfidrStringT name,
                         cfidrInt32T start,
                         cfidrInt32T step,
                         cfidrErrorT *error));

cfidrNetScalarIdT   cfidrNetlistViewCreateNetScalar CFI_ARGS((
                         cfidrNetlistViewIdT owner,
                         cfidrStringT name,
                         cfidrErrorT *error));

cfidrPortBundleIdT  cfidrNetlistViewCreatePortBundle CFI_ARGS((
                         cfidrNetlistViewIdT owner,
                         cfidrStringT name,
                         cfidrInt32T position,
                         cfidrErrorT *error));

cfidrPortBusIdT     cfidrNetlistViewCreatePortBus CFI_ARGS((
                         cfidrNetlistViewIdT owner,
                         cfidrStringT name,
                         cfidrInt32T position,
                         cfidrInt32T start,
                         cfidrInt32T step,
                         cfidrErrorT *error));

cfidrPortScalarIdT  cfidrNetlistViewCreatePortScalar CFI_ARGS((
                         cfidrNetlistViewIdT owner,
                         cfidrStringT name,
                         cfidrInt32T position,
                         cfidrPortDirectionT dir,
                         cfidrErrorT *error));

cfidrInstsIdT       cfidrNetlistViewGetInsts CFI_ARGS((
                         cfidrNetlistViewIdT object, 
                         cfidrErrorT *error));

cfidrNetsIdT        cfidrNetlistViewGetNets CFI_ARGS((
                         cfidrNetlistViewIdT object,
                         cfidrIterModeT mode, 
                         cfidrErrorT *error));

cfidrCellIdT        cfidrViewGetOwner CFI_ARGS((
                         cfidrViewIdT object, 
                         cfidrErrorT *error));

cfidrPortsIdT       cfidrNetlistViewGetPorts CFI_ARGS((
                         cfidrNetlistViewIdT object,
                         cfidrIterModeT mode, 
                         cfidrErrorT *error));

cfidrVoidT          cfidrViewPurge CFI_ARGS((
                         cfidrViewIdT view, 
                         cfidrErrorT *error));

cfidrVoidT          cfidrViewSave CFI_ARGS((
                         cfidrViewIdT view, 
                         cfidrErrorT *error));

cfidrViewIdT        cfidrCellFindViewByName CFI_ARGS((
                         cfidrCellIdT cell, 
                         cfidrStringT name, 
                         cfidrErrorT *error));

cfidrPortInstIdT    cfidrInstFindPortInstByDescriberName CFI_ARGS((
                         cfidrInstIdT inst, 
                         cfidrStringT name, 
                         cfidrErrorT *error));

cfidrCellIdT        cfidrLibFindCellByName CFI_ARGS((
                         cfidrLibIdT library, 
                         cfidrStringT name, 
                         cfidrErrorT *error));

cfidrNetIdT         cfidrNetBundleFindNetByName CFI_ARGS((
                         cfidrNetBundleIdT netBundle, 
                         cfidrStringT name, 
                         cfidrErrorT *error));

cfidrNetBundleIdT   cfidrNetFindNetBundleByName CFI_ARGS((
                         cfidrNetIdT net, 
                         cfidrStringT name, 
                         cfidrErrorT *error));

cfidrInstIdT        cfidrNetlistViewFindInstByName CFI_ARGS((
                         cfidrNetlistViewIdT view, 
                         cfidrStringT name, 
                         cfidrErrorT *error));

cfidrNetIdT         cfidrNetlistViewFindNetByName CFI_ARGS((
                         cfidrNetlistViewIdT view, 
                         cfidrStringT name, 
                         cfidrErrorT *error));

cfidrPortIdT        cfidrNetlistViewFindPortByName CFI_ARGS((
                         cfidrNetlistViewIdT view, 
                         cfidrStringT name, 
                         cfidrErrorT *error));

cfidrPropIdT        cfidrObjectFindPropByName CFI_ARGS((
                         cfidrObjectIdT object, 
                         cfidrStringT name, 
                         cfidrErrorT *error));

cfidrLibIdT         cfidrPIFindLibByName CFI_ARGS((
                         cfidrStringT name, 
                         cfidrErrorT *error));

cfidrPortIdT        cfidrPortBundleFindPortByName CFI_ARGS((
                         cfidrPortBundleIdT portBundle, 
		         cfidrStringT name, 
                         cfidrErrorT *error));

cfidrPortInstIdT    cfidrPortInstBundleFindPortInstByDescriberName CFI_ARGS((
                         cfidrPortInstBundleIdT portInstBundle, 
                         cfidrStringT name, 
                         cfidrErrorT *error));

cfidrLibrarySelectorIdT cfidrIterNextLibrarySelector CFI_ARGS((
                         cfidrLibrarySelectorsIdT iter, 
                         cfidrErrorT *error));

cfidrViewSelectorIdT    cfidrIterNextViewSelector CFI_ARGS((
                         cfidrViewSelectorsIdT iter, 
                         cfidrErrorT *error));

cfidrStringT            cfidrLibSelectorGetLibName CFI_ARGS((
                         cfidrLibrarySelectorIdT selector, 
                         cfidrErrorT *error));
 
cfidrViewIdT            cfidrSelectorSetSelectView CFI_ARGS((
                         cfidrSelectorSetIdT owner, 
                         cfidrCellIdT cell,
                         cfidrErrorT *error));

cfidrSelectorSetIdT     cfidrPICreateSelectorSet CFI_ARGS((
                         cfidrStringT name, 
                         cfidrErrorT *error));

cfidrSelectorSetIdT     cfidrPIOpenSelectorSet CFI_ARGS((
                         cfidrStringT name, 
                         cfidrErrorT *error));

cfidrLibrarySelectorIdT cfidrSelectorSetCreateLibrarySelector CFI_ARGS((
                         cfidrSelectorSetIdT owner, 
                         cfidrStringT name, 
                         cfidrErrorT *error));

cfidrViewNameSelectorIdT cfidrSelectorSetCreateViewNameSelector CFI_ARGS((
                         cfidrSelectorSetIdT owner, 
                         cfidrStringT libname, 
			 cfidrStringT cellname, 
                         cfidrStringT viewname, 
                         cfidrErrorT *error));

cfidrViewTypeSelectorIdT cfidrSelectorSetCreateViewTypeSelector CFI_ARGS((
                         cfidrSelectorSetIdT owner, 
                         cfidrStringT libname, 
			 cfidrStringT cellname, 
                         cfidrStringT viewtype, 
                         cfidrErrorT *error));

cfidrLibrarySelectorsIdT cfidrSelectorSetGetLibrarySelectors CFI_ARGS((
                         cfidrSelectorSetIdT owner, 
                         cfidrErrorT *error));

cfidrViewSelectorsIdT   cfidrSelectorSetGetViewSelectors CFI_ARGS((
                         cfidrSelectorSetIdT owner, 
                         cfidrErrorT *error));

cfidrVoidT              cfidrSelectorSetPurge CFI_ARGS((
                         cfidrSelectorSetIdT owner, 
                         cfidrErrorT *error));

cfidrVoidT              cfidrSelectorSetSave CFI_ARGS((
                         cfidrSelectorSetIdT owner, 
                         cfidrErrorT *error));

cfidrStringT            cfidrViewNameSelectorGetCellName CFI_ARGS((
                         cfidrViewNameSelectorIdT selector, 
                         cfidrErrorT *error));

cfidrStringT            cfidrViewNameSelectorGetLibName CFI_ARGS((
                         cfidrViewNameSelectorIdT selector, 
                         cfidrErrorT *error));

cfidrStringT            cfidrViewNameSelectorGetViewName CFI_ARGS((
                         cfidrViewNameSelectorIdT selector, 
                         cfidrErrorT *error));

cfidrStringT            cfidrViewTypeSelectorGetCellName CFI_ARGS((
                         cfidrViewTypeSelectorIdT selector, 
                         cfidrErrorT *error));

cfidrStringT            cfidrViewTypeSelectorGetLibName CFI_ARGS((
                         cfidrViewTypeSelectorIdT selector, 
                         cfidrErrorT *error));

cfidrStringT            cfidrViewTypeSelectorGetViewType CFI_ARGS((
                         cfidrViewTypeSelectorIdT selector, 
                         cfidrErrorT *error));

/* END OF CFI '92 DR PI Public Functions */

#ifdef __cpluscplus
}
#endif

#endif /* CFIDR_H */



