/*
 * cfidrDefs.h					GMD-ETS, February 14, 1991 (RC)
 *
 * derived from 
 *                 CFI'91 Design Representation 
 *                       Information Model 
 *                             and 
 *                     Programming Interface
 *                          Version 0.7
 *                        February 7, 1991
 */


/*
 * Boolean Values
 */

typedef enum {
    CFIDR_FALSE  = 0,
    CFIDR_TRUE   = 1  
} cfidrBooleanT;


/*
 * 32-bit Integer Values
 */

#define CFIDR_MIN_INT32   		2147483647
#define CFIDR_MAX_INT32   		-2147483648
typedef long int			cfidrInt32T;


/*
 * 32-bit Float Point Values
 */

#define CFIDR_MIN_FLOAT32 		-1E+37
#define CFIDR_MAX_FLOAT32 		1E+37
typedef float				cfidrFloat32T;


/*
 * Character Strings
 */

typedef char*	cfidrStringT;


/*
 * Void Return Values
 */

typedef void	cfidrVoidT;


/*
 * Object Identifiers
 */

typedef long int		 cfidrObjectIdT;
typedef cfidrObjectIdT           cfidrLibIdT;
typedef cfidrObjectIdT           cfidrCellIdT;
typedef cfidrObjectIdT           cfidrPortIdT;
typedef cfidrObjectIdT           cfidrPortBundleIdT;
typedef cfidrObjectIdT           cfidrPortScalarIdT;
typedef cfidrObjectIdT           cfidrInstIdT;
typedef cfidrObjectIdT           cfidrPortInstIdT;
typedef cfidrObjectIdT           cfidrPortInstBundleIdT;
typedef cfidrObjectIdT           cfidrPortInstScalarIdT;
typedef cfidrObjectIdT           cfidrNetIdT;
typedef cfidrObjectIdT           cfidrNetBundleIdT;
typedef cfidrObjectIdT           cfidrNetScalarIdT;
typedef cfidrObjectIdT           cfidrPropIdT;
typedef cfidrObjectIdT           cfidrViewIdT;


/*
 * List Identifiers
 */

typedef long int	cfidrListIdT;
typedef cfidrListIdT	cfidrCellsIdT;
typedef cfidrListIdT	cfidrInstsIdT;
typedef cfidrListIdT	cfidrLibsIdT;
typedef cfidrListIdT	cfidrNetsIdT;
typedef cfidrListIdT	cfidrNetBundlesIdT;
typedef cfidrListIdT	cfidrPortsIdT;
typedef cfidrListIdT	cfidrPortInstsIdT;
typedef cfidrListIdT	cfidrPropsIdT;
typedef cfidrListIdT	cfidrViewsIdT;


/*
 * Enumerated Constants
 */



/*
 * Types of Objects
 */

 typedef enum {
    CFIDR_UNDEFINED_OBJECT	= 0, 
    CFIDR_LIB			= 1,
    CFIDR_CELL			= 2,
    CFIDR_PORTBUNDLE		= 3,
    CFIDR_PORTSCALAR		= 4,
    CFIDR_INST			= 5,
    CFIDR_PORTINSTBUNDLE 	= 6,
    CFIDR_PORTINSTSCALAR	= 7,
    CFIDR_NETBUNDLE	 	= 8,
    CFIDR_NETSCALAR	 	= 9,
    CFIDR_PROP			= 10,
    CFIDR_VIEW			= 11
} cfidrObjectTypeT;


/*
 * Types of Views
 */

 typedef enum {
    CFIDR_UNDEFINED_VIEWTYPE	= 0,
    CFIDR_NETLIST_VIEW		= 1 
} cfidrViewTypeT;


/*
 * Types of Values
 */

 typedef enum {
    CFIDR_UNDEFINED_VALUETYPE	= 0,
    CFIDR_INT32			= 1,
    CFIDR_STRING		= 2,
    CFIDR_FLOAT32		= 3,
    CFIDR_BOOLEAN		= 4
} cfidrValueTypeT;


/*
 * Types of Port Directions
 */

 typedef enum {
    CFIDR_UNDEFINED_PORTDIRECTION	= 0,
    CFIDR_INPUT				= 1,
    CFIDR_OUTPUT			= 2,
    CFIDR_IO				= 3
} cfidrPortDirectionT;


/*
 * Types of Attach Modes
 */

 typedef enum {
    CFIDR_UNDEFINED_ATTACH_MODE			= 0,
    CFIDR_ATTACH_BY_ORDER			= 1,
    CFIDR_ATTACH_BY_NAME			= 2
} cfidrAttachModeT;


/*
 * Types of List Modes
 */

typedef enum {
    CFIDR_UNDEFINED_LIST_MODE		= 0,
    CFIDR_LIST_SCALARS			= 1,
    CFIDR_LIST_BUNDLES			= 2,
    CFIDR_LIST_TOP			= 3,
    CFIDR_LIST_ALL			= 4
} cfidrListModeT;


/*
 * Types of Errors and the Error Text
 */

typedef enum {
    CFIDR_NO_ERROR			=  0,
    CFIDR_CONNECTIVITY_RULE_VIOLATION	=  1,
    CFIDR_DESCRIBER_PORT_NOT_FOUND	=  2,
    CFIDR_DESCRIBER_VIEW_NOT_FOUND	=  3,
    CFIDR_INTERNAL_SYSTEM_ERROR		=  4,
    CFIDR_INVALID_ATTACH_MODE		=  5,
    CFIDR_INVALID_DESCRIBER		=  6,
    CFIDR_INVALID_ERROR_CODE		=  7,
    CFIDR_INVALID_FLOAT32		=  8,
    CFIDR_INVALID_INT32			=  9,
    CFIDR_INVALID_LIST			= 10,
    CFIDR_INVALID_MEMBERS		= 11,
    CFIDR_INVALID_NAME			= 12,
    CFIDR_INVALID_OBJECT_TYPE		= 13,
    CFIDR_INVALID_OID			= 14,
    CFIDR_INVALID_PORT			= 15,
    CFIDR_INVALID_PORTINST		= 16,
    CFIDR_INVALID_VIEWTYPE		= 17,
    CFIDR_LIBRARY_ALREADY_OPEN		= 18,
    CFIDR_NEW_DESCRIBER_PORT_FOUND	= 19,
    CFIDR_NO_NAME_MATCHES		= 20,
    CFIDR_OBJECT_ALREADY_EXISTS		= 21,
    CFIDR_OBJECT_NOT_FOUND		= 22,
    CFIDR_PORTINST_ALREADY_ATTACHED	= 23,
    CFIDR_PORTINST_NOT_ATTACHED		= 24,
    CFIDR_PORT_ALREADY_ATTACHED		= 25,
    CFIDR_PORT_MEMBERS_DIFFER		= 26,
    CFIDR_PORT_NOT_ATTACHED		= 27,
    CFIDR_VIEW_ALREADY_OPEN		= 28,
    CFIDR_OPEN_FOR_UPDATE_FAILED	= 29,
    CFIDR_INVALID_PORTDIRECTION		= 30,
    CFIDR_INVALID_LIST_MODE		= 31
} cfidrErrorT;

/*
cfidrVoidT		cfidrAttachPortInsts ();
cfidrVoidT		cfidrAttachPorts ();
cfidrBooleanT		cfidrCheckDescriber ();
cfidrCellIdT		cfidrCreateCell ();
cfidrInstIdT		cfidrCreateInst ();
cfidrLibIdT		cfidrCreateLib ();
cfidrNetBundleIdT	cfidrCreateNetBundle ();
cfidrNetScalarIdT	cfidrCreateNetScalar ();
cfidrPortBundleIdT	cfidrCreatePortBundle ();
cfidrPortScalarIdT	cfidrCreatePortScalar ();
cfidrPropIdT		cfidrCreateProp ();
cfidrViewIdT		cfidrCreateView ();
cfidrVoidT		cfidrDeleteObject ();
cfidrVoidT		cfidrDetachPortInsts ();
cfidrVoidT		cfidrDetachPorts ();
cfidrBooleanT		cfidrGetBooleanValue ();
cfidrCellsIdT		cfidrGetCells ();
cfidrObjectIdT		cfidrGetDescriber ();
cfidrStringT		cfidrGetErrorText ();
cfidrFloat32T		cfidrGetFloat32Value ();
cfidrInstsIdT		cfidrGetInsts ();
cfidrInt32T		cfidrGetInt32Value ();
cfidrLibsIdT		cfidrGetLibs ();
cfidrStringT		cfidrGetName ();
cfidrNetsIdT		cfidrGetNetBundles ();
cfidrNetIdT		cfidrGetNetScalar ();
cfidrNetsIdT		cfidrGetNets ();
cfidrObjectIdT		cfidrGetNullId ();
cfidrObjectTypeT	cfidrGetObjectType ();
cfidrObjectIdT		cfidrGetOwner ();
cfidrPortDirectionT	cfidrGetPortDirection ();
cfidrPortInstsIdT	cfidrGetPortInsts ();
cfidrPortsIdT		cfidrGetPorts ();
cfidrPropsIdT		cfidrGetProps ();
cfidrStringT		cfidrGetStringValue ();
cfidrValueTypeT		cfidrGetValueType ();
cfidrViewTypeT		cfidrGetViewType ();
cfidrViewsIdT		cfidrGetViews ();
cfidrVoidT		cfidrInitPI ();
cfidrBooleanT		cfidrIsNullObject ();
cfidrBooleanT		cfidrIsSameObject ();
cfidrBooleanT		cfidrIsUsableObject ();
cfidrCellIdT		cfidrNextCell ();
cfidrInstIdT		cfidrNextInst ();
cfidrStringT			cfidrNextLib ();
cfidrNetIdT		cfidrNextNet ();
cfidrNetBundleIdT	cfidrNextNetBundle ();
cfidrPortIdT		cfidrNextPort ();
cfidrPortInstIdT	cfidrNextPortInst ();
cfidrPropIdT		cfidrNextProp ();
cfidrViewIdT		cfidrNextView ();
cfidrLibIdT		cfidrOpenLib ();
cfidrViewIdT		cfidrOpenView ();
cfidrVoidT		cfidrPurgeObject ();
cfidrVoidT		cfidrQuitList ();
cfidrVoidT		cfidrQuitPI ();
cfidrVoidT		cfidrResetList ();
cfidrVoidT		cfidrSaveObject ();
cfidrVoidT		cfidrSetBooleanValue ();
cfidrVoidT		cfidrSetFloat32Value ();
cfidrVoidT		cfidrSetInt32Value ();
cfidrVoidT		cfidrSetName ();
cfidrVoidT		cfidrSetStringValue ();
*/
