#ifdef FRAMEWORK
/******************includes****************************************************/
#include <strings.h>
#include "extern.h"

#include "cfidr.h"			    /* contains typedefs and enums    */
#include "cfidrread.h"
#include "gate_C.h"
#include "line_C.h"
#include "gate_liste_C.h"
#include "line_liste_C.h"
#include "stringliste_C.h"
#include "cfidrprototypes.h"


/******************************************************************************/
/*									      */
/*   local variables							      */
/*									      */
/* ************************************************************************** */
cfidrViewIdT cfidrMainViewId;		    /* Id of the Main View	      */
cfidrErrorT cfidrError;
gate_C *dummy_gate_p = new gate_C("dummy_gate");
extern boolean path_delay_test_b;

/******************************************************************************/
/*                                                                            */
/*   function declarations                                                    */
/*                                                                            */
/* ************************************************************************** */
cfidrVoidT cfidr_GetAllViewLeafs(cfidrViewIdT cfidrViewId);
cfidrVoidT cfidr_GetAllViewLines(cfidrViewIdT cfidrViewId);
cfidrVoidT cfidr_InitLeaf(cfidrViewIdT cfidrViewId);
cfidrVoidT cfidr_FindConnection(cfidrViewIdT cfidrViewId);
cfidrVoidT cfidr_CreateInputsOutputs(void);
cfidrVoidT cfidr_ConnectMacros(cfidrViewIdT cfidrViewId);

/* ************************************************************************** */
/*									      */
/*   FUNCTION NAME: cfidr_ReadCircuitFromDb				      */
/*									      */
/*   AUTHOR: U. Glaeser                                                       */
/*                                                                            */
/*   FUNCTIONAL DESCRIPTION:						      */
/*									      */
/*   cfidr_ReadCircuitFromDb reads the circuit-description from the cfi-data  */
/*   base into the general data structure.				      */
/*									      */
/*   First, the input and output devices for the circuit are created (call    */
/*   of cfidr_CreateInputsOutputs).					      */
/*   									      */
/*   Second, all leafs of the view are visited (call of			      */
/*   cfidr_GetAllViewLeafs). During this process, the blocks are initialised  */
/*   (no connection between blocks and lines is established).	              */
/*									      */
/*   Third, the macros are internally connected by cfidr_ConnectMacros	      */
/*									      */
/*   Then, all nets of the view are visited (call of cfidr_FindConnection).   */
/*									      */
/*   Last, all the lines, which don't point to any block are deleted. The     */
/*   number of outputs of a block is computed and put to the component        */
/*   no_output of the block. Every fanout rises no_output in extest	      */
/*									      */
/*   FORMAL PARAMETERS:							      */
/*									      */
/*	 cfidrViewId: Id of the View					      */
/*									      */
/*   IMPLICIT INPUT/OUTPUT: none					      */
/*									      */
/*   FUNCTION VALUE: none						      */
/*									      */
/*   CREATION DATE: 18.01.90						      */
/*									      */
/*   MODIFICATION HISTORY: not modified					      */
/*									      */
/* ************************************************************************** */
void cfidr_ReadCircuitFromDb(cfidrViewIdT cfidrViewId)
{
   line_C *line_item_p;

   cfidrError = CFIDR_NO_ERROR;
   dummy_gate_p->belege_gate(UNDEFINED_GATE,0,0);
   io_t dummy_io_e;

   cout << " Read Lines \n" << flush;
   cfidr_GetAllViewLines(cfidrViewId);
   cout << " Read Instances \n" << flush;
   cfidr_GetAllViewLeafs(cfidrViewId);
   cout << " Read Inputs and Outputs \n" << flush;
   cfidr_CreateInputsOutputs();
   cout << " Connect Nets \n" << flush;
   cfidr_FindConnection(cfidrViewId);
   cout << " Connect Macros Internal \n" << flush;
   cfidr_ConnectMacros(cfidrViewId);
   cout << " Connect Lines which aren't connected to Dummy\n" << flush;
//   forall(line_item_p,lines_C)
//	if ((line_item_p->nachfolger()->is_empty() == TRUE) && (line_item_p->is_primary_output() == FALSE))
//	   {
//	   cout << line_item_p->get_name();
//	   warning("cfidr_ReadCircuitFromDb: Line is not connected");
//	   dummy_io_e = line_item_p->nachfolger()->verbinde(dummy_gate_p,I1);
//	   }//forall
}  /*cfidr_ReadCircuitFromDb*/

/* ************************************************************************** */
/*									      */
/*   FUNCTION NAME: cfidr_ReadCircuit					      */
/*									      */
/*   AUTHOR: U. Glaeser                                                       */
/*                                                                            */
/*   FUNCTIONAL DESCRIPTION:						      */
/*									      */
/*   cfidr_ReadCircuit  calls cfidr_ReadCircuitFromDb to read the	      */
/*   circuit-description from the CFi-Db into the general data structure.     */
/*   The circuit should be a flat one.					      */
/*   Aditionally the database is intialized and after reading closed	      */
/*									      */
/*   FORMAL PARAMETERS: 						      */
/*									      */
/*	 cellname_s: name of the cell, which should be analysed by extest     */
/*									      */
/*   IMPLICIT INPUT/OUTPUT: none					      */
/*									      */
/*   FUNCTION VALUE: none						      */
/*									      */
/*   CREATION DATE: 16.01.91						      */
/*									      */
/*   MODIFICATION HISTORY: not modified					      */
/*									      */
/* ************************************************************************** */
cfidrViewIdT cfidr_ReadCircuit(char *libraryname_s,char *cellname_s,char *viewname_s)
{
   cfidrStringT cfidrString;
   
	cout << " Open Database \n" << flush;
   cfidrPIInit (&cfidrError);
   if (cfidrError != CFIDR_NO_ERROR)
	{
	cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
	error (cfidrString);
	}  /*if*/
   cfidrMainViewId = cfidrPIOpenView (libraryname_s, cellname_s,
				    viewname_s, CFIDR_UPDATE, &cfidrError);
	/* open db with libraryname, cellname, netlist as arguments */
   if (cfidrError != CFIDR_NO_ERROR)
	{
	cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
	error (cfidrString);
	}/*if*/

   cfidr_ReadCircuitFromDb(cfidrMainViewId);	/* read circuit description from  */
						/* cfi-db		          */
return(cfidrMainViewId);   
}/*cfidr_ReadCircuit*/
   


/* ************************************************************************** */
/*									      */
/*  FUNCTION NAME: cfidr_GetAllViewLines				      */
/*									      */
/*   AUTHOR: U. Glaeser                                                       */
/*                                                                            */
/*  FUNCTIONAL DESCRIPTION:						      */
/*									      */
/*	For each instance found in the instance list of the view,	      */
/*	cfidr_InitLeaf is called.					      */
/*									      */
/*  FORMAL PARAMETERS:							      */
/*									      */
/*	cfidrViewId: Id of the current view == MainView			      */
/*									      */
/*  IMPLICIT INPUT/OUTPUT:						      */
/*									      */
/*	none								      */
/*									      */
/*  FUNCTION VALUE:							      */
/*									      */
/*	none								      */
/*									      */
/*  CREATION DATE: 16.01.91						      */
/*									      */
/*  MODIFICATION HISTORY:						      */
/*									      */
/*	none								      */
/*									      */
/* ************************************************************************** */
void cfidr_GetAllViewLines(cfidrViewIdT cfidrViewId)
{
   cfidrNetsIdT cfidrNetsId;
   cfidrNetIdT cfidrNetId;
   cfidrStringT cfidrString;
   char local_name_s[stringsize_i];
   line_C *line_item_p;


   cfidrError = CFIDR_NO_ERROR;
   cfidrNetsId = cfidrNetlistViewGetNets(cfidrViewId, CFIDR_ITER_SCALARS, &cfidrError); /* Get List of      */
					    /* nets from current view    */
					    /* cfidr_GetAllViewLines	      */
	if (cfidrError != CFIDR_NO_ERROR)
		{
		cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
		error (cfidrString);
		}  /*if*/

   for (;;)
	{
	cfidrNetId = cfidrIterNextNet(cfidrNetsId, &cfidrError);
	if (cfidrError != CFIDR_NO_ERROR)
		{
		cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
		error (cfidrString);
		}  /*if*/

	if (cfidrObjectIsNull(cfidrNetId ,&cfidrError) == CFIDR_TRUE)
	return;
	else
		{
		if (cfidrError != CFIDR_NO_ERROR)
			{
			cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
			error (cfidrString);
			}  /*if*/

		strcpy(local_name_s,cfidrNamedObjectGetName(cfidrNetId, &cfidrError)); /* perform name of line      */
		statistic_u.number_lines_i++;
		line_item_p = new line_C(local_name_s);
		lines_C.fuege_ein(line_item_p);
		}
	}
}//cfidr_GetAllViewLines


/* ************************************************************************** */
/*									      */
/*  FUNCTION NAME: cfidr_GetAllViewLeafs				      */
/*									      */
/*   AUTHOR: U. Glaeser                                                       */
/*                                                                            */
/*  FUNCTIONAL DESCRIPTION:						      */
/*									      */
/*	For each instance found in the instance list of the view,	      */
/*	cfidr_InitLeaf is called.					      */
/*									      */
/*  FORMAL PARAMETERS:							      */
/*									      */
/*	cfidrViewId: Id of the current view == MainView			      */
/*									      */
/*  IMPLICIT INPUT/OUTPUT:						      */
/*									      */
/*	none								      */
/*									      */
/*  FUNCTION VALUE:							      */
/*									      */
/*	none								      */
/*									      */
/*  CREATION DATE: 16.01.91						      */
/*									      */
/*  MODIFICATION HISTORY:						      */
/*									      */
/*	none								      */
/*									      */
/* ************************************************************************** */
void cfidr_GetAllViewLeafs(cfidrViewIdT cfidrViewId)
{
   cfidrInstsIdT cfidrInstsId;
   cfidrInstIdT cfidrInstId;
   cfidrStringT cfidrString;

   cfidrError = CFIDR_NO_ERROR;
   cfidrInstsId = cfidrNetlistViewGetInsts(cfidrViewId, &cfidrError); /* Get List of      */
					    /* instances from current view    */
					    /* cfidr_GetAllViewLeafs	      */
	if (cfidrError != CFIDR_NO_ERROR)
		{
		cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
		error (cfidrString);
		}  /*if*/

   for (;;)
	{
	cfidrInstId = cfidrIterNextInst(cfidrInstsId, &cfidrError);
	if (cfidrError != CFIDR_NO_ERROR)
		{
		cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
		error (cfidrString);
		}  /*if*/

	if (cfidrObjectIsNull(cfidrInstId ,&cfidrError) == CFIDR_TRUE)
	return;
	else
		{
		if (cfidrError != CFIDR_NO_ERROR)
			{
			cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
			error (cfidrString);
			}  /*if*/

		cfidr_InitLeaf(cfidrInstId); /* perform init of leaf      */
		statistic_u.number_gates_i++;
		}
	}
}//cfidr_GetAllViewLeafes


/*************************************************************************/
/*                                                                       */
/*   FUNCTION-NAME: cfidr_TableLookup                                    */
/*                                                                       */
/*   AUTHOR: U. Glaeser                                                  */
/*                                                                       */
/*   FUNCTIONAL-DESCRIPTION:						 */
/*                                                                       */
/*   The cell name is looked up in the table cfidr_Tabelle and for each  */
/*   element the corresponding id is written to el_id_i. For each cell   */
/*   the extest id's of each subcell (for macros) is returned in the     */
/*   function value. A second array extest_name_s is returned with the   */
/*   extest cell names.							 */
/*                                                                       */
/*   FORMAL PARAMETERS:  cfidrString: der Name des Elements              */
/*                                                                       */
/*   IMPLICIT INPUT/OUTPUT: none                                         */
/*                                                                       */
/*   FUNCTION-VALUE: int *: die Id's der Elemente                        */
/*                                                                       */
/*   CREATION-DATE:  Maerz '91                                           */
/*                                                                       */
/*   MODIFICATION-HISTORY: none                                          */
/*                                                                       */
/*************************************************************************/
void cfidr_TableLookup(cfidrStringT cfidrString,cfidrStringT cfidrInstName)
{
int i,el_id_i,no_in_i = 0;
gate_C *gate_item_p;

if ((cfidrString[0] <= '9') && (cfidrString[0] > '0'))
	{
	no_in_i = cfidrString[0] - '0';
	cfidrString = &cfidrString[1];
	}//if
if ((cfidrString[0] <= '9') && (cfidrString[0] > '0'))
	error("cfidr_TableLookup: Some gate has more than nine inputs");
for (i=0;i<baustein_size_i;++i)
	if (strcasecmp(cfidrString,(&cfidr_Tabelle[i])->baust_name) == 0)
		{
		el_id_i = (&cfidr_Tabelle[i])->codenummer;
		break;
		}
if (i == baustein_size_i)
      {
      cout << cfidrString;
      error("Type of element not found in internal library");
      }//if
if ((el_id_i < 12) && (el_id_i > 7) && (no_in_i == 0))
	no_in_i = cfidrString[2] - '0';
switch (el_id_i) {
	case 6: /* driver */
		statistic_u.number_drivers_i++;
		gate_item_p = new gate_C(cfidrInstName);
		gate_item_p->belege_gate(DRIVER,1,1);
		gates_C.fuege_ein(gate_item_p);
		break;
	case 7: /* inverter */
		statistic_u.number_inverters_i++;
		gate_item_p = new gate_C(cfidrInstName);
		gate_item_p->belege_gate(INVERTER,1,1);
		gates_C.fuege_ein(gate_item_p);
		break;
	case 8: /* and gate */
		statistic_u.number_and_gates_i++;
		gate_item_p = new gate_C(cfidrInstName);
		gate_item_p->belege_gate(AND,no_in_i,1);
		gates_C.fuege_ein(gate_item_p);
		break;
	case 9: /* nand gate */
		statistic_u.number_nand_gates_i++;
		gate_item_p = new gate_C(cfidrInstName);
		gate_item_p->belege_gate(NAND,no_in_i,1);
		gates_C.fuege_ein(gate_item_p);
		break;
	case 10: /* or gate */
		statistic_u.number_or_gates_i++;
		gate_item_p = new gate_C(cfidrInstName);
		gate_item_p->belege_gate(OR,no_in_i,1);
		gates_C.fuege_ein(gate_item_p);
		break;
	case 11: /* nor gate */
		statistic_u.number_nor_gates_i++;
		gate_item_p = new gate_C(cfidrInstName);
		gate_item_p->belege_gate(NOR,no_in_i,1);
		gates_C.fuege_ein(gate_item_p);
		break;
	case 12: /* exor gate */
		statistic_u.number_exor_gates_i++;
		gate_item_p = new gate_C(cfidrInstName);
		gate_item_p->belege_gate(EXOR,2,1);
		gates_C.fuege_ein(gate_item_p);
		break;
	case 14: /* simple latch */
		statistic_u.number_flipflops_i++;
		gate_item_p = new gate_C(cfidrInstName);
		if (strcasecmp(cfidrString,"dff1") == 0)
			gate_item_p->belege_gate(D_FLIP_FLOP,1,1);
		else if (strcasecmp(cfidrString,"dtp10") == 0)
			gate_item_p->belege_gate(D_FLIP_FLOP,3,2);
		else if (strcasecmp(cfidrString,"flipflop") == 0)
			gate_item_p->belege_gate(D_FLIP_FLOP,4,2);
		else
			gate_item_p->belege_gate(D_FLIP_FLOP,2,2);
		gates_C.fuege_ein(gate_item_p);
		break;
	case TO010LJ: /* function: out1 = HIGH, out2 = LOW */
		      /* extest function: out1 = in OR NOT(in) */
		      /*		  out2 = in AND NOT(in) */
		gate_item_p = new gate_C(cfidrInstName);
		gate_item_p->belege_gate(UNDEFINED_GATE,0,2);
		gates_C.fuege_ein(gate_item_p);
		break;
	case AO220LJ: /* inputs: A,B,C,D	output: Y */
		      /* function: Y = (A AND B) OR (C AND D) */
		gate_item_p = new gate_C(cfidrInstName);
		gate_item_p->belege_gate(UNDEFINED_GATE,4,1);
		gates_C.fuege_ein(gate_item_p);
		break;
	case AO221LJ: /* inputs: A,B,C,D	output: Y */
		      /* function: Y = (A AND B) NOR (C AND D) */
		gate_item_p = new gate_C(cfidrInstName);
		gate_item_p->belege_gate(UNDEFINED_GATE,4,1);
		gates_C.fuege_ein(gate_item_p);
		break;
	case AO241LJ: /* inputs: A1,A2,B1,B2,C1,C2,D1,D2  output: Y */
		/* function: Y = NOR(A1 AND A2,B1 AND B2,C1 AND C2,D1 AND D2) */
		gate_item_p = new gate_C(cfidrInstName);
		gate_item_p->belege_gate(UNDEFINED_GATE,8,1);
		gates_C.fuege_ein(gate_item_p);
		break;
	case AO421LJ: /* inputs: A1,A2,A3,A4,B1,B2,B3,B4  output: Y */
		      /* function: Y = AND(A1,A2,A3,A4) NOR AND(B1,B2,B3,B4) */
		gate_item_p = new gate_C(cfidrInstName);
		gate_item_p->belege_gate(UNDEFINED_GATE,8,1);
		gates_C.fuege_ein(gate_item_p);
		break;
	case BF001LJ: /* inputs: B1,B2,A1		output: Y */
		      /* function: Y = (B1 AND B2) NOR A1 */
		gate_item_p = new gate_C(cfidrInstName);
		gate_item_p->belege_gate(UNDEFINED_GATE,3,1);
		gates_C.fuege_ein(gate_item_p);
		break;
	case BF006LJ: /* inputs: B1,B2,A1,A2		output: Y */
		      /* function: Y = NOR(B1 AND B2,A1,A2) */
		gate_item_p = new gate_C(cfidrInstName);
		gate_item_p->belege_gate(UNDEFINED_GATE,4,1);
		gates_C.fuege_ein(gate_item_p);
		break;
	case BF011LJ: /* inputs: A1,A2,B1,B2,C1,C2	output: Y */
		      /* function: Y = NOR(A1 AND A2,B1 AND B2,C1 AND C2) */
		gate_item_p = new gate_C(cfidrInstName);
		gate_item_p->belege_gate(UNDEFINED_GATE,6,1);
		gates_C.fuege_ein(gate_item_p);
		break;
	case BF022LJ: /* inputs: C1,C2,B1,B2,A1,A2	output: Y */
		      /* function: Y = AND(C1 OR C2,B1,B2) NOR (A1 AND A2) */
		gate_item_p = new gate_C(cfidrInstName);
		gate_item_p->belege_gate(UNDEFINED_GATE,6,1);
		gates_C.fuege_ein(gate_item_p);
		break;
	case BF051LJ: /* inputs: B1,B2,A1		output: Y */
		      /* function: Y = (B1 OR B2) NAND A1 */
		gate_item_p = new gate_C(cfidrInstName);
		gate_item_p->belege_gate(UNDEFINED_GATE,3,1);
		gates_C.fuege_ein(gate_item_p);
		break;
	case BF053LJ: /* inputs: A1,A2,B1,B2		output: Y */
		      /* function: Y = (A1 OR A2) NAND (B1 OR B2) */
		gate_item_p = new gate_C(cfidrInstName);
		gate_item_p->belege_gate(UNDEFINED_GATE,4,1);
		gates_C.fuege_ein(gate_item_p);
		break;
	case BF056LJ: /* inputs: B1,B2,A1,A2		output: Y */
		      /* function: Y = NAND(B1 OR B2,A1,A2) */
		gate_item_p = new gate_C(cfidrInstName);
		gate_item_p->belege_gate(UNDEFINED_GATE,4,1);
		gates_C.fuege_ein(gate_item_p);
		break;
	case NA311LJ: /* inputs: A,B,C			output: Y */
		      /* function: Y = NAND(NOT A,B,C) */
		gate_item_p = new gate_C(cfidrInstName);
		gate_item_p->belege_gate(UNDEFINED_GATE,3,1);
		gates_C.fuege_ein(gate_item_p);
		break;
	default: /* may also be not a flat circuit as input */
		error("Unknown Gate Type in Database Circuit");
	}/*switch*/

} /* cfidr_TableLookup */

/* ************************************************************************** */
/*									      */
/*  FUNCTION NAME: cfidr_InitLeaf					      */
/*									      */
/*  FUNCTIONAL DESCRIPTION:						      */
/*									      */
/*	cfidr_InitLeaf performs an init procedure of a leaf. The type of the  */
/*	leaf sets the block-type.					      */
/*									      */
/*	First, the name of the cell and of the instance is read from the db.  */
/*	The list of primitiv gates (and,or,exor, ...) is computed by	      */
/*	cfidr_TableLookup, and a blocklist_element is created for every	      */
/*	instance of the el_id_arr_p, the return of cfidr_TableLookup.	      */
/*									      */
/*  FORMAL PARAMETERS:							      */
/*									      */
/*	cfidrViewId: Id of a InstLeaf					      */
/*									      */
/*  IMPLICIT INPUT/OUTPUT:						      */
/*									      */
/*	none			 					      */
/*									      */
/*  FUNCTION VALUE:							      */
/*									      */
/*	none								      */
/*									      */
/*  CREATION DATE: 17.01.90						      */
/*									      */
/*  MODIFICATION HISTORY:						      */
/*									      */
/*	None								      */
/*									      */
/* ************************************************************************** */
void cfidr_InitLeaf(cfidrInstIdT cfidrInstId)
{
cfidrStringT cfidrString,cfidrInstName;
cfidrObjectIdT cfidrViewId, cfidrCellId;
char inst_name_s[stringsize_i],gate_name_s[stringsize_i];

cfidrViewId = cfidrInstGetDescriber(cfidrInstId, &cfidrError);
if (cfidrError != CFIDR_NO_ERROR)
	{
	cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
	error (cfidrString);
	}  /*if*/
cfidrCellId = cfidrViewGetOwner(cfidrViewId, &cfidrError);
if (cfidrError != CFIDR_NO_ERROR)
	{
	cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
	error (cfidrString);
	}  /*if*/
cfidrString = cfidrNamedObjectGetName(cfidrCellId, &cfidrError); /* cell name */
if (cfidrError != CFIDR_NO_ERROR)
	{
	cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
	error (cfidrString);
	}  /*if*/
strcpy(gate_name_s,cfidrString);
cfidrInstName = cfidrNamedObjectGetName(cfidrInstId, &cfidrError); /* instance name */
if (cfidrError != CFIDR_NO_ERROR)
	{
	cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
	error (cfidrString);
	}  /*if*/
strcpy(inst_name_s,cfidrInstName);

		/* extracting the gate type (gate list for macro cells) */
		/* with the name cfidrString by looking up in a table   */
cfidr_TableLookup(gate_name_s,inst_name_s);

}/*cfidr_InitLeaf*/


/* ************************************************************************** */
/*									      */
/*  FUNCTION NAME: cfidr_FindConnection				              */
/*									      */
/*  AUTOR: U. Glaeser							      */
/*									      */
/*  FUNCTIONAL DESCRIPTION:						      */
/*									      */
/*	for all blocks which outputs are not connected yet (for instance in   */
/*	cfidr_ConnectMacros, they were connected) the corresponding instance  */
/*	in the database is found by cfidr_FindCorrespondingObject and the     */
/*	elements are connected with cfidr_ConnectNets.			      */
/*									      */
/*  FORMAL PARAMETERS:							      */
/*									      */
/*	cfidrViewId: Id of the View					      */
/*									      */
/*  IMPLICIT INPUT/OUTPUT:						      */
/*									      */
/*	none								      */
/*									      */
/*  FUNCTION VALUE:							      */
/*									      */
/*	none								      */
/*									      */
/*  CREATION DATE: 23.01.90						      */
/*									      */
/*  MODIFICATION HISTORY:						      */
/*									      */
/*	none								      */
/*									      */
/* ************************************************************************** */
void cfidr_FindConnection(cfidrViewIdT cfidrViewId)
{
cfidrPortInstsIdT cfidrPortInstsId;
cfidrPortInstIdT cfidrPortInstId;
cfidrNetIdT cfidrNetId;
cfidrStringT cfidrString,cfidrInstName,cfidrNetName;
cfidrStringT cfidrPortInstName;
gate_C *gate_item_p;
cfidrInstsIdT cfidrInstsId;
cfidrInstIdT cfidrInstId;
string_C name_sC(stringsize_i);
stringliste_C in_namelist_C,out_namelist_C;

cfidrError = CFIDR_NO_ERROR;
cfidrInstsId = cfidrNetlistViewGetInsts(cfidrViewId, &cfidrError); /* Get List of      */
					    /* instances from current view    */
					    /* cfidr_GetAllViewLeafs	      */
if (cfidrError != CFIDR_NO_ERROR)
	{
	cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
	error (cfidrString);
	}  /*if*/

do
	{
	cfidrInstId = cfidrIterNextInst(cfidrInstsId, &cfidrError);
	if (cfidrError != CFIDR_NO_ERROR)
		{
		cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
		error (cfidrString);
		}  /*if*/

	if (cfidrObjectIsNull(cfidrInstId ,&cfidrError) == CFIDR_TRUE)
		return;
	if (cfidrError != CFIDR_NO_ERROR)
		{
		cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
		error (cfidrString);
		}  /*if*/
	cfidrInstName = cfidrNamedObjectGetName(cfidrInstId, &cfidrError); /* instance name */
	if (cfidrError != CFIDR_NO_ERROR)
		{
		cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
		error (cfidrString);
		}  /*if*/
	name_sC = string_C(cfidrInstName);
	gate_item_p = search(&name_sC,milef_gate_tree_root_p);

	cfidrPortInstsId = cfidrInstGetPortInsts(cfidrInstId,CFIDR_ITER_SCALARS,
					        &cfidrError);
	if (cfidrError != CFIDR_NO_ERROR)
		{
		cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
		error (cfidrString);
		}/*if*/
	do
		{ /* find the right ouput line by name */
		cfidrPortInstId = cfidrIterNextPortInst(cfidrPortInstsId,&cfidrError);
		if (cfidrError != CFIDR_NO_ERROR)
			{
			cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
			error (cfidrString);
			}/*if*/
		if (cfidrObjectIsNull(cfidrPortInstId,&cfidrError) == CFIDR_TRUE)
			break; /* not found */
		if (cfidrError != CFIDR_NO_ERROR)
			{
			cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
			error (cfidrString);
			}/*if*/
		cfidrPortInstName = cfidrPortInstGetDescriberName(cfidrPortInstId,&cfidrError);
		if (cfidrError != CFIDR_NO_ERROR)
			{
			cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
			error (cfidrString);
			}  /*if*/
		cfidrNetId = cfidrPortInstScalarGetNetScalar(cfidrPortInstId,&cfidrError);
		if (cfidrError != CFIDR_NO_ERROR)
			{
			cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
			error (cfidrString);
			}/*if*/
		cfidrNetName = cfidrNamedObjectGetName(cfidrNetId, &cfidrError); /* net name */
		if (cfidrError != CFIDR_NO_ERROR)
			{
			cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
			error (cfidrString);
			}  /*if*/
		switch (cfidrPortInstScalarGetDescriberDirection(cfidrPortInstId,&cfidrError))
			{
			case CFIDR_OUTPUT:
				out_namelist_C.add_element(cfidrNetName);
				break;
			case CFIDR_INPUT:
				in_namelist_C.add_element(cfidrNetName);
				break;
			default:
				error("cfidr_FindConnection: Unknown Port Direction");
			}//switch
			
		}while(cfidrObjectIsNull(cfidrPortInstId,&cfidrError) == CFIDR_FALSE);
	gate_item_p->connect_input(in_namelist_C);
	gate_item_p->connect_output(out_namelist_C);
	}while(cfidrObjectIsNull(cfidrInstId,&cfidrError) == CFIDR_FALSE);
}/*cfidr_FindConnection*/



/* ************************************************************************** */
/*									      */
/*  FUNCTION NAME: cfidr_CreateInputsOutputs				      */
/*									      */
/*  AUTHOR: U. Glaeser                                                        */
/*									      */
/*  FUNCTIONAL DESCRIPTION:						      */
/*									      */
/*	cfidr_CreateInputsOutputs creates the input and outputs of the	      */
/*	circuit.							      */
/*									      */
/*	All the ports of the main view are visited and the port direction     */
/*	determines input or output.					      */
/*									      */
/*									      */
/*  FORMAL PARAMETERS:							      */
/*									      */
/*	none								      */
/*									      */
/*  IMPLICIT INPUT/OUTPUT:						      */
/*									      */
/*	cfidrMainViewId							      */
/*									      */
/*  FUNCTION VALUE:							      */
/*									      */
/*	none								      */
/*									      */
/*  CREATION DATE: 25.01.90						      */
/*									      */
/*  MODIFICATION HISTORY:						      */
/*									      */
/*	none								      */
/*									      */
/* ************************************************************************** */
void cfidr_CreateInputsOutputs()
{
   cfidrPortsIdT cfidrPortsId;
   cfidrPortIdT cfidrPortId;
   cfidrPortDirectionT cfidrPortDirection;
   cfidrNetIdT cfidrNetId;
   cfidrStringT cfidrString;
   line_C *line_item_p;

   cfidrError = CFIDR_NO_ERROR;
   cfidrPortsId = cfidrNetlistViewGetPorts(cfidrMainViewId,CFIDR_ITER_SCALARS,
				&cfidrError);
   if (cfidrError != CFIDR_NO_ERROR)
	{
	cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
	error (cfidrString);
	}  /*if*/
   cfidrPortId = cfidrIterNextPort(cfidrPortsId, &cfidrError);
   if (cfidrError != CFIDR_NO_ERROR)
	{
	cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
	error (cfidrString);
	}  /*if*/

   do
	{ /* for each port in the view */
	  /* first create ouput list with type struct help_io for allocblock function */
	cfidrNetId = cfidrPortScalarGetNetScalar(cfidrPortId,&cfidrError); /* Get List of      */
					    /* nets from current view    */
					    /* cfidr_GetAllViewLines	      */
	if (cfidrError != CFIDR_NO_ERROR)
		{
		cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
		error (cfidrString);
		}  /*if*/

      cfidrString = cfidrNamedObjectGetName(cfidrNetId, &cfidrError);
      string_C name_sC(cfidrString);
      line_item_p = search(&name_sC,milef_line_tree_root_p);
      if (line_item_p == NULL)
		error("cfidr_CreateInputsOutputs: Line not found: " + name_sC);
      if (cfidrError != CFIDR_NO_ERROR)
	{
	error (cfidrString);
	}  /*if*/
	/*    determine input or output						      */
      cfidrPortDirection = cfidrPortScalarGetDirection(cfidrPortId, &cfidrError);
      switch(cfidrPortDirection)
      {
	 case CFIDR_INPUT:
		statistic_u.number_inputs_i++;
		line_item_p->set_primary_input();
		input_lines_C.fuege_ein(line_item_p);
		break;
	 case CFIDR_OUTPUT:
		statistic_u.number_outputs_i++;
		line_item_p->set_primary_output();
		output_lines_C.fuege_ein(line_item_p);
		break;
	 default: error("cfidr_CreateInputsOutputs unknown portdirection"); 
      }	 /*switch*/
      /* name the block */
      cfidrPortId = cfidrIterNextPort(cfidrPortsId, &cfidrError);
   }  while (cfidrObjectIsNull(cfidrPortId, &cfidrError) != CFIDR_TRUE);
if (cfidrError != CFIDR_NO_ERROR)
	{ /* belongs to cfidrObjectIsNull */
	cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
	error (cfidrString);
	}  /*if*/
}  /*cfidr_CreateInputsOutputs*/


/* ************************************************************************** */
/*                                                                            */
/*  FUNCTION NAME: cfidr_ConnectMacros                                        */
/*                                                                            */
/*  AUTHOR: U. Glaeser                                                        */
/*                                                                            */
/*  FUNCTIONAL DESCRIPTION:                                                   */
/*                                                                            */
/*      cfidr_ConnectMacros creates the internal connections of the macro     */
/*      elements in the circuit.                                              */
/*	In the first step, the macro paramaters (component macro_par) are set */
/*	The second step connects all the macros intern, and sets the input    */
/*	and the output names of the lines, which are connected extern	      */
/*	At the end of the internal connection names, INTERN is occured.	      */
/*	Different internal lines have the same name except a different number */
/*	before the INTERN.						      */
/*                                                                            */
/*  FORMAL PARAMETERS:                                                        */
/*                                                                            */
/*      none                                                                  */
/*                                                                            */
/*  IMPLICIT INPUT/OUTPUT:                                                    */
/*									      */
/*	none								      */
/*                                                                            */
/*  FUNCTION VALUE:                                                           */
/*                                                                            */
/*      void                                                                  */
/*                                                                            */
/*  CREATION DATE: 25.01.90                                                   */
/*                                                                            */
/*  MODIFICATION HISTORY:                                                     */
/*                                                                            */
/*      none                                                                  */
/*                                                                            */
/* ************************************************************************** */
void cfidr_ConnectMacros(cfidrViewIdT cfidrViewId)
{
gate_C *gate_item_p,*inverter_p,*and_p,*or_p,*and1_p,*and2_p,*and3_p,*and4_p,*nor_p,*or1_p,*or2_p;
line_C *input_line_p,*line1_p,*line2_p,*line3_p,*line4_p;
char gatter_name_s[stringsize_i],hilf_name_s[stringsize_i];
cfidrInstsIdT cfidrInstsId;
cfidrInstIdT cfidrInstId;
cfidrStringT cfidrString;
cfidrObjectIdT cfidrCellId;
cfidrViewIdT ViewId;
int el_id_i,i;
stringliste_C in_liste_C,out_liste_C;

cfidrError = CFIDR_NO_ERROR;

forall(gate_item_p,gates_C)
	{
	if (gate_item_p->gib_type() != UNDEFINED_GATE)
		continue;
	strcpy(gatter_name_s,(char *)gate_item_p->get_name());
	cfidrInstsId = cfidrNetlistViewGetInsts(cfidrViewId, &cfidrError); /* Get List of      */
					    /* instances from current view    */
					    /* cfidr_GetAllViewLeafs	      */
	if (cfidrError != CFIDR_NO_ERROR)
		{
		cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
		error (cfidrString);
		}  /*if*/

	do
		{
		cfidrInstId = cfidrIterNextInst(cfidrInstsId, &cfidrError);
		if (cfidrError != CFIDR_NO_ERROR)
			{
			cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
			error (cfidrString);
			}  /*if*/

		if (cfidrObjectIsNull(cfidrInstId ,&cfidrError) == CFIDR_TRUE)
			error("cfidr_ConnectMacros: Instance not found");
		if (cfidrError != CFIDR_NO_ERROR)
			{
			cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
			error (cfidrString);
			}  /*if*/
		cfidrString = cfidrNamedObjectGetName(cfidrInstId ,&cfidrError);
		if (cfidrError != CFIDR_NO_ERROR)
			{
			cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
			error (cfidrString);
			}  /*if*/
		}while(strcmp(gatter_name_s,cfidrString) != 0);
	ViewId = cfidrInstGetDescriber(cfidrInstId, &cfidrError);
	if (cfidrError != CFIDR_NO_ERROR)
		{
		cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
		error (cfidrString);
		}  /*if*/
	cfidrCellId = cfidrViewGetOwner(ViewId, &cfidrError);
	if (cfidrError != CFIDR_NO_ERROR)
		{
		cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
		error (cfidrString);
		}  /*if*/
	cfidrString = cfidrNamedObjectGetName(cfidrCellId, &cfidrError); /* cell name */
	if (cfidrError != CFIDR_NO_ERROR)
		{
		cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
		error (cfidrString);
		}  /*if*/
	for (i=0;i<baustein_size_i;++i)
		if (strcasecmp(cfidrString,(&cfidr_Tabelle[i])->baust_name) == 0)
			{
			el_id_i = (&cfidr_Tabelle[i])->codenummer;
			break;
			}

	switch (el_id_i)
	   {
	   case TO010LJ: /* create dummy input element */
		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%dummyinput");
		input_line_p = new line_C(hilf_name_s);
		input_line_p->set_primary_input();
		lines_C.fuege_ein(input_line_p);
		input_lines_C.fuege_ein(input_line_p);

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%inverter");
		inverter_p = new gate_C(hilf_name_s);
		inverter_p->belege_gate(INVERTER,1,1);
		gates_C.fuege_ein(inverter_p);

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%and");
		and_p = new gate_C(hilf_name_s);
		and_p->belege_gate(AND,2,1);
		gates_C.fuege_ein(and_p);

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%i1");
		line1_p = new line_C(hilf_name_s);
		lines_C.fuege_ein(line1_p);

		in_liste_C.add_element(input_line_p->get_name());
		out_liste_C.add_element(line1_p->get_name());
		inverter_p->connect_input(in_liste_C);
		inverter_p->connect_output(out_liste_C);

		in_liste_C.add_element(input_line_p->get_name());
		in_liste_C.add_element(line1_p->get_name());
		and_p->connect_input(in_liste_C);
		line2_p = (*gate_item_p->gib_outputs())[1];
		and_p->output_p->liste_a[0] = line2_p;
		and_p->output_p->number_elements_i = 1;
		line2_p->vorgaenger_p->liste_a[0] = and_p;
		line2_p->vorgaenger_p->number_elements_i = 1;

		or_p = gate_item_p;
		or_p->gate_e = OR;
		delete or_p->input_p;
		or_p->input_p = new line_liste_C(2,INPUT_PORT);
		or_p->no_input_i = 2;
		line2_p = or_p->output_p->liste_a[0];
		delete or_p->output_p;
		or_p->output_p = new line_liste_C(1,OUTPUT_PORT);
		or_p->no_output_i = 1;
		or_p->output_p->liste_a[0] = line2_p;
		or_p->output_p->number_elements_i = 1;
		line2_p->vorgaenger_p->liste_a[0] = or_p;
		line2_p->vorgaenger_p->number_elements_i = 1;
		in_liste_C.add_element(input_line_p->get_name());
		in_liste_C.add_element(line1_p->get_name());
		or_p->connect_input(in_liste_C);
		break;
	   case AO220LJ: /* inputs: A,B,C,D			output: Y */
			 /* gate list: AND2, AND2, OR2			  */
			 /* function: Y = (A * B) + (C * D)		  */
		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%and1");
		and1_p = new gate_C(hilf_name_s);
		and1_p->belege_gate(AND,2,1);
		gates_C.fuege_ein(and1_p);

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%and2");
		and2_p = new gate_C(hilf_name_s);
		and2_p->belege_gate(AND,2,1);
		gates_C.fuege_ein(and2_p);

		or_p = gate_item_p;
		or_p->gate_e = OR;

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%i1");
		line1_p = new line_C(hilf_name_s);
		lines_C.fuege_ein(line1_p);

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%i2");
		line2_p = new line_C(hilf_name_s);
		lines_C.fuege_ein(line2_p);

		gate_item_p->input_p->liste_a[0]->nachfolger_p->change_connection(gatter_name_s,and1_p);
		gate_item_p->input_p->liste_a[1]->nachfolger_p->change_connection(gatter_name_s,and1_p);
		gate_item_p->input_p->liste_a[2]->nachfolger_p->change_connection(gatter_name_s,and2_p);
		gate_item_p->input_p->liste_a[3]->nachfolger_p->change_connection(gatter_name_s,and2_p);
		and1_p->input_p->liste_a[0] = gate_item_p->input_p->liste_a[0];
		and1_p->input_p->liste_a[1] = gate_item_p->input_p->liste_a[1];
		and1_p->input_p->number_elements_i = 2;
		and2_p->input_p->liste_a[0] = gate_item_p->input_p->liste_a[2];
		and2_p->input_p->liste_a[1] = gate_item_p->input_p->liste_a[3];
		and2_p->input_p->number_elements_i = 2;
		out_liste_C.add_element(line1_p->get_name());
		and1_p->connect_output(out_liste_C);
		out_liste_C.add_element(line2_p->get_name());
		and2_p->connect_output(out_liste_C);
		delete or_p->input_p;
		or_p->input_p = new line_liste_C(2,INPUT_PORT);
		or_p->no_input_i = 2;
		in_liste_C.add_element(line1_p->get_name());
		in_liste_C.add_element(line2_p->get_name());
		or_p->connect_input(in_liste_C);
		break;
	   case AO221LJ: /* inputs: A,B,C,D                     output: Y */
			 /* gate list: AND2, AND2, NOR2                   */
			 /* function: Y = (A * B) NOR (C * D)             */
		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%and1");
		and1_p = new gate_C(hilf_name_s);
		and1_p->belege_gate(AND,2,1);
		gates_C.fuege_ein(and1_p);

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%and2");
		and2_p = new gate_C(hilf_name_s);
		and2_p->belege_gate(AND,2,1);
		gates_C.fuege_ein(and2_p);

		or_p = gate_item_p;
		or_p->gate_e = NOR;

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%i1");
		line1_p = new line_C(hilf_name_s);
		lines_C.fuege_ein(line1_p);

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%i2");
		line2_p = new line_C(hilf_name_s);
		lines_C.fuege_ein(line2_p);

		gate_item_p->input_p->liste_a[0]->nachfolger_p->change_connection(gatter_name_s,and1_p);
		gate_item_p->input_p->liste_a[1]->nachfolger_p->change_connection(gatter_name_s,and1_p);
		gate_item_p->input_p->liste_a[2]->nachfolger_p->change_connection(gatter_name_s,and2_p);
		gate_item_p->input_p->liste_a[3]->nachfolger_p->change_connection(gatter_name_s,and2_p);
		and1_p->input_p->liste_a[0] = gate_item_p->input_p->liste_a[0];
		and1_p->input_p->liste_a[1] = gate_item_p->input_p->liste_a[1];
		and1_p->input_p->number_elements_i = 2;
		and2_p->input_p->liste_a[0] = gate_item_p->input_p->liste_a[2];
		and2_p->input_p->liste_a[1] = gate_item_p->input_p->liste_a[3];
		and2_p->input_p->number_elements_i = 2;
		out_liste_C.add_element(line1_p->get_name());
		and1_p->connect_output(out_liste_C);
		out_liste_C.add_element(line2_p->get_name());
		and2_p->connect_output(out_liste_C);
		delete or_p->input_p;
		or_p->input_p = new line_liste_C(2,INPUT_PORT);
		or_p->no_input_i = 2;
		in_liste_C.add_element(line1_p->get_name());
		in_liste_C.add_element(line2_p->get_name());
		or_p->connect_input(in_liste_C);
		break;
	   case AO241LJ: /* inputs: A1,A2,B1,B2,C1,C2,D1,D2     output: Y */
			 /* gate list: AND2, AND2, AND2, AND2, NOR4       */
			 /* function: Y = NOR(A1 * A2,B1 * B2,C1 * C2,D1 * D2)  */
		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%and1");
		and1_p = new gate_C(hilf_name_s);
		and1_p->belege_gate(AND,2,1);
		gates_C.fuege_ein(and1_p);

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%and2");
		and2_p = new gate_C(hilf_name_s);
		and2_p->belege_gate(AND,2,1);
		gates_C.fuege_ein(and2_p);

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%and3");
		and3_p = new gate_C(hilf_name_s);
		and3_p->belege_gate(AND,2,1);
		gates_C.fuege_ein(and3_p);

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%and4");
		and4_p = new gate_C(hilf_name_s);
		and4_p->belege_gate(AND,2,1);
		gates_C.fuege_ein(and4_p);

		or_p = gate_item_p;
		or_p->gate_e = NOR;

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%i1");
		line1_p = new line_C(hilf_name_s);
		lines_C.fuege_ein(line1_p);

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%i2");
		line2_p = new line_C(hilf_name_s);
		lines_C.fuege_ein(line2_p);

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%i3");
		line3_p = new line_C(hilf_name_s);
		lines_C.fuege_ein(line3_p);

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%i4");
		line4_p = new line_C(hilf_name_s);
		lines_C.fuege_ein(line4_p);

		gate_item_p->input_p->liste_a[0]->nachfolger_p->change_connection(gatter_name_s,and1_p);
		gate_item_p->input_p->liste_a[1]->nachfolger_p->change_connection(gatter_name_s,and1_p);
		gate_item_p->input_p->liste_a[2]->nachfolger_p->change_connection(gatter_name_s,and2_p);
		gate_item_p->input_p->liste_a[3]->nachfolger_p->change_connection(gatter_name_s,and2_p);
		gate_item_p->input_p->liste_a[4]->nachfolger_p->change_connection(gatter_name_s,and3_p);
		gate_item_p->input_p->liste_a[5]->nachfolger_p->change_connection(gatter_name_s,and3_p);
		gate_item_p->input_p->liste_a[6]->nachfolger_p->change_connection(gatter_name_s,and4_p);
		gate_item_p->input_p->liste_a[7]->nachfolger_p->change_connection(gatter_name_s,and4_p);
		and1_p->input_p->liste_a[0] = gate_item_p->input_p->liste_a[0];
		and1_p->input_p->liste_a[1] = gate_item_p->input_p->liste_a[1];
		and1_p->input_p->number_elements_i = 2;
		and2_p->input_p->liste_a[0] = gate_item_p->input_p->liste_a[2];
		and2_p->input_p->liste_a[1] = gate_item_p->input_p->liste_a[3];
		and2_p->input_p->number_elements_i = 2;
		and3_p->input_p->liste_a[0] = gate_item_p->input_p->liste_a[4];
		and3_p->input_p->liste_a[1] = gate_item_p->input_p->liste_a[5];
		and3_p->input_p->number_elements_i = 2;
		and4_p->input_p->liste_a[0] = gate_item_p->input_p->liste_a[6];
		and4_p->input_p->liste_a[1] = gate_item_p->input_p->liste_a[7];
		and4_p->input_p->number_elements_i = 2;
		out_liste_C.add_element(line1_p->get_name());
		and1_p->connect_output(out_liste_C);
		out_liste_C.add_element(line2_p->get_name());
		and2_p->connect_output(out_liste_C);
		out_liste_C.add_element(line3_p->get_name());
		and3_p->connect_output(out_liste_C);
		out_liste_C.add_element(line4_p->get_name());
		and4_p->connect_output(out_liste_C);
		delete or_p->input_p;
		or_p->input_p = new line_liste_C(4,INPUT_PORT);
		or_p->no_input_i = 4;
		in_liste_C.add_element(line1_p->get_name());
		in_liste_C.add_element(line2_p->get_name());
		in_liste_C.add_element(line3_p->get_name());
		in_liste_C.add_element(line4_p->get_name());
		or_p->connect_input(in_liste_C);
		break;
	   case AO421LJ: /* inputs: A1,A2,A3,A4,B1,B2,B3,B4     output: Y */
			 /* gate list: AND4, AND4, NOR2			  */
			 /* function: Y = AND(A1,A2,A3,A4) NOR AND(B1,B2,B3,B4) */
		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%and1");
		and1_p = new gate_C(hilf_name_s);
		and1_p->belege_gate(AND,4,1);
		gates_C.fuege_ein(and1_p);

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%and2");
		and2_p = new gate_C(hilf_name_s);
		and2_p->belege_gate(AND,4,1);
		gates_C.fuege_ein(and2_p);

		or_p = gate_item_p;
		or_p->gate_e = NOR;

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%i1");
		line1_p = new line_C(hilf_name_s);
		lines_C.fuege_ein(line1_p);

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%i2");
		line2_p = new line_C(hilf_name_s);
		lines_C.fuege_ein(line2_p);

		gate_item_p->input_p->liste_a[0]->nachfolger_p->change_connection(gatter_name_s,and1_p);
		gate_item_p->input_p->liste_a[1]->nachfolger_p->change_connection(gatter_name_s,and1_p);
		gate_item_p->input_p->liste_a[2]->nachfolger_p->change_connection(gatter_name_s,and1_p);
		gate_item_p->input_p->liste_a[3]->nachfolger_p->change_connection(gatter_name_s,and1_p);
		gate_item_p->input_p->liste_a[4]->nachfolger_p->change_connection(gatter_name_s,and2_p);
		gate_item_p->input_p->liste_a[5]->nachfolger_p->change_connection(gatter_name_s,and2_p);
		gate_item_p->input_p->liste_a[6]->nachfolger_p->change_connection(gatter_name_s,and2_p);
		gate_item_p->input_p->liste_a[7]->nachfolger_p->change_connection(gatter_name_s,and2_p);
		and1_p->input_p->liste_a[0] = gate_item_p->input_p->liste_a[0];
		and1_p->input_p->liste_a[1] = gate_item_p->input_p->liste_a[1];
		and1_p->input_p->liste_a[2] = gate_item_p->input_p->liste_a[2];
		and1_p->input_p->liste_a[3] = gate_item_p->input_p->liste_a[3];
		and1_p->input_p->number_elements_i = 4;
		and2_p->input_p->liste_a[0] = gate_item_p->input_p->liste_a[4];
		and2_p->input_p->liste_a[1] = gate_item_p->input_p->liste_a[5];
		and2_p->input_p->liste_a[2] = gate_item_p->input_p->liste_a[6];
		and2_p->input_p->liste_a[3] = gate_item_p->input_p->liste_a[7];
		and2_p->input_p->number_elements_i = 4;
		out_liste_C.add_element(line1_p->get_name());
		and1_p->connect_output(out_liste_C);
		out_liste_C.add_element(line2_p->get_name());
		and2_p->connect_output(out_liste_C);
		delete or_p->input_p;
		or_p->input_p = new line_liste_C(2,INPUT_PORT);
		or_p->no_input_i = 2;
		in_liste_C.add_element(line1_p->get_name());
		in_liste_C.add_element(line2_p->get_name());
		or_p->connect_input(in_liste_C);
		break;
	   case BF001LJ: /* inputs: B1,B2,A1			output: Y */
			 /* gate list: AND2, NOR2			  */
			 /* function: Y = (B1 * B2) NOR A1		  */
		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%and1");
		and1_p = new gate_C(hilf_name_s);
		and1_p->belege_gate(AND,2,1);
		gates_C.fuege_ein(and1_p);

		or_p = gate_item_p;
		or_p->gate_e = NOR;

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%i1");
		line1_p = new line_C(hilf_name_s);
		lines_C.fuege_ein(line1_p);

		gate_item_p->input_p->liste_a[0]->nachfolger_p->change_connection(gatter_name_s,and1_p);
		gate_item_p->input_p->liste_a[1]->nachfolger_p->change_connection(gatter_name_s,and1_p);
		and1_p->input_p->liste_a[0] = gate_item_p->input_p->liste_a[0];
		and1_p->input_p->liste_a[1] = gate_item_p->input_p->liste_a[1];
		and1_p->input_p->number_elements_i = 2;
		line2_p = gate_item_p->input_p->liste_a[2];
		delete or_p->input_p;
		or_p->input_p = new line_liste_C(2,INPUT_PORT);
		or_p->no_input_i = 2;
		line2_p->nachfolger_p->change_connection(gatter_name_s,or_p);
		or_p->input_p->liste_a[0] = line2_p;
		out_liste_C.add_element(line1_p->get_name());
		and1_p->connect_output(out_liste_C);
		or_p->input_p->liste_a[1] = line1_p;
		line1_p->nachfolger_p->liste_a[0] = or_p;
		line1_p->nachfolger_p->number_elements_i = 1;
		or_p->input_p->number_elements_i = 2;
		break;
	   case BF006LJ: /* inputs: B1,B2,A1,A2                 output: Y */
			 /* gate list: AND2, NOR3                         */
			 /* function: Y = NOR(B1 * B2,A1,A2)              */
		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%and1");
		and1_p = new gate_C(hilf_name_s);
		and1_p->belege_gate(AND,2,1);
		gates_C.fuege_ein(and1_p);

		or_p = gate_item_p;
		or_p->gate_e = NOR;

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%i1");
		line1_p = new line_C(hilf_name_s);
		lines_C.fuege_ein(line1_p);

		gate_item_p->input_p->liste_a[0]->nachfolger_p->change_connection(gatter_name_s,and1_p);
		gate_item_p->input_p->liste_a[1]->nachfolger_p->change_connection(gatter_name_s,and1_p);
		and1_p->input_p->liste_a[0] = gate_item_p->input_p->liste_a[0];
		and1_p->input_p->liste_a[1] = gate_item_p->input_p->liste_a[1];
		and1_p->input_p->number_elements_i = 2;
		line2_p = gate_item_p->input_p->liste_a[2];
		line3_p = gate_item_p->input_p->liste_a[3];
		delete or_p->input_p;
		or_p->input_p = new line_liste_C(3,INPUT_PORT);
		or_p->no_input_i = 3;
		or_p->input_p->liste_a[0] = line1_p;
		line1_p->nachfolger_p->liste_a[0] = or_p;
		line1_p->nachfolger_p->number_elements_i = 1;
		line2_p->nachfolger_p->change_connection(gatter_name_s,or_p);
		line3_p->nachfolger_p->change_connection(gatter_name_s,or_p);
		or_p->input_p->liste_a[1] = line2_p;
		or_p->input_p->liste_a[2] = line3_p;
		out_liste_C.add_element(line1_p->get_name());
		and1_p->connect_output(out_liste_C);
		or_p->input_p->number_elements_i = 3;
		break;
	   case BF011LJ: /* inputs: A1,A2,B1,B2,C1,C2           output: Y */
			 /* gate list: AND2, AND2, AND2, NOR3             */
			 /* function: Y = NOR(A1 * A2,B1 * B2,C1 * C2)    */
		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%and1");
		and1_p = new gate_C(hilf_name_s);
		and1_p->belege_gate(AND,2,1);
		gates_C.fuege_ein(and1_p);

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%and2");
		and2_p = new gate_C(hilf_name_s);
		and2_p->belege_gate(AND,2,1);
		gates_C.fuege_ein(and2_p);

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%and3");
		and3_p = new gate_C(hilf_name_s);
		and3_p->belege_gate(AND,2,1);
		gates_C.fuege_ein(and3_p);

		or_p = gate_item_p;
		or_p->gate_e = NOR;

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%i1");
		line1_p = new line_C(hilf_name_s);
		lines_C.fuege_ein(line1_p);

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%i2");
		line2_p = new line_C(hilf_name_s);
		lines_C.fuege_ein(line2_p);

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%i3");
		line3_p = new line_C(hilf_name_s);
		lines_C.fuege_ein(line3_p);

		gate_item_p->input_p->liste_a[0]->nachfolger_p->change_connection(gatter_name_s,and1_p);
		gate_item_p->input_p->liste_a[1]->nachfolger_p->change_connection(gatter_name_s,and1_p);
		gate_item_p->input_p->liste_a[2]->nachfolger_p->change_connection(gatter_name_s,and2_p);
		gate_item_p->input_p->liste_a[3]->nachfolger_p->change_connection(gatter_name_s,and2_p);
		gate_item_p->input_p->liste_a[4]->nachfolger_p->change_connection(gatter_name_s,and3_p);
		gate_item_p->input_p->liste_a[5]->nachfolger_p->change_connection(gatter_name_s,and3_p);
		and1_p->input_p->liste_a[0] = gate_item_p->input_p->liste_a[0];
		and1_p->input_p->liste_a[1] = gate_item_p->input_p->liste_a[1];
		and1_p->input_p->number_elements_i = 2;
		and2_p->input_p->liste_a[0] = gate_item_p->input_p->liste_a[2];
		and2_p->input_p->liste_a[1] = gate_item_p->input_p->liste_a[3];
		and2_p->input_p->number_elements_i = 2;
		and3_p->input_p->liste_a[0] = gate_item_p->input_p->liste_a[4];
		and3_p->input_p->liste_a[1] = gate_item_p->input_p->liste_a[5];
		and3_p->input_p->number_elements_i = 2;
		out_liste_C.add_element(line1_p->get_name());
		and1_p->connect_output(out_liste_C);
		out_liste_C.add_element(line2_p->get_name());
		and2_p->connect_output(out_liste_C);
		out_liste_C.add_element(line3_p->get_name());
		and3_p->connect_output(out_liste_C);
		delete or_p->input_p;
		or_p->input_p = new line_liste_C(3,INPUT_PORT);
		or_p->no_input_i = 3;
		in_liste_C.add_element(line1_p->get_name());
		in_liste_C.add_element(line2_p->get_name());
		in_liste_C.add_element(line3_p->get_name());
		or_p->connect_input(in_liste_C);
		break;
	   case BF022LJ: /* inputs: C1,C2,B1,B2,A1,A2           output: Y */
			 /* gate list: OR2, AND3, AND2, NOR2              */
			 /* function: Y = AND(C1 + C2,B1,B2) NOR (A1 * A2) */
		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%or1");
		or_p = new gate_C(hilf_name_s);
		or_p->belege_gate(OR,2,1);
		gates_C.fuege_ein(or_p);

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%and1");
		and1_p = new gate_C(hilf_name_s);
		and1_p->belege_gate(AND,3,1);
		gates_C.fuege_ein(and1_p);

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%and2");
		and2_p = new gate_C(hilf_name_s);
		and2_p->belege_gate(AND,2,1);
		gates_C.fuege_ein(and2_p);

		nor_p = gate_item_p;
		nor_p->gate_e = NOR;

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%i1");
		line1_p = new line_C(hilf_name_s);
		lines_C.fuege_ein(line1_p);

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%i2");
		line2_p = new line_C(hilf_name_s);
		lines_C.fuege_ein(line2_p);

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%i3");
		line3_p = new line_C(hilf_name_s);
		lines_C.fuege_ein(line3_p);

		gate_item_p->input_p->liste_a[0]->nachfolger_p->change_connection(gatter_name_s,or_p);
		gate_item_p->input_p->liste_a[1]->nachfolger_p->change_connection(gatter_name_s,or_p);
		gate_item_p->input_p->liste_a[2]->nachfolger_p->change_connection(gatter_name_s,and1_p);
		gate_item_p->input_p->liste_a[3]->nachfolger_p->change_connection(gatter_name_s,and1_p);
		gate_item_p->input_p->liste_a[4]->nachfolger_p->change_connection(gatter_name_s,and2_p);
		gate_item_p->input_p->liste_a[5]->nachfolger_p->change_connection(gatter_name_s,and2_p);
		or_p->input_p->liste_a[0] = gate_item_p->input_p->liste_a[0];
		or_p->input_p->liste_a[1] = gate_item_p->input_p->liste_a[1];
		or_p->input_p->number_elements_i = 2;
		and1_p->input_p->liste_a[0] = gate_item_p->input_p->liste_a[2];
		and1_p->input_p->liste_a[1] = gate_item_p->input_p->liste_a[3];
		and1_p->input_p->liste_a[2] = line1_p;
		and1_p->input_p->number_elements_i = 3;
		line1_p->nachfolger_p->liste_a[0] = and1_p;
		line1_p->nachfolger_p->number_elements_i = 1;
		and2_p->input_p->liste_a[0] = gate_item_p->input_p->liste_a[4];
		and2_p->input_p->liste_a[1] = gate_item_p->input_p->liste_a[5];
		and2_p->input_p->number_elements_i = 2;
		out_liste_C.add_element(line1_p->get_name());
		or_p->connect_output(out_liste_C);
		out_liste_C.add_element(line2_p->get_name());
		and1_p->connect_output(out_liste_C);
		out_liste_C.add_element(line3_p->get_name());
		and2_p->connect_output(out_liste_C);
		delete nor_p->input_p;
		nor_p->input_p = new line_liste_C(2,INPUT_PORT);
		nor_p->no_input_i = 2;
		in_liste_C.add_element(line2_p->get_name());
		in_liste_C.add_element(line3_p->get_name());
		nor_p->connect_input(in_liste_C);
		break;
	   case BF051LJ: /* inputs: B1,B2,A1			output: Y */
			 /* gate list: OR2, NAND2			  */
			 /* function: Y = (B1 + B2) NAND A1		  */
		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%or1");
		or_p = new gate_C(hilf_name_s);
		or_p->belege_gate(OR,2,1);
		gates_C.fuege_ein(or_p);

		and_p = gate_item_p;
		and_p->gate_e = NAND;

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%i1");
		line1_p = new line_C(hilf_name_s);
		lines_C.fuege_ein(line1_p);

		gate_item_p->input_p->liste_a[0]->nachfolger_p->change_connection(gatter_name_s,or_p);
		gate_item_p->input_p->liste_a[1]->nachfolger_p->change_connection(gatter_name_s,or_p);
		or_p->input_p->liste_a[0] = gate_item_p->input_p->liste_a[0];
		or_p->input_p->liste_a[1] = gate_item_p->input_p->liste_a[1];
		or_p->input_p->number_elements_i = 2;
		line2_p = gate_item_p->input_p->liste_a[2];
		delete and_p->input_p;
		and_p->input_p = new line_liste_C(2,INPUT_PORT);
		and_p->no_input_i = 2;
		line2_p->nachfolger_p->change_connection(gatter_name_s,and_p);
		and_p->input_p->liste_a[1] = line2_p;
		out_liste_C.add_element(line1_p->get_name());
		or_p->connect_output(out_liste_C);
		and_p->input_p->liste_a[0] = line1_p;
		line1_p->nachfolger_p->liste_a[0] = and_p;
		line1_p->nachfolger_p->number_elements_i = 1;
		and_p->input_p->number_elements_i = 2;
		break;
	   case BF053LJ: /* inputs: A1,A2,B1,B2                 output: Y */
			 /* gate list: OR2, OR2, NAND2                    */
			 /* function: Y = (A1 + A2) NAND (B1 + B2)        */
		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%or1");
		or1_p = new gate_C(hilf_name_s);
		or1_p->belege_gate(AND,2,1);
		gates_C.fuege_ein(or1_p);

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%or2");
		or2_p = new gate_C(hilf_name_s);
		or2_p->belege_gate(AND,2,1);
		gates_C.fuege_ein(or2_p);

		and_p = gate_item_p;
		and_p->gate_e = NAND;

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%i1");
		line1_p = new line_C(hilf_name_s);
		lines_C.fuege_ein(line1_p);

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%i2");
		line2_p = new line_C(hilf_name_s);
		lines_C.fuege_ein(line2_p);

		gate_item_p->input_p->liste_a[0]->nachfolger_p->change_connection(gatter_name_s,or1_p);
		gate_item_p->input_p->liste_a[1]->nachfolger_p->change_connection(gatter_name_s,or1_p);
		gate_item_p->input_p->liste_a[2]->nachfolger_p->change_connection(gatter_name_s,or2_p);
		gate_item_p->input_p->liste_a[3]->nachfolger_p->change_connection(gatter_name_s,or2_p);
		or1_p->input_p->liste_a[0] = gate_item_p->input_p->liste_a[0];
		or1_p->input_p->liste_a[1] = gate_item_p->input_p->liste_a[1];
		or1_p->input_p->number_elements_i = 2;
		or2_p->input_p->liste_a[0] = gate_item_p->input_p->liste_a[2];
		or2_p->input_p->liste_a[1] = gate_item_p->input_p->liste_a[3];
		or2_p->input_p->number_elements_i = 2;
		out_liste_C.add_element(line1_p->get_name());
		or1_p->connect_output(out_liste_C);
		out_liste_C.add_element(line2_p->get_name());
		or2_p->connect_output(out_liste_C);
		delete and_p->input_p;
		and_p->input_p = new line_liste_C(2,INPUT_PORT);
		and_p->no_input_i = 2;
		in_liste_C.add_element(line1_p->get_name());
		in_liste_C.add_element(line2_p->get_name());
		and_p->connect_input(in_liste_C);
		break;
	   case BF056LJ: /* inputs: B1,B2,A1,A2                 output: Y */
			 /* gate list: OR2, NAND3			  */
			 /* function: Y = NAND(B1 + B2,A1,A2)		  */
		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%and1");
		or_p = new gate_C(hilf_name_s);
		or_p->belege_gate(OR,2,1);
		gates_C.fuege_ein(or_p);

		and_p = gate_item_p;
		and_p->gate_e = NAND;

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%i1");
		line1_p = new line_C(hilf_name_s);
		lines_C.fuege_ein(line1_p);

		gate_item_p->input_p->liste_a[0]->nachfolger_p->change_connection(gatter_name_s,or_p);
		gate_item_p->input_p->liste_a[1]->nachfolger_p->change_connection(gatter_name_s,or_p);
		or_p->input_p->liste_a[0] = gate_item_p->input_p->liste_a[0];
		or_p->input_p->liste_a[1] = gate_item_p->input_p->liste_a[1];
		or_p->input_p->number_elements_i = 2;
		line2_p = gate_item_p->input_p->liste_a[2];
		line3_p = gate_item_p->input_p->liste_a[3];
		delete and_p->input_p;
		and_p->input_p = new line_liste_C(3,INPUT_PORT);
		and_p->no_input_i = 3;
		line2_p->nachfolger_p->change_connection(gatter_name_s,and_p);
		line3_p->nachfolger_p->change_connection(gatter_name_s,and_p);
		and_p->input_p->liste_a[0] = line2_p;
		and_p->input_p->liste_a[1] = line3_p;
		out_liste_C.add_element(line1_p->get_name());
		or_p->connect_output(out_liste_C);
		and_p->input_p->liste_a[2] = line1_p;
		line1_p->nachfolger_p->liste_a[0] = and_p;
		line1_p->nachfolger_p->number_elements_i = 1;
		and_p->input_p->number_elements_i = 3;
		break;
	   case NA311LJ: /* inputs: A,B,C			output: Y */
			 /* gate list: NOT, NAND3                         */
			 /* function: Y = NAND(NOT A,B,C)		  */
		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%and1");
		or_p = new gate_C(hilf_name_s);
		or_p->belege_gate(INVERTER,1,1);
		gates_C.fuege_ein(or_p);

		and_p = gate_item_p;
		and_p->gate_e = NAND;

		strcpy(hilf_name_s,gatter_name_s);
		strcat(hilf_name_s,"%i1");
		line1_p = new line_C(hilf_name_s);
		lines_C.fuege_ein(line1_p);

		gate_item_p->input_p->liste_a[0]->nachfolger_p->change_connection(gatter_name_s,or_p);
		or_p->input_p->liste_a[0] = gate_item_p->input_p->liste_a[0];
		or_p->input_p->number_elements_i = 1;
		and_p->input_p->liste_a[0] = line1_p;
		line1_p->nachfolger_p->liste_a[0] = and_p;
		line1_p->nachfolger_p->number_elements_i = 1;
		out_liste_C.add_element(line1_p->get_name());
		or_p->connect_output(out_liste_C);
		break;
	   default: /* error state macro not found */
		cout << cfidrString << "\n";
		error("Unknown Element type in cfidr_ConnectMacros");
	   }/*switch*/
	}/*forall*/
}/* cfidr_ConnectMacros */

void cfidrCreatePatternBit(cfidrPortIdT cfidrPortId,char *time_s,char *state_s,cfidrErrorT *cfidrError_p)
   {
   cfidrPropIdT prop_id;
   cfidrStringT cfidrString;

   prop_id = cfidrObjectCreateProp(cfidrPortId,time_s,cfidrError_p);
   if (*cfidrError_p != CFIDR_NO_ERROR)
	 {
	 cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
	 error (cfidrString);
	 }  /*if*/
   cfidrPropSetStringValue(prop_id,state_s,cfidrError_p);
   }//cfidrCreatePatternBit

/* ************************************************************************** */
/*                                                                            */
/*  FUNCTION NAME: cfidr_PrintTestpattern				      */
/*                                                                            */
/*  AUTHOR: U. Glaeser                                                        */
/*                                                                            */
/*  FUNCTIONAL DESCRIPTION:                                                   */
/*                                                                            */
//	Diese Funktion gibt die Testpattern an die Ports aus, nach der
//	Schnittstellendefinition von Ansgar und Susanne
/*                                                                            */
/*  FORMAL PARAMETERS:                                                        */
/*                                                                            */
/*      none                                                                  */
/*                                                                            */
/*  IMPLICIT INPUT/OUTPUT:                                                    */
/*									      */
/*	none								      */
/*                                                                            */
/*  FUNCTION VALUE:                                                           */
/*                                                                            */
/*      void                                                                  */
/*                                                                            */
/*  CREATION DATE: 07.02.92                                                   */
/*                                                                            */
/*  MODIFICATION HISTORY:                                                     */
/*                                                                            */
/*      none                                                                  */
/*                                                                            */
/* ************************************************************************** */
void cfidr_PrintTestpattern(int time_i)
{
   cfidrPortsIdT cfidrPortsId;
   cfidrPortIdT cfidrPortId;
   cfidrPortDirectionT cfidrPortDirection;
   cfidrNetIdT cfidrNetId;
   cfidrStringT cfidrString;
   line_C *line_item_p;
   boolean gefunden_b;
   char time_s[stringsize_i],state_s[2];

   sprintf(time_s,"%i",time_i);
   cfidrError = CFIDR_NO_ERROR;
   cfidrPortsId = cfidrNetlistViewGetPorts(cfidrMainViewId,CFIDR_ITER_SCALARS,
				&cfidrError);
   if (cfidrError != CFIDR_NO_ERROR)
	{
	cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
	error (cfidrString);
	}  /*if*/
   cfidrPortId = cfidrIterNextPort(cfidrPortsId, &cfidrError);
   if (cfidrError != CFIDR_NO_ERROR)
	{
	cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
	error (cfidrString);
	}  /*if*/

   do
	{ /* for each port in the view */
	  /* first create ouput list with type struct help_io for allocblock function */
	cfidrNetId = cfidrPortScalarGetNetScalar(cfidrPortId,&cfidrError); /* Get List of      */
					    /* nets from current view    */
					    /* cfidr_GetAllViewLines	      */
	if (cfidrError != CFIDR_NO_ERROR)
		{
		cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
		error (cfidrString);
		}  /*if*/

      cfidrString = cfidrNamedObjectGetName(cfidrNetId, &cfidrError);
      if (cfidrError != CFIDR_NO_ERROR)
	{
	error (cfidrString);
	}  /*if*/
	/*    determine input or output						      */
      string_C name_sC(cfidrString);
      cfidrPortDirection = cfidrPortScalarGetDirection(cfidrPortId, &cfidrError);
      if (cfidrError != CFIDR_NO_ERROR)
	{
	error (cfidrString);
	}  /*if*/
      gefunden_b = FALSE;
      switch(cfidrPortDirection)
      {
	 case CFIDR_INPUT:
		if (path_delay_test_b == TRUE)
		  {
		  for (int i=0;i<3;i++)
		   {
		   forall(line_item_p,input_lines_C)
			if (line_item_p->get_name() == name_sC)
				{
				gefunden_b = TRUE;
				if ((line_item_p->get_good_value() == VAL_0) || (line_item_p->get_good_value() == VAL_X))
					strcpy(state_s,"0");
				else if (line_item_p->get_good_value() == VAL_D_BAR)
				   if (i == 1)
					strcpy(state_s,"1");
				   else
					strcpy(state_s,"0");
				else if (line_item_p->get_good_value() == VAL_1)
					strcpy(state_s,"1");
				else if (line_item_p->get_good_value() == VAL_D)
				   if (i == 1)
					strcpy(state_s,"0");
				   else
					strcpy(state_s,"1");
				else
					error("cfidr_PrintTestpattern: Unknown Value in Line Output");
				sprintf(time_s,"%i",3*time_i+i);
				cfidrCreatePatternBit(cfidrPortId,time_s,state_s,&cfidrError);
				if (cfidrError != CFIDR_NO_ERROR)
					{
					error (cfidrString);
					}/*if*/
				}//if
		   }//for
		  }//if
		else
		   {
		   forall(line_item_p,input_lines_C)
			if (line_item_p->get_name() == name_sC)
				{
				if (gefunden_b == TRUE)
					error("cfidr_PrintTestpattern: 2 Lines with the same name");
				else
					gefunden_b = TRUE;
				if ((line_item_p->get_good_value() == VAL_0) || (line_item_p->get_good_value() == VAL_X))
					strcpy(state_s,"0");
				else if (line_item_p->get_good_value() == VAL_D_BAR)
					strcpy(state_s,"0");
				else if (line_item_p->get_good_value() == VAL_1)
					strcpy(state_s,"1");
				else if (line_item_p->get_good_value() == VAL_D)
					strcpy(state_s,"1");
				else
					error("cfidr_PrintTestpattern: Unknown Value in Line Output");
				cfidrCreatePatternBit(cfidrPortId,time_s,state_s,&cfidrError);
				if (cfidrError != CFIDR_NO_ERROR)
					{
					error (cfidrString);
					}/*if*/
				}//if
		   }//else
//		if (gefunden_b == FALSE)
//			error("cfidr_PrintTestpattern: Lines not found");
		break;
	 case CFIDR_OUTPUT:
		{forall(line_item_p,output_lines_C)
			if (line_item_p->get_name() == name_sC)
				{
				if (gefunden_b == TRUE)
					error("cfidr_PrintTestpattern: 2 Lines with the same name");
				else
					gefunden_b = TRUE;
				if ((line_item_p->get_good_value() == VAL_0)|| (line_item_p->get_good_value() == VAL_D_BAR) || (line_item_p->get_good_value() == VAL_X))
					strcpy(state_s,"0");
				else if ((line_item_p->get_good_value() == VAL_1) || (line_item_p->get_good_value() == VAL_D))
					strcpy(state_s,"1");
				else
					error("cfidr_PrintTestpattern: Unknown Value in Line Output");
				cfidrCreatePatternBit(cfidrPortId,time_s,state_s,&cfidrError);
				if (cfidrError != CFIDR_NO_ERROR)
					{
					error (cfidrString);
					}/*if*/
				}}//if
//		if (gefunden_b == FALSE)
//			error("cfidr_PrintTestpattern: Lines not found");
		break;
	 default: error("cfidr_CreateInputsOutputs unknown portdirection"); 
      }	 /*switch*/
      /* name the block */
      cfidrPortId = cfidrIterNextPort(cfidrPortsId, &cfidrError);
   }  while (cfidrObjectIsNull(cfidrPortId, &cfidrError) != CFIDR_TRUE);
if (cfidrError != CFIDR_NO_ERROR)
	{ /* belongs to cfidrObjectIsNull */
	cfidrString = cfidrPIGetErrorText(cfidrError, &cfidrError);
	error (cfidrString);
	}  /*if*/
}//cfidr_PrintTestpattern
#endif





