//*********************************************************************
//
//   FILE-NAME: check.c
//
//   CREATION-DATE: August 91
//
//   FILE-DESCRIPTION: Dieses File beinhaltet Routinen zur Ueberpruefung der
//	Implikations Funktion
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
#include "extern.h"
#include "protocol_stack_C.h"
#include "line_C.h"
#include "gate_C.h"
#include "line_liste_C.h"
#include "gate_liste_C.h"

extern line_liste_C implication_line_C;
extern line_liste_C unjustified_bound_line_C;
extern protocol_stack_C protocol_C;
extern boolean only_iddq_test_b;

inline double gib_contby(line_C *line_p){return(line_p->gib_zero_contby()*line_p->gib_one_contby());}
/* ************************************************************************** */
/*									      */
/*  FUNCTION NAME: milef_search_line_tree				      */
/*									      */
/*  FUNCTIONAL DESCRIPTION:						      */
/*									      */
/*	milef_search_line_tree searchs for a line named linename_s in a    */
/*	tree of lines. (The tree is a fully balanced AVL-tree.)	      */
/*									      */
/*	If the line is not found, a new tree element is created.	      */
/*									      */
/*  FORMAL PARAMETERS:							      */
/*									      */
/*	linename_s: name of the line to search for			      */
/*									      */
/*	tree_p: pointer to the root of the line tree		      */
/*									      */
/*	H: boolean variable, must be FALSE at call			      */
/*									      */
/*	line_tree_p: pointer to the found tree element, or pointer to the     */
/*	new tree element, if not found. (Then line_tree_p-> line_p equals     */
/*	NULL.)								      */
/*									      */
/*  IMPLICIT INPUT/OUTPUT:						      */
/*									      */
/*	none								      */
/*									      */
/*  FUNCTION VALUE:							      */
/*									      */
/*	none								      */
/*									      */
/*  CREATION DATE: 20.06.91						      */
/*									      */
/*									      */
/*  MODIFICATION HISTORY:						      */
/*									      */
/*	none								      */
/*									      */
/* ************************************************************************** */
void milef_search_line_tree(line_C *line_p,line_tree_u *&tree_p,boolean &H,line_tree_u *&line_tree_p,
			    boolean use_testaby_val_b)
{
  line_tree_u *tree_p1 = NULL, *tree_p2 = NULL;
  char *linename_s,*actual_name_s;

  linename_s = line_p->get_name();
  if (tree_p != NULL)
     actual_name_s = tree_p->line_p->get_name();
  else
  {
      H = TRUE;
      try {
      tree_p = new line_tree_u;
      }
      catch(...) {
      error("No more memory available");
      }
      if (tree_p == NULL)
      error("no free memory available");
      tree_p->left = NULL;
      tree_p->right = NULL;
      tree_p->bal_i = 0;
      tree_p->line_p = NULL;
      tree_p->left_done_b = FALSE;
      tree_p->right_done_b = FALSE;
      tree_p->element_done_b = FALSE;
      line_tree_p = tree_p;
      return;
  }   /*if*/

  if  (((use_testaby_val_b == FALSE) && (strcmp(linename_s, actual_name_s) < 0)) ||
       ((use_testaby_val_b == TRUE) && (only_iddq_test_b == FALSE) && (line_p->gib_testaby() <= tree_p->line_p->gib_testaby())) ||
       ((use_testaby_val_b == TRUE) && (only_iddq_test_b == TRUE) && (gib_contby(line_p) <= gib_contby(tree_p->line_p))))
  {
    milef_search_line_tree(line_p, tree_p->left, H, line_tree_p,use_testaby_val_b);
    if (H == FALSE)
      return;
    switch (tree_p->bal_i) {

    case 1:
      tree_p->bal_i = 0;
      H = FALSE;
      break;

    case 0:
      tree_p->bal_i = -1;
      break;

    case -1:  /*erneutes Ausgleichen*/
      tree_p1 = tree_p->left;
      if (tree_p1->bal_i == -1) {  /*einfache LL Rotation*/
	tree_p->left = tree_p1->right;
	tree_p1->right = tree_p;
	tree_p->bal_i = 0;
	tree_p = tree_p1;
      }
      else {  /*doppelte LR Rotation*/
	tree_p2 = tree_p1->right;
	tree_p1->right = tree_p2->left;
	tree_p2->left = tree_p1;
	tree_p->left = tree_p2->right;
	tree_p2->right = tree_p;
	if (tree_p2->bal_i == -1)
	  tree_p->bal_i = 1;
	else
	  tree_p->bal_i = 0;
	if (tree_p2->bal_i == 1)
	  tree_p1->bal_i = -1;
	else
	  tree_p1->bal_i = 0;
	tree_p = tree_p2;
      }
      tree_p->bal_i = 0;
      H = FALSE;
      break;
    }
    return;
  }
  if  ((use_testaby_val_b == FALSE) && (strcmp(linename_s,actual_name_s) <= 0))
  {
    H = FALSE;
    line_tree_p = tree_p;
    return;
  }
  milef_search_line_tree(line_p,tree_p->right,H,line_tree_p,use_testaby_val_b);
  if (H == FALSE)
    return;
  switch (tree_p->bal_i) {

  case -1:
    tree_p->bal_i = 0;
    H = FALSE;
    break;

  case 0:
    tree_p->bal_i = 1;
    break;

  case 1:  /*erneutes Ausgleichen*/
    tree_p1 = tree_p->right;
    if (tree_p1->bal_i == 1) {  /*einfache LL Rotation*/
      tree_p->right = tree_p1->left;
      tree_p1->left = tree_p;
      tree_p->bal_i = 0;
      tree_p = tree_p1;
    }
    else {  /*doppelte LR Rotation*/
      tree_p2 = tree_p1->left;
      tree_p1->left = tree_p2->right;
      tree_p2->right = tree_p1;
      tree_p->right = tree_p2->left;
      tree_p2->left = tree_p;
      if (tree_p2->bal_i == 1)
	tree_p->bal_i = -1;
      else
	tree_p->bal_i = 0;
      if (tree_p2->bal_i == -1)
	tree_p1->bal_i = 1;
      else
	tree_p1->bal_i = 0;
      tree_p = tree_p2;
    }
    tree_p->bal_i = 0;
    H = FALSE;
    break;
  }
}  /*milef_search_line_tree*/

/* ************************************************************************** */
/*									      */
/*  FUNCTION NAME: milef_add_to_line_tree					      */
/*									      */
/*  FUNCTIONAL DESCRIPTION:						      */
/*									      */
/*	milef_add_to_line_tree adds the line_p to the tree of lines with      */
/*	root line_tree_root_p.						      */
/*									      */
/*  FORMAL PARAMETERS:							      */
/*									      */
/*	line_tree_root_p: pointer to root of a line tree		      */
/*									      */
/*	line_p: pointer to a line to add				      */
/*									      */
/*  IMPLICIT INPUT/OUTPUT:						      */
/*									      */
/*	none								      */
/*									      */
/*  FUNCTION VALUE:							      */
/*									      */
/*	none								      */
/*									      */
/*  CREATION DATE: 25.07.91						      */
/*									      */
/*  MODIFICATION HISTORY:						      */
/*									      */
/*	none								      */
/*									      */
/* ************************************************************************** */
void milef_add_to_line_tree(line_tree_u *&line_tree_root_p,line_C *line_p,boolean use_testaby_val_b)
{
   line_tree_u *new_line_tree_p;
   boolean h_b = FALSE;

   milef_search_line_tree(line_p,line_tree_root_p,h_b,new_line_tree_p,use_testaby_val_b);
   if (new_line_tree_p->line_p == NULL)
      new_line_tree_p->line_p = line_p;
   else
      {
      cout << "\n\n\n\nSignal name: " << line_p->get_name() << "\n";
      error("milef_add_to_line_tree: Line was already in the list");
      }//else
}/*milef_add_to_line_tree*/

/* ************************************************************************** */
/*									      */
/*  FUNCTION NAME: milef_search_gate_tree				      */
/*									      */
/*  FUNCTIONAL DESCRIPTION:						      */
/*									      */
/*	milef_search_gate_tree searchs for a gate named gatename_s in a    */
/*	tree of gates. (The tree is a fully balanced AVL-tree.)	      */
/*									      */
/*	If the gate is not found, a new tree element is created.	      */
/*									      */
/*  FORMAL PARAMETERS:							      */
/*									      */
/*	gatename_s: name of the gate to search for			      */
/*									      */
/*	tree_p: pointer to the root of the gate tree		      */
/*									      */
/*	H: boolean variable, must be FALSE at call			      */
/*									      */
/*	gate_tree_p: pointer to the found tree element, or pointer to the     */
/*	new tree element, if not found. (Then gate_tree_p-> gate_p equals     */
/*	NULL.)								      */
/*									      */
/*  IMPLICIT INPUT/OUTPUT:						      */
/*									      */
/*	none								      */
/*									      */
/*  FUNCTION VALUE:							      */
/*									      */
/*	none								      */
/*									      */
/*  CREATION DATE: 20.06.91						      */
/*									      */
/*									      */
/*  MODIFICATION HISTORY:						      */
/*									      */
/*	none								      */
/*									      */
/* ************************************************************************** */
void milef_search_gate_tree(char *gatename_s,gate_tree_u *&tree_p,boolean &H,gate_tree_u *&gate_tree_p)
{
  gate_tree_u *tree_p1 = NULL, *tree_p2 = NULL;

  if (tree_p == NULL)
  {
      H = TRUE;
      try {
      tree_p = new gate_tree_u;
      }
      catch(...) {
      error("No more memory available");
      }
      if (tree_p == NULL)
	error("no free memory available");
      tree_p->left = NULL;
      tree_p->right = NULL;
      tree_p->bal_i = 0;
      tree_p->gate_p = NULL;
      gate_tree_p = tree_p;
      return;
  }   /*if*/

  if  (strcmp(gatename_s,tree_p->gate_p->get_name()) < 0)
  {
    milef_search_gate_tree(gatename_s, tree_p->left, H, gate_tree_p);
    if (H == FALSE)
      return;
    switch (tree_p->bal_i) {

    case 1:
      tree_p->bal_i = 0;
      H = FALSE;
      break;

    case 0:
      tree_p->bal_i = -1;
      break;

    case -1:  /*erneutes Ausgleichen*/
      tree_p1 = tree_p->left;
      if (tree_p1->bal_i == -1) {  /*einfache LL Rotation*/
	tree_p->left = tree_p1->right;
	tree_p1->right = tree_p;
	tree_p->bal_i = 0;
	tree_p = tree_p1;
      }
      else {  /*doppelte LR Rotation*/
	tree_p2 = tree_p1->right;
	tree_p1->right = tree_p2->left;
	tree_p2->left = tree_p1;
	tree_p->left = tree_p2->right;
	tree_p2->right = tree_p;
	if (tree_p2->bal_i == -1)
	  tree_p->bal_i = 1;
	else
	  tree_p->bal_i = 0;
	if (tree_p2->bal_i == 1)
	  tree_p1->bal_i = -1;
	else
	  tree_p1->bal_i = 0;
	tree_p = tree_p2;
      }
      tree_p->bal_i = 0;
      H = FALSE;
      break;
    }
    return;
  }
  if  (strcmp(gatename_s,tree_p->gate_p->get_name()) <= 0)
  {
    H = FALSE;
    gate_tree_p = tree_p;
    return;
  }
  milef_search_gate_tree(gatename_s,tree_p->right,H,gate_tree_p);
  if (H == FALSE)
    return;
  switch (tree_p->bal_i) {

  case -1:
    tree_p->bal_i = 0;
    H = FALSE;
    break;

  case 0:
    tree_p->bal_i = 1;
    break;

  case 1:  /*erneutes Ausgleichen*/
    tree_p1 = tree_p->right;
    if (tree_p1->bal_i == 1) {  /*einfache LL Rotation*/
      tree_p->right = tree_p1->left;
      tree_p1->left = tree_p;
      tree_p->bal_i = 0;
      tree_p = tree_p1;
    }
    else {  /*doppelte LR Rotation*/
      tree_p2 = tree_p1->left;
      tree_p1->left = tree_p2->right;
      tree_p2->right = tree_p1;
      tree_p->right = tree_p2->left;
      tree_p2->left = tree_p;
      if (tree_p2->bal_i == 1)
	tree_p->bal_i = -1;
      else
	tree_p->bal_i = 0;
      if (tree_p2->bal_i == -1)
	tree_p1->bal_i = 1;
      else
	tree_p1->bal_i = 0;
      tree_p = tree_p2;
    }
    tree_p->bal_i = 0;
    H = FALSE;
    break;
  }
}  /*milef_search_gate_tree*/

/* ************************************************************************** */
/*									      */
/*  FUNCTION NAME: milef_add_to_gate_tree					      */
/*									      */
/*  FUNCTIONAL DESCRIPTION:						      */
/*									      */
/*	milef_add_to_gate_tree adds the gate_p to the tree of gates with      */
/*	root gate_tree_root_p.						      */
/*									      */
/*  FORMAL PARAMETERS:							      */
/*									      */
/*	gate_tree_root_p: pointer to root of a gate tree		      */
/*									      */
/*	gate_p: pointer to a gate to add				      */
/*									      */
/*  IMPLICIT INPUT/OUTPUT:						      */
/*									      */
/*	none								      */
/*									      */
/*  FUNCTION VALUE:							      */
/*									      */
/*	none								      */
/*									      */
/*  CREATION DATE: 25.07.91						      */
/*									      */
/*  MODIFICATION HISTORY:						      */
/*									      */
/*	none								      */
/*									      */
/* ************************************************************************** */
void milef_add_to_gate_tree(gate_tree_u *&gate_tree_root_p,gate_C *gate_p)
{
   gate_tree_u *new_gate_tree_p;
   boolean h_b = FALSE;


   milef_search_gate_tree((char *)gate_p->get_name(),gate_tree_root_p,h_b,new_gate_tree_p);
   if (new_gate_tree_p->gate_p == NULL)
      new_gate_tree_p->gate_p = gate_p;
   else
      {
      cout << "\n\n\n\nGate name: " << gate_p->get_name() << "\n";
      error("milef_add_to_gate_tree: gate was already in the list");
      }//else
}/*milef_add_to_gate_tree*/

/* ************************************************************************** */
/*									      */
/*  FUNCTION NAME: proof					      */
/*									      */
/*  FUNCTIONAL DESCRIPTION:						      */
/*									      */
//	proof checks the io List of gates and lines
/*									      */
/*  FORMAL PARAMETERS:	None						      */
/*									      */
/*									      */
/*  IMPLICIT INPUT/OUTPUT:						      */
/*									      */
/*	none								      */
/*									      */
/*  FUNCTION VALUE:							      */
/*									      */
/*	void								      */
/*									      */
/*  CREATION DATE: 15.10.91						      */
/*									      */
/*  MODIFICATION HISTORY:						      */
/*									      */
/*	none								      */
/*									      */
/* ************************************************************************** */
void proof()
{
line_C *line_item_p;
gate_C *gate_item_p;
{forall(line_item_p,lines_C)
	{
	if ((line_item_p->is_primary_input() == FALSE) && (line_item_p->is_pseudo_input() == FALSE) &&
	    (line_item_p->vorgaenger()->no_elements() == 0))
		error(string_C("proof: line without sender must be input:  ") + line_item_p->get_name());
	if ((line_item_p->is_primary_output() == FALSE) && (line_item_p->is_pseudo_output() == FALSE) &&
	    (line_item_p->nachfolger()->no_elements() == 0))
		error(string_C("proof: not connected line must be output:  ") + line_item_p->get_name());
	if (line_item_p->nachfolger()->no_elements() > 1)
		error(string_C("proof: too many output at:  ") + line_item_p->get_name());
	if (line_item_p->vorgaenger()->no_elements() > 1)
		error(string_C("proof: too many sender of:  ") + line_item_p->get_name());
	}}//forall
forall(gate_item_p,gates_C)
	{
	if ((gate_item_p->gib_outputs()->no_elements() > 1) && (gate_item_p->gib_type() != FANOUT))
		error(string_C("proof: too many outputs at:  ") + gate_item_p->get_name());
	if (gate_item_p->gib_outputs()->no_elements() < 1)
		error(string_C("proof: too less outputs at:  ") + gate_item_p->get_name());
	if (gate_item_p->gib_inputs()->no_elements() < 1)
		error(string_C("proof: too less inputs at:  ") + gate_item_p->get_name());
	}//forall
}//proof

//*********************************************************************
//
//   FUNCTION-NAME: search
//
//   CLASS: line_liste_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Diese Funktion sucht rekursiv eine Line in dem AVL Baum, die 
//	den vorgegebenen Namen hat.
//
//   FORMAL PARAMETERS: string_C name_sC: der String, der gesucht wird
//			line_tree_u *tree_p: AVL Baum mit Lines
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: line_C*: das Gatter mit entsprechendem Namen
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
line_C *search(char *name_s,line_tree_u *tree_p)
{
if (tree_p == NULL)
	return(NULL);

int i = strcmp(name_s,tree_p->line_p->get_name());
if (i == 0)
   return(tree_p->line_p);
if (i < 0)
   return(search(name_s,tree_p->left));
else // >0 bleibt uebrig
   return(search(name_s,tree_p->right));
}/*element_in_liste*/

gate_C *search(char *name_s,gate_tree_u *tree_p)
{
if (tree_p == NULL)
	return(NULL);

int i = strcmp(name_s,tree_p->gate_p->get_name());
if (i == 0)
   return(tree_p->gate_p);
if (i < 0)
   return(search(name_s,tree_p->left));
else // >0 bleibt uebrig
   return(search(name_s,tree_p->right));
}/*element_in_liste*/






















