//*********************************************************************
//
//   FILE-NAME: fan_alg.c
//
//   CREATION-DATE: August 91
//
//   FILE-DESCRIPTION: Dieses File beinhaltet die Hauptroutinen fuer den
//	Fan Algorithmus
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
#include "extern.h"
#include "line_liste_C.h"
#include "gate_liste_C.h"
#include "line_value_liste_C.h"
#include "protocol_stack_C.h"
#include "objective_liste_C.h"
#include "line_C.h"
#include "gate_C.h"


	// Prototypen fuer Hauptprogramme des Fan Algorithmus
objective_type_u determine_final_objective();
entry_exit_t multiple_backtrace(entry_exit_t);

	// Prototypen fuer Unterprogramme des Fan Algorithmus
void add_gate_in_d_front_to_initial_objectives();
void add_n0_and_n1_to_fanout_objective(objective_type_u&);
void add_objective_to_head_objectives(objective_type_u&);
void assign_value_to_final_objective(objective_type_u&);
boolean continuation_of_backtrace_meaningful();
void determine_next_objective_and_add_them_to_current_objectives(objective_type_u&);
boolean exist_untried_combination_of_values();
boolean robustness_test(line_C*,value_t);
fan_result_t implication();
void justify_free_lines();
void let_objectives_be_empty();
void let_unjustified_lines_be_initial_objectives();
boolean n0p_and_n1p_nonequal_zero(objective_type_u&);
int number_of_gates_in_d_front();
boolean objective_line_follows_fanout(objective_type_u&);
void set_untried_combination_of_values();
objective_type_u take_out_fanout_objective_p_closest_to_output();
objective_type_u take_out_head_objective();
fan_result_t unique_sensitisation();
fan_result_t dyn_learning();
void reset_dyn_learning();
extern boolean in_kegel(line_C*,line_C*);
extern void clear_hazard_bits(gate_C*);

	// Globale Variablen
extern boolean input_kegel_changed_b;
extern boolean init_pattern_b;
extern boolean weak_test_b;
extern boolean phase2_b;
extern boolean phase3_b;
boolean hazard_test_ein_b;
boolean hazard_occured_b;
boolean backtrace_flag_b;
boolean ujust_bound_changed_b = TRUE;
boolean d_front_changed_b = FALSE;
boolean only_iddq_test_b = FALSE;
boolean path_delay_test_b = FALSE;
boolean transition_test_b = FALSE;
boolean stuck_open_test_b = FALSE;
boolean iddq_stuck_on_test_b = FALSE;
boolean only_stuck_at_test_b = FALSE;
boolean bridging_iddq_b = FALSE;
boolean stuck_at_test_without_branches_b = FALSE;
extern boolean ctest_pattern_b;
int max_number_backtracks_i; // Default ist 10 s. main.c
int number_backtracks_i;
int backtrack_counter_i = 0;
int reject_counter_i = 0;
line_C *ziel_line_p;
extern gate_liste_C blocks_C;
line_liste_C unjustified_bound_line_C;
line_liste_C unjustified_free_line_C;
line_liste_C d_front_C;
line_liste_C implication_line_C;
line_liste_C tmp_line_liste_C;
line_liste_C free_lines_C;
line_liste_C off_path_signals_C;
line_liste_C off_path_signals_save_C;
line_liste_C off_path_signals_tmp_C;
extern line_value_liste_C zielfehler_C;
line_value_liste_C hilf_liste_C;
gate_liste_C tmp_gate_liste_C;
protocol_stack_C protocol_C;
extern objective_liste_C current_objectives_C;
extern objective_liste_C head_objectives_C;
extern objective_liste_C fanout_objectives_C;
extern objective_type_u final_objective_u;
#if DEBUG
opt_line_liste_C optional_lines_C;
#endif
extern gate_C *d_front_gate_in_initial_objectives_p;

	// Inline Functions
boolean set_of_fanout_objectives_empty(){return(fanout_objectives_C.is_empty());}
boolean set_of_head_objectives_empty(){return(head_objectives_C.is_empty());}
boolean set_of_current_objectives_empty(){return(current_objectives_C.is_empty());}
objective_type_u &take_out_current_objective(){return(*current_objectives_C.pop());}
objective_type_u fanout_point_objective_is_final_objective(){return(final_objective_u);}
boolean p_reachable_from_fault_line(objective_type_u &objective_u){return(objective_u.line_p->in_output_kegel());}
boolean head_line_unspecified(objective_type_u &objective_u){return((objective_u.line_p->get_good_value() == VAL_X)?TRUE:FALSE);}
boolean is_objective_head_line(objective_type_u &objective_u){return((objective_u.line_p->gib_type() == HEAD_LINE)?TRUE:FALSE);}
boolean exist_unjustified_bound_line(){return((unjustified_bound_line_C.is_empty()==TRUE)?FALSE:TRUE);}
void let_initial_objectives_be_current_objectives(){}
//****************************** Macros ******************************
	//Value Invertierung
value_t mynot(value_t val_e)
{
switch (val_e)
	{
	case VAL_0: return(VAL_1);
	case VAL_1: return(VAL_0);
	case VAL_0S: return(VAL_1S);
	case VAL_1S: return(VAL_0S);
	case VAL_0X: return(VAL_1X);
	case VAL_1X: return(VAL_0X);
	case VAL_D: return(VAL_D_BAR);
	case VAL_D_BAR: return(VAL_D);
	case VAL_X: return(VAL_X);
	case UNDEFINED_VALUE: error("not: undefined values as input");
	}//switch
}//not
//*********************************************************************
//
//   FUNCTION-NAME: fault_signal_at_output
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Stellt fest, ob das Fehlersignal am Ausgang ist
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: boolean: TRUE, wenn D oder D_BAR am Ausgang anliegt
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
boolean fault_signal_at_output()
{
line_C *line_item_p;
int i;

if ((only_iddq_test_b == TRUE) || (iddq_stuck_on_test_b == TRUE) ||
    (bridging_iddq_b == TRUE))
	return(TRUE);
for (i=0;i<d_front_C.no_elements();i++)
	{
	line_item_p = d_front_C[i];
	if (line_item_p->is_primary_output() == TRUE)
		return(TRUE);
	}//for
return(FALSE);
}//fault_signal_at_output

//*********************************************************************
//
//   FUNCTION-NAME: fan_alg
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Hauptprogramm des Fan Algorithmus. Fig. 9 von Fujiwara,
//	Shimono: On the Acceleration of Test Generation Algorithms findet sich
//	hier wieder.
//
//   FORMAL PARAMETERS: line_C *line_item_p: Die als naechstes zu testende Line
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: fan_result_t: Ergebnis des Fan (SUCCESS,ABORTED_FAULT,REDUNDANT_FAULT)
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
fan_result_t fan_alg(line_C *line_item_p,value_t ziel_val_e)
{
fan_result_t implication_result_e = SUCCESS;
objective_type_u objective_u;
line_C *hilf_line_p;
int i;
protocol_type_C *dummy_p;
gate_C *gate_item_p;

ziel_line_p = line_item_p;
d_front_gate_in_initial_objectives_p = NULL;
d_front_C.fuege_ein(ziel_line_p);
if (ctest_pattern_b == FALSE)
	{
	if (((stuck_open_test_b == TRUE) || (iddq_stuck_on_test_b == TRUE)) &&
	    ((ziel_val_e == VAL_U0) || (ziel_val_e == VAL_U1)) &&
	    (ziel_line_p->nachfolger()->gib_only_gate()->gib_type() != EXOR))
		{
		gate_C *follow_p = ziel_line_p->nachfolger()->gib_only_gate();
	
				//nicht kontrollierende Werte muessen gesetzt werden am Nachfolger
		{forall(hilf_line_p,*follow_p->gib_inputs())
		   {
		   switch ((*ziel_line_p->nachfolger())[0]->gib_type())
			{
			case AND:
			case NAND:
				if (hilf_line_p != ziel_line_p)
					hilf_line_p->set_good_value(VAL_1);
				break;
			case OR:
			case NOR:
				if (hilf_line_p != ziel_line_p)
					hilf_line_p->set_good_value(VAL_0);
				break;
			default:
				break;
			}//switch
		   if (hilf_line_p->gib_type() == BOUND_LINE)
			unjustified_bound_line_C.fuege_ein(hilf_line_p);
		   }}//forall
		}//if
	else
		if (ziel_line_p->gib_type() == BOUND_LINE)
			unjustified_bound_line_C.fuege_ein(ziel_line_p);
	}//if
else
	{forall(hilf_line_p,*(*ziel_line_p->vorgaenger())[0]->gib_inputs())
	   if (hilf_line_p->get_good_value() != VAL_X)
		if (hilf_line_p->gib_type() == BOUND_LINE)
			unjustified_bound_line_C.fuege_ein(hilf_line_p);}
implication_line_C.fuege_ein(ziel_line_p);
for (i=0;i<lines_C.no_elements();i++)
	{
	hilf_line_p = lines_C[i];
	if ((hilf_line_p->get_only_value() != VAL_X) && (hilf_line_p->in_output_kegel() == FALSE))
	   if ((hilf_line_p->get_good_value() == VAL_X) ||
	       (hilf_line_p->get_good_value() == hilf_line_p->get_only_value()))
		{
		hilf_line_p->set_good_value(hilf_line_p->get_only_value());
		implication_line_C.fuege_ein(hilf_line_p);
		}//if
	   else
		return(INCONSISTENT);
	}//for

	// jetzt kommt das Diagramm Figure 9
input_kegel_changed_b = FALSE;
backtrace_flag_b = TRUE;
for(;;)
	{
	if (implication_result_e == SUCCESS)
		{
		implication_result_e = implication();
		implication_line_C.loesche_liste();
		}//if
	if ((hazard_test_ein_b == TRUE) && (ziel_val_e != VAL_0) && (ziel_val_e != VAL_1) &&
	    (init_pattern_b == FALSE) && (implication_result_e == SUCCESS) &&
	    (input_kegel_changed_b == TRUE) && (robustness_test(ziel_line_p,ziel_val_e) == TRUE))
		{
		ziel_line_p->hazard_occured();
		hazard_occured_b = TRUE;
		implication_result_e = INCONSISTENT;
		}//if
	input_kegel_changed_b = FALSE;
	if (implication_result_e == INCONSISTENT)
		{
		if (exist_untried_combination_of_values() == TRUE)
			{
			if (number_backtracks_i >= max_number_backtracks_i)
				return(ABORTED_FAULT);
			if (phase3_b == TRUE)
				reset_dyn_learning();
			number_backtracks_i++;
			set_untried_combination_of_values();
			backtrace_flag_b = TRUE;
			implication_result_e = SUCCESS;
			}//if
		else
			return(REDUNDANT_FAULT);
		}//if
	else
		{
		if (continuation_of_backtrace_meaningful() == FALSE)
			backtrace_flag_b = TRUE;
		if ((init_pattern_b == TRUE) || (fault_signal_at_output() == TRUE))
			if (exist_unjustified_bound_line() == TRUE)
				{
				if (phase3_b == TRUE)
					implication_result_e = dyn_learning();
				else
					implication_result_e = NO_IMPLICATION;
				if (implication_result_e == NO_IMPLICATION)
					{
					objective_u = determine_final_objective();
					assign_value_to_final_objective(objective_u);
					implication_result_e = SUCCESS;
					}//if
				}//if
			else
				{
				justify_free_lines();
				if (weak_test_b == FALSE)
				   {
				   forall(gate_item_p,blocks_C)
					if ((*gate_item_p->gib_outputs())[0]->in_output_kegel() == TRUE)
						{
						(*gate_item_p->gib_outputs())[0]->set_bad_value(VAL_X);
						gate_item_p->fast_forward_implication();
						dummy_p = protocol_C.pop();
						}//if
				   }//if
				return(SUCCESS);
				}//else
		else
		   if (number_of_gates_in_d_front() == 0)
			if (exist_untried_combination_of_values() == TRUE)
				{
				if (number_backtracks_i >= max_number_backtracks_i)
					return(ABORTED_FAULT);
				if (phase3_b == TRUE)
					reset_dyn_learning();
				number_backtracks_i++;
				set_untried_combination_of_values();
				backtrace_flag_b = TRUE;
				implication_result_e = SUCCESS;
				}//if
			else
				return(REDUNDANT_FAULT);
		   else
			{
			implication_result_e = unique_sensitisation();
			if (implication_result_e == NO_IMPLICATION)
				{
				if (phase3_b == TRUE)
					implication_result_e = dyn_learning();
				if (implication_result_e == NO_IMPLICATION)
					{
					objective_u = determine_final_objective();
					assign_value_to_final_objective(objective_u);
					implication_result_e = SUCCESS;
					}//if
				}//if
			}//else
		}//else
	}//forever
}//fan_alg

//*********************************************************************
//
//   FUNCTION-NAME: determine_final_objective
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Hauptprogramm vom Auffinden eines Objectives. Fig. 10
//	von Fujiwara, Shimono: On the Acceleration of Test Generation Algorithms
//	findet sich hier wieder.
//
//   FORMAL PARAMETERS: NONE
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: objective_type_u: The final Objective for to assign a Value
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
objective_type_u determine_final_objective()
{
objective_type_u objective_u;
entry_exit_t return_value_e;
entry_exit_t entry_value_e;
boolean multibacktrace_entry_flag_b = FALSE; // Fuer den ersten Aufruf von Multibacktrace

for (;;)
	{
	if (backtrace_flag_b == TRUE)
		{
		backtrace_flag_b = FALSE;
		ujust_bound_changed_b = FALSE;
		d_front_changed_b = FALSE;
		let_objectives_be_empty();
		if (exist_unjustified_bound_line() == FALSE)
			add_gate_in_d_front_to_initial_objectives();
		else
			{
			let_unjustified_lines_be_initial_objectives();
			if ((fault_signal_at_output() == FALSE) && (init_pattern_b == FALSE))
				add_gate_in_d_front_to_initial_objectives();
			}//else
		entry_value_e = ENTRY_A;
		multibacktrace_entry_flag_b = TRUE;
		}//if
	else
		{
		if (set_of_fanout_objectives_empty() == TRUE)
			entry_value_e = ENTRY_B;
		multibacktrace_entry_flag_b = TRUE;
		}//else
	if (multibacktrace_entry_flag_b == TRUE)
		{
		return_value_e = multiple_backtrace(ENTRY_A);
		if (return_value_e == EXIT_D)
			{
			return(fanout_point_objective_is_final_objective());
			}//if
		}//if
	for(;;)		// Wert von multibacktrace ist EXIT_C oder Multibacktrace
		{	// ist noch nicht aufgerufen worden
		if (set_of_head_objectives_empty() == FALSE)
			objective_u = take_out_head_objective();
		else
			{
			backtrace_flag_b = TRUE; // nicht in Fig. 10, wird aber sofort
						 // zurueckgesetzt nach der Abfrage
			break;
			}//else
		if (head_line_unspecified(objective_u) == TRUE)
			{	// head Objective is Final Objective
			return(objective_u);
			}//if
		}//forever
	}//forever
}//determine_final_objective

//*********************************************************************
//
//   FUNCTION-NAME: multiple_backtrace
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Hauptprogramm von dem Multibacktrace Algorithmus. Fig. 8
//	von Fujiwara, Shimono: On the Acceleration of Test Generation Algorithms
//	findet sich hier wieder.
//
//   FORMAL PARAMETERS: entry_exit_t entry_value_e: Eingang der Funktion (ENTRY_A oder
//							ENTRY_B)
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: entry_exit_t: Ausgang von Multibacktrace (EXIT_A oder EXIT_B)
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
entry_exit_t multiple_backtrace(entry_exit_t entry_value_e)
{
boolean b_path_flag_b = FALSE;	// Fuer den Pfad ab entry_b bis exit_d
entry_exit_t entry_e = entry_value_e;
objective_type_u objective_u;

if (entry_e == ENTRY_A)
	let_initial_objectives_be_current_objectives();
for(;;)
	{
	if (entry_e == ENTRY_A)
		if (set_of_current_objectives_empty() == FALSE)
			objective_u = take_out_current_objective();
		else
			if (set_of_fanout_objectives_empty() == TRUE)
				return(EXIT_C);
			else
				{
				objective_u = take_out_fanout_objective_p_closest_to_output();
				b_path_flag_b = TRUE;
				}//else
	if (entry_e == ENTRY_B)
		{ // darf nur einmal ausgefuehrt werden
		objective_u = take_out_fanout_objective_p_closest_to_output();
		b_path_flag_b = TRUE;
		entry_e = ENTRY_A;
		}//if
	if (b_path_flag_b == TRUE)
		{
		b_path_flag_b = FALSE;
		if (p_reachable_from_fault_line(objective_u) == FALSE)
			if (n0p_and_n1p_nonequal_zero(objective_u) == TRUE)
				return(EXIT_D);
		}//if
	if (is_objective_head_line(objective_u) == TRUE)
		add_objective_to_head_objectives(objective_u);
	else
		if (objective_line_follows_fanout(objective_u) == TRUE)
			add_n0_and_n1_to_fanout_objective(objective_u);
		else
			determine_next_objective_and_add_them_to_current_objectives(objective_u);
	}//forever
}//multiple_backtrace

//*********************************************************************
//
//   FUNCTION-NAME: exist_untried_combination_of_values
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Stellt fest, ob ein Backtrack moeglich ist.
//	Sucht zusaetzlich die neue D Front
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: setzt Implikationen gegebenenfalls zurueck
//
//   FUNCTION-VALUE: boolean: TRUE, wenn Backtrack moeglich ist
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
boolean exist_untried_combination_of_values()
{
protocol_type_C *prot_item_p;
for(;;)
	{
	prot_item_p = protocol_C.pop();
	if (prot_item_p == NULL)
		return(FALSE);
	if (prot_item_p->line_p == NULL)
		error("exist_untried_combination_of_values: Line Pointer is zero");
	if (prot_item_p->optional() == TRUE)
		break;
        if ((prot_item_p->new_good_value() == VAL_X) && (prot_item_p->new_bad_value() == UNDEFINED_VALUE))
                error("exist_untried_combination_of_values: combination of good- and bad-value is invalid");
	if ((prot_item_p->line_p->get_good_value() != prot_item_p->new_good_value()) &&
	    (prot_item_p->new_good_value() != UNDEFINED_VALUE) && (prot_item_p->new_good_value() != VAL_X))
		error("exist_untried_combination_of_values: Test Value doesn't match billeting");
	if ((prot_item_p->old_good_value() != UNDEFINED_VALUE) && (prot_item_p->new_good_value() != VAL_X))
		{
		prot_item_p->line_p->set_good_value(prot_item_p->old_good_value());
		if ((prot_item_p->line_p->is_primary_output() == FALSE) &&
		    (prot_item_p->line_p->is_pseudo_output() == FALSE))
			clear_hazard_bits(prot_item_p->line_p->nachfolger()->gib_only_gate());
		}//if
	if ((prot_item_p->line_p->get_bad_value() != prot_item_p->new_bad_value()) &&
	    (prot_item_p->new_bad_value() != UNDEFINED_VALUE))
		error("exist_untried_combination_of_values:  Test Value doesn't match billeting");
	if (prot_item_p->old_bad_value() != UNDEFINED_VALUE)
		prot_item_p->line_p->set_bad_value(prot_item_p->old_bad_value());
	if ((prot_item_p->new_good_value() != UNDEFINED_VALUE) && (prot_item_p->new_good_value() != VAL_X))
		{
		if (prot_item_p->line_p->is_primary_output() == FALSE)
			prot_item_p->line_p->nachfolger()->gib_only_gate()->implikation_durchgefuehrt_b = FALSE;
		if (prot_item_p->line_p->is_primary_input() == FALSE)
			prot_item_p->line_p->vorgaenger()->gib_only_gate()->implikation_durchgefuehrt_b = FALSE;
		}//if
	if (prot_item_p->new_bad_value() != UNDEFINED_VALUE)
		{
		if (prot_item_p->line_p->is_primary_output() == FALSE)
			prot_item_p->line_p->nachfolger()->gib_only_gate()->implikation_durchgefuehrt_b = FALSE;
		if (prot_item_p->line_p->is_primary_input() == FALSE)
			prot_item_p->line_p->vorgaenger()->gib_only_gate()->implikation_durchgefuehrt_b = FALSE;
		}//if
	if (prot_item_p->ujust_bound_in() == TRUE)
		unjustified_bound_line_C.fuege_aus(prot_item_p->line_p);
	if (prot_item_p->ujust_bound_out() == TRUE)
		unjustified_bound_line_C.fuege_ein(prot_item_p->line_p);
	if (prot_item_p->d_front_in() == TRUE)
		d_front_C.fuege_aus(prot_item_p->line_p);
	if (prot_item_p->d_front_out() == TRUE)
		d_front_C.fuege_ein(prot_item_p->line_p);
	}//forever

protocol_C.push(prot_item_p->line_p,FALSE,prot_item_p->ujust_bound_in(),prot_item_p->ujust_bound_out(),
		prot_item_p->d_front_in(),prot_item_p->d_front_out(),
		prot_item_p->old_good_value(),prot_item_p->new_good_value(),
		prot_item_p->old_bad_value(),prot_item_p->new_bad_value());
return(TRUE);
}//exist_untried_combination_of_values

//*********************************************************************
//
//   FUNCTION-NAME: set_untried_combination_of_values
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Der Backtrack wird vollzogen
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void set_untried_combination_of_values()
{
protocol_type_C *prot_item_p;

implication_line_C.loesche_liste();
prot_item_p = protocol_C.pop();
if (prot_item_p == NULL)
	error("set_untried_combination_of_values: Null Protocol Item");
if (prot_item_p->line_p == NULL)
	error("exist_untried_combination_of_values: Line Pointer = zero");
prot_item_p->line_p->set_good_value(mynot(prot_item_p->new_good_value()));
if (prot_item_p->line_p->in_input_kegel() == TRUE)
	input_kegel_changed_b = TRUE;
protocol_C.push(prot_item_p->line_p,FALSE,prot_item_p->ujust_bound_in(),prot_item_p->ujust_bound_out(),
		prot_item_p->d_front_in(),prot_item_p->d_front_out(),
		prot_item_p->old_good_value(),mynot(prot_item_p->new_good_value()));
if (prot_item_p->line_p->gib_type() == BOUND_LINE)
	{
	unjustified_bound_line_C.fuege_ein(prot_item_p->line_p);
	protocol_C.push(prot_item_p->line_p,FALSE,TRUE);
	}//if
implication_line_C.fuege_ein(prot_item_p->line_p);
#if DEBUG
optional_lines_C.insert(prot_item_p->line_p,prot_item_p->line_p->get_test_value(),FALSE);
#endif
}//set_untried_combination_of_values

//*********************************************************************
//
//   FUNCTION-NAME: search_off_path_signals
//
//   CLASS: gate_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: Juli 92
//
//   FUNCTIONAL-DESCRIPTION: X-Pfade zu einem Primaerasugang werden gesucht und in eine Liste
//	off_path_signals_tmp_C eingetragen.
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: boolean: TRUE, wenn X-Pfad existiert.
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
boolean gate_C::search_off_path_signals(boolean &previous_path_b)
{
line_C *line_item_p;

if (output_p->no_elements() <= actual_branch_i)
	{
	if (dyn_path_exist_b == FALSE)
		return(FALSE);
	if (previous_path_b == TRUE)
		return(TRUE);
	forall(line_item_p,off_path_signals_save_C)
		if ((*(*line_item_p->nachfolger())[0]->gib_outputs())[0]->gib_plane_no() >
		    (*output_p)[0]->gib_plane_no())
			off_path_signals_tmp_C.fuege_ein(line_item_p);
	return(TRUE);
	}//if
for(;;)
   {
   if (output_p->no_elements() <= actual_branch_i)
	return(FALSE);
   line_item_p = (*output_p)[actual_branch_i];
   if (line_item_p->search_off_path_signals(previous_path_b) == TRUE)
	{
	dyn_path_exist_b = TRUE;
	if ((gate_e == FANOUT) && (previous_path_b == TRUE))
		{
		previous_path_b = FALSE;
		actual_branch_i++;
		if (output_p->no_elements() <= actual_branch_i)
		        return(FALSE);
		while ((*output_p)[actual_branch_i]->search_off_path_signals(previous_path_b) == FALSE)
			{
			actual_branch_i++;
			if (output_p->no_elements() <= actual_branch_i)
				return(FALSE);
			}//while
		}//if
	if (previous_path_b == FALSE)
	  {
	  forall(line_item_p,*input_p)
	    if ((line_item_p->in_output_kegel() == FALSE) && (line_item_p->get_good_value() == VAL_X) &&
		(gate_e != BLOCK) && (gate_e != EXOR))
		off_path_signals_tmp_C.fuege_ein(line_item_p);
	  }//if
	return(TRUE);
	}//if
   actual_branch_i++;
   }//forever
}//search_off_path_signals

//*********************************************************************
//
//   FUNCTION-NAME: search_off_path_signals
//
//   CLASS: line_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: Maerz 92
//
//   FUNCTIONAL-DESCRIPTION: Die Funktion ruft search_off_path_signals fuer das folgende Gatter auf,
//	wenn die aktuelle Line kein Primaerausgang ist.
//
//   FORMAL PARAMETERS: NONE
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: boolean: Pfad gefunden
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
boolean line_C::search_off_path_signals(boolean &previous_path_b)
{
if ((good_value_e != VAL_X) && (bad_value_e != VAL_X))
	return(FALSE);
if (prime_output_b == TRUE)
	return(TRUE);
return(nachfolger_p->gib_only_gate()->search_off_path_signals(previous_path_b));
}//search_off_path_signals

//*********************************************************************
//
//   FUNCTION-NAME: unique_sensitisation
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Ein einzelnes D_Signal in der D Front wird bis
//	zum naechsten fanout oder zum Ausgang propagiert. Aenderungen fuer Block
//	und EXOR sind notwendig, da fuer diese unique sensitisation im Sinne des
//	FAN Algorithmus nicht definiert ist.
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
fan_result_t unique_sensitisation()
{
line_C *line_item_p = NULL,*line2_p,*hilf_line_p,*line_p;
gate_C *gate_item_p = NULL,*gate2_p;
value_t controlling_value_e = VAL_X;
boolean uni_sens_entry_b = TRUE,hilf_b = FALSE;
fan_result_t result_e = NO_IMPLICATION;
static gate_liste_C hilf_C;

if (path_delay_test_b == TRUE)
	return(result_e);
forall(line2_p,d_front_C)
   {
   gate2_p = line2_p->nachfolger()->gib_only_gate();
   forall(hilf_line_p,*gate2_p->gib_outputs())
	if ((hilf_line_p->get_good_value() == VAL_X) || (hilf_line_p->get_bad_value() == VAL_X))
	     if ((gate_item_p == NULL) || (gate_item_p == gate2_p))
		{
		line_item_p = line2_p;
		gate_item_p = gate2_p;
		}//if
	     else
		uni_sens_entry_b = FALSE;
   }//forall

if (uni_sens_entry_b == TRUE)
   while ((line_item_p->is_primary_output() == FALSE) && (number_of_gates_in_d_front() == 1))
	{
	{forall(hilf_line_p,d_front_C)
		protocol_C.push(hilf_line_p,FALSE,FALSE,FALSE,FALSE,TRUE);}
	d_front_C.loesche_liste();
	gate_item_p = NULL;
	gate2_p = line_item_p->nachfolger()->gib_only_gate();
	forall(line2_p,*gate2_p->gib_outputs())
	   if ((line2_p->get_good_value() == VAL_X) || (line2_p->get_bad_value() == VAL_X))
		if ((gate_item_p != NULL) && (gate_item_p != gate2_p))
			error("unique_sensitisation: more than one gate in D Front");
		else
			gate_item_p = gate2_p;
	string_C name_sC(gate_item_p->get_name());
	line_p = (*gate_item_p->gib_outputs())[0];
	if ((gate_item_p->gib_type() != EXOR) && (gate_item_p->gib_type() != BLOCK))
		result_e = SUCCESS;
	switch (gate_item_p->gib_type())
		{
		case AND:
		case OR:
		case DRIVER:
			if (gate_item_p->gib_outputs()->gib_only_line()->get_good_value() == VAL_X)
			   {
			   protocol_C.push(gate_item_p->gib_outputs()->gib_only_line(),FALSE,FALSE,FALSE,FALSE,
					FALSE,gate_item_p->gib_outputs()->gib_only_line()->get_good_value(),
					line_item_p->get_good_value());
			   (*gate_item_p->gib_outputs())[0]->set_good_value(line_item_p->get_good_value());
			   result_e = SUCCESS;
                           if ((*gate_item_p->gib_outputs())[0]->gib_type() == BOUND_LINE)
                                 {
                                 unjustified_bound_line_C.fuege_ein((*gate_item_p->gib_outputs())[0]);
                                 protocol_C.push((*gate_item_p->gib_outputs())[0],FALSE,TRUE);
				 }//if
			   }//if
			else
			   if ((*gate_item_p->gib_outputs())[0]->get_good_value() != line_item_p->get_good_value())
				return(INCONSISTENT);
			if (gate_item_p->gib_outputs()->gib_only_line()->get_bad_value() == VAL_X)
			   {
			   protocol_C.push(gate_item_p->gib_outputs()->gib_only_line(),FALSE,FALSE,FALSE,FALSE,
					FALSE,UNDEFINED_VALUE,UNDEFINED_VALUE,
					gate_item_p->gib_outputs()->gib_only_line()->get_bad_value(),
					line_item_p->get_bad_value());
			   (*gate_item_p->gib_outputs())[0]->set_bad_value(line_item_p->get_bad_value());
			   result_e = SUCCESS;
                           if ((*gate_item_p->gib_outputs())[0]->gib_type() == BOUND_LINE)
                                 {
                                 unjustified_bound_line_C.fuege_ein((*gate_item_p->gib_outputs())[0]);
                                 protocol_C.push((*gate_item_p->gib_outputs())[0],FALSE,TRUE);
				 }//if
			   }//if
			else
			   if ((*gate_item_p->gib_outputs())[0]->get_bad_value() != line_item_p->get_bad_value())
				return(INCONSISTENT);
			if (gate_item_p->gib_type() == OR)
				controlling_value_e = VAL_1;
			else
				controlling_value_e = VAL_0;
			d_front_changed_b = TRUE;
			break;
		case NAND:
		case NOR:
		case INVERTER:
			if (gate_item_p->gib_outputs()->gib_only_line()->get_good_value() == VAL_X)
			   {
			   protocol_C.push(gate_item_p->gib_outputs()->gib_only_line(),FALSE,FALSE,FALSE,FALSE,
					FALSE,gate_item_p->gib_outputs()->gib_only_line()->get_good_value(),
					mynot(line_item_p->get_good_value()));
			   (*gate_item_p->gib_outputs())[0]->set_good_value(mynot(line_item_p->get_good_value()));
			   result_e = SUCCESS;
                           if ((*gate_item_p->gib_outputs())[0]->gib_type() == BOUND_LINE)
                                 {
                                 unjustified_bound_line_C.fuege_ein((*gate_item_p->gib_outputs())[0]);
                                 protocol_C.push((*gate_item_p->gib_outputs())[0],FALSE,TRUE);
				 }//if
			   }//if
			else
			   if ((*gate_item_p->gib_outputs())[0]->get_good_value() != mynot(line_item_p->get_good_value()))
				return(INCONSISTENT);
			if (gate_item_p->gib_outputs()->gib_only_line()->get_bad_value() == VAL_X)
			   {
			   protocol_C.push(gate_item_p->gib_outputs()->gib_only_line(),FALSE,FALSE,FALSE,FALSE,
					FALSE,UNDEFINED_VALUE,UNDEFINED_VALUE,
					gate_item_p->gib_outputs()->gib_only_line()->get_bad_value(),
					mynot(line_item_p->get_bad_value()));
			   (*gate_item_p->gib_outputs())[0]->set_bad_value(mynot(line_item_p->get_bad_value()));
			   result_e = SUCCESS;
                           if ((*gate_item_p->gib_outputs())[0]->gib_type() == BOUND_LINE)
                                 {
                                 unjustified_bound_line_C.fuege_ein((*gate_item_p->gib_outputs())[0]);
                                 protocol_C.push((*gate_item_p->gib_outputs())[0],FALSE,TRUE);
				 }//if
			   }//if
			else
			   if ((*gate_item_p->gib_outputs())[0]->get_bad_value() != mynot(line_item_p->get_bad_value()))
				return(INCONSISTENT);
			if (gate_item_p->gib_type() == NOR)
				controlling_value_e = VAL_1;
			else
				controlling_value_e = VAL_0;
			d_front_changed_b = TRUE;
			break;
		case EXOR:
			controlling_value_e = VAL_X;
			if (line_p->get_good_value() != VAL_X)
			   {
			   if (line_p->get_bad_value() == VAL_X)
				{
				result_e = SUCCESS;
				line_p->set_bad_value(mynot(line_p->get_good_value()));
				protocol_C.push(line_p,FALSE,FALSE,FALSE,FALSE,FALSE,VAL_X,VAL_X,VAL_X,mynot(line_p->get_good_value()));
                           if (line_p->gib_type() == BOUND_LINE)
                                 {
                                 unjustified_bound_line_C.fuege_ein(line_p);
                                 protocol_C.push(line_p,FALSE,TRUE);
				 }//if
				break;
				}//if
			   }//if
			else if (line_p->get_bad_value() != VAL_X)
				{
				line_p->set_good_value(mynot(line_p->get_bad_value()));
				protocol_C.push(line_p,FALSE,FALSE,FALSE,FALSE,FALSE,VAL_X,mynot(line_p->get_bad_value()));
				if (line_p->gib_type() == BOUND_LINE)
                                 {
                                 unjustified_bound_line_C.fuege_ein(line_p);
                                 protocol_C.push(line_p,FALSE,TRUE);
				 }//if
				result_e = SUCCESS;
				break;
				}//if
			break;
		case BLOCK:
		  {
			cube_list_t *cube_list_p;
			belegung_t *belegung_p;
			int i,j,k;
			value_t set_val_e = VAL_X;
			int cube_parity_i = -1;

			controlling_value_e = VAL_X;
			if (line_p->get_good_value() != VAL_X)
			   {
			   if (line_p->get_bad_value() == VAL_X)
				{
				result_e = SUCCESS;
				line_p->set_bad_value(mynot(line_p->get_good_value()));
				protocol_C.push(line_p,FALSE,FALSE,FALSE,FALSE,FALSE,VAL_X,VAL_X,VAL_X,mynot(line_p->get_good_value()));
				if ((line_p->gib_type() == BOUND_LINE) && (unjustified_bound_line_C.element_in_liste(line_p) == FALSE))
                                 {
                                 unjustified_bound_line_C.fuege_ein(line_p);
                                 protocol_C.push(line_p,FALSE,TRUE);
				 }//if
				break;
				}//if
			   }//if
			else if (line_p->get_bad_value() != VAL_X)
				{ // nicht bei gut Wert durchfuehren => zerstoerung der weak tests
				line_p->set_good_value(mynot(line_p->get_bad_value()));
				protocol_C.push(line_p,FALSE,FALSE,FALSE,FALSE,FALSE,VAL_X,mynot(line_p->get_bad_value()));
				result_e = SUCCESS;
				if ((line_p->gib_type() == BOUND_LINE) && (unjustified_bound_line_C.element_in_liste(line_p) == FALSE))
                                 {
                                 unjustified_bound_line_C.fuege_ein(line_p);
                                 protocol_C.push(line_p,FALSE,TRUE);
				 }//if
				break;
				}//if
			if (line_item_p->get_good_value() == VAL_X)
				break;
			for(k=0;k<2;k++)
				{
				for (cube_list_p = gate_item_p->gib_cubes((k==0)?VAL_0:VAL_1);cube_list_p!=NULL;cube_list_p = cube_list_p->next_p)
				   {
				   for (j=0,belegung_p=cube_list_p->to_belegung_p;j<gate_item_p->no_inputs();j++,belegung_p=belegung_p->next_p)
					{
					if (((*gate_item_p->gib_inputs())[j]->get_good_value() == VAL_X) ||
					    (belegung_p->val_e == VAL_X))
						continue;
					if ((belegung_p->val_e == VAL_0) && ((*gate_item_p->gib_inputs())[j]->get_good_value() != VAL_0) && ((*gate_item_p->gib_inputs())[j]->get_good_value() != VAL_0S))
						break;
					if ((belegung_p->val_e == VAL_1) && ((*gate_item_p->gib_inputs())[j]->get_good_value() != VAL_1) && ((*gate_item_p->gib_inputs())[j]->get_good_value() != VAL_1S))
						break;
					}//for
				   if (j==gate_item_p->no_inputs())
				     for (i=0,belegung_p=cube_list_p->to_belegung_p;i<gate_item_p->no_inputs();i++,belegung_p=belegung_p->next_p)
					{
					if (belegung_p->val_e == VAL_X)
						continue;
					if ((*gate_item_p->gib_inputs())[i]->get_bad_value() != VAL_X)
						if (((k==0) && (belegung_p->val_e == VAL_0)) ||
						    ((k==1) && (belegung_p->val_e == VAL_1)))
							if (cube_parity_i < 1)
								cube_parity_i = 0;
							else
								cube_parity_i = 2;
						else if (((k==0) && (belegung_p->val_e == VAL_1)) ||
							 ((k==1) && (belegung_p->val_e == VAL_0)))
							if ((cube_parity_i == -1) || (cube_parity_i == 1))
								cube_parity_i = 1;
							else
								cube_parity_i = 2;
					}//for
				   if (cube_parity_i == 2)
					break;
				   }//for
				if ((cube_parity_i < 2) && (cube_parity_i > -1))
					break;
				cube_parity_i = -1;
				}//for
			if (cube_parity_i == 2)
				break;
			if (cube_parity_i == -1)
				return(INCONSISTENT);
			if (cube_parity_i == 0)
				set_val_e = line_item_p->get_good_value();
			else
				set_val_e = mynot(line_item_p->get_good_value());
			line_p->set_good_value(set_val_e);
			line_p->set_bad_value(mynot(set_val_e));
			if ((line_p->gib_type() == BOUND_LINE) && (unjustified_bound_line_C.element_in_liste(line_p) == FALSE))
                                 {
                                 unjustified_bound_line_C.fuege_ein(line_p);
                                 protocol_C.push(line_p,FALSE,TRUE);
				 }//if
			protocol_C.push(line_p,FALSE,FALSE,FALSE,FALSE,FALSE,VAL_X,set_val_e,VAL_X,mynot(set_val_e));
			result_e = SUCCESS;
			break;
		  }
		case D_FLIP_FLOP:
		case D_LATCH:
			error("unique_sensitisation: not implemented for sequential elements");
		case FANOUT:
			error("unique_sensitisation: not feasible at fanout-gates");
		case UNDEFINED_GATE:
			error("unique_sensitisation: undefined gate in circuit");
		}//switch
	if (controlling_value_e == VAL_X)
	   {
	   if ((*gate_item_p->gib_outputs())[0]->get_good_value() == VAL_X)
		{	// EXOR, BLOCK
		d_front_C.fuege_ein(line_item_p);
		protocol_C.push(line_item_p,FALSE,FALSE,FALSE,TRUE);
		break;
		}//if
	   }//if
	else			// setzen der off-path Signale
	   {forall(hilf_line_p,*gate_item_p->gib_inputs())
		{
		if (hilf_line_p == line_item_p)
			continue;
		if (hilf_line_p->get_good_value() != VAL_X)
			if ((hilf_line_p->get_good_value() == mynot(controlling_value_e)) ||
			    (hilf_line_p->get_bad_value() == line_item_p->get_bad_value()) ||
			    ((hilf_line_p->in_output_kegel() == TRUE) &&
			     (hilf_line_p->get_good_value() == line_item_p->get_good_value())))
				continue;
			else
				return(INCONSISTENT);
		else if ((hilf_line_p->in_output_kegel() == FALSE) ||
			 (line_item_p->get_good_value() == mynot(controlling_value_e)))
			{
			hilf_line_p->set_good_value(mynot(controlling_value_e));
			result_e = SUCCESS;
			if (hilf_line_p->gib_type() == BOUND_LINE)
				{
				unjustified_bound_line_C.fuege_ein(hilf_line_p);
				protocol_C.push(hilf_line_p,FALSE,TRUE,FALSE,FALSE,FALSE,VAL_X,mynot(controlling_value_e));
				}//if
			else
				protocol_C.push(hilf_line_p,FALSE,FALSE,FALSE,FALSE,FALSE,VAL_X,mynot(controlling_value_e));
			}//else if
		}}//forall
	line_item_p = gate_item_p->gib_outputs()->gib_only_line();
	if (line_item_p->gib_type() == BOUND_LINE)
		{
		unjustified_bound_line_C.fuege_ein(line_item_p);
		protocol_C.push(line_item_p,FALSE,TRUE,FALSE,TRUE);
		}//if
	else
		protocol_C.push(line_item_p,FALSE,FALSE,FALSE,TRUE);
	implication_line_C.fuege_ein(line_item_p);
	d_front_C.fuege_ein(line_item_p);
	if ((line_item_p->is_primary_output() == FALSE) &&
	    (line_item_p->nachfolger()->gib_only_gate()->gib_type() == FANOUT))
		break;//while
	}//while
if ((result_e == SUCCESS) || (d_front_changed_b == FALSE) || (phase2_b == FALSE))
	return(result_e);
off_path_signals_C.loesche_liste();
off_path_signals_tmp_C.loesche_liste();
off_path_signals_save_C.loesche_liste();
{forall(gate_item_p,gates_C)
	{
	gate_item_p->reset_branch();
	gate_item_p->dyn_path_exist_b = FALSE;
	}}//forall
{forall(line_item_p,d_front_C)
	{
	if (line_item_p->nachfolger()->gib_only_gate()->search_off_path_signals(hilf_b) == FALSE)
		continue;
			// Index kann hier noch gespeichert werden
	hilf_b = TRUE;
	break;
	}}//forall
if (hilf_b == FALSE)
	return(INCONSISTENT);
if (off_path_signals_tmp_C.is_empty() == TRUE)
	return(result_e);
{forall(line_item_p,off_path_signals_tmp_C)
	off_path_signals_save_C.fuege_ein(line_item_p);
	}//forall
off_path_signals_tmp_C.loesche_liste();
{forall(line_item_p,d_front_C)
	while (line_item_p->nachfolger()->gib_only_gate()->search_off_path_signals(hilf_b) == TRUE)
		{
		if (hilf_b == TRUE)
			{
			hilf_b = FALSE;
			break;
			}//if
		else
			hilf_b = TRUE;
		forall(hilf_line_p,off_path_signals_save_C)
			{
			forall(line2_p,off_path_signals_tmp_C)
				{
				if (((hilf_line_p->is_primary_input() == TRUE) &&
				     (line2_p->is_primary_input() == FALSE)) ||
				    ((hilf_line_p->is_primary_input() == FALSE) &&
				     (line2_p->is_primary_input() == TRUE)))
					continue;
				if ((hilf_line_p->is_primary_input() == TRUE) &&
				    (line2_p->is_primary_input() == TRUE))
					{
					if (hilf_line_p == line2_p)
						off_path_signals_C.fuege_ein(hilf_line_p);
					continue;
					}//if
				if ((hilf_line_p->vorgaenger()->gib_only_gate() ==
				     line2_p->vorgaenger()->gib_only_gate()) &&
				    ((hilf_line_p->nachfolger()->gib_only_gate()->gib_type() ==
				      line2_p->nachfolger()->gib_only_gate()->gib_type()) ||
				     ((hilf_line_p->nachfolger()->gib_only_gate()->gib_type() == NAND) &&
				      (line2_p->nachfolger()->gib_only_gate()->gib_type() == AND)) ||
				     ((hilf_line_p->nachfolger()->gib_only_gate()->gib_type() == AND) &&
				      (line2_p->nachfolger()->gib_only_gate()->gib_type() == NAND)) ||
				     ((hilf_line_p->nachfolger()->gib_only_gate()->gib_type() == NOR) &&
				      (line2_p->nachfolger()->gib_only_gate()->gib_type() == OR)) ||
				     ((hilf_line_p->nachfolger()->gib_only_gate()->gib_type() == OR) &&
				      (line2_p->nachfolger()->gib_only_gate()->gib_type() == NOR))))
					off_path_signals_C.fuege_ein(hilf_line_p);
				}//forall
			}//forall
		if (off_path_signals_C.is_empty() == TRUE)
			return(result_e);
		off_path_signals_save_C.loesche_liste();
		{forall(hilf_line_p,off_path_signals_C)
		   if (off_path_signals_save_C.element_in_liste(hilf_line_p) == FALSE)
			off_path_signals_save_C.fuege_ein(hilf_line_p);}
		off_path_signals_C.loesche_liste();
		off_path_signals_tmp_C.loesche_liste();
		}//while
	}//forall
{forall(line_item_p,off_path_signals_save_C)
	{
	gate_item_p = (*line_item_p->nachfolger())[0];
	switch (gate_item_p->gib_type())
		{
		case AND:
		case NAND:
			controlling_value_e = VAL_0;
			break;
		case OR:
		case NOR:
			controlling_value_e = VAL_1;
			break;
		case EXOR:
		case BLOCK:
		case FANOUT:
		case DRIVER:
		case INVERTER:
			controlling_value_e = VAL_X;
			break;
		case D_FLIP_FLOP:
		case D_LATCH:
			error("unique_sensitisation: not implemented for sequential elements");
		case UNDEFINED_GATE:
			error("unique_sensitisation: undefined gate in circuit");
		}//switch
	if (controlling_value_e != VAL_X)		// nicht EXOR, BLOCK, FANOUT
				// setzen der off-path Signale
		{
		result_e = SUCCESS;
		line_item_p->set_good_value(mynot(controlling_value_e));
		if (line_item_p->gib_type() == BOUND_LINE)
			{
			unjustified_bound_line_C.fuege_ein(line_item_p);
			protocol_C.push(line_item_p,FALSE,TRUE,FALSE,FALSE,FALSE,VAL_X,mynot(controlling_value_e));
			}//if
		else
			protocol_C.push(line_item_p,FALSE,FALSE,FALSE,FALSE,FALSE,VAL_X,mynot(controlling_value_e));
		}//if
	}}//forall
return(result_e);
}//unique_sensitisation

//*********************************************************************
//
//   FUNCTION-NAME: justify_free_lines
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Alle Unjustified Free Lines werden justified
//	Hier kann es nicht mehr zu Wiederspruechen kommen.
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void justify_free_lines()
{
line_C *line_item_p;
gate_C *gate_item_p;
value_t val_e;

{forall(line_item_p,free_lines_C)
   if ((line_item_p->is_primary_input() == FALSE) && (line_item_p->get_good_value() != VAL_X))
	if (line_item_p->vorgaenger()->gib_only_gate()->backward_implication() == INCONSISTENT)
		error("justify_free_lines: implication's not in consistency");}
{forall(line_item_p,free_lines_C)
	if (line_item_p->get_good_value() != VAL_X)
		unjustified_free_line_C.fuege_ein(line_item_p);}
{forall(line_item_p,unjustified_free_line_C)
	{
	line_C *hilf_line_p;

	if (line_item_p->is_primary_input() == TRUE)
		continue;
	gate_item_p = line_item_p->vorgaenger()->gib_only_gate();
	val_e = line_item_p->get_good_value();
	if (val_e == VAL_X)
		warning("justify_free_lines: X at the output of some gate");
	else if (val_e == UNDEFINED_VALUE)
		error("justify_free_lines: undefined value found");
	switch (gate_item_p->gib_type())
		{
		case DRIVER:
		case AND:
			{forall(hilf_line_p,*gate_item_p->gib_inputs())
				if (hilf_line_p->get_good_value() == VAL_X)
					{
					hilf_line_p->set_good_value(val_e);
					if (hilf_line_p->is_primary_input() == FALSE)
						unjustified_free_line_C.fuege_ein(hilf_line_p);
					}//if
			}//forall
			break;
		case INVERTER:
		case NAND:
			{forall(hilf_line_p,*gate_item_p->gib_inputs())
				if (hilf_line_p->get_good_value() == VAL_X)
					{
					hilf_line_p->set_good_value(mynot(val_e));
					if (hilf_line_p->is_primary_input() == FALSE)
						unjustified_free_line_C.fuege_ein(hilf_line_p);
					}//if
			}//forall
			break;
		case OR:
			{forall(hilf_line_p,*gate_item_p->gib_inputs())
				if (hilf_line_p->get_good_value() == VAL_X)
					{
					hilf_line_p->set_good_value(val_e);
					if (hilf_line_p->is_primary_input() == FALSE)
						unjustified_free_line_C.fuege_ein(hilf_line_p);
					}//if
			}//forall
			break;
		case NOR:
			{forall(hilf_line_p,*gate_item_p->gib_inputs())
				if (hilf_line_p->get_good_value() == VAL_X)
					{
					hilf_line_p->set_good_value(mynot(val_e));
					if (hilf_line_p->is_primary_input() == FALSE)
						unjustified_free_line_C.fuege_ein(hilf_line_p);
					}//if
			}//forall
			break;
		case EXOR:
			if (val_e == VAL_1)
				{
				boolean zweiter_lauf_b = FALSE;
					
				{forall(hilf_line_p,*gate_item_p->gib_inputs())
					if (hilf_line_p->get_good_value() != VAL_X)
						continue;
					else
						{
						if (zweiter_lauf_b == FALSE)
							hilf_line_p->set_good_value(VAL_1);
						else
							hilf_line_p->set_good_value(VAL_0);
						if (hilf_line_p->is_primary_input() == FALSE)
							unjustified_free_line_C.fuege_ein(hilf_line_p);
						}//else
					}//forall
				}//if
			else
				{forall(hilf_line_p,*gate_item_p->gib_inputs())
					if (hilf_line_p->get_good_value() != VAL_X)
						continue;
					else
						{
						hilf_line_p->set_good_value(val_e);
						if (hilf_line_p->is_primary_input() == FALSE)
							unjustified_free_line_C.fuege_ein(hilf_line_p);
						}//else
				}
			break;
		case BLOCK:
			cube_list_t *cube_list_p;
			belegung_t *belegung_p;
			int l;
			
			switch(val_e)
				{
				case VAL_1:
					cube_list_p = gate_item_p->gib_cubes(VAL_1);
					break;
				case VAL_0:
					cube_list_p = gate_item_p->gib_cubes(VAL_0);
					break;
				}//switch
			for (;cube_list_p != NULL;cube_list_p = cube_list_p->next_p)
			   {
			   for (l=0,belegung_p = cube_list_p->to_belegung_p;belegung_p != NULL;l++,belegung_p = belegung_p->next_p)
				{
				if ((belegung_p->val_e == VAL_X) ||
				    ((*gate_item_p->gib_inputs())[l]->get_good_value() == VAL_X))
					continue;
				if ((*gate_item_p->gib_inputs())[l]->get_good_value() == belegung_p->val_e)
					continue;
				if ((*gate_item_p->gib_inputs())[l]->get_good_value() == belegung_p->val_e)
					continue;
				break;
				}//for
			   if (belegung_p == NULL)
				break;
			   }//for
			if (cube_list_p == NULL)
				error("justify_free_lines:  no cube fits to the actual billeting");
			for (l=0,belegung_p = cube_list_p->to_belegung_p;belegung_p != NULL;l++,belegung_p = belegung_p->next_p)
			   {
			   hilf_line_p = (*gate_item_p->gib_inputs())[l];
			   if (hilf_line_p->get_good_value() == VAL_X)
				{
				val_e = belegung_p->val_e;
				if (val_e == VAL_X)
					val_e = VAL_0;
				hilf_line_p->set_good_value(val_e);
				if (hilf_line_p->is_primary_input() == FALSE)
					unjustified_free_line_C.fuege_ein(hilf_line_p);
				}//if
			   }//for
			break;
		case D_FLIP_FLOP:
		case D_LATCH:
			error("justify_free_lines:  sequential gates can`t  be managed");
		case FANOUT:
			error("justify_free_lines: fanout-gates aren't allowed to have free lines as followers");
		case UNDEFINED_GATE:
			error("justify_free_lines: undefined gate found");
		}//switch
	}}//forall
unjustified_free_line_C.loesche_liste();
}//justify_free_lines

//*********************************************************************
//
//   FUNCTION-NAME: continuation_of_backtrace_meaningful
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Es wird festgestellt, ob es Sinvoll ist, die Multibacktrace
//	Funktion mit den aktuellen Objectives weiter durchzufuehren.
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: boolean: TRUE, wenn Multibacktrace weiter durchgefuehrt wird.
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
boolean continuation_of_backtrace_meaningful()	
{
line_C *line_item_p,*line2_p = NULL,*nachfolger_item_p,*nachfolger2_p = NULL;
gate_C *gate_item_p;

if (ujust_bound_changed_b == TRUE)
	return(FALSE);
if (fault_signal_at_output() == TRUE)
	return(TRUE);
forall(line_item_p,d_front_C)
	{
	gate_item_p = line_item_p->nachfolger()->gib_only_gate();
	forall(nachfolger_item_p,*gate_item_p->gib_outputs())
		{
		if ((nachfolger_item_p->get_good_value() != VAL_X) && (nachfolger_item_p->get_bad_value() != VAL_X))
			continue;
		if (line2_p == NULL)
			{
			line2_p = line_item_p;
			nachfolger2_p = nachfolger_item_p;
			}//if
		if (nachfolger2_p->gib_observaby() > nachfolger_item_p->gib_observaby())
			{
			line2_p = line_item_p;
			nachfolger2_p = nachfolger_item_p;
			}//if
		}//if
	}//forall
if ((nachfolger2_p == NULL) ||
    (nachfolger2_p->vorgaenger()->gib_only_gate() == d_front_gate_in_initial_objectives_p))
	return(TRUE);
return(FALSE);
}//continuation_of_backtrace_meaningful

//*********************************************************************
//
//   FUNCTION-NAME: assign_value_to_final_objective
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Der Line mit der Final Objective wird der Wert
//	zugewiesen, der den meisten Anforderungen entspricht.
//
//   FORMAL PARAMETERS: objective_type_u objective_u: die entsprechende Objective
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: boolean: TRUE, wenn Multibacktrace weiter durchgefuehrt wird.
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void assign_value_to_final_objective(objective_type_u &objective_u)
{
if (objective_u.line_p->get_good_value() != VAL_X)
	error("assign_value_to_final_objective: signal was already covered");
if ((objective_u.n0_i < objective_u.n1_i) || ((objective_u.n0_i == objective_u.n1_i) &&
    (objective_u.line_p->gib_one_contby() < objective_u.line_p->gib_zero_contby())))
	{
	protocol_C.push(objective_u.line_p,TRUE,FALSE,FALSE,FALSE,FALSE,objective_u.line_p->get_good_value(),VAL_1);
	objective_u.line_p->set_good_value(VAL_1);
	}//if
else
	{
	protocol_C.push(objective_u.line_p,TRUE,FALSE,FALSE,FALSE,FALSE,objective_u.line_p->get_good_value(),VAL_0);
	objective_u.line_p->set_good_value(VAL_0);
	}//else
if (objective_u.line_p->gib_type() == BOUND_LINE)
	{
	unjustified_bound_line_C.fuege_ein(objective_u.line_p);
	protocol_C.push(objective_u.line_p,FALSE,TRUE);
	}//if
implication_line_C.fuege_ein(objective_u.line_p);
#if DEBUG
optional_lines_C.insert(objective_u.line_p,objective_u.line_p->get_test_value(),TRUE);
#endif
if (objective_u.line_p->in_input_kegel() == TRUE)
	input_kegel_changed_b = TRUE;
}//assign_value_to_final_objective

//*********************************************************************
//
//   FUNCTION-NAME: number_of_gates_in_d_front
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Die Anzahl der Gatter in der D Front wird bestimmt. Wenn
//	sie groesser als 1 ist, werden Strukturbezogene Sensibilisierungsmassnahmen
//	durchgefuehrt.
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: boolean: int: Gatteranzahl in D Front
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
int number_of_gates_in_d_front()
{
gate_C *gate_item_p;
line_C *line_item_p,*line2_p;
int ergebnis_i = 0;
value_t non_contr_val_e = UNDEFINED_VALUE;
boolean erfolg_b = FALSE;

tmp_gate_liste_C.loesche_liste();

{forall(line2_p,d_front_C)
	{
	gate_item_p = line2_p->nachfolger()->gib_only_gate();
	{forall(line_item_p,*gate_item_p->gib_outputs())
		if (((line_item_p->get_good_value() == VAL_X) || (line_item_p->get_bad_value() == VAL_X)) &&
		    (tmp_gate_liste_C.element_in_liste(gate_item_p) == FALSE))
			{
			tmp_gate_liste_C.fuege_ein(gate_item_p);
			ergebnis_i++;
			}//if
	}//forall
	}}//forall
return(ergebnis_i);
}//number_of_gates_in_d_front

//*********************************************************************
//
//   FUNCTION-NAME: dyn_learn
//
//   CLASS: line_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: Juli 92
//
//   FUNCTIONAL-DESCRIPTION: Ueberpruefung des Lernkriteriums fuer ein Signal.
//	Gegebenenfalls Lernen der globalen Implikation.
//
//   FORMAL PARAMETERS: line_C *line_item_p: Die Line, fuer die Implikation durchgefuehrt wurde.
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
fan_result_t line_C::dyn_learn(line_C *line_item_p)
{
int i = (good_value_e == VAL_1)?0:1;

if ((prime_input_b == TRUE) || (line_item_p == this) ||
    (good_value_e == VAL_X) || (set_in_forw_imp_b == FALSE) ||
    (global_implications[i].element_in_liste(line_item_p,mynot(line_item_p->get_good_value())) == TRUE) ||
    (tmp_global_implications[i].element_in_liste(line_item_p,mynot(line_item_p->get_good_value())) == TRUE))
	return(NO_IMPLICATION);
switch ((*vorgaenger_p)[0]->gib_type())
	{
	case AND:
	case NOR:
		if (good_value_e == VAL_0)
                        return(NO_IMPLICATION);
                break;
	case NAND:
	case OR:
		if (good_value_e == VAL_1)
                        return(NO_IMPLICATION);
                break;
	case EXOR:
	case BLOCK:
		if ((*vorgaenger_p)[0]->no_inputs() == 1)
                        return(NO_IMPLICATION);
		break;
	case INVERTER:
	case DRIVER:
	case FANOUT:
		return(NO_IMPLICATION);
	default:
		error("learn: Unknown gate type found");
	}//switch
tmp_global_implications[i].fuege_ein(line_item_p,mynot(line_item_p->get_good_value()));
return(SUCCESS);
}//dyn_learn

//*********************************************************************
//
//   FUNCTION-NAME: dyn_try_learning
//
//   CLASS: line_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: Juli 92
//
//   FUNCTIONAL-DESCRIPTION: Lernen der globalen Implikation fuer ein Signal
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: fan_result_t: SUCCESS, falls gelernt werden konnte
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
fan_result_t line_C::dyn_try_learning()
{
protocol_type_C *prot_item_p;
boolean zero_inconsistent_b = FALSE,one_inconsistent_b = FALSE;
fan_result_t res_e = NO_IMPLICATION;

	// VAL_0
set_good_value(VAL_0);
protocol_C.push(this,TRUE,FALSE,FALSE,FALSE,FALSE,VAL_X,VAL_0);
if (((prime_output_b == FALSE) && (nachfolger()->gib_only_gate()->forward_implication() == INCONSISTENT)) ||
    ((prime_input_b == FALSE) && (vorgaenger()->gib_only_gate()->backward_implication() == INCONSISTENT)))
	zero_inconsistent_b = TRUE;
	//clear
for(;;)
	{
	prot_item_p = protocol_C.pop();
	if (prot_item_p == NULL)
		error("dyn_learning: something is wrong");
	if (prot_item_p->optional() == TRUE)
		break;
	if (prot_item_p->old_good_value() != UNDEFINED_VALUE)
		{
		prot_item_p->line_p->set_good_value(prot_item_p->old_good_value());
		if (prot_item_p->line_p->dyn_learn(this) == SUCCESS)
			res_e = SUCCESS;
		prot_item_p->line_p->reset_in_forw_imp();
		}//if
	if (prot_item_p->old_bad_value() != UNDEFINED_VALUE)
		prot_item_p->line_p->set_bad_value(prot_item_p->old_bad_value());
	if (prot_item_p->ujust_bound_in() == TRUE)
		unjustified_bound_line_C.fuege_aus(prot_item_p->line_p);
	if (prot_item_p->ujust_bound_out() == TRUE)
		unjustified_bound_line_C.fuege_ein(prot_item_p->line_p);
	if (prot_item_p->d_front_in() == TRUE)
		d_front_C.fuege_aus(prot_item_p->line_p);
	if (prot_item_p->d_front_out() == TRUE)
		d_front_C.fuege_ein(prot_item_p->line_p);
	}//forever
	// VAL_1
set_good_value(VAL_1);
protocol_C.push(this,TRUE,FALSE,FALSE,FALSE,FALSE,VAL_X,VAL_1);
if (((prime_output_b == FALSE) && (nachfolger()->gib_only_gate()->forward_implication() == INCONSISTENT)) ||
    ((prime_input_b == FALSE) && (vorgaenger()->gib_only_gate()->backward_implication() == INCONSISTENT)))
	one_inconsistent_b = TRUE;
	//clear
for(;;)
	{
	prot_item_p = protocol_C.pop();
	if (prot_item_p == NULL)
		error("dyn_learning: something is wrong");
	if (prot_item_p->optional() == TRUE)
		break;
	if (prot_item_p->old_good_value() != UNDEFINED_VALUE)
		{
		prot_item_p->line_p->set_good_value(prot_item_p->old_good_value());
		if (prot_item_p->line_p->dyn_learn(this) == SUCCESS)
			res_e = SUCCESS;
		prot_item_p->line_p->reset_in_forw_imp();
		}//if
	if (prot_item_p->old_bad_value() != UNDEFINED_VALUE)
		prot_item_p->line_p->set_bad_value(prot_item_p->old_bad_value());
	if (prot_item_p->ujust_bound_in() == TRUE)
		unjustified_bound_line_C.fuege_aus(prot_item_p->line_p);
	if (prot_item_p->ujust_bound_out() == TRUE)
		unjustified_bound_line_C.fuege_ein(prot_item_p->line_p);
	if (prot_item_p->d_front_in() == TRUE)
		d_front_C.fuege_aus(prot_item_p->line_p);
	if (prot_item_p->d_front_out() == TRUE)
		d_front_C.fuege_ein(prot_item_p->line_p);
	}//forever
set_good_value(VAL_X);
if ((one_inconsistent_b == TRUE) && (zero_inconsistent_b == TRUE))
	return(INCONSISTENT);
if (zero_inconsistent_b == TRUE)
	{
	set_good_value(VAL_1);
	protocol_C.push(this,FALSE,FALSE,FALSE,FALSE,FALSE,VAL_X,VAL_1);
	}//if
if (one_inconsistent_b == TRUE)
	{
	set_good_value(VAL_0);
	protocol_C.push(this,FALSE,FALSE,FALSE,FALSE,FALSE,VAL_X,VAL_0);
	}//if
if ((one_inconsistent_b == TRUE) || (zero_inconsistent_b == TRUE))
	{
	implication_line_C.fuege_ein(this);
	if (gib_type() == BOUND_LINE)
		{
		unjustified_bound_line_C.fuege_ein(this);
		protocol_C.push(this,FALSE,TRUE);
		}//if
	res_e = SUCCESS;
	}//if
return(res_e);
}//dyn_try_learning

//*********************************************************************
//
//   FUNCTION-NAME: dyn_learning
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: Juli 92
//
//   FUNCTIONAL-DESCRIPTION: Dynamische Lernphase nach Schulz/Auth, mit der die
//	Redundanzidentifizierung schneller moeglich ist.
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: fan_result_t: INCONSISTENT, wenn backtracking erforderlich ist.
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
fan_result_t dyn_learning()
{
line_C *line_item_p;
fan_result_t res_e,result_e = NO_IMPLICATION;
line_value_t *hilf_p;

hilf_liste_C.loesche_liste();
{forall(line_item_p,lines_C)
	line_item_p->reset_in_forw_imp();}
forall(line_item_p,lines_C)
	if (((line_item_p->is_primary_input() == TRUE) ||
	    (((*line_item_p->vorgaenger())[0]->gib_type() != FANOUT) &&
	     ((*line_item_p->vorgaenger())[0]->gib_type() != INVERTER) &&
	     ((*line_item_p->vorgaenger())[0]->gib_type() != DRIVER))) &&
	    (line_item_p->get_good_value() == VAL_X))
		{
		res_e = line_item_p->dyn_try_learning();
		if (res_e == INCONSISTENT)
			return(INCONSISTENT);
		else if (res_e == SUCCESS)
			result_e = SUCCESS;
		}//if
if (hilf_liste_C.is_empty() == FALSE)
   {forall(hilf_p,hilf_liste_C)
	{
	hilf_p->line_p->set_good_value(hilf_p->val_e);
	if (hilf_p->line_p->gib_type() == BOUND_LINE)
                {
                unjustified_bound_line_C.fuege_ein(hilf_p->line_p);
		protocol_C.push(hilf_p->line_p,FALSE,TRUE,FALSE,FALSE,FALSE,VAL_X,hilf_p->val_e);
		}//if
	else
		protocol_C.push(hilf_p->line_p,FALSE,FALSE,FALSE,FALSE,FALSE,VAL_X,hilf_p->val_e);
	implication_line_C.fuege_ein(hilf_p->line_p);
	}}//forall
return(result_e);
}//dyn_learning_implications

//*********************************************************************
//
//   FUNCTION-NAME: reset_dyn_learning
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: Juli 92
//
//   FUNCTIONAL-DESCRIPTION: Werte, die beim dynamischen Lernen entstanden sind werden
//	zurueckgesetzt.
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: fan_result_t: INCONSISTENT, wenn backtracking erforderlich ist.
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void reset_dyn_learning()
{
line_C *line_item_p;

forall(line_item_p,lines_C)
	line_item_p->reset_dyn_learning();
}//reset_dyn_learning
