//*********************************************************************
//
//   FILE-NAME: final_objective.c
//
//   CREATION-DATE: August 91
//
//   FILE-DESCRIPTION: Dieses File beinhaltet die Funktionen fuer das Auffinden der
//	final objectives.
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
#include "extern.h"
#include "protocol_stack_C.h"
#include "objective_liste_C.h"
#include "line_liste_C.h"
#include "gate_liste_C.h"
#include "line_C.h"
#include "gate_C.h"

	// Globale Variablen
extern line_C *ziel_line_p;
extern line_liste_C unjustified_bound_line_C;
extern line_liste_C d_front_C;
extern line_liste_C implication_line_C;
extern protocol_stack_C protocol_C;
objective_liste_C current_objectives_C;
objective_liste_C head_objectives_C;
objective_liste_C fanout_objectives_C;
objective_type_u final_objective_u;
gate_C *d_front_gate_in_initial_objectives_p;

	// Inline Functions
extern boolean phase2_b;
extern value_t mynot(value_t);
extern void add_constraints(gate_C*,int[]);
//*********************************************************************
//
//   FUNCTION-NAME: let_objectives_be_empty
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Initialisert alle Objective Listen leer
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void let_objectives_be_empty()
{
current_objectives_C.loesche_liste();
head_objectives_C.loesche_liste();
fanout_objectives_C.loesche_liste();
}//let_objectives_be_empty

//*********************************************************************
//
//   FUNCTION-NAME: let_unjustified_lines_be_initial_objectives
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Initialisert Intiale Objectives mit Unjustified Lines
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void let_unjustified_lines_be_initial_objectives()
{
line_C *line_item_p;
value_t val_e;

forall(line_item_p,unjustified_bound_line_C)
	{
	val_e = line_item_p->get_good_value();
	if ((val_e == VAL_X) || (val_e == UNDEFINED_VALUE))
		error("let_unjustified_lines_be_initial_objectives: Unjustified Bound Line has valueX or undefined value");
	if (((*line_item_p->vorgaenger())[0]->gib_type() == BLOCK) && (line_item_p->in_output_kegel() == TRUE) &&
	    (line_item_p->get_bad_value() == mynot(val_e)))
	      {
	      current_objectives_C.fuege_ein(line_item_p,1,1);
	      continue;
	      }//if
	if (val_e == VAL_1)
		current_objectives_C.fuege_ein(line_item_p,0,1);
	else
		current_objectives_C.fuege_ein(line_item_p,1,0);
	}//forall
}//let_unjustified_lines_be_initial_objectives

//*********************************************************************
//
//   FUNCTION-NAME: add_gate_in_d_front_to_initial_objectives
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Der Nachfolger der Lines in der D Front mit der besten Observability
//	wird in die Initialen Objectives eingefuegt.
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void add_gate_in_d_front_to_initial_objectives()
{
line_C *line_item_p,*line2_p = NULL,*nachfolger_item_p,*nachfolger2_p = NULL;
gate_C *gate_item_p;

forall(line_item_p,d_front_C)
	{
	gate_item_p = line_item_p->nachfolger()->gib_only_gate();
	forall(nachfolger_item_p,*gate_item_p->gib_outputs())
		{
		if ((nachfolger_item_p->get_good_value() != VAL_X) && (nachfolger_item_p->get_bad_value() != VAL_X))
			continue;
		if (line2_p == NULL)
			{
			line2_p = line_item_p;
			nachfolger2_p = nachfolger_item_p;
			}//if
		if (nachfolger2_p->gib_observaby() > nachfolger_item_p->gib_observaby())
			{
			line2_p = line_item_p;
			nachfolger2_p = nachfolger_item_p;
			}//if
		}//if
	}//forall
if (line2_p == NULL)
	error("add_gate_in_d_front_to_initial_objectives: D Front empty");
else
   switch (nachfolger2_p->vorgaenger()->gib_only_gate()->gib_type())
	{
	case AND:
	case NOR:
		current_objectives_C.fuege_ein(nachfolger2_p,0,1);
		break;
	case OR:
	case NAND:
		current_objectives_C.fuege_ein(nachfolger2_p,1,0);
		break;
	case EXOR:
		if (line2_p == (*nachfolger2_p->vorgaenger()->gib_only_gate()->gib_inputs())[1])
			line_item_p = (*nachfolger2_p->vorgaenger()->gib_only_gate()->gib_inputs())[0];
		else
			line_item_p = (*nachfolger2_p->vorgaenger()->gib_only_gate()->gib_inputs())[1];
		if (line_item_p->gib_one_contby() < line_item_p->gib_zero_contby())
			current_objectives_C.fuege_ein(nachfolger2_p,0,1);
		else
			current_objectives_C.fuege_ein(nachfolger2_p,1,0);
		break;
	case BLOCK:
		if (line2_p->get_good_value() == VAL_1)
			current_objectives_C.fuege_ein(nachfolger2_p,1,0);
		else
			current_objectives_C.fuege_ein(nachfolger2_p,0,1);
		break;
	case DRIVER:
	case INVERTER:
		error("add_gate_in_d_front_to_initial_objectives: inverter and driver are not valid `cause of implication");
	case D_FLIP_FLOP:
	case D_LATCH:
		error("add_gate_in_d_front_to_initial_objectives: not implemented for sequential elements");
	case FANOUT:
		error("add_gate_in_d_front_to_initial_objectives: fanout gates cannot be here");
	case UNDEFINED_GATE:
		error("add_gate_in_d_front_to_initial_objectives: undefined gate in circuit");
	}//switch
d_front_gate_in_initial_objectives_p = nachfolger2_p->vorgaenger()->gib_only_gate();
}//add_gate_in_d_front_to_initial_objectives

//*********************************************************************
//
//   FUNCTION-NAME: objective_line_follows_fanout
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Stellt fest, ob die Line des Objectives ein
//	Nachfolger von einem Fanout ist.
//
//   FORMAL PARAMETERS: objective_type_u objective_u: das entsprechende Objective
//
//   IMPLICIT INPUT/OUTPUT: void
//
//   FUNCTION-VALUE: boolean: TRUE, wenn die Vorgaenger Line ein Fanout ist.
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
boolean objective_line_follows_fanout(objective_type_u &objective_u)
{
if ((objective_u.line_p->is_primary_input() == TRUE) ||
    (objective_u.line_p->vorgaenger()->gib_only_gate()->gib_type() != FANOUT))
	return(FALSE);
else
	return(TRUE);
}//objective_line_follows_fanout

//*********************************************************************
//
//   FUNCTION-NAME: add_n0_and_n1_to_fanout_objective
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Fuegt das Objective, das den Fanout hat in die Fanout
//	Objectives ein, falls es nicht schon dort vorhanden ist. Sonst werden die 0
//	und die 1 Anforderungen zu dem entsprechenden Objective addiert.
//	Fuer alle Vorgaenger der Objective Line werden die n0 und n1
//	berechnet.
//
//   FORMAL PARAMETERS: objective_type_u objective_u: das entsprechende Objective
//
//   IMPLICIT INPUT/OUTPUT: void
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void add_n0_and_n1_to_fanout_objective(objective_type_u &objective_u)
{
line_C *line_item_p;
int upper_i=fanout_objectives_C.no_elements(),lower_i=-1,middle_i=-1;

line_item_p = objective_u.line_p->vorgaenger()->gib_only_gate()->gib_inputs()->gib_only_line();
while (lower_i < upper_i - 1)
	{
	middle_i = (upper_i+lower_i)/2;
	if ((fanout_objectives_C[middle_i]->line_p->gib_plane_no() < line_item_p->gib_plane_no()) ||
	    ((fanout_objectives_C[middle_i]->line_p->gib_plane_no() == line_item_p->gib_plane_no()) &&
	     (fanout_objectives_C[middle_i]->line_p->gib_number() < line_item_p->gib_number())))
		lower_i = middle_i;
	else if (fanout_objectives_C[middle_i]->line_p == line_item_p)
		{
		upper_i = middle_i;
		break;
		}//else if
	else
		upper_i = middle_i;
	}//while
if ((middle_i != -1) && (fanout_objectives_C[middle_i]->line_p == line_item_p))
	{
	fanout_objectives_C[middle_i]->n1_i += objective_u.n1_i;
	fanout_objectives_C[middle_i]->n0_i += objective_u.n0_i;
	}//if
else
	fanout_objectives_C.fuege_ein(line_item_p,objective_u.n0_i,objective_u.n1_i);
}//add_n0_and_n1_to_fanout_objective

//*********************************************************************
//
//   FUNCTION-NAME: determine_next_objective_and_add_them_to_current_objectives
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Diese Funktion macht das gleiche, wie
//	add_n0_and_n1_to_fanout_objective mit dem Unterschied, dass keine Fanout
//	Line Vorgaenger der aktuellen Objective Line sein kann.
//
//   FORMAL PARAMETERS: objective_type_u objective_u: das entsprechende Objective
//
//   IMPLICIT INPUT/OUTPUT: void
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void determine_next_objective_and_add_them_to_current_objectives(objective_type_u &objective_u)
{
gate_C *gate_item_p;
line_C *line_item_p;
line_C *lowest_1_contby_line_p = NULL,*lowest_0_contby_line_p = NULL;
line_C *line0_p = NULL,*line1_p = NULL; //Fuer EXOR benoetigt
objective_type_u o0_u,o1_u; //Fuer EXOR und BLOCK benoetigt
int belegung_a[max_number_gate_inputs_i];
int min_cost_0_belegung_a[max_number_gate_inputs_i],min_cost_1_belegung_a[max_number_gate_inputs_i];
double min_0_cost_f = 0.0,min_1_cost_f = 0.0,cost_f;
boolean good_x_value_exist_b = FALSE;

gate_item_p = objective_u.line_p->vorgaenger()->gib_only_gate();
{forall(line_item_p,*gate_item_p->gib_inputs())
   {
   if (line_item_p->get_good_value() == VAL_X)
	good_x_value_exist_b = TRUE;
   if ((line_item_p->get_good_value() == VAL_X) || ((line_item_p->in_output_kegel() == TRUE) &&
						    (line_item_p->get_bad_value() == VAL_X)))
	if (lowest_1_contby_line_p == NULL)
		{
		lowest_1_contby_line_p = line_item_p;
		lowest_0_contby_line_p = line_item_p;
		}//if
	else
		{
		if (lowest_1_contby_line_p->gib_one_contby() > line_item_p->gib_one_contby())
			lowest_1_contby_line_p = line_item_p;
		if (lowest_0_contby_line_p->gib_zero_contby() > line_item_p->gib_zero_contby())
			lowest_0_contby_line_p = line_item_p;
		}//else
   }}//forall
{forall(line_item_p,*gate_item_p->gib_inputs())
	if ((line_item_p->get_good_value() == VAL_X) || ((line_item_p->in_output_kegel() == TRUE) &&
	     (line_item_p->get_bad_value() == VAL_X) && (good_x_value_exist_b == FALSE)) ||
	    (gate_item_p->gib_type() == EXOR))
	   switch(gate_item_p->gib_type())
		{
		case FANOUT:
			error("determine_next_objective_and_add_them_to_current_objectives: Fanout not allowed");
		case AND:
		case DRIVER:
			if (line_item_p == lowest_0_contby_line_p)
				current_objectives_C.fuege_ein(line_item_p,objective_u.n0_i,objective_u.n1_i);
			else // if (objective_u.n1_i > 0)
				current_objectives_C.fuege_ein(line_item_p,0,objective_u.n1_i);
			break;
		case OR:
			if (line_item_p == lowest_1_contby_line_p)
				current_objectives_C.fuege_ein(line_item_p,objective_u.n0_i,objective_u.n1_i);
			else // if (objective_u.n0_i > 0)
				current_objectives_C.fuege_ein(line_item_p,objective_u.n0_i,0);
			break;
		case NAND:
		case INVERTER:
			if (line_item_p == lowest_0_contby_line_p)
				current_objectives_C.fuege_ein(line_item_p,objective_u.n1_i,objective_u.n0_i);
			else // if (objective_u.n0_i > 0)
				current_objectives_C.fuege_ein(line_item_p,0,objective_u.n0_i);
			break;
		case NOR:
			if (line_item_p == lowest_1_contby_line_p)
				current_objectives_C.fuege_ein(line_item_p,objective_u.n1_i,objective_u.n0_i);
			else // if (objective_u.n1_i > 0)
				current_objectives_C.fuege_ein(line_item_p,objective_u.n1_i,0);
			break;
		case EXOR:
			if (line0_p == NULL)
				{
				line0_p = line_item_p;
				break;
				}//if
			line1_p = line_item_p;
				// Objectives bestimmen nach Schulz Seite 117
			if (((line0_p->gib_zero_contby() + line1_p->gib_zero_contby()
						< line0_p->gib_one_contby() + line1_p->gib_one_contby()) &&
			     (line0_p->get_good_value() != VAL_1) && (line1_p->get_good_value() != VAL_1)) ||
			    (line0_p->get_good_value() == VAL_0) || (line1_p->get_good_value() == VAL_0))
				{
				o0_u.n0_i = objective_u.n0_i;
				o1_u.n0_i = objective_u.n0_i;
				o0_u.n1_i = 0;
				o1_u.n1_i = 0;
				}//if
			else
				{
				o0_u.n1_i = objective_u.n0_i;
				o1_u.n1_i = objective_u.n0_i;
				o0_u.n0_i = 0;
				o1_u.n0_i = 0;
				}//else
			if (((line0_p->gib_zero_contby() + line1_p->gib_one_contby()
						< line0_p->gib_one_contby() + line1_p->gib_zero_contby()) &&
			     (line0_p->get_good_value() != VAL_1) && (line1_p->get_good_value() != VAL_0)) ||
			    (line0_p->get_good_value() == VAL_0) || (line1_p->get_good_value() == VAL_1))
				{
				o0_u.n0_i += objective_u.n1_i;
				o1_u.n1_i += objective_u.n1_i;
				}//if
			else
				{
				o0_u.n1_i += objective_u.n1_i;
				o1_u.n0_i += objective_u.n1_i;
				}//else
			current_objectives_C.fuege_ein(line0_p,o0_u.n0_i,o0_u.n1_i);
			current_objectives_C.fuege_ein(line1_p,o1_u.n0_i,o1_u.n1_i);
			break;
		case BLOCK:
			cube_list_t *cube_list_p;
			belegung_t *cube_p;
			int i;

			if (min_0_cost_f == 0.0)
			   { // Wenn noch nicht berechnet wurde, wird jetzt berechnet
			   for (i=0;i<max_number_gate_inputs_i;i++)
				{
				belegung_a[i] = 0;
				min_cost_0_belegung_a[i] = 0;
				}//for
			   for (cube_list_p = gate_item_p->gib_cubes(VAL_0);cube_list_p != NULL;cube_list_p = cube_list_p->next_p)
			     {
			     for (i=0,cube_p = cube_list_p->to_belegung_p;cube_p != NULL;i++,cube_p= cube_p->next_p)
				switch (cube_p->val_e)
				   {
				   case VAL_0:
					belegung_a[i] = 0;
					break;
				   case VAL_1:
					belegung_a[i] = 1;
					break;
				   case VAL_X:
					belegung_a[i] = 2;
					break;
				   default:
					error("add_n0_and_n1_to_fanout_objective: wrong value found in block cube");
				   }//switch
			     add_constraints(gate_item_p,belegung_a);
			     for (i=0,cube_p = cube_list_p->to_belegung_p;cube_p != NULL;i++,cube_p= cube_p->next_p)
				{
				line_C *line_p;
				value_t val_e;
				
				line_p = (*gate_item_p->gib_inputs())[i];
				if ((line_p->get_good_value() == VAL_X) || (belegung_a[i] == 2))
					continue;
				val_e = line_p->get_good_value();
				if (((val_e == VAL_1) && (belegung_a[i] == 1)) || ((val_e == VAL_0) && (belegung_a[i] == 0)))
					continue;
				break;
				}//for
			     if (cube_p != NULL)
				continue;
			     cost_f = 0.0;
			     for (i=0,cube_p = cube_list_p->to_belegung_p;cube_p != NULL;i++,cube_p= cube_p->next_p)
				switch (belegung_a[i])
				   {
				   case 0:
					cost_f += (*gate_item_p->gib_inputs())[i]->gib_zero_contby();
					break;
				   case 1:
					cost_f += (*gate_item_p->gib_inputs())[i]->gib_one_contby();
					break;
				   case 2:
					if ((*gate_item_p->gib_inputs())[i]->gib_zero_contby() < (*gate_item_p->gib_inputs())[i]->gib_one_contby())
						cost_f += (*gate_item_p->gib_inputs())[i]->gib_zero_contby();
					else
						cost_f += (*gate_item_p->gib_inputs())[i]->gib_one_contby();
					break;
				   default:
					error("add_n0_and_n1_to_fanout_objective: wrong value found in block cube");
				   }//switch
			     if ((min_0_cost_f == 0.0) || (cost_f < min_0_cost_f))
				{
				min_0_cost_f = cost_f;
				for (i=0;i<max_number_gate_inputs_i;i++)
					min_cost_0_belegung_a[i] = belegung_a[i];
				}//if
			     }//for
			   for (i=0;i<max_number_gate_inputs_i;i++)
				{
				belegung_a[i] = 0;
				min_cost_1_belegung_a[i] = 0;
				}//for
			   for (cube_list_p = gate_item_p->gib_cubes(VAL_1);cube_list_p != NULL;cube_list_p = cube_list_p->next_p)
			     {
			     for (i=0,cube_p = cube_list_p->to_belegung_p;cube_p != NULL;i++,cube_p= cube_p->next_p)
				switch (cube_p->val_e)
				   {
				   case VAL_0:
					belegung_a[i] = 0;
					break;
				   case VAL_1:
					belegung_a[i] = 1;
					break;
				   case VAL_X:
					belegung_a[i] = 2;
					break;
				   default:
					error("add_n0_and_n1_to_fanout_objective: wrong value found in block cube");
				   }//switch
			     add_constraints(gate_item_p,belegung_a);
			     for (i=0,cube_p = cube_list_p->to_belegung_p;cube_p != NULL;i++,cube_p= cube_p->next_p)
				{
				line_C *line_p;
				value_t val_e;
				
				line_p = (*gate_item_p->gib_inputs())[i];
				if ((line_p->get_good_value() == VAL_X) || (belegung_a[i] == 2))
					continue;
				val_e = line_p->get_good_value();
				if (((val_e == VAL_1) && (belegung_a[i] == 1)) || ((val_e == VAL_0) && (belegung_a[i] == 0)))
					continue;
				break;
				}//for
			     if (cube_p != NULL)
				continue;
			     cost_f = 0.0;
			     for (i=0,cube_p = cube_list_p->to_belegung_p;cube_p != NULL;i++,cube_p= cube_p->next_p)
				switch (belegung_a[i])
				   {
				   case 0:
					cost_f += (*gate_item_p->gib_inputs())[i]->gib_zero_contby();
					break;
				   case 1:
					cost_f += (*gate_item_p->gib_inputs())[i]->gib_one_contby();
					break;
				   case 2:
					if ((*gate_item_p->gib_inputs())[i]->gib_zero_contby() < (*gate_item_p->gib_inputs())[i]->gib_one_contby())
						cost_f += (*gate_item_p->gib_inputs())[i]->gib_zero_contby();
					else
						cost_f += (*gate_item_p->gib_inputs())[i]->gib_one_contby();
					break;
				   default:
					error("add_n0_and_n1_to_fanout_objective: wrong value found in block cube");
				   }//switch
			     if ((min_1_cost_f == 0.0) || (cost_f < min_1_cost_f))
				{
				min_1_cost_f = cost_f;
				for (i=0;i<max_number_gate_inputs_i;i++)
					min_cost_1_belegung_a[i] = belegung_a[i];
				}//if
			     }//for
			   }//if
			for (i=0;i<max_number_gate_inputs_i+1;i++)
				if ((*gate_item_p->gib_inputs())[i] == line_item_p)
					break;
				else if (i == max_number_gate_inputs_i)
					error("add_n0_and_n1_to_fanout_objective:  Line not found in input-list");
			o0_u.n0_i = 0;
			o0_u.n1_i = 0;
			if (min_cost_1_belegung_a[i] == 1)
				o0_u.n1_i += objective_u.n1_i;
			else if (min_cost_1_belegung_a[i] == 0)
				o0_u.n0_i += objective_u.n1_i;
			if (min_cost_0_belegung_a[i] == 1)
				o0_u.n1_i += objective_u.n0_i;
			else if (min_cost_1_belegung_a[i] == 0)
				o0_u.n0_i += objective_u.n0_i;
			current_objectives_C.fuege_ein(line_item_p,o0_u.n0_i,o0_u.n1_i);
			break;
		case D_LATCH:
		case D_FLIP_FLOP:
			error("add_n0_and_n1_to_fanout_objective: not implemented for sequential gates");
		case UNDEFINED_GATE:
			error("add_n0_and_n1_to_fanout_objective: undefined gate found");
		}//switch
}
}//determine_next_objective_and_add_them_to_current_objectives

//*********************************************************************
//
//   FUNCTION-NAME: take_out_fanout_objective_p_closest_to_output
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Das Fanout Objective mit der hoechsten Ebenennummer
//	wird entnommen und zurueckgegeben. Nicht optimal !!!!
//
//   FORMAL PARAMETERS: NONE
//
//   IMPLICIT INPUT/OUTPUT: void
//
//   FUNCTION-VALUE: objective_type_u objective_u: das entsprechende Objective
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
objective_type_u take_out_fanout_objective_p_closest_to_output()
{
objective_type_u *objective_p = fanout_objectives_C[fanout_objectives_C.no_elements() - 1];

fanout_objectives_C.fuege_aus(objective_p->line_p);
return(*objective_p);
}//take_out_fanout_objective_p_closest_to_output

//*********************************************************************
//
//   FUNCTION-NAME: n0p_and_n1p_nonequal_zero
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Gibt TRUE zurueck, wenn n0_i und n1_i vom Eingabe
//	Objective beide nicht 0 sind und weist in diesem Fall die Objective
//	der Final Objective zu.
//
//   FORMAL PARAMETERS: objective_type_u objective_u: das entsprechende Objective
//
//   IMPLICIT INPUT/OUTPUT: void
//
//   FUNCTION-VALUE: boolean: TRUE wenn n0 und n1 nicht 0 sind
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
boolean n0p_and_n1p_nonequal_zero(objective_type_u &objective_u)
{
#ifndef FO_DECISION
   if ((objective_u.n0_i != 0) && (objective_u.n1_i != 0) && (objective_u.line_p->get_good_value() == VAL_X))
	{
	final_objective_u = objective_u;
	return(TRUE);
	}//if
#endif
return(FALSE);
}//n0p_and_n1p_nonequal_zero

//*********************************************************************
//
//   FUNCTION-NAME: add_objective_to_head_objectives
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: fuegt ein Objective in die head_objectives ein,
//	falls es nicht schon dort eingefuegt ist.
//
//   FORMAL PARAMETERS: objective_type_u objective_u: das entsprechende Objective
//
//   IMPLICIT INPUT/OUTPUT: void
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void add_objective_to_head_objectives(objective_type_u &objective_u)
{
if (head_objectives_C.element_in_liste(objective_u.line_p) == FALSE)
	head_objectives_C.fuege_ein(objective_u.line_p,objective_u.n0_i,objective_u.n1_i);
}//add_objective_to_head_objectives

//*********************************************************************
//
//   FUNCTION-NAME: take_out_head_objective
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: fuegt head Objective mit den hoechsten Anforderungen
//	aus.
//
//   FORMAL PARAMETERS: NONE
//
//   IMPLICIT INPUT/OUTPUT: void
//
//   FUNCTION-VALUE: objective_type_u: das entsprechende Objective
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
objective_type_u take_out_head_objective()
{
objective_type_u *objective_p,*ret_objective_p = NULL;

forall(objective_p,head_objectives_C)
	if (ret_objective_p == NULL)
		ret_objective_p = objective_p;
	else
		if ((ret_objective_p->n0_i + ret_objective_p->n1_i) < (objective_p->n0_i + objective_p->n1_i))
			ret_objective_p = objective_p;
head_objectives_C.fuege_aus(ret_objective_p->line_p);
return(*ret_objective_p);
}//take_out_head_objective
