#include "gate_C.h"
#include "line_liste_C.h"
#include "stringliste_C.h"
#include "line_C.h"
#include "modify.h"

extern line_C *search(char*,line_tree_u*);
//*********************************************************************
//
//   CLASS-NAME: gate_C: Unterklasse von object_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   CLASS-DESCRIPTION: Diese Klasse definiert Gatter und alle notwendigen
//	Operationen auf Gattern.
//
//   OPERATOREN: =: Zuweisung
//
//   FUNKTIONEN: Konstruktor mit char* Argument fuer objet_C Konstruktor
//		 Destruktor macht nichts
//		 print_all: Debug Routine, gibt Inhalt des Gates aus
//		 belege_gate: siehe friends.h
//		 connect_input: siehe friends.h
//		 connect_output: siehe friends.h
//		 compute_contby: siehe friends.h
//		 compute_observaby: siehe friends.h
//		 gib_zero_contby: gibt den Wert von zero_contby_f aus
//		 gib_one_contby: gibt den Wert von one_contby_f aus
//		 gib_observaby: gibt den Wert von observaby_f aus
//		 calculate_planes: siehe friends.h
//		 calculate_type: siehe friends.h
//		 gib_input: gibt die Liste der Eingaenge des Gatters aus
//		 gib_outputs: gibt die Liste der Ausgaenge des Gatters aus
//		 set_reconvergence: Gatter ist Endpunkt einer Pfadrekonvergenz
//			reconvergence_b wird auf TRUE gesetzt
//		 set_feedback: Gatter ist in Rueckkopplung, feedback_b wird auf
//			TRUE gesetzt
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void gate_C::gib_constraints(cube_list_t*&source_p,cube_list_t*&target_p)
{
source_p = constraints_a[0];
target_p = constraints_a[1];
}//gib_constraints

void gate_C::print_constr()
{
ofstream datei_F("Constraints");
int i;
cube_list_t *source_p,*target_p;
belegung_t *bel_p;

if (constraints_a[0] != NULL)
	{
	source_p = constraints_a[0];
	target_p = constraints_a[1];
	datei_F << name_sC << "\n";
	while (source_p != NULL)
	   {
	   bel_p = source_p->to_belegung_p;
	   i=1;
	   while (bel_p != NULL)
		{
		if (bel_p->val_e != VAL_X)
			datei_F << "I" << i << "=" << bel_p->val_e << " => ";
		i++;
		bel_p = bel_p->next_p;
		}//while
	   bel_p = target_p->to_belegung_p;
	   while (bel_p != NULL)
		{
		if (bel_p->val_e != VAL_X)
			datei_F << "I" << i << "=" << bel_p->val_e;
		i++;
		bel_p = bel_p->next_p;
		}//while
	   datei_F << "\n";
	   source_p = source_p->next_p;
	   target_p = target_p->next_p;
	   }//for
	}//if
datei_F.close();
}//print_constr

gate_C::gate_C(string_C s_sC):object_C(s_sC)
{
object_type_e = GATE;
actual_branch_i = 0;
input_p = NULL;
output_p = NULL;
constraints_a[0] = NULL;
constraints_a[1] = NULL;
block_cubes_a[0] = NULL;
block_cubes_a[1] = NULL;
no_input_i = 0;
no_output_i = 0;
gate_e = UNDEFINED_GATE;
feedback_b = FALSE;
implikation_durchgefuehrt_b = FALSE;
hazard_bit_c = 0;
}

gate_C::~gate_C()
{
delete input_p;
delete output_p;
}

//*********************************************************************
//
//   FUNCTION-NAME: belege_gate
//
//   CLASS: gate_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Dem Gatter werden Werte zugewiesen. Gatter Typ
//	Anzahl Ein- und Ausgaenge werden belegt, sowie Stringlisten dafuer
//	Alloziiert.
//
//   FORMAL PARAMETERS: gate_t gate_type_e: der Typ des Gatters (AND, OR,...)
//		int no_inputs_i: Anzahl der Eingaenge
//		int no_outputs_i: Anzahl der Ausgaenge
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void gate_C::belege_gate(gate_t gate_type_e,int no_inputs_i,int no_outputs_i)
{
if (((gate_type_e == AND) || (gate_type_e == OR)) && (no_inputs_i == 1))
   gate_e = DRIVER;
else if (((gate_type_e == NAND) || (gate_type_e == NOR)) && (no_inputs_i == 1))
   gate_e = INVERTER;
else
   gate_e = gate_type_e;
no_input_i = no_inputs_i;
no_output_i = no_outputs_i;
input_p = new line_liste_C(no_inputs_i);
if (input_p == NULL)
	error("no memory available");
output_p = new line_liste_C(no_outputs_i);
if (output_p == NULL)
	error("no memory available");
return;
}

// *********************************************************************
//
//   FUNCTION-NAME: connect_input
//
//   CLASS: gate_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Die Inputs eines Gatters werden angeschlossen.
//	Es wird festgestellt ob die Anzahl der Inputs mit der Definition
//	uebereinstimmt.
//
//   FORMAL PARAMETERS: stringliste_C &namen_sC: die Namen der Eingaenge
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
// *********************************************************************
void gate_C::connect_input(line_liste_C& ios_C)
{
line_C *line_item_p;

if (ios_C.no_elements() != no_input_i)
   {
   if (ios_C.no_elements() < no_input_i)
      error("element has less io's than defined");
   else
      error("element has more io's than defined");
   }//if
forall(line_item_p,ios_C)
   {
   line_item_p->verbinde(this,'I');
   input_p->verbinde(line_item_p,forall_index+1);
   }//forall
}//connect_input

// *********************************************************************
//
//   FUNCTION-NAME: connect_output
//
//   CLASS: gate_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Die Outputs eines Gatters werden angeschlossen.
//	Es wird festgestellt ob die Anzahl der Outputs mit der Definition
//	uebereinstimmt.
//
//   FORMAL PARAMETERS: stringliste_C &namen_sC: die Namen der Ausgaenge
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
// *********************************************************************
void gate_C::connect_output(line_liste_C& ios_C)
{
line_C *line_item_p;

if (ios_C.no_elements() != no_output_i)
   {
   if (ios_C.no_elements() < no_input_i)
      error("element has less io's than defined");
   else
      error("element has more io's than defined");
   }//if
forall(line_item_p,ios_C)
   {
   line_item_p->verbinde(this,'O');
   output_p->verbinde(line_item_p,forall_index+1);
   }//forall
}//connect_output

void gate_C::print_all(ofstream &datei_F)
{ // gibt die Belegung des Gatters aus (DEBUG)
datei_F << "Adresse   " << this << "\n";
datei_F << "Name:   " << name_sC << "\n";
   	datei_F << "Inputliste:   " << input_p << "\n";
datei_F << "Outputliste:   " << output_p << "\n";
datei_F << "Number Inputs:   " << no_input_i << "\n";
datei_F << "Number Outputs:   " << no_output_i << "\n";
datei_F << "Gate Type:   " << gate_e << "\n";
datei_F << "Gate in Feedback:   " << feedback_b << "\n";
datei_F << "Element in Feedback:   " << feedback_b << "\n";
datei_F << "Pfadrekonvergenz an Element:   " << reconvergence_b << "\n";
datei_F << "##### Inputs\n";
input_p->print_all(datei_F);
datei_F << "##### Outputs\n";
output_p->print_all(datei_F);
datei_F << "\n\n***********\n";
}//print_all

//*********************************************************************
//
//   FUNCTION-NAME: calculate_planes
//
//   CLASS: gate_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Fuer alle Nachfolgenden Lines wird calculate_planes
//	aufgerufen.
//
//   FORMAL PARAMETERS: keine
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void gate_C::calculate_planes()
{
line_C *line_item_p;
boolean erfolg_b = FALSE;

forall(line_item_p,*output_p)
	{
	erfolg_b = TRUE;
	line_item_p->calculate_planes();
	}//forall
if (erfolg_b == FALSE)
	{
	cout << name_sC;
	error("gate_C::calculate_planes: no output detected");
	}//if
}//calculate_planes

//*********************************************************************
//
//   FUNCTION-NAME: calculate_type
//
//   CLASS: gate_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Fuer alle Nachfolgenden Lines wird calculate_type
//	aufgerufen.
//
//   FORMAL PARAMETERS: keine
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void gate_C::calculate_type()
{
line_C *line_item_p;

forall(line_item_p,*output_p)
	line_item_p->calculate_type();
}//calculate_type
