#ifndef gate_C_h
#define gate_C_h
#include "object_C.h"
#ifdef FRAMEWORK
#include "cfidr.h"
#endif

//*********************************************************************
//
//   CLASS-NAME: gate_C: Unterklasse von object_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   CLASS-DESCRIPTION: Diese Klasse definiert Gatter und alle notwendigen
//	Operationen auf Gattern.
//
//   OPERATOREN: =: Zuweisung
//
//   FUNKTIONEN: Konstruktor mit char* Argument fuer objet_C Konstruktor
//		 Destruktor macht nichts
//		 print_all: Debug Routine, gibt Inhalt des Gates aus
//		 belege_gate: siehe friends.h
//		 connect_input: siehe friends.h
//		 connect_output: siehe friends.h
//		 compute_contby: siehe friends.h
//		 compute_observaby: siehe friends.h
//		 gib_zero_contby: gibt den Wert von zero_contby_f aus
//		 gib_one_contby: gibt den Wert von one_contby_f aus
//		 gib_observaby: gibt den Wert von observaby_f aus
//		 calculate_planes: siehe friends.h
//		 calculate_type: siehe friends.h
//		 gib_input: gibt die Liste der Eingaenge des Gatters aus
//		 gib_outputs: gibt die Liste der Ausgaenge des Gatters aus
//		 set_reconvergence: Gatter ist Endpunkt einer Pfadrekonvergenz
//			reconvergence_b wird auf TRUE gesetzt
//		 set_feedback: Gatter ist in Rueckkopplung, feedback_b wird auf
//			TRUE gesetzt
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
class gate_C: public object_C {
public:
   gate_C(string_C);
   ~gate_C();
   void search_next_path(line_C*,boolean&);
   boolean search_off_path_signals(boolean&);
   void get_block_cube();
   cube_list_t *block_cubes(int i){return(block_cubes_a[i]);}
   void put_constraint_cube();
   void print_all(ofstream &);
   void belege_gate(gate_t gate_type_e,int no_inputs_i,int no_outputs_i);
   void connect_input(line_liste_C&);
   void connect_output(line_liste_C&);
   void compute_contby();
   void compute_observaby();
   void prob_compute_contby();
   void prob_compute_observaby();
   void level_compute_contby();
   void level_compute_observaby();
   void reko_compute_contby();
   void reko_compute_observaby();
   void calculate_planes();
   void calculate_type();
   void invert_type();
   void gib_constraints(cube_list_t*&source_p,cube_list_t*&target_p);
   inline line_liste_C *gib_inputs(){return(input_p);}
   inline line_liste_C *gib_outputs(){return(output_p);}
   inline void set_feedback(){feedback_b = TRUE;}
   gate_t gib_type(){return(gate_e);}
   fan_result_t forward_implication();
   fan_result_t fast_forward_implication();
   fan_result_t backward_implication(boolean = TRUE);
   fan_result_t try_global_imp();
   void try_constraint_ana(ofstream&);
   void gate_C::mark_robust_signals();
   void analyse_result(int,ofstream&);
   void add_cube(line_C*,line_C*);
   int no_inputs(){return(no_input_i);}
   int no_outputs(){return(no_output_i);}
   cube_list_t *gib_cubes(value_t val_e);
   void reset_branch(){actual_branch_i = 0;}
   void print_constr();
   void update_no_inputs();
   void update_no_outputs();
   boolean hazard_checked(){return((hazard_bit_c == 0)?FALSE:TRUE);}
   boolean hazard_occured(){return((hazard_bit_c == 2)?TRUE:FALSE);}
   void set_hazard_bits(unsigned char c){hazard_bit_c = c;}
#ifdef FRAMEWORK
   friend void cfidr_ConnectMacros(cfidrViewIdT);
#endif
   boolean implikation_durchgefuehrt_b;
   boolean dyn_path_exist_b;
private:
   int actual_branch_i;
   line_liste_C *input_p;
   line_liste_C *output_p;
   cube_list_t *constraints_a[2];
   cube_list_t *block_cubes_a[2];
   int no_input_i;
   int no_output_i;
   gate_t gate_e;
   unsigned char hazard_bit_c;
   boolean feedback_b;	// Element in feedback
};
#endif
