#include "gate_liste_C.h"
#include "string_C.h"
#include "gate_C.h"

extern gate_liste_C gates_C;
extern gate_C *search(char*,gate_tree_u*);
extern void milef_add_to_gate_tree(gate_tree_u*&,gate_C*);
// *********************************************************************
//
//   CLASS-NAME: gate_liste_C Unterklasse von liste_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   CLASS-DESCRIPTION: Diese Klasse definiert Gatterlisten und alle notwendigen
//	Operationen auf Gatterlisten.
//
//   OPERATOREN: keine
//
//   FUNKTIONEN: 2 Konstruktoren mit Integer und port sowie ohne Argument
//			ohne Argumente => generische Liste
//		 Destruktor loescht Liste
//		 print_all: gibt Inhalt aller Listenelemente aus (DEBUG)
//		 fuege_ein: fuegt Element ein (siehe friends.h)
//		 fuege_aus: fuegt Element aus Liste aus (gate_C* Argument)
//		 element_in_liste: prueft, ob Element gate_C* in Liste ist
//		 search: siehe friends.h
//		 next: gibt aktuelles Element von Liste aus (aktuell entspricht
//			aktual_no_i), wenn Liste geschlossen ist (closed_b == TRUE)
//			beim letzten Element wird Liste geoeffnet
//		 verbinde: siehe friends.h
//		 gib_only_gate: wenn ein Element in der Liste ist, wird dieses
//			Ausgegeben, sonst Fehlermeldung
//
//   MODIFICATION-HISTORY:
//
// *********************************************************************
gate_liste_C::gate_liste_C(int listengroesse_i):liste_C(listengroesse_i)
{
liste_a = new gate_C *[listengroesse_i];
if (liste_a == NULL)
	error("No more memory available");
}

void gate_liste_C::init(int size_i)
{
liste_a = new gate_C *[size_i];
if (liste_a == NULL)
	error("No more memory available");
}

void gate_liste_C::resize()
{
gate_liste_C *tmp_C;
int i;

tmp_C = new gate_liste_C;
tmp_C->init(number_elements_i);
for (i=0;i<number_elements_i;i++)
	tmp_C->fuege_ein(liste_a[i]);
delete[] liste_a;
liste_a = new gate_C *[number_elements_i];
for (i=0;i<number_elements_i;i++)
	liste_a[i] = (*tmp_C)[i];
delete tmp_C;
}

void gate_liste_C::print_all(ofstream& datei_F)
{
liste_C::print_all(datei_F);
for (int i=0;i<number_elements_i;i++)
datei_F << "\t" << "element[" << i << "]:   " << liste_a[i]->get_name() << "\n";
}

gate_C *gate_liste_C::gib_only_gate()
{
if (number_elements_i != 1)
	error("gib_only_gate: more than 1 gate in list");
return(liste_a[0]);
}//gib_only_gate

void gate_liste_C::loesche_liste()
{
for (int i=0;i<number_elements_i;i++)
	liste_a[i] = NULL;
number_elements_i = 0;
}//loesche_liste

void gate_liste_C::loesche_liste_und_pointer()
{
delete[] liste_a;
number_elements_i = 0;
liste_a = new gate_C *[1];
max_elements_i = 1;
generisch_b = FALSE;
}//loesche_liste

void gate_liste_C::change_connection(char *name_s,gate_C *gate_item_p)
{
int i;

for (i=0;i<number_elements_i;i++)
   if (strcasecmp(liste_a[i]->get_name(),name_s) == 0)
	{
	liste_a[i] = gate_item_p;
	break;
	}//if
}//change_connection

//*********************************************************************
//
//   FUNCTION-NAME: element_in_liste
//
//   CLASS: gate_liste_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Diese Funktion stellt fest, ob ein Gate in der Liste
//	vorhanden ist.
//
//   FORMAL PARAMETERS: gate_C *gate_item_p: das Gatter, das gesucht wird
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: boolean: TRUE, wenn das Gatter in der Liste ist
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
boolean gate_liste_C::element_in_liste(gate_C *gate_item_p,boolean fuege_aus_b)
{
int i=0;
gate_C *ausgefuegtes_element_p;

if (this == &gates_C)
	{
	if (fuege_aus_b == TRUE)
		error("element_in_liste: fuege_aus in gates_C nicht erlaubt");
	gate_C *gate_item_p = search(gate_item_p->get_name(),milef_gate_tree_root_p);
	if (gate_item_p == NULL)
		return FALSE;
	return TRUE;
	}//if
for (;i < number_elements_i;i++)
   if (liste_a[i] == gate_item_p)
	if (fuege_aus_b == FALSE)
		return TRUE;
	else
		{
                ausgefuegtes_element_p = liste_a[i];
                break;
		}//else
if ((fuege_aus_b == FALSE) || (i >= number_elements_i))
	return FALSE;
liste_C::fuege_aus();
for(;i<number_elements_i;i++)
	liste_a[i] = liste_a[i+1];
liste_a[number_elements_i] = ausgefuegtes_element_p;
return TRUE;
}/*element_in_liste*/

//*********************************************************************
//
//   FUNCTION-NAME: fuege_ein
//
//   CLASS: gate_liste_C (line_liste_C)
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Ein(e) Gate (Line) wird ans Ende der Liste eingefuegt
//
//   FORMAL PARAMETERS: gate_C *gate_item_p: das entsprechende Gatter
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void gate_liste_C::fuege_ein(gate_C *gate_item_p)
{
if (this == &gates_C) // baut den AVL Baum fuer gates_C auf
	milef_add_to_gate_tree(milef_gate_tree_root_p,gate_item_p);
liste_C::fuege_ein();
liste_a[number_elements_i - 1] = gate_item_p;
return;
}/*fuege_ein*/

//*********************************************************************
//
//   FUNCTION-NAME: fuege_aus
//
//   CLASS: gate_liste_C (line_liste_C)
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Ein(e) Gate (Line) wird aus der Liste ausgefuegt
//
//   FORMAL PARAMETERS: line_C *line_item_p: die entsprechende Line
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void gate_liste_C::fuege_aus(gate_C *gate_item_p)
{
int i;
gate_C *ausgefuegtes_element_p;

liste_C::fuege_aus();
for (i=0;i <= number_elements_i;i++)
	if (liste_a[i] == gate_item_p)
		{
		ausgefuegtes_element_p = liste_a[i];
		break;
		}//if
if (i > number_elements_i)
	error("fuege_aus: element not found in list");
for(;i<number_elements_i;i++)
	liste_a[i] = liste_a[i+1];
liste_a[number_elements_i] = ausgefuegtes_element_p;
return;
}/*fuege_aus*/

//*********************************************************************
//
//   FUNCTION-NAME: verbinde
//
//   CLASS: gate_liste_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Ein Gatter wird in die Input/Output Liste einer
//	Line eingefuegt. Der Port des Gatters wird in connected_to_a an der
//	entsprechenden Stelle vermerkt.
//
//   FORMAL PARAMETERS: gate_C *gate_item_p: das entsprechende Gatter
//		io_t gate_io_e: Der Port der Gatters, wo die Line angeschlossen wird
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: io_t: Port, an dem das Gatter angeschlossen wurde
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void gate_liste_C::verbinde(gate_C *gate_item_p)
{
gate_C **tmp_liste_a;

number_elements_i++;
if (number_elements_i > max_no_io_i)
	error("too many fanout-branches at some signal");
if (number_elements_i >= max_elements_i)
	{
        try {
        tmp_liste_a = new gate_C *[2 * max_elements_i];
        }
        catch(...) {
          error("No more memory available");
        }
	for (int i=0;i<number_elements_i - 1;i++)
		tmp_liste_a[i] = liste_a[i];
	delete[] liste_a;
	liste_a = tmp_liste_a;
	max_elements_i = 2 * max_elements_i;
	}//if
liste_a[number_elements_i - 1] = gate_item_p;
}//verbinde
