#include "line_C.h"
#include "gate_liste_C.h"
#include "gate_C.h"
#include "line_liste_C.h"
#include "modify.h"

extern line_liste_C free_lines_C;
//*********************************************************************
//
//   CLASS-NAME: line_C: Unterklasse von object_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   CLASS-DESCRIPTION: Diese Klasse definiert Lines und alle notwendigen
//	Operationen auf Lines.
//
//   OPERATOREN: =: Zuweisung
//
//   FUNKTIONEN: Konstruktor mit char* Argument: siehe friends.h
//		 Destruktor macht nichts
//		 set_primary_input: Line ist Primaereingang: prime_input_b
//			wird auf TRUE gesetzt
//		 set_primary_output: Line ist Primaerausgang: prime_output_b
//                      wird auf TRUE gesetzt
//		 print_all: alle Belegungen werden Ausgegeben fuer DEBUG
//		 verbinde: siehe friends.h
//		 compute_contby: siehe friends.h
//		 compute_observaby: siehe friends.h
//		 vorgaenger: gibt Liste mit Gattern aus, die Line als Ausgang
//			haben. (Die Liste sollte ein Element haben)
//		 nachfolger: gibt Gatterliste aus, die einen Input auf der Line
//			haben. (Elemente > 1 <==> Line hat fanout)
//		 calculate_planes: siehe friends.h
//		 calculate_type: siehe friends.h
//		 gib_plane_no: Die Nummer der Schaltungsebene, in der sich die
//			Line befindet wird Ausgegeben. Variable ist plane_no_i
//		 gib_type: Es wird Ausgegeben, ob die Line FREE_LINE, BOUND_LINE
//			oder HEAD_LINE ist. Variable line_type_e
//		 set_feedback: Line ist in Rueckkopplung. feedback_b wird auf
//			TRUE gesetzt.
//		 reset_value: test_value_e wird auf VAL_X gesetzt. init_value_e
//			wird auf letzte test_value_e gesetzt.
//		 should_be_tested: Belegt Line fuer den naechsten Rising oder Falling Test
//		 must_be_tested: Initialisiert  Line fuer den naechsten Rising oder
//			Falling Test
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void line_C::reset_dyn_learning()
{
tmp_global_implications[0].loesche_liste();
tmp_global_implications[1].loesche_liste();
}//reset_dyn_learning

void line_C::reset_value(fan_result_t result_e)
{
if (result_e == SUCCESS)
	init_value_e = good_value_e;
good_value_e = VAL_X;
bad_value_e = VAL_X;
}//reset_value

void line_C::reset_non_robust()
{
hazard_occured_b = FALSE;
if (rising_non_robust_b == TRUE)
	rising_aborted_b = FALSE;
if (falling_non_robust_b == TRUE)
	falling_aborted_b = FALSE;
rising_non_robust_b = FALSE;
falling_non_robust_b = FALSE;
}//reset_non_robust

void line_C::reset_aborted_faults()
{
s_a_0_aborted_b = FALSE;
s_a_1_aborted_b = FALSE;
rising_aborted_b = FALSE;
falling_aborted_b = FALSE;
}//reset_aborted_faults

boolean line_C::lookup_testability()
{
if (((good_value_e == VAL_1) && (s_a_0_untestable_b == TRUE)) ||
    ((bad_value_e == VAL_0) && (s_a_1_untestable_b == TRUE)))
	return(FALSE);
else
	return(TRUE);
}//lookup_testability

void line_C::reset_testaby()
{
one_contby_f = 0.0;
zero_contby_f = 0.0;
contby_f = 0.0;
observaby_f = 0.0;
testaby_f = 0.0;
}//reset_testaby

void line_C::berechne_testaby_kehrwert()
{
one_contby_f = 1.0 / one_contby_f;
zero_contby_f = 1.0 / zero_contby_f;
contby_f = 1.0 / contby_f;
observaby_f = 1.0 / observaby_f;
testaby_f = 1.0 / testaby_f;
}//berechne_testaby_kehrwert

line_C::line_C(char *s_sC,int number_of_outputs_i):object_C(s_sC)
{
  // Konstruktor fuer Lines, Initialisierung der line
aequiklass_at_0_nr_i = 0;
aequiklass_at_1_nr_i = 0;
number_i = 0;
output_depth_i = 0;
path_blocked_b = FALSE;
input_kegel_b = FALSE;
output_kegel_b = FALSE;
prime_input_b = FALSE;
prime_output_b = FALSE;
pseudo_input_b = FALSE;
pseudo_output_b = FALSE;
fanout_b = FALSE;
set_in_forw_imp_b = FALSE;
feedback_b = FALSE;
hazard_occured_b = FALSE;
rising_test_b = FALSE;
falling_test_b = FALSE;
rising_aborted_b = FALSE;
falling_aborted_b = FALSE;
rising_untestable_b = FALSE;
falling_untestable_b = FALSE;
s_a_0_untestable_b = FALSE;
s_a_1_untestable_b = FALSE;
s_a_0_aborted_b = FALSE;
s_a_1_aborted_b = FALSE;
rising_non_robust_b = FALSE;
falling_non_robust_b = FALSE;
s_a_0_test_b = FALSE;
s_a_1_test_b = FALSE;
line_type_e = UNDEFINED_LINE;
init_value_e = VAL_X;
good_value_e = VAL_X;
bad_value_e = VAL_X;
only_value_e = VAL_X;
vorgaenger_p = new gate_liste_C(1);
if (vorgaenger_p == NULL)
	error("no memory available");
nachfolger_p = new gate_liste_C(number_of_outputs_i);
if (nachfolger_p == NULL)
	error("no memory available");
object_type_e = LINE;
plane_no_i = 0;
geprueft_b = FALSE;
signal_path_exist_b = FALSE;
test_done_b = FALSE;
stuck_at_test_b = FALSE;
one_contby_f = 0.0;
zero_contby_f = 0.0;
contby_f = 0.0;
observaby_f = 0.0;
testaby_f = 0.0;
static_dominator_p = NULL;
dynamic_dominator_p = NULL;
}

line_C::~line_C()
{
delete vorgaenger_p;
delete nachfolger_p;
}

//*********************************************************************
//
//   FUNCTION-NAME: verbinde
//
//   CLASS: line_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Ein Gate wird an eine Line angeschlossen. In abhaengigkeit
//	von port_e ist das Gate Vorgaenger oder Nachfolger. Wenn die Line schon
//	einen nachfolger hat und das Gate ist auch Nachfolger wird fanout_b gesetzt.
//
//   FORMAL PARAMETERS: gate_C *gate_item_p: das entsprechende Gatter
//		port_t port_e: INPUT fuer Vorgaenger, OUTPUT fuer Nachfolger
//		io_t gate_io_e: Der Port, an dem die Line an das Gatter angeschlossen wurde.
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: io_t: Der Port, an dem angeschlossen wird in Vorgaenger/Nachfolger
//				Liste
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void line_C::verbinde(gate_C *gate_item_p, char io_c)
{
if (io_c == 'I')
        {
        nachfolger_p->verbinde(gate_item_p);
        if ((nachfolger_p->no_elements() > 1) || (prime_output_b == TRUE))
                fanout_b = TRUE;
        }//if
else
        vorgaenger_p->verbinde(gate_item_p);
}//verbinde

void line_C::print_all(ofstream &datei_F)
{ // gibt die Belegung der Line aus (DEBUG)
datei_F << "Adresse   " << this << "\n";
datei_F << "Name:   " << name_sC << "\n";
datei_F << "Nummer:   " << number_i << "\n";
datei_F << "Im Eingangskegel:   " << input_kegel_b << "\n";
datei_F << "Im Ausgangskegel:   " << output_kegel_b << "\n";
datei_F << "Primaer Eingang:   " << prime_input_b << "\n";
datei_F << "Primaer Ausgang:   " << prime_output_b << "\n";
datei_F << "Fanout:   " << fanout_b << "\n";
datei_F << "Plane Number:   " << plane_no_i << "\n";
datei_F << "Output Depth:   " << output_depth_i << "\n";
datei_F << "Line Type:   " << line_type_e << "\n";
datei_F << "Element in Feedback:   " << feedback_b << "\n";
datei_F << "Init Value:   " << init_value_e << "\n";
datei_F << "Good Value:   " << good_value_e << "\n";
datei_F << "Bad Value:   " << bad_value_e << "\n";
datei_F << "Only Value:   " << only_value_e << "\n";
datei_F << "Vorgaenger:   " << vorgaenger_p << "\n";
datei_F << "Nachfolger:   " << nachfolger_p << "\n";
datei_F << "Rising Test erledigt:   " << rising_test_b << "\n";
datei_F << "Falling Test erledigt:   " << falling_test_b << "\n";
datei_F << "Stuck at 0 redundant:   " << s_a_0_untestable_b << "\n";
datei_F << "Stuck at 1 redundant:   " << s_a_1_untestable_b << "\n";
datei_F << "Stuck at 0 aborted:   " << s_a_0_aborted_b << "\n";
datei_F << "Stuck at 1 aborted:   " << s_a_1_aborted_b << "\n";
datei_F << "Test erledigt:   " << test_done_b << "\n";
datei_F << "Stuck at Test erledigt:   " << stuck_at_test_b << "\n";
datei_F << "Stuck at 0 Test erledigt:   " << s_a_0_test_b << "\n";
datei_F << "Stuck at 1 Test erledigt:   " << s_a_1_test_b << "\n";
datei_F << "0 Controllability:   " << zero_contby_f << "\n";
datei_F << "1 Controllability:   " << one_contby_f << "\n";
datei_F << "Gesamt Controllability:   " << contby_f << "\n";
datei_F << "Observability:   " << observaby_f << "\n";
datei_F << "Testability:   " << testaby_f << "\n";
datei_F << "##### Vorgaenger\n";
vorgaenger_p->print_all(datei_F);
datei_F << "##### Nachfolger\n";
nachfolger_p->print_all(datei_F);
datei_F << "\n\n***********\n";
}//print_all

//*********************************************************************
//
//   FUNCTION-NAME: calculate_planes
//
//   CLASS: line_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Fuer alle Nachfolgenden Gatter wird calculate_planes
//	aufgerufen. Die Nummer der Schaltungsebene, in der sich die Line befindet,
//	ist die hoechste Nummer des Vorgaengers + 1
//
//   FORMAL PARAMETERS: keine
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void line_C::calculate_planes()
{
static int line_number_i = 1;
int worst_plane_no_i = 0;
boolean fail_b = FALSE;
gate_C *gate_item_p;
line_C *line_item_p;

if (plane_no_i != 0)
	return;
if (prime_input_b == TRUE)
	plane_no_i = 1;
else
	{
	forall(line_item_p,*vorgaenger_p->gib_only_gate()->gib_inputs())
		{
		int akt_plane_no_i;
		
		akt_plane_no_i = line_item_p->gib_plane_no();
		if (akt_plane_no_i == 0)
			fail_b = TRUE; // nicht return, da Liste abgeschlossen werden muss
		worst_plane_no_i = (worst_plane_no_i > akt_plane_no_i) ? worst_plane_no_i
									: akt_plane_no_i;
			// Die schlechteste plane Nummer der Vorgaenger ermitteln
		}//for
	if (fail_b == TRUE)
		return; // plane_no_i eines Vorgaengers noch nicht berechnet
	plane_no_i = worst_plane_no_i + 1;
	}//else
number_i = line_number_i++;

	// Aufruf fuer alle Nachfolger
if (prime_output_b == FALSE)
	{
	gate_item_p = nachfolger_p->gib_only_gate();
	gate_item_p->calculate_planes();
	}//if
}//calculate_planes

//*********************************************************************
//
//   FUNCTION-NAME: calculate_depth_to_output
//
//   CLASS: line_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Fuer alle Vorgaenger Lines wird calculate_depth_to_output
//	aufgerufen. Die Entfernung der Schaltungsebene, in der sich die Line befindet,
//	zum Ausgang ist die niedrigste Nummer der Nachfolger + 1
//
//   FORMAL PARAMETERS: keine
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void line_C::calculate_depth_to_output()
{
int best_depth_no_i = max_number_lines_i;
boolean fail_b = FALSE;
gate_C *gate_item_p;
line_C *line_item_p;

if (output_depth_i != 0)
	return;
if (prime_output_b == TRUE)
	output_depth_i = 1;
else
	{
	forall(gate_item_p,*nachfolger_p)
	   {forall(line_item_p,*gate_item_p->gib_outputs())
		{
		int akt_output_depth_i;
		
		akt_output_depth_i = line_item_p->gib_output_depth();
		if (akt_output_depth_i == 0)
			fail_b = TRUE; // nicht return, da Liste abgeschlossen werden muss
		best_depth_no_i = (best_depth_no_i < akt_output_depth_i) ? best_depth_no_i
									: akt_output_depth_i;
			// Die beste Tiefe der Nachfolger ermitteln
		}}//forall
	if (fail_b == TRUE)
		return; // output_depth_i eines Nachfolgers noch nicht berechnet
	output_depth_i = best_depth_no_i + 1;
	}//else

	// Aufruf fuer alle Vorgaenger
if (prime_input_b == FALSE)
   {
   gate_item_p = vorgaenger_p->gib_only_gate();
   for (int i=0;i<gate_item_p->no_inputs();i++)
	(*gate_item_p->gib_inputs())[i]->calculate_depth_to_output();
   }//if
}//calculate_depth_to_output

//*********************************************************************
//
//   FUNCTION-NAME: calculate_type
//
//   CLASS: line_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Fuer alle Nachfolgenden Gatter wird calculate_type
//	aufgerufen. Es wird berechnet, ob die entsprechende Line eine FREE_LINE,
//	eine BOUND_LINE oder eine HEAD_LINE ist.
//
//   FORMAL PARAMETERS: keine
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void line_C::calculate_type()
{
line_t akt_line_type_e = FREE_LINE;
boolean fail_b = FALSE;
line_C *line_item_p;
gate_C *gate_item_p;

if (prime_input_b == TRUE)
	{  // Input ist immer FREE_LINE, wenn er Fanout hat sogar HEAD_LINE
	if ((prime_output_b == FALSE) && (nachfolger_p->gib_only_gate()->gib_type() == FANOUT))
		{
		line_type_e = HEAD_LINE;
		}//if
	else
		line_type_e = FREE_LINE;
	free_lines_C.fuege_ein(this);
	}//if
else
	{  // Line Typ des Vorgaengers ermitteln
	forall(line_item_p,*vorgaenger_p->gib_only_gate()->gib_inputs())
		switch (line_item_p->gib_type())
		   {
		   case UNDEFINED_LINE:
			fail_b = TRUE; // nicht return, da Liste abgeschlossen werden muss
			break;
		   case BOUND_LINE:
		   case HEAD_LINE:
			akt_line_type_e = BOUND_LINE;
			break;
		   case FREE_LINE:
			break; // nothing
		   }//switch
	if (fail_b == TRUE)
		return; // line_type_e eines Vorgaengers noch nicht berechnet
		// Line Typ berechnen
	if ((akt_line_type_e == FREE_LINE) && (prime_output_b == FALSE) &&
	    (nachfolger_p->gib_only_gate()->gib_type() == FANOUT))
		{
		line_type_e = HEAD_LINE;
		free_lines_C.fuege_ein(this);
		}//if
	else
		line_type_e = akt_line_type_e;
	if (line_type_e == FREE_LINE)
		free_lines_C.fuege_ein(this);
	}//else

	// Aufruf fuer alle Nachfolger
if (prime_output_b == FALSE)
	{
	gate_item_p = nachfolger_p->gib_only_gate();
	gate_item_p->calculate_type();
	}//if
}//calculate_type

//*********************************************************************
//
//   FUNCTION-NAME: test_done_or_aborted
//
//   CLASS: line_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Diese Funktion stellt fest, ob ein Element nicht mehr als
//	Zielfehler auftreten kann.
//
//   FORMAL PARAMETERS: keine
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: boolean: TRUE, wenn Element getestet oder alle Tests aborted
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
boolean line_C::test_done_or_aborted()
{
if (test_done_b == TRUE)	// Test bereits generiert
	return(TRUE);
if ((prime_input_b == FALSE) && ((*vorgaenger_p)[0]->gib_type() == BLOCK))
	return(FALSE);
if ((s_a_1_untestable_b == FALSE) && (s_a_1_aborted_b == FALSE))
   if ((s_a_0_untestable_b == TRUE) || (s_a_0_aborted_b == TRUE))
	if ((s_a_1_test_b == FALSE) && (s_a_1_untestable_b == FALSE) && (s_a_1_aborted_b == FALSE))
		return(FALSE);
	else
		return(TRUE);
if ((s_a_0_untestable_b == FALSE) && (s_a_0_aborted_b == FALSE))
   if ((s_a_1_untestable_b == TRUE) || (s_a_1_aborted_b == TRUE))
	if ((s_a_0_test_b == FALSE) && (s_a_0_untestable_b == FALSE) && (s_a_0_aborted_b == FALSE))
		return(FALSE);
	else
		return(TRUE);
if ((s_a_1_untestable_b == FALSE) && (s_a_1_aborted_b == FALSE))
   if (rising_test_b == FALSE)
	return(FALSE);
if ((s_a_0_untestable_b == FALSE) && (s_a_0_aborted_b == FALSE))
   if (falling_test_b == FALSE)
	return(FALSE);
return(TRUE);
}//test_done_or_aborted
