#ifndef line_C_h
#define line_C_h
#include "object_C.h"
#include "line_value_liste_C.h"
#ifdef FRAMEWORK
#include "cfidr.h"
#endif

//*********************************************************************
//
//   CLASS-NAME: line_C: Unterklasse von object_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   CLASS-DESCRIPTION: Diese Klasse definiert Lines und alle notwendigen
//	Operationen auf Lines.
//
//   OPERATOREN: =: Zuweisung
//
//   FUNKTIONEN: Konstruktor mit char* Argument: siehe friends.h
//		 Destruktor macht nichts
//		 set_primary_input: Line ist Primaereingang: prime_input_b
//			wird auf TRUE gesetzt
//		 set_primary_output: Line ist Primaerausgang: prime_output_b
//                      wird auf TRUE gesetzt
//		 print_all: alle Belegungen werden Ausgegeben fuer DEBUG
//		 verbinde: siehe friends.h
//		 compute_contby: siehe friends.h
//		 compute_observaby: siehe friends.h
//		 vorgaenger: gibt Liste mit Gattern aus, die Line als Ausgang
//			haben. (Die Liste sollte ein Element haben)
//		 nachfolger: gibt Gatterliste aus, die einen Input auf der Line
//			haben. (Elemente > 1 <==> Line hat fanout)
//		 calculate_planes: siehe friends.h
//		 calculate_type: siehe friends.h
//		 gib_plane_no: Die Nummer der Schaltungsebene, in der sich die
//			Line befindet wird Ausgegeben. Variable ist plane_no_i
//		 gib_type: Es wird Ausgegeben, ob die Line FREE_LINE, BOUND_LINE
//			oder HEAD_LINE ist. Variable line_type_e
//		 set_feedback: Line ist in Rueckkopplung. feedback_b wird auf
//			TRUE gesetzt.
//		 reset_value: test_value_e wird auf VAL_X gesetzt. init_value_e
//			wird auf letzte test_value_e gesetzt.
//		 should_be_tested: Belegt Line fuer den naechsten Rising oder Falling Test
//		 must_be_tested: Initialisiert  Line fuer den naechsten Rising oder
//			Falling Test
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
class line_C: public object_C {
public:
   line_C(char*,int = max_no_fanout_branches_i);
   ~line_C();
   boolean search_next_path(boolean&);
   boolean search_off_path_signals(boolean&);
   void set_primary_input(){prime_input_b = TRUE;}
   void set_primary_output(boolean b = TRUE){prime_output_b = b;}
   void set_pseudo_input(){pseudo_input_b = TRUE;}
   void set_pseudo_output(boolean b = TRUE){pseudo_output_b = b;}
   boolean is_primary_input(){return(prime_input_b);}
   boolean is_primary_output(){return(prime_output_b);}
   boolean is_pseudo_input(){return(pseudo_input_b);}
   boolean is_pseudo_output(){return(pseudo_output_b);}
   void print_all(ofstream &datei_F);
   void verbinde(gate_C*,char);
   void compute_contby();
   void compute_observaby();
   void prob_compute_contby();
   void prob_compute_observaby();
   void level_compute_contby();
   void level_compute_observaby();
   void reko_compute_contby();
   void reko_compute_observaby();
   inline double gib_zero_contby(){return(zero_contby_f);}
   inline double gib_one_contby(){return(one_contby_f);}
   inline double gib_observaby(){return(observaby_f);}
   inline double gib_testaby(){return(testaby_f);}
   inline gate_liste_C *vorgaenger(){return(vorgaenger_p);}
   inline gate_liste_C *nachfolger(){return(nachfolger_p);}
   void calculate_planes();
   void calculate_depth_to_output();
   void calculate_type();
   void learn(line_C *);
   void try_learning();
   fan_result_t dyn_learn(line_C *);
   fan_result_t dyn_try_learning();
   fan_result_t global_implication();
   void reset_dyn_learning();
   inline int gib_plane_no(){return(plane_no_i);}
   inline int gib_output_depth(){return(output_depth_i);}
   inline line_t gib_type(){return(line_type_e);}
   friend void line_structure();
   inline void set_feedback(){feedback_b = TRUE;}
   void reset_value(fan_result_t result_e);
   boolean gib_fanout(){return(fanout_b);}
   boolean is_set_in_forw_imp(){return(set_in_forw_imp_b);}
   void set_in_forw_imp(){set_in_forw_imp_b = TRUE;}
   void reset_in_forw_imp(){set_in_forw_imp_b = FALSE;}
   void set_good_value(value_t val_e){good_value_e = val_e;}
   void set_bad_value(value_t val_e){bad_value_e = val_e;}
   boolean must_be_tested(value_t);
   boolean should_be_tested(value_t);
   boolean test_done_or_aborted();
   value_t get_good_value(){return(good_value_e);}
   value_t get_bad_value(){return(bad_value_e);}
   value_t get_init_value(){return(init_value_e);}
   boolean check_generated_test(line_C *,value_t);
   void check_iddq_test(value_t);
   void reset_non_robust();
   void reset_aborted_faults();
   void set_s_a_0_aborted(boolean b){s_a_0_aborted_b = b;}
   void set_s_a_1_aborted(boolean b){s_a_1_aborted_b = b;}
   void set_s_a_0_untestable(){s_a_0_untestable_b = TRUE;}
   void set_s_a_1_untestable(){s_a_1_untestable_b = TRUE;}
   void set_s_a_0_test(){s_a_0_test_b = TRUE;}
   void set_s_a_1_test(){s_a_1_test_b = TRUE;}
   void set_stuck_at_test(){stuck_at_test_b = TRUE;}
   boolean get_s_a_0_aborted(){return(s_a_0_aborted_b);}
   boolean get_s_a_1_aborted(){return(s_a_1_aborted_b);}
   boolean get_s_a_0_untestable(){return(s_a_0_untestable_b);}
   boolean get_s_a_1_untestable(){return(s_a_1_untestable_b);}
   boolean get_s_a_0_test(){return(s_a_0_test_b);}
   boolean get_s_a_1_test(){return(s_a_1_test_b);}
   boolean get_stuck_at_test(){return(stuck_at_test_b);}
   void mark_result(fan_result_t,value_t);
   friend void print_result();
   friend void nachanalyse();
   void hazard_occured(){hazard_occured_b = TRUE;}
   boolean in_input_kegel(){return(input_kegel_b);}
   boolean in_output_kegel(){return(output_kegel_b);}
   boolean path_blocked(){return(path_blocked_b);}
   void set_path_blocked(){path_blocked_b = TRUE;}
   void reset_path_blocked(){path_blocked_b = FALSE;}
   void mark_input_kegel(){input_kegel_b = TRUE;}
   void mark_output_kegel(){output_kegel_b = TRUE;}
   void clear_kegel(){input_kegel_b = FALSE;output_kegel_b = FALSE;}
   boolean lookup_testability();
   value_t get_only_value(){return(only_value_e);}
   void set_only_value(value_t val_e){only_value_e = val_e;}
   boolean geprueft_b; // wird in feedback Analyse und Fehlersimulation gebraucht
   boolean signal_path_exist_b;	// wird in Fehlersimulation gebraucht
   void reset_test_done(){test_done_b = FALSE;}
   void set_test_done(){test_done_b = TRUE;}
   boolean get_test_done(){return(test_done_b);}
   void reset_testaby();
   void berechne_testaby_kehrwert();
   void set_static_dominator(line_C *line_item_p){static_dominator_p = line_item_p;}
   void set_dynamic_dominator(line_C *line_item_p){dynamic_dominator_p = line_item_p;}
   void clear_dynamic_dominator(){dynamic_dominator_p = NULL;}
   line_C *get_static_dominator(){return(static_dominator_p);}
   line_C *get_dynamic_dominator(){return(dynamic_dominator_p);}
   friend boolean exist_signal_path_to_output(line_C*,value_t);
   friend void analyse_signal_path(line_C*);
   int gib_number(){return(number_i);}
   int gib_aequiklass_at_0_nr_i(){return(aequiklass_at_0_nr_i);}
   int gib_aequiklass_at_1_nr_i(){return(aequiklass_at_1_nr_i);}
#ifdef FRAMEWORK
   friend void cfidr_ConnectMacros(cfidrViewIdT);
#endif
   int aequiklass_at_0_nr_i;
   int aequiklass_at_1_nr_i;
private:
   int number_i;
   int output_depth_i;		// Abstand zum Primaerausgang
   boolean path_blocked_b;	// Ist Pfad zum Ausgang blockiert
   boolean input_kegel_b;	// Ist Line im Input Kegel der Ziel Line
   boolean output_kegel_b;	// Ist Line im Output Kegel der Ziel Line
   boolean prime_input_b;	// Ist Line Primaereingang
   boolean prime_output_b;	// Ist Line Primaerausgang
   boolean pseudo_input_b;	// Ist Line Pseudoeingang
   boolean pseudo_output_b;	// Ist Line Pseudoausgang
   boolean fanout_b;		// Hat Line Fanout
   boolean set_in_forw_imp_b;	// Hat Line Fanout
   boolean feedback_b;		// Ist Line in Rueckkopplung
   boolean hazard_occured_b;	// Ist Line Hazardbehaftet
   boolean rising_test_b;	// Ist rising Test durchgefuehrt
   boolean falling_test_b;	// Ist falling Test durchgefuehrt
   boolean rising_aborted_b;	// Ist rising Test aborted
   boolean falling_aborted_b;	// Ist falling Test aborted
   boolean rising_untestable_b;	// Ist rising Test unmoeglich
   boolean falling_untestable_b;// Ist falling Test unmoeglich
   boolean s_a_0_untestable_b;	// Ist Line nicht testbar auf stuck-at 0
   boolean s_a_1_untestable_b;	// Ist Line nicht testbar auf stuck-at 1
   boolean s_a_0_aborted_b;	// Ist Line aborted Fault auf stuck-at 0
   boolean s_a_1_aborted_b;	// Ist Line aborted Fault auf stuck-at 1
   boolean falling_non_robust_b;// Ist Line non_robust auf Falling
   boolean rising_non_robust_b;	// Ist Line non_robust auf Rising
   boolean test_done_b;		// Ist beabsichtigter Test durchgefuehrt
   boolean stuck_at_test_b;	// Ist stuck at Test durchgefuehrt
   boolean s_a_0_test_b;	// Ist stuck at 0 Test durchgefuehrt
   boolean s_a_1_test_b;	// Ist stuck at 1 Test durchgefuehrt
   line_t line_type_e;		// Line Typ (FREE_LINE, BOUND_LINE, HEAD_LINE)
   value_t init_value_e;	// Init Value
   value_t good_value_e;	// Gut-Wert
   value_t bad_value_e;		// Fehler-Wert
   value_t only_value_e;	// Einzig moegliche Value
   gate_liste_C *vorgaenger_p;	// Vorgaengerliste
   gate_liste_C *nachfolger_p;	// Nachfolger Liste
   line_value_liste_C global_implications[2];	// globale Implikationen
   line_value_liste_C tmp_global_implications[2];	// temporaere globale Implikationen
   int plane_no_i;		// Nummer der Ebene, in der Line sich befindet
   double one_contby_f;
   double zero_contby_f;
   double contby_f;
   double observaby_f;
   double testaby_f;
   line_C *static_dominator_p;
   line_C *dynamic_dominator_p;
};
#endif
