#include "line_liste_C.h"
#include "gate_liste_C.h"
#include "string_C.h"
#include "line_C.h"
#include "gate_C.h"
#include "modify.h"

extern line_liste_C lines_C;
extern line_liste_C tmp_line_liste_C;
extern line_C *search(char*,line_tree_u*);
extern void milef_add_to_line_tree(line_tree_u*&,line_C*,boolean = FALSE);
// *********************************************************************
//
//   CLASS-NAME: line_liste_C Unterklasse von liste_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   CLASS-DESCRIPTION: Diese Klasse definiert Linelisten und alle notwendigen
//	Operationen auf Linelisten.
//
//   OPERATOREN: keine
//
//   FUNKTIONEN: 2 Konstruktoren mit Integer und port sowie ohne Argument
//			ohne Argumente => generische Liste
//		 Destruktor loescht Liste
//		 print_all: gibt Inhalt aller Listenelemente aus (DEBUG)
//		 fuege_ein: fuegt Element ein (siehe friends.h)
//		 fuege_aus: fuegt Element aus Liste aus (gate_C* Argument)
//		 element_in_liste: prueft, ob Element gate_C* in Liste ist
//		 next: gibt aktuelles Element von Liste aus (aktuell entspricht
//			aktual_no_i), wenn Liste geschlossen ist (closed_b == TRUE)
//			beim letzten Element wird Liste geoeffnet
//		 search: siehe friends.h
//		 verbinde: siehe friends.h
//
//   MODIFICATION-HISTORY:
//
// *********************************************************************
line_liste_C::line_liste_C(int listengroesse_i):liste_C(listengroesse_i)
{
liste_a = new line_C *[listengroesse_i];
if (liste_a == NULL)
	error("No more memory available");
}

void line_liste_C::init(int size_i)
{
liste_a = new line_C *[size_i];
if (liste_a == NULL)
	error("No more memory available");
}

void line_liste_C::resize()
{
line_liste_C *tmp_C;
int i;

tmp_C = new line_liste_C;
tmp_C->init(number_elements_i);
for (i=0;i<number_elements_i;i++)
	tmp_C->fuege_ein(liste_a[i]);
delete[] liste_a;
liste_a = new line_C *[number_elements_i];
for (i=0;i<number_elements_i;i++)
	liste_a[i] = (*tmp_C)[i];
delete tmp_C;
}

void line_liste_C::print_all(ofstream& datei_F)
{
liste_C::print_all(datei_F);
for (int i=0;i<number_elements_i;i++)
datei_F << "\t" << "element[" << i << "]:   " << liste_a[i]->get_name() << "\n";
}

line_C *line_liste_C::gib_only_line()
{
if (number_elements_i != 1)
	error("gib_only_line: more than 1 gate in list");
return(liste_a[0]);
}//gib_only_line

void line_liste_C::change_connection(char *name_s,line_C *line_item_p)
{
int i;

for (i=0;i<number_elements_i;i++)
   if (strcasecmp(liste_a[i]->get_name(),name_s) == 0)
	{
	liste_a[i] = line_item_p;
	break;
	}//if
}//change_connection

//*********************************************************************
//
//   FUNCTION-NAME: element_in_liste
//
//   CLASS: line_liste_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Diese Funktion stellt fest, ob eine Line in der Liste
//	vorhanden ist.
//
//   FORMAL PARAMETERS: line_C *line_item_p: die Line, die gesucht wird
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: boolean: TRUE, wenn die Line in der Liste ist
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
boolean line_liste_C::element_in_liste(line_C *line_item_p,boolean fuege_aus_b)
{
int i=0;
line_C *ausgefuegtes_element_p;

if (this == &lines_C)
	{
	if (fuege_aus_b == TRUE)
		error("element_in_liste: fuege_aus in lines_C isn't allowed");
	line_C *line_item_p = search(line_item_p->get_name(),milef_line_tree_root_p);
	if (line_item_p == NULL)
		return FALSE;
	return TRUE;
	}//if
for (;i < number_elements_i;i++)
   if (liste_a[i] == line_item_p)
	if (fuege_aus_b == FALSE)
		return TRUE;
	else
		{
                ausgefuegtes_element_p = liste_a[i];
                break;
		}//else
if ((fuege_aus_b == FALSE) || (i >= number_elements_i))
	return FALSE;
liste_C::fuege_aus();
for(;i<number_elements_i;i++)
	liste_a[i] = liste_a[i+1];
liste_a[number_elements_i] = ausgefuegtes_element_p;
return TRUE;
}/*element_in_liste*/

void line_liste_C::fuege_ein(line_C *line_item_p)
{
#if DEBUG
if (this->element_in_liste(line_item_p) == TRUE)
	{
	warning("fuege_ein: Line is already in list, is not inserted");
	return;
	}//if
if (this == &unjustified_bound_line_C)
	if (line_item_p->get_test_value() == VAL_X)
		error("line_liste_C::fuege_ein: Unjustified bound Line has Test Value VAL_X");
	else if (element_in_liste(line_item_p) == TRUE)
		error("line_liste_C::fuege_ein: Unjustified bound Line was already in Liste");
#endif
if (this == &lines_C) // baut den AVL Baum fuer lines_C auf
	milef_add_to_line_tree(milef_line_tree_root_p,line_item_p);
liste_C::fuege_ein();
liste_a[number_elements_i - 1] = line_item_p;
if (this == &tmp_line_liste_C)
   {
   int upper_i=number_elements_i-1,lower_i=-1,middle_i=0;
   line_C *hilf_p;

   hilf_p = liste_a[number_elements_i-1];
   while (lower_i < upper_i - 1)
   	{
   	middle_i = (upper_i+lower_i)/2;
   	if (((*liste_a[middle_i]->nachfolger()->gib_only_gate()->gib_outputs())[0]->gib_plane_no() > (*line_item_p->nachfolger()->gib_only_gate()->gib_outputs())[0]->gib_plane_no()) ||
   	    (((*liste_a[middle_i]->nachfolger()->gib_only_gate()->gib_outputs())[0]->gib_plane_no() == (*line_item_p->nachfolger()->gib_only_gate()->gib_outputs())[0]->gib_plane_no()) &&
   	     (liste_a[middle_i]->gib_number() < line_item_p->gib_number())))
   		lower_i = middle_i;
   	else
   		upper_i = middle_i;
   	}//while
   for (middle_i=number_elements_i-1;middle_i>upper_i;middle_i--)
   	liste_a[middle_i] = liste_a[middle_i-1];
   liste_a[upper_i] = hilf_p;
   }//if
return;
}/*fuege_ein*/

void line_liste_C::fuege_aus(line_C *line_item_p,boolean ignore_closed_b)
{
int i;
line_C *ausgefuegtes_element_p;

if ((ignore_closed_b == FALSE) || (number_elements_i == 0))
   liste_C::fuege_aus();
else
   number_elements_i--;
if (liste_a[number_elements_i] == line_item_p)
	return;
for (i=0;i <= number_elements_i;i++)
	if (liste_a[i] == line_item_p)
		{
		ausgefuegtes_element_p = liste_a[i];
		break;
		}//if
if (i > number_elements_i)
	error("fuege_aus: element not found in list");
for(;i<number_elements_i;i++)
	liste_a[i] = liste_a[i+1];
liste_a[number_elements_i] = ausgefuegtes_element_p;
return;
}/*fuege_aus*/

//*********************************************************************
//
//   FUNCTION-NAME: verbinde
//
//   CLASS: line_liste_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Ein Line wird in die Input/Output Liste eines
//	Gatters eingefuegt. Der Port, der Line wird in connected_to_a an der
//	entsprechenden Stelle vermerkt
//
//   FORMAL PARAMETERS: line_C *line_item_p: die entsprechende Line
//		io_t conn_io_e: Der Port der Line, wo das Gatter angeschlossen wird
//		int io_no_i: Der Index, an den die Line angeschlossen wird in Liste
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void line_liste_C::verbinde(line_C *line_item_p, int io_no_i)
{
number_elements_i++;
if (number_elements_i > max_no_io_i)
	error("some gate has too many inputs or outputs");
liste_a[io_no_i - 1] = line_item_p;
}//verbinde

void line_liste_C::merge_list(line_liste_C *list_p)
{
line_C **hilf_liste_p;
int i;

number_elements_i += list_p->no_elements()-1;
hilf_liste_p = new line_C *[number_elements_i];
for (i=0;i<number_elements_i-list_p->no_elements()+1;i++)
	hilf_liste_p[i] = liste_a[i];
for (i=number_elements_i-list_p->no_elements()+1;i<number_elements_i;i++)
	hilf_liste_p[i] = (*list_p)[i-number_elements_i+list_p->no_elements()];
delete[] liste_a;
liste_a = hilf_liste_p;
}
