//*********************************************************************
//
//   FILE-NAME: praeproc.c
//
//   CREATION-DATE: August 91
//
//   FILE-DESCRIPTION: Dieses File beinhaltet Praeprocessing Routinen
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
#include "extern.h"
#include <sys/types.h>
#include <sys/times.h>
#include "protocol_stack_C.h"
#include "gate_liste_C.h"
#include "gate_C.h"
#include "line_liste_C.h"
#include "line_line_liste_C.h"
#include "line_C.h"
#include "modify.h"

void feedback_analyse();
void line_structure();
void suche_dominatoren();
void calculate_planes();
void read_bridging_fault_list();
void calculate_bridgings();
void test_feedback(line_liste_C&,line_C*);
void mark_feedback(line_liste_C&,line_C*);
void learning_implications();
void constraint_analysis();
extern void reset_net(fan_result_t);
extern value_t mynot(value_t);
extern void set_time();
extern void add_time(clock_t&);
extern void gib_aequi_klassen_zahl();
extern int compare_zielfehler(const void*,const void*);
extern line_C *signal_to_propagate(line_C *, line_C *);

extern line_liste_C unjustified_bound_line_C;
extern line_liste_C implication_line_C;
extern line_value_liste_C zielfehler_C;
extern line_line_liste_C zielfehler2_C;
extern protocol_stack_C protocol_C;
extern line_liste_C tmp_line_liste_C;
long int imp_zeit_l;
long int constr_zeit_l;
int no_imp_calls_i = 0;
extern boolean only_iddq_test_b;
extern boolean only_stuck_at_test_b;
extern boolean stuck_at_test_without_branches_b;
extern boolean transition_test_b;
extern boolean stuck_open_test_b;
extern boolean bridging_iddq_b;
extern boolean bridging_b;
extern boolean iddq_stuck_on_test_b;
boolean flt_read_b = FALSE;
boolean something_learned_b = TRUE;
int learned_implications_i = 0;
int nr_redFaults = 0;

//*********************************************************************
//
//   FUNCTION-NAME: read_flt
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: Mai 93
//
//   FUNCTIONAL-DESCRIPTION: Einlesen der Fehlerliste (<circuit>.flt)
//
//   FORMAL PARAMETERS: ifstream& flt_F: Der File
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void read_flt(ifstream& flt_F)
{
line_C *line_item_p;
boolean line_found_b;

flt_read_b = TRUE;
char s1[stringsize_i],s2[stringsize_i],s3[stringsize_i];
flt_F >> s1 >> s2;
while (!flt_F.eof())
	{
	if (((only_stuck_at_test_b == TRUE) || (only_iddq_test_b == TRUE)) && 
	    (strcasecmp(s1,"s-a-0") != 0) && (strcasecmp(s1,"s-a-1") != 0))
		{
		flt_F >> s1 >> s2;
		continue;
		}//if
	if (((stuck_open_test_b == TRUE) || (transition_test_b == TRUE)) && 
	    (strcasecmp(s1,"falling") != 0) && (strcasecmp(s1,"rising") != 0))
		{
		flt_F >> s1 >> s2;
		continue;
		}//if
	if ((iddq_stuck_on_test_b == TRUE) && 
	    (strcasecmp(s1,"i-s-on-0") != 0) && (strcasecmp(s1,"i-s-on-1") != 0))
		{
		flt_F >> s1 >> s2;
		continue;
		}//if
	line_found_b = FALSE;
	forall(line_item_p,lines_C)
	  {
	  strcpy(s3,(char*)line_item_p->get_name());
	  if ((line_item_p->is_primary_input() == FALSE) && (line_item_p->is_pseudo_input() == FALSE) &&
	      ((*line_item_p->vorgaenger())[0]->gib_type() == FANOUT) && (strlen(s3) > strlen(s2)))
	    {
	    int i;

	    for (i=strlen(s3);s3[i]!='_';i--);
	    s3[i] = '\0';
	    }//if
	  if (strcmp(s3,s2) == 0)
	    {
	    line_found_b = TRUE;
	    if (((only_stuck_at_test_b == TRUE) || (only_iddq_test_b == TRUE)) && 
		(strcasecmp(s1,"s-a-0") == 0))
	      zielfehler_C.fuege_ein(line_item_p,VAL_0);
	    else if (((only_stuck_at_test_b == TRUE) || (only_iddq_test_b == TRUE)) &&
		     (strcasecmp(s1,"s-a-1") == 0))
	      zielfehler_C.fuege_ein(line_item_p,VAL_1);
	    else if ((only_stuck_at_test_b == FALSE) && (only_iddq_test_b == FALSE) &&
		     ((strcasecmp(s1,"falling") == 0) || (strcasecmp(s1,"i-s-on-0") == 0)))
	      zielfehler_C.fuege_ein(line_item_p,VAL_U0);
	    else if ((only_stuck_at_test_b == FALSE) && (only_iddq_test_b == FALSE) &&
		     ((strcasecmp(s1,"rising") == 0) || (strcasecmp(s1,"i-s-on-1") == 0)))
	      zielfehler_C.fuege_ein(line_item_p,VAL_U1);
	    }//if
	  }//forall
	if (line_found_b == FALSE)
	  warning(string_C(s2) + " not found in circuit");
	flt_F >> s1 >> s2;
	}//while
cout << "\t\t" << zielfehler_C.no_elements() << " target faults demanded\n" << flush;
qsort(zielfehler_C.gib_liste(),zielfehler_C.no_elements(),sizeof(line_value_t*),compare_zielfehler);
}//read_flt

//********************************************************
//
// FUNCTION NAME: search_line_in_tree 
//
// CREATION DATE: Januar 1994
//
// FUNCTIONAL DESCRIPTION: Suchen einer `line` im AVL-Baum
//
//*******************************************************
line_C* search_line_in_tree(line_tree_u *tree_p, char *linename_s)
{
if (tree_p == NULL)
   return(NULL);

char *actual_name_s = (char*)tree_p->line_p->get_name();
int i;

if ((i = strcmp (linename_s, actual_name_s)) == 0)
   return(tree_p->line_p);

else if (i > 0)
   return(search_line_in_tree(tree_p->right, linename_s));
 
else if (i < 0)
   return(search_line_in_tree(tree_p->left, linename_s));

}//search_line_in_tree

//*********************************************************************
//
//   FUNCTION-NAME: praeprocessing
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Praeprocessing Steuerung: Aufruf von feedback_analyse,
//	print_statistic, calculate_planes und line_structure
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
//extern boolean in_kegel(line_C *, line_C *);

void praeprocessing(boolean fault_simu_b)
{
gate_C *gate_item_p;
int learning_phases_i = 0;
ifstream fliste_F;

cout << "\tFeedback Analysis\n" << flush;
feedback_analyse();
cout << "\tPrint Circuit Statistic\n" << flush;
print_statistic();
cout << "\tFind Planes\n" << flush;
calculate_planes();
if ((bridging_iddq_b == TRUE) || (bridging_b == TRUE))
   {
   fliste_F.open((char*)(string_C(schaltungsname_s) + ".BF_flt"));  // (eventuell auch Zeile 314 aendern)
   if (!fliste_F)
      {
      cout << "\tCalculate Bridging-Fault List\n" << flush;
      calculate_bridgings();
      }//if	
   else
      {
      cout << "\tRead Bridging-Fault List\n" << flush;
      read_bridging_fault_list();
      }//else
   }//if
cout << "\tFind Free-Lines, Bound-Lines and Head-Lines\n" << flush;
line_structure();
cout << "\tFind Dominators to Fanouts\n" << flush;
suche_dominatoren();
//if (statistic_u.number_flipflops_i > 0)
//	exit(0);
cout << "\tGet Block Cubes, if there are any Blocks\n" << flush;
{forall(gate_item_p,gates_C)
   if (gate_item_p->gib_type() == BLOCK)
	gate_item_p->get_block_cube();
}
if ((bridging_iddq_b == FALSE) && (bridging_b == FALSE))
   {
   fliste_F.open((char*)(string_C(schaltungsname_s) + ".flt"));
   if (fliste_F == NULL)
	gib_aequi_klassen_zahl();
   else
	{
	flt_read_b = TRUE;
	read_flt(fliste_F);
	}//else
   }//if
if ((only_iddq_test_b == FALSE) && (fault_simu_b == FALSE))
	{
	while (something_learned_b == TRUE)
		{
		something_learned_b = FALSE;
		cout << "\tLearning Implications\n" << flush;
		learning_implications();
		learning_phases_i++;
		}//while
	cout << "\t" << learning_phases_i << " learn phases used\n";
	cout << "\t" << learned_implications_i << " global implications learned\n";
	cout << "\tReconvergency Analysis for Blocks\n" << flush;
	constraint_analysis();
	}//if
cout << "\tSend Constraints to Switch-Level\n" << flush;
{forall(gate_item_p,gates_C)
   if (gate_item_p->gib_type() == BLOCK)
	gate_item_p->put_constraint_cube();
}
//datei_F.open("Constraints");
//datei_F.close();
//forall(gate_item_p,gates_C)
//	gate_item_p->print_constr();
}

//*********************************************************************
//
//   FUNCTION-NAME: calculate_planes
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Schaltungsebenen werden ermittelt. Fuer alle
//	Primaereingaenge wird calculate_planes aufgerufen
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void calculate_planes()
{
line_C *line_item_p;
boolean error_b = FALSE;

forall(line_item_p,input_lines_C)
	line_item_p->calculate_planes();
{forall(line_item_p,output_lines_C)
	line_item_p->calculate_depth_to_output();}
{forall(line_item_p,lines_C)
	{
	if (line_item_p->gib_plane_no() == 0) // || (line_item_p->gib_output_depth() == 0))
		{
		cout << line_item_p->get_name();
		error_b = TRUE;
		warning("calculate_planes: not all line planes are discovered");
		}//if
	if (line_item_p->gib_output_depth() == 0)
		{
		cout << line_item_p->get_name();
		error_b = TRUE;
		warning("calculate_planes: not all output depths are discovered");
		}//if
	}}//forall
if (error_b == TRUE)
	error(" ");
}//calculate_planes

//*************************************************************
//
// FUNCTION NAME: read_bridging_fault_list 
//
// CREATION DATE: Januar 1994
//
// FUNCTIONAL DESCRIPTION: Einlesen der Bridging-Fault Zielfehlerliste
//
//*************************************************************
void read_bridging_fault_list()
{
line_C *line1_item_p, *line2_item_p;
char s1[stringsize_i],s2[stringsize_i],s3[stringsize_i];

ifstream fliste_F;
fliste_F.open((char*)(string_C(schaltungsname_s) + ".BF_flt"));  // (eventuell auch Zeile 207 aendern)

fliste_F >> s1 >> s2 >> s3;
for(;;)
   {
    if ((bridging_iddq_b == TRUE) || (bridging_b == TRUE) && (strcasecmp(s1,"bridge") == 0))
        {
	line1_item_p = search_line_in_tree(milef_line_tree_root_p,s2);
	line2_item_p = search_line_in_tree(milef_line_tree_root_p,s3);
        if ((line1_item_p != NULL) && (line2_item_p != NULL))
 	   zielfehler2_C.fuege_ein(line1_item_p,line2_item_p);
        fliste_F >> s1 >> s2 >> s3;
	if (fliste_F.eof())
	   break;
        }//if
    }//for
cout << "\t\t" << zielfehler2_C.no_elements() << " target faults demanded\n" << flush;
}//read_bridging_fault_list

//**************************************************************
//
// FUNCTION NAME: calculate_bridgings 
//
// CREATION DATE: Januar 1994
//
// FUNCTIONAL DESCRIPTION: Berechnung der Bridging-Fault Zielfehlerliste
//
//*************************************************************
extern boolean in_kegel(line_C *, line_C *);

void calculate_bridgings()
{
    line_C *line1_item_p, *line2_item_p;
    line_C *line_weak_item_p;
    int hilf_i, nr_BrFaults = 0;
    ofstream datei_F;

    datei_F.open((char *)(string_C(schaltungsname_s) + ".stat"),ios::app);
    datei_F << "\n\n\nElectric Untestable Faults:" << "\n";

    forall(line1_item_p,lines_C)                      // Jeder Knoten einer Schaltung wird mit jedem
       {                                              // anderen verglichen (verschachteltes `forall`)
       if ((line1_item_p->is_primary_input() == FALSE) && ((*line1_item_p->vorgaenger())[0]->gib_type() == FANOUT))
	  continue;
       hilf_i = forall_index;
         
       {forall(line2_item_p,lines_C)
         {
         if (forall_index == 0)
	    {
	    forall_index = hilf_i;
  	    continue;
  	    }//if
         if ((line2_item_p->is_primary_input() == FALSE) && ((*line2_item_p->vorgaenger())[0]->gib_type() == FANOUT))
	    continue;

	 if (bridging_iddq_b == TRUE)
	    zielfehler2_C.fuege_ein(line1_item_p,line2_item_p);
	 else
	    {
	    line_weak_item_p = signal_to_propagate(line1_item_p, line2_item_p);
	    if (line_weak_item_p == NULL)
	       {
	       nr_redFaults++;
	       datei_F << line1_item_p->get_name() << "  <-->  " << line2_item_p->get_name() << "\n";
	       continue;
	       }//if
	    else
	       {
	       zielfehler2_C.fuege_ein(line1_item_p,line2_item_p);
	       nr_BrFaults++;
	       }//else
	    }//else
       }//forall
    }
   }//forall
datei_F.close();
}//calculate_bridgings

//*******************************************************************
//
//   FUNCTION-NAME: line_structure
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Line Typen werden ermittelt. Fuer alle
//	Primaereingaenge wird calculate_type aufgerufen
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void line_structure()
{
line_C *line_item_p,*line2_p;

{forall(line_item_p,input_lines_C)
	line_item_p->calculate_type();}
forall(line_item_p,lines_C)
	if (line_item_p->gib_type() == UNDEFINED_LINE)
		error("line_structure: not all lines are discovered");
	else if ((line_item_p->gib_type() == FREE_LINE) && (line_item_p->is_primary_output() == FALSE))
		{forall(line2_p,*line_item_p->nachfolger()->gib_only_gate()->gib_outputs())
			if (line2_p->gib_type() == BOUND_LINE)
				line_item_p->line_type_e = HEAD_LINE;
				}//if
}//line_structure

//*********************************************************************
//
//   FUNCTION-NAME: feedback_analyse
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Rueckkopplungen werden ermittelt. Fuer alle
//	Primaereingaenge wird test_feedback aufgerufen. Hierbei wird eine
//	Liste uebergeben, die fuer die feedback Analyse gebraucht wird.
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void feedback_analyse()
{
line_liste_C feedback_liste_C;
line_C *line_item_p;

feedback_liste_C.init(statistic_u.number_lines_i);
forall(line_item_p,input_lines_C)
	{
	line_C *hilf_line_p;
	
	test_feedback(feedback_liste_C,line_item_p);
	forall(hilf_line_p,lines_C)
		hilf_line_p->geprueft_b = FALSE;
	}//for
}//feedback_analyse

//*********************************************************************
//
//   FUNCTION-NAME: test_feedback
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Rekursive Funktion, die nach einem Algorithmus von
//	Gundlach alle Rueckkopplungen und alle Pfadrekonvergenzen findet.
//	(Literatur bei mir oder Uwe Huebner)
//
//   FORMAL PARAMETERS: line_liste_C& feedback_liste_C: Hilfsliste fuer Feedback finden
//		line_C *line_item_p: die aktuelle Line
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void test_feedback(line_liste_C& feedback_liste_C,line_C *line_item_p)
{ // Algorithmus von Gundlach
gate_C *gate_item_p;
line_C *line2_p,*hilf_line_p;

if (line_item_p->geprueft_b == TRUE)
	{
	gate_item_p = line_item_p->vorgaenger()->gib_only_gate();
	gate_item_p->set_reconvergence();
	forall(hilf_line_p,*(gate_item_p->gib_inputs()))
		if (hilf_line_p->geprueft_b == TRUE)
			hilf_line_p->set_reconvergence();
	feedback_liste_C[feedback_liste_C.no_elements()-1]->set_reconvergence();
	return; // bereits auf feedback ueberprueft
	}//if

if (feedback_liste_C.element_in_liste(line_item_p) == TRUE)
	{
	statistic_u.feedback_b = TRUE;
	mark_feedback(feedback_liste_C,line_item_p); // feedback gefunden
	return;
	}//if

feedback_liste_C.fuege_ein(line_item_p);

	// rekursiver Aufruf
if (line_item_p->is_primary_output() == FALSE)
	{
	int i;

	gate_item_p = line_item_p->nachfolger()->gib_only_gate();
	for (i=0;i<gate_item_p->no_outputs();i++)
		{
		line2_p = (*gate_item_p->gib_outputs())[i];
		test_feedback(feedback_liste_C,line2_p);
		}//for
	}//if

feedback_liste_C.fuege_aus(line_item_p);
line_item_p->geprueft_b = TRUE;
}//test_feedback

//*********************************************************************
//
//   FUNCTION-NAME: mark_feedback
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Alle Elemente, die in der Rueckkopplung sind werden
//	markiert mittels der feedback Liste. Das erste Element in der Rueckkopplung
//	ist line_item_p.
//
//   FORMAL PARAMETERS: line_liste_C& feedback_liste_C: Hilfsliste fuer Feedback markieren
//		line_C *line_item_p: Das erste Element in der Rueckkopplung
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void mark_feedback(line_liste_C& feedback_liste_C,line_C *line_item_p)
{
line_C *hilf_line_p;
boolean after_line_b = FALSE;

line_item_p->vorgaenger()->gib_only_gate()->set_feedback();
line_item_p->set_feedback();

forall(hilf_line_p,feedback_liste_C)
        {
                //Suche nach erstem Element in feedback = line_item_p
        if ((after_line_b == FALSE) && (hilf_line_p != line_item_p))
                continue;
        after_line_b = TRUE;
                // setzen des Feedback flags in Line
        hilf_line_p->vorgaenger()->gib_only_gate()->set_feedback();
        hilf_line_p->set_feedback();
	}//forall
}//mark_feedback

//*********************************************************************
//
//   FUNCTION-NAME: suche_dominator
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Fuer alle Fanout-Staemme wird sche_dominator aufgerufen
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void suche_dominator(line_C *line_p)
{
line_C *line_item_p;
gate_C *gate_item_p;
boolean dummy_b;

tmp_line_liste_C.loesche_liste();
forall(line_item_p,*(*line_p->nachfolger())[0]->gib_outputs())
   if (line_item_p->is_primary_output() == TRUE)
	return;
   else
	tmp_line_liste_C.fuege_ein(line_item_p);
for(;;)
	{
	if (tmp_line_liste_C.no_elements() == 1)
		{
		line_p->set_static_dominator(tmp_line_liste_C[0]);
		break;
		}//if
	line_item_p = tmp_line_liste_C[tmp_line_liste_C.no_elements()-1];
	if (line_item_p->is_primary_output() == TRUE)
		break;
	gate_item_p = (*line_item_p->nachfolger())[0];
	forall(line_item_p,*gate_item_p->gib_inputs())
		dummy_b = tmp_line_liste_C.element_in_liste(line_item_p,TRUE);
	{forall(line_item_p,*gate_item_p->gib_outputs())
		if (line_item_p->is_primary_output() == FALSE)
			tmp_line_liste_C.fuege_ein(line_item_p);
		else
			if (tmp_line_liste_C.no_elements() == 0)
				{
				line_p->set_static_dominator(tmp_line_liste_C[0]);
				return;
				}//if
			else
				return;}
	}//forever
}//suche_dominator

//*********************************************************************
//
//   FUNCTION-NAME: suche_dominatoren
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Fuer alle Fanout-Staemme wird sche_dominator aufgerufen
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void suche_dominatoren()
{
line_C *line_item_p;

forall(line_item_p,lines_C)
	if ((line_item_p->is_primary_output() == FALSE) &&
	    ((*line_item_p->nachfolger())[0]->gib_type() == FANOUT))
		suche_dominator(line_item_p);
}//suche_dominatoren

//*********************************************************************
//
//   FUNCTION-NAME: output_kegel
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Markiert den Ausgangskegel der Line rekursiv
//
//   FORMAL PARAMETERS: line_C *line_p: Die Line
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void output_kegel(line_C *line_p)
{
line_C *line_item_p;
gate_C *gate_item_p;

if (line_p->in_output_kegel() == TRUE)
	return;
line_p->mark_output_kegel();
if (line_p->is_primary_output() == FALSE)
	{
	gate_item_p = line_p->nachfolger()->gib_only_gate();
	forall(line_item_p,*gate_item_p->gib_outputs())
		output_kegel(line_item_p);
	}//if
}//output_kegel

//*********************************************************************
//
//   FUNCTION-NAME: input_kegel
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Markiert den Ausgangskegel der Line rekursiv
//
//   FORMAL PARAMETERS: line_C *line_p: Die Line
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void input_kegel(line_C *line_p)
{
line_C *line_item_p;
gate_C *gate_item_p;

if (line_p->in_input_kegel() == TRUE)
	return;
line_p->mark_input_kegel();
if (line_p->is_primary_input() == FALSE)
	{
	gate_item_p = line_p->vorgaenger()->gib_only_gate();
	forall(line_item_p,*gate_item_p->gib_inputs())
		input_kegel(line_item_p);
	}//if
}//input_kegel

//*********************************************************************
//
//   FUNCTION-NAME: in_kegel
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 92
//
//   FUNCTIONAL-DESCRIPTION: stellt fest, ob ein Signal s2 im Ausgangskegel von
//	einem anderen Signal s1 liegt
//
//   FORMAL PARAMETERS: line_C *line1_p,line2_p: Die Signale s1 und s2
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: boolean: TRUE, wenn s2 im Ausgangskegel von s1
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
boolean in_kegel(line_C *line1_p,line_C *line2_p)
{
line_C *line_item_p;
gate_C *gate_item_p;

if (line1_p == line2_p)
        return(TRUE);
if ((line1_p->gib_plane_no() > line2_p->gib_plane_no()) || (line1_p->is_primary_output() == TRUE))
	return(FALSE);
gate_item_p = line1_p->nachfolger()->gib_only_gate();
forall(line_item_p,*gate_item_p->gib_outputs())
	if (in_kegel(line_item_p,line2_p) == TRUE)
		return(TRUE);
return(FALSE);
}//in_kegel

//*********************************************************************
//
//   FUNCTION-NAME: learn
//
//   CLASS: line_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: Juli 92
//
//   FUNCTIONAL-DESCRIPTION: Ueberpruefung des Lernkriteriums fuer ein Signal.
//	Gegebenenfalls Lernen der globalen Implikation.
//
//   FORMAL PARAMETERS: line_C *line_item_p: Die Line, fuer die Implikation durchgefuehrt wurde.
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void line_C::learn(line_C *line_item_p)
{
int i = (good_value_e == VAL_1)?0:1;

if ((prime_input_b == TRUE) || (line_item_p == this) ||
    (good_value_e == VAL_X) || (set_in_forw_imp_b == FALSE) ||
    (global_implications[i].element_in_liste(line_item_p,mynot(line_item_p->get_good_value())) == TRUE))
	return;
switch ((*vorgaenger_p)[0]->gib_type())
	{
	case AND:
	case NOR:
		if (good_value_e == VAL_0)
			return;
		break;
	case NAND:
	case OR:
		if (good_value_e == VAL_1)
			return;
		break;
	case INVERTER:
	case DRIVER:
	case FANOUT:
		return;
	case EXOR:
	case BLOCK:
		if ((*vorgaenger_p)[0]->no_inputs() == 1)
			return;
		break;
	default:
		error("learn: Unknown gate type found");
	}//switch
something_learned_b = TRUE;
global_implications[i].fuege_ein(line_item_p,mynot(line_item_p->get_good_value()));
learned_implications_i++;
}//learn

//*********************************************************************
//
//   FUNCTION-NAME: try_learning
//
//   CLASS: line_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: Juli 92
//
//   FUNCTIONAL-DESCRIPTION: Lernen der globalen Implikation fuer ein Signal
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void line_C::try_learning()
{
protocol_type_C *prot_item_p;

	// VAL_1
set_good_value(VAL_1);
protocol_C.push(this,FALSE,FALSE,FALSE,FALSE,FALSE,VAL_X,VAL_1);
if (((prime_output_b == FALSE) && (nachfolger()->gib_only_gate()->forward_implication() == INCONSISTENT)) ||
    ((prime_input_b == FALSE) && (vorgaenger()->gib_only_gate()->backward_implication() == INCONSISTENT)))
	{
	set_only_value(VAL_0);
	mark_result(REDUNDANT_FAULT,VAL_0);
	}//if
	//clear
unjustified_bound_line_C.loesche_liste();
implication_line_C.loesche_liste();
for(;;)
	{
	prot_item_p = protocol_C.pop();
	if (prot_item_p == NULL)
		break;
	if (prot_item_p->line_p == NULL)
		error("try_learning: Line Pointer is Null");
	prot_item_p->line_p->learn(this);
	prot_item_p->line_p->reset_in_forw_imp();
	prot_item_p->line_p->set_good_value(VAL_X);
	}//forever
	// VAL_0
set_good_value(VAL_0);
protocol_C.push(this,FALSE,FALSE,FALSE,FALSE,FALSE,VAL_X,VAL_0);
if (((prime_output_b == FALSE) && (nachfolger()->gib_only_gate()->forward_implication() == INCONSISTENT)) ||
    ((prime_input_b == FALSE) && (vorgaenger()->gib_only_gate()->backward_implication() == INCONSISTENT)))
	{
	set_only_value(VAL_1);
	mark_result(REDUNDANT_FAULT,VAL_1);
	}//if
	//clear
unjustified_bound_line_C.loesche_liste();
implication_line_C.loesche_liste();
for(;;)
	{
	prot_item_p = protocol_C.pop();
	if (prot_item_p == NULL)
		break;
	if (prot_item_p->line_p == NULL)
		error("try_learning: Line Pointer is Null");
	prot_item_p->line_p->learn(this);
	prot_item_p->line_p->reset_in_forw_imp();
	prot_item_p->line_p->set_good_value(VAL_X);
	}//forever
}//try_learning

//*********************************************************************
//
//   FUNCTION-NAME: learning_implications
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: Juli 92
//
//   FUNCTIONAL-DESCRIPTION: Lernphase analog zu SOCRATES fuer globale Implikationen
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void learning_implications()
{
line_C *line_item_p;

reset_net(SUCCESS);
{forall(line_item_p,lines_C)
	line_item_p->reset_in_forw_imp();}
forall(line_item_p,lines_C)
	if ((line_item_p->is_primary_input() == TRUE) ||
	    (((*line_item_p->vorgaenger())[0]->gib_type() != FANOUT) &&
	     ((*line_item_p->vorgaenger())[0]->gib_type() != INVERTER) &&
	     ((*line_item_p->vorgaenger())[0]->gib_type() != DRIVER)))
		line_item_p->try_learning();
reset_net(SUCCESS);
}//learning_implications

//*********************************************************************
//
//   FUNCTION-NAME: add_cube
//
//   CLASS: gate_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Fuegt eine Cube an die Cubliste an. Die entstehenden
//	Kuben entsprechen den Werten, die nicht angelegt werden duerfen.
//
//   FORMAL PARAMETERS: gate_C *line1_p,line2_p: die Lines mit dem Constraint
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void gate_C::add_cube(line_C *line1_p,line_C *line2_p)
{
int i,k,j;
boolean line1_belegt_b = FALSE;
boolean constr1_done_b = FALSE,constr2_done_b = FALSE;
cube_list_t *source_p,*target_p;
belegung_t *source_bel_p,*target_bel_p;

if (line2_p == NULL)
   {
   for (source_p = constraints_a[0],target_p = constraints_a[1];source_p != NULL;
        source_p = source_p->next_p,target_p = target_p->next_p)
      for (j=0,source_bel_p=source_p->to_belegung_p,target_bel_p=target_p->to_belegung_p;j<no_input_i;
	   j++,source_bel_p=source_bel_p->next_p,target_bel_p=target_bel_p->next_p)
	if (((*input_p)[j] == line1_p) && ((*input_p)[j]->get_good_value() == source_bel_p->val_e) &&
	    (target_bel_p->val_e == IMPOSSIBLE_VALUE))
		return;

   if (constraints_a[0] == NULL)
	{
	constraints_a[0] = new cube_list_t;
	if (constraints_a[0] == NULL)
		error("No more memory available");
	constraints_a[1] = new cube_list_t;
	if (constraints_a[1] == NULL)
		error("No more memory available");
	source_p = constraints_a[0];
	target_p = constraints_a[1];
	}
   else
	{
	source_p = constraints_a[0];
	target_p = constraints_a[1];
	while (source_p->next_p != NULL)
	   {
	   source_p = source_p->next_p;
	   target_p = target_p->next_p;
	   }//while
	source_p->next_p = new cube_list_t;
	if (source_p->next_p == NULL)
		error("No more memory available");
	target_p->next_p = new cube_list_t;
	if (target_p->next_p == NULL)
		error("No more memory available");
	source_p = source_p->next_p;
	target_p = target_p->next_p;
	}//else
   source_p->to_belegung_p = NULL;
   source_p->next_p = NULL;
   target_p->to_belegung_p = NULL;
   target_p->next_p = NULL;
   for (i=0,source_bel_p=source_p->to_belegung_p,target_bel_p=target_p->to_belegung_p;i<no_input_i;
	i++,source_bel_p=source_p->to_belegung_p,target_bel_p=target_p->to_belegung_p)
	{
	if (source_p->to_belegung_p == NULL)
		{
		source_p->to_belegung_p = new belegung_t;
		if (source_p->to_belegung_p == NULL)
			error("No more memory available");
		target_p->to_belegung_p = new belegung_t;
		if (target_p->to_belegung_p == NULL)
			error("No more memory available");
		source_bel_p=source_p->to_belegung_p;
		target_bel_p=target_p->to_belegung_p;
		}//if
	else
		{
		while(source_bel_p->next_p != NULL)
			{
			source_bel_p = source_bel_p->next_p;
			target_bel_p = target_bel_p->next_p;
			}//while
		source_bel_p->next_p = new belegung_t;
		if (source_bel_p->next_p == NULL)
			error("No more memory available");
		target_bel_p->next_p = new belegung_t;
		if (target_bel_p->next_p == NULL)
			error("No more memory available");
		source_bel_p = source_bel_p->next_p;
		target_bel_p = target_bel_p->next_p;
		}//else
	source_bel_p->next_p = NULL;
	target_bel_p->next_p = NULL;
	if ((*input_p)[i] == line1_p)
		{
		target_bel_p->val_e = IMPOSSIBLE_VALUE;
		source_bel_p->val_e = line1_p->get_good_value();
		}//if
	else
		{
		source_bel_p->val_e = VAL_X;
		target_bel_p->val_e = VAL_X;
		}//else
	}//for
   return;
   }//if

for (source_p = constraints_a[0],target_p = constraints_a[1];source_p != NULL;
     source_p = source_p->next_p,target_p = target_p->next_p)
   for (j=0,source_bel_p=source_p->to_belegung_p;j<no_input_i;j++,source_bel_p=source_bel_p->next_p)
	{
	if (((*input_p)[j] == line1_p) && ((*input_p)[j]->get_good_value() == source_bel_p->val_e))
		for (k=0,target_bel_p=target_p->to_belegung_p;k<no_input_i;k++,target_bel_p=target_bel_p->next_p)
			if ((*input_p)[k] == line2_p)
				{
				constr1_done_b = TRUE;
				target_bel_p->val_e = line2_p->get_good_value();
				}//if
	if (((*input_p)[j] == line2_p) && ((*input_p)[j]->get_good_value() == mynot(source_bel_p->val_e)))
		for (k=0,target_bel_p=target_p->to_belegung_p;k<no_input_i;k++,target_bel_p=target_bel_p->next_p)
			if ((*input_p)[k] == line1_p)
				{
				constr2_done_b = TRUE;
				target_bel_p->val_e = mynot(line1_p->get_good_value());
				}//if
	}//for
if (constr1_done_b == FALSE)
   {
   if (constraints_a[0] == NULL)
	{
	constraints_a[0] = new cube_list_t;
	if (constraints_a[0] == NULL)
		error("No more memory available");
	constraints_a[1] = new cube_list_t;
	if (constraints_a[1] == NULL)
		error("No more memory available");
	source_p = constraints_a[0];
	target_p = constraints_a[1];
	}
   else
	{
	source_p = constraints_a[0];
	target_p = constraints_a[1];
	while (source_p->next_p != NULL)
	   {
	   source_p = source_p->next_p;
	   target_p = target_p->next_p;
	   }//while
	source_p->next_p = new cube_list_t;
	if (source_p->next_p == NULL)
		error("No more memory available");
	target_p->next_p = new cube_list_t;
	if (target_p->next_p == NULL)
		error("No more memory available");
	source_p = source_p->next_p;
	target_p = target_p->next_p;
	}//else
   source_p->to_belegung_p = NULL;
   source_p->next_p = NULL;
   target_p->to_belegung_p = NULL;
   target_p->next_p = NULL;
   for (i=0,source_bel_p=source_p->to_belegung_p,target_bel_p=target_p->to_belegung_p;i<no_input_i;
	i++,source_bel_p=source_p->to_belegung_p,target_bel_p=target_p->to_belegung_p)
	{
	if (source_p->to_belegung_p == NULL)
		{
		source_p->to_belegung_p = new belegung_t;
		if (source_p->to_belegung_p == NULL)
			error("No more memory available");
		target_p->to_belegung_p = new belegung_t;
		if (target_p->to_belegung_p == NULL)
			error("No more memory available");
		source_bel_p=source_p->to_belegung_p;
		target_bel_p=target_p->to_belegung_p;
		}//if
	else
		{
		while(source_bel_p->next_p != NULL)
			{
			source_bel_p = source_bel_p->next_p;
			target_bel_p = target_bel_p->next_p;
			}//while
		source_bel_p->next_p = new belegung_t;
		if (source_bel_p->next_p == NULL)
			error("No more memory available");
		target_bel_p->next_p = new belegung_t;
		if (target_bel_p->next_p == NULL)
			error("No more memory available");
		source_bel_p = source_bel_p->next_p;
		target_bel_p = target_bel_p->next_p;
		}//else
	source_bel_p->next_p = NULL;
	target_bel_p->next_p = NULL;
	if (((*input_p)[i] == line1_p) && (line1_belegt_b == FALSE))
		{
		target_bel_p->val_e = VAL_X;
		source_bel_p->val_e = line1_p->get_good_value();
		line1_belegt_b = TRUE; //wird benoetigt, wenn line1 == line2
		}//if
	else if ((*input_p)[i] == line2_p)
		{
		source_bel_p->val_e = VAL_X;
		target_bel_p->val_e = line2_p->get_good_value();
		}//else if
	else
		{
		source_bel_p->val_e = VAL_X;
		target_bel_p->val_e = VAL_X;
		}//else
	}//for
   }//if
line1_belegt_b = FALSE;
if (constr2_done_b == FALSE)
   {
   if (constraints_a[0] == NULL)
	{
	constraints_a[0] = new cube_list_t;
	if (constraints_a[0] == NULL)
		error("No more memory available");
	constraints_a[1] = new cube_list_t;
	if (constraints_a[1] == NULL)
		error("No more memory available");
	source_p = constraints_a[0];
	target_p = constraints_a[1];
	}
   else
	{
	source_p = constraints_a[0];
	target_p = constraints_a[1];
	while (source_p->next_p != NULL)
	   {
	   source_p = source_p->next_p;
	   target_p = target_p->next_p;
	   }//while
	source_p->next_p = new cube_list_t;
	if (source_p->next_p == NULL)
		error("No more memory available");
	target_p->next_p = new cube_list_t;
	if (target_p->next_p == NULL)
		error("No more memory available");
	source_p = source_p->next_p;
	target_p = target_p->next_p;
	}//else
   source_p->to_belegung_p = NULL;
   source_p->next_p = NULL;
   target_p->to_belegung_p = NULL;
   target_p->next_p = NULL;
   for (i=0,source_bel_p=source_p->to_belegung_p,target_bel_p=target_p->to_belegung_p;i<no_input_i;
	i++,source_bel_p=source_p->to_belegung_p,target_bel_p=target_p->to_belegung_p)
	{
	if (source_p->to_belegung_p == NULL)
		{
		source_p->to_belegung_p = new belegung_t;
		if (source_p->to_belegung_p == NULL)
			error("No more memory available");
		target_p->to_belegung_p = new belegung_t;
		if (target_p->to_belegung_p == NULL)
			error("No more memory available");
		source_bel_p=source_p->to_belegung_p;
		target_bel_p=target_p->to_belegung_p;
		}//if
	else
		{
		while(source_bel_p->next_p != NULL)
			{
			source_bel_p = source_bel_p->next_p;
			target_bel_p = target_bel_p->next_p;
			}//while
		source_bel_p->next_p = new belegung_t;
		if (source_bel_p->next_p == NULL)
			error("No more memory available");
		target_bel_p->next_p = new belegung_t;
		if (target_bel_p->next_p == NULL)
			error("No more memory available");
		source_bel_p = source_bel_p->next_p;
		target_bel_p = target_bel_p->next_p;
		}//else
	source_bel_p->next_p = NULL;
	target_bel_p->next_p = NULL;
	if (((*input_p)[i] == line1_p) && (line1_belegt_b == FALSE))
		{
		source_bel_p->val_e = VAL_X;
		target_bel_p->val_e = mynot(line1_p->get_good_value());
		line1_belegt_b = TRUE; //wird benoetigt, wenn line1 == line2
		}//if
	else if ((*input_p)[i] == line2_p)
		{
		target_bel_p->val_e = VAL_X;
		source_bel_p->val_e = mynot(line2_p->get_good_value());
		}//else if
	else
		{
		source_bel_p->val_e = VAL_X;
		target_bel_p->val_e = VAL_X;
		}//else
	}//for
   }//if
}//add_cube

//*********************************************************************
//
//   FUNCTION-NAME: analyse_result
//
//   CLASS: gate_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Analyse der Implikation
//
//   FORMAL PARAMETERS: gate_C *gate_item_p: das Gatter
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void gate_C::analyse_result(int index_i,ofstream& datei_F)
{
gate_C *gate_item_p;

for (int i = 0;i<no_input_i;i++)
	{
	if (i == index_i)
		continue;
	if ((*input_p)[i]->get_good_value() != VAL_X)
		{
		datei_F << name_sC << ": I" << index_i+1 << " = " << (*input_p)[index_i]->get_good_value()
			<< "  =>  I" << i+1 << " = " << (*input_p)[i]->get_good_value() << "\n";
		datei_F << name_sC << ": I" << i+1 << " = " << mynot((*input_p)[i]->get_good_value())
			<< "  =>  I" << index_i+1 << " = " << mynot((*input_p)[index_i]->get_good_value()) << "\n";
		add_cube((*input_p)[index_i],(*input_p)[i]);
		if ((*input_p)[index_i]->get_good_value() == mynot((*input_p)[i]->get_good_value()))
		   switch (gate_e)
			{
			case AND:
				if ((*input_p)[index_i]->get_good_value() == VAL_1)
					{
					(*output_p)[0]->set_good_value(VAL_1);
					(*output_p)[0]->set_only_value(VAL_0);
					(*output_p)[0]->mark_result(REDUNDANT_FAULT,VAL_0);
					datei_F << "Redundant Fault for VAL_1 at Line: " << (*output_p)[0]->get_name() << "\n";
					if ((*output_p)[0]->is_primary_output() == TRUE)
						break;
					gate_item_p = (*output_p)[0]->nachfolger()->gib_only_gate();
					gate_item_p->add_cube((*output_p)[0],NULL);
					}//if
				break;
			case NAND:
				if ((*input_p)[index_i]->get_good_value() == VAL_1)
					{
					(*output_p)[0]->set_good_value(VAL_0);
					(*output_p)[0]->set_only_value(VAL_1);
					(*output_p)[0]->mark_result(REDUNDANT_FAULT,VAL_1);
					datei_F << "Redundant Fault for VAL_0 at Line: " << (*output_p)[0]->get_name() << "\n";
					if ((*output_p)[0]->is_primary_output() == TRUE)
						break;
					gate_item_p = (*output_p)[0]->nachfolger()->gib_only_gate();
					gate_item_p->add_cube((*output_p)[0],NULL);
					}//if
				break;
			case OR:
				if ((*input_p)[index_i]->get_good_value() == VAL_0)
					{
					(*output_p)[0]->set_good_value(VAL_0);
					(*output_p)[0]->set_only_value(VAL_1);
					(*output_p)[0]->mark_result(REDUNDANT_FAULT,VAL_1);
					datei_F << "Redundant Fault for VAL_0 at Line: " << (*output_p)[0]->get_name() << "\n";
					if ((*output_p)[0]->is_primary_output() == TRUE)
						break;
					gate_item_p = (*output_p)[0]->nachfolger()->gib_only_gate();
					gate_item_p->add_cube((*output_p)[0],NULL);
					}//if
				break;
			case NOR:
				if ((*input_p)[index_i]->get_good_value() == VAL_0)
					{
					(*output_p)[0]->set_good_value(VAL_1);
					(*output_p)[0]->set_only_value(VAL_0);
					(*output_p)[0]->mark_result(REDUNDANT_FAULT,VAL_0);
					datei_F << "Redundant Fault for VAL_1 at Line: " << (*output_p)[0]->get_name() << "\n";
					if ((*output_p)[0]->is_primary_output() == TRUE)
						break;
					gate_item_p = (*output_p)[0]->nachfolger()->gib_only_gate();
					gate_item_p->add_cube((*output_p)[0],NULL);
					}//if
				break;
			default:
				break;
			}//switch
		}//if
	}//for
}//analyse_result

//*********************************************************************
//
//   FUNCTION-NAME: try_constraint_ana
//
//   CLASS: gate_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Constraint Analyse an einem Gatter
//
//   FORMAL PARAMETERS: gate_C *gate_item_p: das Gatter
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void gate_C::try_constraint_ana(ofstream& datei_F)
{
gate_t type_e;
protocol_type_C *prot_item_p;

for (int i=0;i<no_input_i;i++)
	{
	if ((*input_p)[i]->get_reconvergence() == FALSE)
		continue;
	type_e = (*input_p)[i]->vorgaenger()->gib_only_gate()->gib_type();
			// VAL_1
	if ((type_e == AND) || (type_e == NOR) || (type_e == INVERTER) ||
	    (type_e == DRIVER) || (type_e == BLOCK) || (type_e == FANOUT))
	   {
	   (*input_p)[i]->set_good_value(VAL_1);
	   protocol_C.push((*input_p)[i],FALSE,FALSE,FALSE,FALSE,FALSE,VAL_X,VAL_1);
	   if ((*input_p)[i]->vorgaenger()->gib_only_gate()->backward_implication() == INCONSISTENT)
		{
		(*input_p)[i]->set_only_value(VAL_0);
		(*input_p)[i]->mark_result(REDUNDANT_FAULT,VAL_0);
		datei_F << "Redundant Fault for VAL_1 at Line: " << (*input_p)[i]->get_name() << "\n";
		}//if
	   else
		this->analyse_result(i,datei_F);
			//clear
	   unjustified_bound_line_C.loesche_liste();
	   implication_line_C.loesche_liste();
	   for(;;)
                {
                prot_item_p = protocol_C.pop();
                if (prot_item_p == NULL)
                        break;
                if (prot_item_p->line_p == NULL)
                        error("try_constraint_ana: Line Pointer is Null");
                prot_item_p->line_p->set_good_value(VAL_X);
		}//forever
	   }//if
			// VAL_0
	if ((type_e == NAND) || (type_e == OR) || (type_e == INVERTER) ||
	    (type_e == DRIVER) || (type_e == BLOCK) || (type_e == FANOUT))
	   {
	   (*input_p)[i]->set_good_value(VAL_0);
	   protocol_C.push((*input_p)[i],FALSE,FALSE,FALSE,FALSE,FALSE,VAL_X,VAL_0);
	   if ((*input_p)[i]->vorgaenger()->gib_only_gate()->backward_implication() == INCONSISTENT)
		{
		(*input_p)[i]->set_only_value(VAL_1);
		(*input_p)[i]->mark_result(REDUNDANT_FAULT,VAL_1);
		datei_F << "Redundant Fault for VAL_0 at Line: " << (*input_p)[i]->get_name() << "\n";
		}//if
	   else
		this->analyse_result(i,datei_F);
			//clear
	   unjustified_bound_line_C.loesche_liste();
	   implication_line_C.loesche_liste();
	   for(;;)
                {
                prot_item_p = protocol_C.pop();
                if (prot_item_p == NULL)
                        break;
                if (prot_item_p->line_p == NULL)
                        error("try_constraint_ana: Line Pointer is Null");
                prot_item_p->line_p->set_good_value(VAL_X);
		}//forever
	   }//if
	}//for
}//try_constraint_ana

//*********************************************************************
//
//   FUNCTION-NAME: constraint_analysis
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Constraint Analyse an Pfadrekonvergenzen
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void constraint_analysis()
{
string_C s_sC(".constr_ana");
ofstream datei_F(string_C(schaltungsname_s) + s_sC);
gate_C *gate_item_p;

//set_time();
reset_net(SUCCESS);
forall(gate_item_p,gates_C)
	if ((gate_item_p->get_reconvergence() == TRUE) && (gate_item_p->gib_type() == BLOCK))
		gate_item_p->try_constraint_ana(datei_F);
reset_net(SUCCESS);
datei_F.close();
//add_time(constr_zeit_l);
//cout << "Zeit Implikation:  " << imp_zeit_l << "\nZeit Konstraint Analyse:  " << constr_zeit_l << "\n";
//cout << "Anzahl der Aufrufe von implication:  " << no_imp_calls_i << "\n";
}//constraint_analysis









