//*********************************************************************
//
//   FILE-NAME: pretest.c
//
//   CREATION-DATE: Februar 92
//
//   FILE-DESCRIPTION: Dieses File beinhaltet die Hauptroutinen fuer die
//      Signalauswahl fuer den FAN Algorithmus
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
#include <sys/types.h>
#include <sys/times.h>
#include "extern.h"
#include "line_value_liste_C.h"
#include "line_line_liste_C.h"
#include "line_liste_C.h"
#include "protocol_stack_C.h"
#include "gate_liste_C.h"
#include "line_C.h"
#include "gate_C.h"

	// Prototypen
void reset_net(fan_result_t);
line_value_t *gen_fault();
line_line_C *gen_br_fault();
line_C *signal_to_propagate(line_C *, line_C *);
line_C *signal_to_propagate_2nd(line_C *, line_C *);
fan_result_t fan_alg(line_C *,value_t);
void iddq_fault_simulation();
void fault_simulation(line_C *,value_t);
extern void output_kegel(line_C*);
extern void input_kegel(line_C*);
void nachanalyse();
extern void mark_all(line_C *,AEQUIVALENZ_TYP);
extern void ziel_F_append();
extern void compute_contby();
extern void compute_observaby();
extern void prob_compute_contby();
extern void prob_compute_observaby();
extern void level_compute_contby();
extern void level_compute_observaby();
extern void reko_compute_contby();
extern void reko_compute_observaby();
extern void print_result();
extern fan_result_t implication();
line_C *search_next_path(void);
extern value_t mynot(value_t);
#if FRAMEWORK
void cfidr_PrintTestpattern(int);
#endif

	// Globale Variablen
boolean use_3_phases_b = TRUE;
boolean init_pattern_b;
boolean signal_not_to_propagate_b = FALSE;
extern boolean weak_test_b;
int no_red_classes_i = 0;
int ctest_reject_level_i = 0;
int inits_i = 0,tests_i = 0;
int act_bridging_list_index_i = 0;
extern boolean hazard_test_ein_b;
extern boolean hazard_occured_b;
boolean test_generated_b;
boolean ctest_pattern_b;
boolean phase2_b = FALSE;
boolean phase3_b = FALSE;
boolean testing_phase_b = FALSE;
extern boolean only_iddq_test_b;
extern boolean path_delay_test_b;
extern boolean transition_test_b;
extern boolean stuck_open_test_b;
extern boolean iddq_stuck_on_test_b;
extern boolean only_stuck_at_test_b;
extern boolean bridging_iddq_b;
boolean bridging_b = FALSE;
extern boolean stuck_at_test_without_branches_b;
extern int max_number_backtracks_i; // Default ist 10 s. main.c
extern int number_backtracks_i;
extern int backtrack_counter_i;
extern int reject_counter_i;
extern line_liste_C unjustified_bound_line_C;
extern line_liste_C d_front_C;
extern line_liste_C implication_line_C;
extern ofstream simu_F;
line_liste_C pseudo_inputs_C;
line_liste_C pseudo_outputs_C;
line_liste_C red_abo_liste_C; //Eb!!!30.6
line_value_liste_C zielfehler_C;
line_value_liste_C hilf_C;
line_line_liste_C zielfehler2_C;
extern protocol_stack_C protocol_C;
#if DEBUG
extern opt_line_liste_C optional_lines_C;
#endif
int no_testpattern_i = 0;
int no_weak_tests_i = 0;
int no_redundant_path_i = 0;
int no_aborted_path_i = 0;
int no_non_robust_pattern_i = 0;
int elements_to_test_i;
clock_t cpu_tpg_l = 0,cpu_fs_l = 0,cpu_sons_l = 0;
tms actual_time_u;
gate_liste_C selected_path_C;
extern gate_liste_C tested_blocks_C;
extern boolean simulate_pattern_called_b;

	// Inline Functions
void set_time(){times(&actual_time_u);}
void add_time(clock_t &ct){tms time_u;times(&time_u);ct+=time_u.tms_utime-actual_time_u.tms_utime;set_time();}

//*********************************************************************
//
//   FUNCTION-NAME: print_testpattern
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Druckt die Testpattern in das File <name>.test
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void print_testpattern(line_C *line_p,int lauf_i)
{
static boolean erster_eingang_b = TRUE;
static ofstream pattern_F;
static boolean robust_test_b = TRUE,last_init_set_b = FALSE;
static int bekannter_lauf_i = 0;
static value_t *last_init_p;
line_C *line_item_p;

#if FRAMEWORK
cfidr_PrintTestpattern(no_testpattern_i + no_non_robust_pattern_i - 1);
#endif
if ((lauf_i == 1) && (bekannter_lauf_i < 1))
	{
//	pattern_F << "* RERETAT Pattern\n";
	bekannter_lauf_i = 1;
	}//if

if ((lauf_i == 2) && (bekannter_lauf_i < 2))
	{
//	pattern_F << "* EXTEST Pattern\n";
	bekannter_lauf_i = 2;
	}//if

if (erster_eingang_b == TRUE)
	{
	erster_eingang_b = FALSE;
	pattern_F.open((char *)(string_C(schaltungsname_s) + ".pat"),ios::out);
	pattern_F << "* Input_lines: ";
	{forall(line_item_p,input_lines_C)
		pattern_F << line_item_p->get_name() << " ";}
	pattern_F << "\n";
	pattern_F << "* Output_lines: ";
	{forall(line_item_p,output_lines_C)
		pattern_F << line_item_p->get_name() << " ";}
	pattern_F << "\n";
	last_init_p = new value_t[input_lines_C.no_elements()+output_lines_C.no_elements()];
	}//if

if ((hazard_test_ein_b == FALSE) && (robust_test_b == TRUE))
	{
	robust_test_b = FALSE;
	pattern_F << "* Non Robust Pattern:\n";
	bekannter_lauf_i = 3;
	}//if

if (init_pattern_b == TRUE)
	{
	last_init_set_b = TRUE;
	{forall(line_item_p,input_lines_C)
		switch(line_item_p->get_good_value())
			{
			case VAL_X:
				last_init_p[forall_index] = VAL_X;
				break;
			case VAL_0:
				last_init_p[forall_index] = VAL_0;
				break;
			case VAL_1:
				last_init_p[forall_index] = VAL_1;
				break;
			case UNDEFINED_VALUE:
				error("print_testpattern: Undefined value in pattern");
			}//switch
	}
	forall(line_item_p,output_lines_C)
		switch(line_item_p->get_good_value())
			{
			case VAL_X:
				last_init_p[forall_index+input_lines_C.no_elements()] = VAL_X;
				break;
			case VAL_0:
				last_init_p[forall_index+input_lines_C.no_elements()] = VAL_0;
				break;
			case VAL_1:
				last_init_p[forall_index+input_lines_C.no_elements()] = VAL_1;
				break;
			case UNDEFINED_VALUE:
				error("print_testpattern: Undefined value in pattern");
			}//switch
	return;
	}//if
if (last_init_set_b == TRUE)
	{
	inits_i++;
	last_init_set_b = FALSE;
	{forall(line_item_p,input_lines_C)
		switch(last_init_p[forall_index])
			{
			case VAL_X:
				pattern_F << "X";
				break;
			case VAL_0:
				pattern_F << "0";
				break;
			case VAL_1:
				pattern_F << "1";
				break;
			case UNDEFINED_VALUE:
				error("print_testpattern: Undefined value in pattern");
			}//switch
	}
	pattern_F << "   ";
	{forall(line_item_p,output_lines_C)
		switch(last_init_p[forall_index+input_lines_C.no_elements()])
			{
			case VAL_X:
				pattern_F << "X";
				break;
			case VAL_0:
				pattern_F << "0";
				break;
			case VAL_1:
				pattern_F << "1";
				break;
			case UNDEFINED_VALUE:
				error("print_testpattern: Undefined value in pattern");
			}//switch
	}
	pattern_F << " * " << line_p->get_name() << "  Init-Pattern\n";
	simu_F << "\n" << flush;
	if (weak_test_b == TRUE)
		no_weak_tests_i++;
	else if ((hazard_test_ein_b == TRUE) || (only_iddq_test_b == TRUE) ||
                 (only_stuck_at_test_b == TRUE) || (iddq_stuck_on_test_b == TRUE) ||
                 (transition_test_b == TRUE))
		no_testpattern_i++;
	else
		no_non_robust_pattern_i++;
	}//if
tests_i++;
if (weak_test_b == TRUE)
	no_weak_tests_i++;
else if ((hazard_test_ein_b == TRUE) || (only_iddq_test_b == TRUE) ||
         (only_stuck_at_test_b == TRUE) || (iddq_stuck_on_test_b == TRUE) ||
         (transition_test_b == TRUE) || (bridging_iddq_b == TRUE) || (bridging_b == TRUE))
	no_testpattern_i++;
else
	no_non_robust_pattern_i++;
{forall(line_item_p,input_lines_C)
	switch(line_item_p->get_good_value())
		{
		case VAL_X:
			pattern_F << "X";
			break;
		case VAL_0:
			pattern_F << "0";
			break;
		case VAL_1:
			pattern_F << "1";
			break;
		case UNDEFINED_VALUE:
			error("print_testpattern: Undefined value in pattern");
		}//switch
}
pattern_F << "   ";
{forall(line_item_p,output_lines_C)
	switch(line_item_p->get_good_value())
		{
		case VAL_X:
			pattern_F << "X";
			break;
		case VAL_0:
			pattern_F << "0";
			break;
		case VAL_1:
			pattern_F << "1";
			break;
		case UNDEFINED_VALUE:
			error("print_testpattern: Undefined value in pattern");
		}//switch
}
pattern_F << " * " << line_p->get_name() << "  " << number_backtracks_i << " Backtracks";
if (path_delay_test_b == TRUE)
	pattern_F << "  Delay(units): " << selected_path_C.no_elements();
if (weak_test_b == TRUE)
	pattern_F << "   WEAK TEST !!!";
pattern_F << "\n";
simu_F << "\n" << flush;
}//print_testpattern

//*********************************************************************
//
//   FUNCTION-NAME: erstelle_zielfehlerliste
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Die Zielfehlerliste soll erstellt werden. Die
//	Reihenfolge wird durch die Testability Werte bestimmt.
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void erstelle_zielfehlerliste(line_value_liste_C &ziel_C)
{
extern dominanz_result_t dominanz(gate_C *);
extern void zaehl_aequivalenzen(line_C *);
extern int aequivalenzen_zahl_i;
line_C *line_item_p;
dominanz_result_t ergebnis_e;

forall(line_item_p,lines_C)
   if ((ziel_C.element_in_liste(line_item_p,VAL_0,FALSE) == TRUE) &&
       (ziel_C.element_in_liste(line_item_p,VAL_1,FALSE) == TRUE))
	continue;
#ifdef O_FS
   else
        {
	ziel_C.fuege_ein(line_item_p,VAL_0,FALSE);
	ziel_C.fuege_ein(line_item_p,VAL_1,FALSE);
	if ((line_item_p->is_primary_input() == FALSE) &&
	    ((*line_item_p->vorgaenger())[0]->gib_type() == BLOCK) &&
	    ((stuck_open_test_b == TRUE) || (iddq_stuck_on_test_b == TRUE)))
		ziel_C.fuege_ein(line_item_p,VAL_X,FALSE);
	}//else
#else
   else if (iddq_stuck_on_test_b == TRUE)
	{
	if ((line_item_p->is_primary_input() == FALSE) &&
	    ((*line_item_p->vorgaenger())[0]->gib_type() == BLOCK))
		ziel_C.fuege_ein(line_item_p,VAL_X,FALSE);
	if ((line_item_p->is_primary_output() == FALSE) &&
	    (line_item_p->nachfolger()->gib_only_gate()->no_inputs() > 1) && // kein Invert. oder Treiberinp.
	    (line_item_p->nachfolger()->gib_only_gate()->gib_type() != BLOCK))
		{
		ziel_C.fuege_ein(line_item_p,VAL_U0,FALSE);
		ziel_C.fuege_ein(line_item_p,VAL_U1,FALSE);
		}//if
	}//if
   else if ((stuck_open_test_b == TRUE) || (transition_test_b == TRUE))
	{
	if ((line_item_p->is_primary_input() == FALSE) &&
	    ((*line_item_p->vorgaenger())[0]->gib_type() == BLOCK))
		ziel_C.fuege_ein(line_item_p,VAL_X,FALSE);
	if ((line_item_p->is_primary_output() == FALSE) &&
	    (line_item_p->nachfolger()->gib_only_gate()->no_inputs() > 1) && // kein Invert. oder Treiberinp.
	    (line_item_p->nachfolger()->gib_only_gate()->gib_type() != BLOCK))
		{
		ziel_C.fuege_ein(line_item_p,VAL_U0,FALSE);
		ziel_C.fuege_ein(line_item_p,VAL_U1,FALSE);
//		ziel_C.fuege_ein(line_item_p,VAL_0,FALSE);
//		ziel_C.fuege_ein(line_item_p,VAL_1,FALSE);
		}//if
	}//if
   else if (only_stuck_at_test_b == FALSE)
	{
	ziel_C.fuege_ein(line_item_p,VAL_0,FALSE);
	ziel_C.fuege_ein(line_item_p,VAL_1,FALSE);
	}//else if

   else if ((stuck_at_test_without_branches_b == FALSE) && (only_stuck_at_test_b == TRUE) &&
            ((line_item_p->is_primary_input() == TRUE) ||
	     ((*line_item_p->vorgaenger())[0]->gib_type() == FANOUT) ||
	     ((*line_item_p->vorgaenger())[0]->gib_type() == EXOR)) )          // EXOR zerstoert dominanz

		{
		ziel_C.fuege_ein(line_item_p,VAL_0,FALSE);
		ziel_C.fuege_ein(line_item_p,VAL_1,FALSE);
		}//else if
   else if ((stuck_at_test_without_branches_b == FALSE) && (only_stuck_at_test_b == TRUE) &&
	    ((*line_item_p->vorgaenger())[0]->gib_type() == BLOCK))
		{
		ergebnis_e = dominanz((*line_item_p->vorgaenger())[0]);
		if (ergebnis_e.VAL_0_b == FALSE) ziel_C.fuege_ein(line_item_p,VAL_0,FALSE);
		if (ergebnis_e.VAL_1_b == FALSE) ziel_C.fuege_ein(line_item_p,VAL_1,FALSE);
	}//else if
	

   else if ((only_stuck_at_test_b == TRUE) && (stuck_at_test_without_branches_b == TRUE) &&
	    ((line_item_p->is_primary_input() == TRUE) ||
	     ((*line_item_p->vorgaenger())[0]->gib_type() != FANOUT)))
	{
	ziel_C.fuege_ein(line_item_p,VAL_0,FALSE);
	ziel_C.fuege_ein(line_item_p,VAL_1,FALSE);
	}//else if
#endif
}//erstelle_zielfehlerliste

//*********************************************************************
//
//   FUNCTION-NAME: test_circuit
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Hauptprogramm der Testmustergenerierung
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void test_circuit()
{
line_C *line_item_p, *line1_item_p, *line2_item_p;
line_C *line_weak_item_p;
line_value_t *act_liva_p;
line_line_C *act_lili_p, *lili_p;
fan_result_t result_e, fan_result_e;
int input_max_no_backtracks_i = max_number_backtracks_i;

simu_F.open(string_C(string_C(schaltungsname_s).basename())+".tests",ios::out);
if (path_delay_test_b == FALSE)
	max_number_backtracks_i = 0;
cout << "\tCompute Controllability\n" << flush;
::prob_compute_contby();
cout << "\tCompute Observability\n" << flush;
::prob_compute_observaby();
cout << "Compute Testpattern\n" << flush;
reset_net(SUCCESS);
if ((bridging_iddq_b == TRUE) || (bridging_b == TRUE)) 
   elements_to_test_i = zielfehler2_C.no_elements();
testing_phase_b = TRUE;
for (int j=0;j<6;j++)
  {
  for (int i=0;i<2;i++)
   {
   if (path_delay_test_b == FALSE)
     {
     int eins_i = 1,null_i = 0;

     ctest_reject_level_i++;
     forall(line_item_p,lines_C)
	if ((line_item_p->is_primary_input() == FALSE) &&
	    ((*line_item_p->vorgaenger())[0]->gib_type() == BLOCK))
		line_item_p->reset_test_done();
     if (ctest_reject_level_i == 2)
        set_test_reject_level(&ctest_reject_level_i,&eins_i);
     else
        set_test_reject_level(&ctest_reject_level_i,&null_i);
     hazard_test_ein_b = ((j==1) || (j==3) || (j==5) || (only_iddq_test_b == TRUE) ||
			  (only_stuck_at_test_b == TRUE) || (iddq_stuck_on_test_b == TRUE) ||
			  (bridging_iddq_b == TRUE) || (bridging_b == TRUE) ||
                          (transition_test_b == TRUE)) ? FALSE : TRUE;
     if ((i==0) && (j!=0))
	{
	forall(line_item_p,lines_C)
		{
		line_item_p->reset_aborted_faults();
		line_item_p->reset_testaby();
		line_item_p->reset_non_robust();
		}//forall
	::prob_compute_contby();
	::prob_compute_observaby();
	}//if
     if (i==1)
	{
	cout << "\nLEVEL Patterns\n" << flush;
	forall(line_item_p,lines_C)
		{
		line_item_p->reset_aborted_faults();
		line_item_p->reset_testaby();
		line_item_p->reset_non_robust();
		}//forall
	::level_compute_contby();
	::level_compute_observaby();
	}//if
     if (i==2)
	{
	cout << "\nRERETAT Patterns\n" << flush;
	forall(line_item_p,lines_C)
		{
		line_item_p->reset_aborted_faults();
		line_item_p->reset_testaby();
		line_item_p->reset_non_robust();
		}//forall
	::reko_compute_contby();
	::reko_compute_observaby();
	}//if
     if (i==3)
	{
	cout << "\nEXTEST Patterns\n" << flush;
	forall(line_item_p,lines_C)
		{
		line_item_p->reset_aborted_faults();
		line_item_p->reset_testaby();
		line_item_p->reset_non_robust();
		}//forall
	::compute_contby();
	::compute_observaby();
	}//if
     if (phase3_b == TRUE)
	i++;
     tested_blocks_C.loesche_liste();
     }//if
   else
     {
     hazard_test_ein_b = FALSE;
     while ((line_item_p = search_next_path()) != NULL)
     	{
//     	line_C *hilf_line_p;
  
     	add_time(cpu_sons_l);
	output_kegel(line_item_p);
	if ((hazard_test_ein_b == TRUE) && (init_pattern_b == FALSE))
		input_kegel(line_item_p);
	weak_test_b = FALSE;
	result_e = fan_alg(line_item_p,line_item_p->get_bad_value());
     	add_time(cpu_tpg_l);
     	backtrack_counter_i += number_backtracks_i;
     	if (result_e == SUCCESS)
     	        {
		no_testpattern_i++;
		test_generated_b = TRUE;
//		forall(hilf_line_p,input_lines_C)
//     			if (hilf_line_p->get_test_value() == VAL_X)
//    				{
//     				if ((hilf_line_p->get_init_value() == VAL_1) ||
//				    (hilf_line_p->get_init_value() == VAL_D))
//     					hilf_line_p->set_value(VAL_1);
//     				else
//     					hilf_line_p->set_value(VAL_0);
//     				implication_line_C.fuege_ein(hilf_line_p);
//     				}//if
//     		if (implication() != SUCCESS)
//     			error("test_circuit: Implication after test generation inconsistent");
		print_testpattern(line_item_p,i);
     		}//if
             else
		{
     		if (result_e == ABORTED_FAULT)
     			no_aborted_path_i++;
     		else
     			no_redundant_path_i++;
                     test_generated_b = FALSE;
                     unjustified_bound_line_C.loesche_liste();
     		}//else
     	reset_net(result_e);
     	}//while
     break;
     }//else
   if ((bridging_iddq_b == TRUE) || (bridging_b == TRUE))
	 {
	 if (i != 0)
	    {
	    forall(lili_p,zielfehler2_C)
	       if ((lili_p->is_tested() == FALSE) && (lili_p->is_redundant() == FALSE))
		  lili_p->reset_aborted();
	    }//else
	 cout << "Remaining Faults: " << elements_to_test_i << "          " << flush;
	 cout << "\b\b\b\b\b\b\b\b\b\b\b";
	 if (elements_to_test_i >= 10000000)
	 	cout << "\b";
	 if (elements_to_test_i >= 1000000)
	 	cout << "\b";
	 if (elements_to_test_i >= 100000)
	 	cout << "\b";
	 if (elements_to_test_i >= 10000)
	 	cout << "\b";
	 if (elements_to_test_i >= 1000)
	 	cout << "\b";
	 if (elements_to_test_i >= 100)
	 	cout << "\b";
	 if (elements_to_test_i >= 10)
	 	cout << "\b";
	 cout << "\b" << flush;
	 act_bridging_list_index_i = 0;
   
         while ((act_lili_p = gen_br_fault()) != NULL)
	       {
	       line1_item_p = act_lili_p->line1_p;      // Zugriff auf den ersten Knoten
	       line2_item_p = act_lili_p->line2_p;      // Zugriff auf den zweiten Knoten
	       if (line1_item_p->gib_type() == BOUND_LINE)
		  unjustified_bound_line_C.fuege_ein(line1_item_p);
	       if (line2_item_p->gib_type() == BOUND_LINE)
		  unjustified_bound_line_C.fuege_ein(line2_item_p);
	       implication_line_C.fuege_ein(line1_item_p);
	       implication_line_C.fuege_ein(line2_item_p);

	       if (bridging_iddq_b == TRUE)
	          {
	       	  line1_item_p->set_good_value(VAL_0);
	       	  line2_item_p->set_good_value(VAL_1);
	          }//if
	       else              // if (bridging_b == TRUE)
		  {
		  line_weak_item_p = signal_to_propagate(line1_item_p, line2_item_p);
		  if (line_weak_item_p == NULL)
		     {
		     result_e = REDUNDANT_FAULT;
     		     elements_to_test_i--;
		     no_red_classes_i++;
		     act_lili_p->set_redundant();
		     test_generated_b = FALSE;
		     unjustified_bound_line_C.loesche_liste();
		     d_front_C.loesche_liste();
		     reset_net(result_e);
		     cout << " " << elements_to_test_i << "          " << flush;
		     cout << "\b\b\b\b\b\b\b\b\b\b\b";
		     if (elements_to_test_i >= 10000000)
			cout << "\b";
		     if (elements_to_test_i >= 1000000)
			cout << "\b";
		     if (elements_to_test_i >= 100000)
			cout << "\b";
		     if (elements_to_test_i >= 10000)
			cout << "\b";
		     if (elements_to_test_i >= 1000)
			cout << "\b";
		     if (elements_to_test_i >= 100)
			cout << "\b";
		     if (elements_to_test_i >= 10)
			cout << "\b";
		     cout << "\b" << flush;
		     continue;
		     }//if
		  output_kegel(line_weak_item_p);
		  line_weak_item_p->set_bad_value(mynot(line_weak_item_p->get_good_value()));
                  }//else

	       add_time(cpu_sons_l);
	       weak_test_b = FALSE;

	       if (bridging_iddq_b == TRUE)
	       	  result_e = fan_alg(line1_item_p,VAL_X);
	       else                                // if (bridging_b == TRUE) 
	       	  result_e = fan_alg(line_weak_item_p,VAL_X);

	       fan_result_e = result_e;
	       add_time(cpu_tpg_l);
	       backtrack_counter_i += number_backtracks_i;

	       if (fan_result_e == SUCCESS) 
		  {
		  test_generated_b = TRUE;
		  if (bridging_iddq_b == TRUE)
		     iddq_fault_simulation();
		  else
		     fault_simulation(line_weak_item_p,VAL_U);
		  
		  add_time(cpu_fs_l);
		  simulate_pattern_called_b = FALSE;

		  if (bridging_iddq_b == TRUE)	     
		     print_testpattern(line1_item_p,i);
		  else                             // if (bridging_b == TRUE)
		     print_testpattern(line_weak_item_p,i);

		  }//if
	       else
		     {
		             // Die Werte an den Knoten koennen nur vertauscht werden, wenn:
		     if ((bridging_iddq_b == TRUE) || ((bridging_b == TRUE) &&
			((line1_item_p->is_primary_input() == FALSE) &&
		         (line2_item_p->is_primary_input() == FALSE)) &&
       		      ((((*line1_item_p->vorgaenger())[0]->gib_type() == NAND) &&
		        ((*line2_item_p->vorgaenger())[0]->gib_type() == NAND)) ||
		       (((*line1_item_p->vorgaenger())[0]->gib_type() == NOR) &&
		        ((*line2_item_p->vorgaenger())[0]->gib_type() == NOR))))) 
			{
			reset_net(SUCCESS);
			unjustified_bound_line_C.loesche_liste();
			d_front_C.loesche_liste();                  // loesche D-Front
			if (line1_item_p->gib_type() == BOUND_LINE)
			   unjustified_bound_line_C.fuege_ein(line1_item_p);
		     	if (line2_item_p->gib_type() == BOUND_LINE)
			   unjustified_bound_line_C.fuege_ein(line2_item_p);
			implication_line_C.fuege_ein(line1_item_p);
			implication_line_C.fuege_ein(line2_item_p);
			}//if
		     else
			{
			if (result_e == REDUNDANT_FAULT)
			   {
			   no_red_classes_i++;
			   act_lili_p->set_redundant();
			   elements_to_test_i--;
			   }//if
			else // if (result_e == ABORTED_FAULT)
			   {
			   no_aborted_path_i++;
			   act_lili_p->set_aborted();
			   }//if
			test_generated_b = FALSE;
			unjustified_bound_line_C.loesche_liste();
			d_front_C.loesche_liste();
			reset_net(result_e);
			cout << " " << elements_to_test_i << "          " << flush;
			cout << "\b\b\b\b\b\b\b\b\b\b\b";
			if (elements_to_test_i >= 10000000)
			   cout << "\b";
			if (elements_to_test_i >= 1000000)
			   cout << "\b";
			if (elements_to_test_i >= 100000)
	                   cout << "\b";
			if (elements_to_test_i >= 10000)
			   cout << "\b";
	                if (elements_to_test_i >= 1000)
	                   cout << "\b";
	                if (elements_to_test_i >= 100)
	                   cout << "\b";
	                if (elements_to_test_i >= 10)
	                   cout << "\b";
	                cout << "\b" << flush;
		        continue;
			}//else

		     if (bridging_iddq_b == TRUE)
        		{
			line1_item_p->set_good_value(VAL_1);
	                line2_item_p->set_good_value(VAL_0);
			}//if
		     else                   // if (bridging_b == TRUE) 
        		{
			         // Fkt. aufrufen, die die Werte an den Knoten vertauscht
			line_weak_item_p = signal_to_propagate_2nd(line1_item_p, line2_item_p);
			if (line_weak_item_p == NULL)
			   {
			   if (result_e == REDUNDANT_FAULT)
			      {
			      no_red_classes_i++;
			      act_lili_p->set_redundant();
			      elements_to_test_i--;
			      }//if
			   else // if (result_e == ABORTED_FAULT)
			      {
			      no_aborted_path_i++;
			      act_lili_p->set_aborted();
			      }//if
			   no_red_classes_i++;
			   act_lili_p->set_redundant();
			   test_generated_b = FALSE;
		           unjustified_bound_line_C.loesche_liste();
		           d_front_C.loesche_liste();
		           reset_net(result_e);
			   cout << " " << elements_to_test_i << "          " << flush;
	                   cout << "\b\b\b\b\b\b\b\b\b\b\b";
			   if (elements_to_test_i >= 10000000)
	                      cout << "\b";
	                   if (elements_to_test_i >= 1000000)
	                      cout << "\b";
	                   if (elements_to_test_i >= 100000)
	                      cout << "\b";
                           if (elements_to_test_i >= 10000)
	                      cout << "\b";
	                   if (elements_to_test_i >= 1000)
	                      cout << "\b";
	                   if (elements_to_test_i >= 100)
	                      cout << "\b";
	                   if (elements_to_test_i >= 10)
	                      cout << "\b";
	                   cout << "\b" << flush;
			   continue;
			   }//if
			output_kegel(line_weak_item_p);
		     	line_weak_item_p->set_bad_value(mynot(line_weak_item_p->get_good_value()));
			}//else
	             add_time(cpu_sons_l);
	             weak_test_b = FALSE;

		     if (bridging_iddq_b == TRUE)
	       	      	result_e = fan_alg(line1_item_p,VAL_X);
	       	     else                                // if (bridging_b == TRUE)
			result_e = fan_alg(line_weak_item_p,VAL_X);
	             add_time(cpu_tpg_l);
	             backtrack_counter_i += number_backtracks_i;

		     if (result_e == SUCCESS)
			   {
			   test_generated_b = TRUE;
		           if (bridging_iddq_b == TRUE)
			      iddq_fault_simulation();
			   else
			      fault_simulation(line_weak_item_p,VAL_U);
		           add_time(cpu_fs_l);
		           simulate_pattern_called_b = FALSE;
			   if (bridging_iddq_b == TRUE)
		              print_testpattern(line1_item_p,i);
			   else
			      print_testpattern(line_weak_item_p,i);
			   }//if
		     else
			   {
			   if (result_e == REDUNDANT_FAULT)
				if (fan_result_e == ABORTED_FAULT)
				   {
				   no_aborted_path_i++;
				   act_lili_p->set_aborted();
				   }//if
				else
				   {
				   elements_to_test_i--;
				   no_red_classes_i++;
				   act_lili_p->set_redundant();
				   }//else
			   else
				{
				no_aborted_path_i++;
				act_lili_p->set_aborted();
				}//else
			   test_generated_b = FALSE;
			   unjustified_bound_line_C.loesche_liste();
			   d_front_C.loesche_liste();
			   }//else
		     }//else
	       reset_net(result_e);
	       cout << " " << elements_to_test_i << "          " << flush;
	       cout << "\b\b\b\b\b\b\b\b\b\b\b";
	       if (elements_to_test_i >= 10000000)
	          cout << "\b";
	       if (elements_to_test_i >= 1000000)
	          cout << "\b";
	       if (elements_to_test_i >= 100000)
	          cout << "\b";
               if (elements_to_test_i >= 10000)
	          cout << "\b";
	       if (elements_to_test_i >= 1000)
	          cout << "\b";
	       if (elements_to_test_i >= 100)
	          cout << "\b";
	       if (elements_to_test_i >= 10)
	          cout << "\b";
	       cout << "\b" << flush;
	       }//while
	 }//if
   else
	 {
	 {line_value_t *liva_p;
	 boolean dummy_b;
	 forall(liva_p,hilf_C)
		dummy_b = zielfehler_C.element_in_liste(liva_p->line_p,liva_p->val_e,TRUE);}
	 hilf_C.loesche_liste();
	 cout << "Remaining Faults: " << zielfehler_C.no_elements() << "          " << flush;
	 cout << "\b\b\b\b\b\b\b\b\b\b\b";
	 if (zielfehler_C.no_elements() >= 100000)
	 	cout << "\b";
	 if (zielfehler_C.no_elements() >= 10000)
	 	cout << "\b";
	 if (zielfehler_C.no_elements() >= 1000)
	 	cout << "\b";
	 if (zielfehler_C.no_elements() >= 100)
	 	cout << "\b";
	 if (zielfehler_C.no_elements() >= 10)
	 	cout << "\b";
	 cout << "\b" << flush;
	   
	 while ((act_liva_p = gen_fault()) != NULL)
	       {                     // Es muss ein weiterer Test generiert werden
	       line_item_p = act_liva_p->line_p;
	       add_time(cpu_sons_l);
	       if ((only_iddq_test_b == FALSE) && (iddq_stuck_on_test_b == FALSE))
		  output_kegel(line_item_p);
	       weak_test_b = FALSE;
	       result_e = fan_alg(line_item_p,act_liva_p->val_e);
	       add_time(cpu_tpg_l);
	       backtrack_counter_i += number_backtracks_i;
	       if (result_e == SUCCESS)
		     {
		     test_generated_b = TRUE;
		     if ((only_iddq_test_b == TRUE) || (iddq_stuck_on_test_b == TRUE))
			iddq_fault_simulation();
		     else
			fault_simulation(line_item_p,act_liva_p->val_e);
		     add_time(cpu_fs_l);
		     if ((hilf_C.no_elements() > 0) || (simulate_pattern_called_b == TRUE))
			init_pattern_b = FALSE;
		     simulate_pattern_called_b = FALSE;
		     print_testpattern(line_item_p,i);
		     }//if
	       else
		     {
		     if (result_e == REDUNDANT_FAULT)
			no_red_classes_i++;
		     test_generated_b = FALSE;
		     line_item_p->mark_result(result_e,act_liva_p->val_e);
		     unjustified_bound_line_C.loesche_liste();
		}//else
	       reset_net(result_e);
	       {line_value_t *liva_p;
		boolean dummy_b;
		forall(liva_p,hilf_C)
		   dummy_b = zielfehler_C.element_in_liste(liva_p->line_p,liva_p->val_e,TRUE);}
	       hilf_C.loesche_liste();
	       cout << " " << zielfehler_C.no_elements() << "          " << flush;
	       cout << "\b\b\b\b\b\b\b\b\b\b\b";
	       if (zielfehler_C.no_elements() >= 100000)
		  cout << "\b";
	       if (zielfehler_C.no_elements() >= 10000)
		  cout << "\b";
	       if (zielfehler_C.no_elements() >= 1000)
		  cout << "\b";
	       if (zielfehler_C.no_elements() >= 100)
		  cout << "\b";
	       if (zielfehler_C.no_elements() >= 10)
		  cout << "\b";
	       cout << "\b" << flush;
	       }//while
	 }//else
#if DEBUG
   cout << "post analysis\n" << flush;
   nachanalyse();
#endif   
   }//for
  cout << "\n";
  if ((bridging_iddq_b == FALSE) && (bridging_b == FALSE))
      {
      if (zielfehler_C.no_elements() == 0)
	break;
      }//if
  else
      if (elements_to_test_i == 0)
	 break;
  no_aborted_path_i = 0;
  if (hazard_test_ein_b == TRUE)
	{
	forall(line_item_p,lines_C)
		line_item_p->reset_non_robust();
	if (phase2_b == TRUE)
		{
		print_result();
		cout << "Non Robust Testpattern\n" << flush;
		continue;
		}//if
	else
		j++;
	}//if
  if (phase2_b == FALSE)
	{
	max_number_backtracks_i = input_max_no_backtracks_i;
	phase2_b = TRUE;
	print_result();
	cout << "2. Phase of Test Generation\n" << flush;
	if ((bridging_iddq_b == FALSE) && (bridging_b == FALSE))
	   {
	   forall(line_item_p,lines_C)
		line_item_p->reset_aborted_faults();
	   }//if
	else
	   {
	   forall(lili_p,zielfehler2_C)
	      if ((lili_p->is_tested() == FALSE) && (lili_p->is_redundant() == FALSE))
		lili_p->reset_aborted();
	   }//else
	}//if
  else if ((phase3_b == FALSE) && (use_3_phases_b == TRUE))
	{
	phase3_b = TRUE;
	j++;
	print_result();
	cout << "3. Phase of Test Generation and Redundancy Identification\n" << flush;
	if ((bridging_iddq_b == FALSE) && (bridging_b == FALSE))
	   {
	   forall(line_item_p,lines_C)
		line_item_p->reset_aborted_faults();
	   }//if
	else
	   {
	   forall(lili_p,zielfehler2_C)
	      if ((lili_p->is_tested() == FALSE) && (lili_p->is_redundant() == FALSE))
		lili_p->reset_aborted();
	   }//else
	}//if
  else
	break;
  }//for
simu_F.close();
print_result();
add_time(cpu_sons_l);
}//test_circuit

//*********************************************************************
//
//   FUNCTION-NAME: test_signal
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Testen eines Signals
//
//   FORMAL PARAMETERS: name_sC: Name des Signals
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void test_signal(char *name_s)
{
line_C *line_item_p,*line_p;
fan_result_t result_e;
//string_C name_sC(name_s);

forall(line_item_p,lines_C)
	{
	if (strcmp(line_item_p->get_name(),name_s) == 0)
		line_p = line_item_p;
	}//forall
reset_net(SUCCESS);
hazard_test_ein_b = FALSE;
line_p->set_good_value(VAL_0);
line_p->set_bad_value(VAL_1);
output_kegel(line_p);
weak_test_b = FALSE;
result_e = fan_alg(line_p,VAL_1);
cout << "S_a_1 Test war: " << result_e << "\n" << flush;
unjustified_bound_line_C.loesche_liste();
reset_net(SUCCESS);
line_p->set_good_value(VAL_1);
line_p->set_bad_value(VAL_0);
output_kegel(line_p);
weak_test_b = FALSE;
result_e = fan_alg(line_p,VAL_0);
cout << "S_a_0 Test war: " << result_e << "\n" << flush;
unjustified_bound_line_C.loesche_liste();
}//test_signal

//*********************************************************************
//
//   FUNCTION-NAME: reset_net
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Initialisierung aller Lines in der Schaltung mit
//	inline Funktion reset_value.
//
//   FORMAL PARAMETERS: fan_result_t result_e: Resultat des letzten Fan Aufrufs
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void reset_net(fan_result_t result_e)
{
line_C *line_item_p;

implication_line_C.loesche_liste();
forall(line_item_p,lines_C)
	{
	line_item_p->reset_value(result_e);
	line_item_p->clear_kegel();
	}//forall
for (int i=0;i<gates_C.no_elements();i++)
	{
	gates_C[i]->implikation_durchgefuehrt_b = FALSE;
	gates_C[i]->set_hazard_bits(0);
	}//for
number_backtracks_i = 0;
#if DEBUG
optional_lines_C.loesche_liste();
#endif
d_front_C.loesche_liste();
protocol_C.loesche_liste();
}//reset_net

//**************************************************************
//
//   FUNCTION-NAME: signal_to_propagate
//
//   CREATION-DATE: 24.2.1994
//
//   Stellt fest, welches Signal 'stark' und welches 'schwach' ist.
//   Das 'schwache' Signal wird propagiert.
//   
//
//**************************************************************
line_C* signal_to_propagate(line_C *line1_item_p, line_C *line2_item_p)
{
gate_t gate1_e, gate2_e;

if (line1_item_p->is_primary_input() == FALSE)        // Abfrage nach Primaereingang   
   gate1_e = (*line1_item_p->vorgaenger())[0]->gib_type();
else
   gate1_e = DRIVER;
if (line2_item_p->is_primary_input() == FALSE)
   gate2_e = (*line2_item_p->vorgaenger())[0]->gib_type();
else
   gate2_e = DRIVER;

if ((gate1_e == FANOUT) || (gate2_e == FANOUT) ||
    (gate1_e == EXOR) || (gate2_e == EXOR) ||
    (gate1_e == BLOCK) || (gate2_e == BLOCK) ||
    (gate1_e == D_FLIP_FLOP) || (gate2_e == D_FLIP_FLOP) ||
    (gate1_e == D_LATCH) || (gate2_e == D_LATCH) ||
    (gate1_e == UNDEFINED_GATE) || (gate2_e == UNDEFINED_GATE))
   return(NULL);

else if ((gate1_e == INVERTER) || (gate1_e == DRIVER) ||
	 (gate1_e == AND) || (gate1_e == OR))

	if (gate2_e == NOR)
	   {
	   line1_item_p->set_good_value(VAL_0);
	   line2_item_p->set_good_value(VAL_1);
	   return(line2_item_p);
	   }//if
	else if (gate2_e == NAND)
	   {
	   line1_item_p->set_good_value(VAL_1);
	   line2_item_p->set_good_value(VAL_0);
	   return(line2_item_p);
	   }//elif
	else
	   return(NULL);

else if (gate1_e == NOR)
   if ((gate2_e == AND) || (gate2_e == OR) ||
       (gate2_e == INVERTER) || (gate2_e == DRIVER))
      {
      line1_item_p->set_good_value(VAL_1);
      line2_item_p->set_good_value(VAL_0);
      return(line1_item_p);
      }//if
   else if (gate2_e == NOR)
      {
      line1_item_p->set_good_value(VAL_1);
      line2_item_p->set_good_value(VAL_0);
      return(line1_item_p);
      }//elif
   else
      return(NULL);

else if (gate1_e == NAND)
   if ((gate2_e == AND) || (gate2_e == OR) ||
       (gate2_e == INVERTER) || (gate2_e == DRIVER))
      {
      line1_item_p->set_good_value(VAL_0);
      line2_item_p->set_good_value(VAL_1);
      return(line1_item_p);
      }//if
   else if (gate2_e == NAND)
      {
      line1_item_p->set_good_value(VAL_0);
      line2_item_p->set_good_value(VAL_1);
      return(line1_item_p);
      }//elif
   else
      return(NULL);
else
   return(NULL);
}//signal_to_propagate

//**************************************************************
//
//   FUNCTION-NAME: signal_to_propagate_2nd
//
//   CREATION-DATE: 24.2.1994
//
//   Sind die Vorgaengergatter entweder beides NAND- oder beides
//   NOR-Gatter, so muss noch einmal festgestellt werden,
//   welches Signal 'stark' und welches 'schwach' ist.
//   Das 'schwache' Signal wird propagiert.
//   
//
//**************************************************************
line_C* signal_to_propagate_2nd(line_C *line1_item_p, line_C *line2_item_p)
{
gate_t gate1_e, gate2_e;

gate1_e = (*line1_item_p->vorgaenger())[0]->gib_type();
gate2_e = (*line2_item_p->vorgaenger())[0]->gib_type();

if ((gate1_e == NOR) && (gate2_e == NOR))
   {
   line1_item_p->set_good_value(VAL_0);
   line2_item_p->set_good_value(VAL_1);
   return(line2_item_p);
   }//if

else if ((gate1_e == NAND) && (gate2_e == NAND))
   {
   line1_item_p->set_good_value(VAL_1);
   line2_item_p->set_good_value(VAL_0);
   return(line2_item_p);
   }//elif
else
   return(NULL);
}//signal_to_propagate_2nd

//**************************************************************
//
//   FUNCTION-NAME: gen_br_fault
//
//   CREATION-DATE: 15.12.1993
//
//   Nimmt fuer das Bridgin-Fault Fehlermodell den naechsten Zielfehler
//   aus der Zielfehlerliste
//
//**************************************************************
line_line_C *gen_br_fault()
{
line_line_C *lili_p;

forall(lili_p,zielfehler2_C)
   if ((forall_index == 0) && (act_bridging_list_index_i != 0))
      forall_index = act_bridging_list_index_i;
   else if (lili_p->test_required() == TRUE)
      {
      act_bridging_list_index_i = forall_index;
      return(lili_p);
      }//elif
return(NULL);
}//gen_br_fault

//*********************************************************************
//
//   FUNCTION-NAME: gen_fault
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Der naechste Test wird initialisiert. Zunaechst werden
//	die input Lines getestet, anschliessend die restlichen Lines. Wenn es kein
//	zu Testtenden Wert mit aktuellen Init Pattern gibt, muss ein Initmuster
//	fuer ein zu testendes Element bestimmt werden. Wenn die Line gefunden wird
//	muss die geoeffnete Liste noch abgeschlossen werden.
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: line_C*: Die Line, die als naechstes zu testen ist.
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
line_value_t *gen_fault()
{
line_value_t *liva_p = NULL;

init_pattern_b = FALSE;
if ((only_iddq_test_b == FALSE) && (only_stuck_at_test_b == FALSE))
	{
	forall(liva_p,zielfehler_C)
		if (liva_p->line_p->should_be_tested(liva_p->val_e) == TRUE)
			return(liva_p);
	}//if

forall(liva_p,zielfehler_C)
	if (liva_p->line_p->must_be_tested(liva_p->val_e) == TRUE)
		return(liva_p);
return(NULL);
}//gen_fault

//*********************************************************************
//
//   FUNCTION-NAME: test_complex_gate
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: Februar 92
//
//   FUNCTIONAL-DESCRIPTION: Diese Funktion stellt fest, ob mit CTEST ein
//	Block noch getestet werden muss oder ob der Test schon erledigt ist.
//
//   FORMAL PARAMETERS: line_C *line_p: Ausgang des Blockes
//			boolean init_allowed_b: Ist Initialisierung erlaubt
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: boolean: TRUE, wenn Testmuster gefunden wurde
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
boolean test_complex_gate(line_C *line_p,boolean init_allowed_b)
{
/*
line_C *line_item_p;
ctest_belegung_t *it_belegung_p,*cube_p;
ctest_belegung_t *cube_bel_p;
int init_allowed_i = (init_allowed_b == TRUE)?1:0;
int is_init_i;
int init_value_i;
gate_C *test_block_p = (*line_p->vorgaenger())[0];
int block_id_i,i,output_id_i;
boolean test_generated_b;
boolean first_entry_b;

if (tested_blocks_C.element_in_liste(test_block_p) == TRUE)
	return(FALSE);
	// Test of switch level Macros
block_id_i = test_block_p->get_name().gib_id();
if (block_id_i == -1)
	{
	int instance_id_i,status_i;

	test_block_p->get_name().gib_mixed_circuit_id(instance_id_i,block_id_i);
	select_instance(&instance_id_i,&status_i);
	if (status_i != 0)
		error("test_complex_gate: select_instance terminated with error");
	}//if
output_id_i = line_p->get_name().gib_id();
switch ((*test_block_p->gib_outputs())[0]->get_init_value())
	{
	case VAL_X:
		if (init_allowed_b == FALSE)
			return(FALSE);
		init_value_i = -1;
		break;
	case VAL_0:
		init_value_i = 0;
		break;
	case VAL_1:
		init_value_i = 1;
		break;
	default:
		error("line_C::gen_fault: Error in output value");
	}
it_belegung_p = new ctest_belegung_t;
cube_p = it_belegung_p;
first_entry_b = TRUE;
forall(line_item_p,*test_block_p->gib_inputs())
	{
	if (first_entry_b == FALSE)
		{
		cube_p->next_p = new ctest_belegung_t;
		cube_p = cube_p->next_p;
		}//if
	else
		first_entry_b = FALSE;
	switch (line_item_p->get_init_value())
		{
		case VAL_X:
			cube_p->val_c = 'X';
			break;
		case VAL_0:
			cube_p->val_c = '0';
			break;
		case VAL_1:
			cube_p->val_c = '1';
			break;
		default:
			error("line_C::gen_fault: Error in input value");
		}//switch
	}//forall
cube_p->next_p = NULL;
cube_p = it_belegung_p;
generate_test(&block_id_i,&output_id_i,&init_value_i,&init_allowed_i,&it_belegung_p,&is_init_i);
if (!init_allowed_i)
	hilf_C.fuege_ein(line_p,VAL_X);
if (it_belegung_p == NULL)
	test_generated_b = FALSE;
else
   {
   test_generated_b = TRUE;
   for (cube_bel_p=it_belegung_p,i=0;cube_bel_p!=NULL;i++,cube_bel_p=cube_bel_p->next_p)
	{
	value_t val_e;
	
	switch(cube_bel_p->val_c)
		{
		case '1':
			val_e = VAL_1;
			break;
		case '0':
			val_e = VAL_0;
			break;
		case 'X':
			val_e = VAL_X;
			break;
		default:
			error("test_complex_gate: Error in cube value");
		}//switch
	(*test_block_p->gib_inputs())[i]->set_good_value(val_e);
	}//for
   if ((only_iddq_test_b == FALSE) && (iddq_stuck_on_test_b == FALSE))
	if (init_value_i == 0)
		{
		(*test_block_p->gib_outputs())[0]->set_good_value(VAL_0);
		(*test_block_p->gib_outputs())[0]->set_bad_value(VAL_1);
		}//if
	else
		{
		(*test_block_p->gib_outputs())[0]->set_good_value(VAL_1);
		(*test_block_p->gib_outputs())[0]->set_bad_value(VAL_0);
		}//else
   else
	if (init_value_i == 0)
		(*test_block_p->gib_outputs())[0]->set_good_value(VAL_0);
	else
		(*test_block_p->gib_outputs())[0]->set_good_value(VAL_1);
   }//else
cube_bel_p = cube_p;
while (cube_p->next_p != NULL)
	{
	ctest_belegung_t *old_cube_p = cube_p;
	
	cube_bel_p = cube_p;
	while (cube_bel_p->next_p != NULL)
		{
		old_cube_p = cube_bel_p;
		cube_bel_p = cube_bel_p->next_p;
		}//while
	if (old_cube_p->next_p != NULL)
		{
		delete old_cube_p->next_p;
		old_cube_p->next_p = NULL;
		}//if
	}//while
delete cube_p;
return (test_generated_b);
*/
}//test_complex_gate

//*********************************************************************
//
//   FUNCTION-NAME: must_be_tested
//
//   CLASS: line_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Diese Funktion stellt fest, ob ein Element auf stuck-at
//	getestet werden kann, bzw. fuer einen Rising oder Falling Test initialisiert
//	werden kann. In diesen Faellen ist der Wert der Funktion TRUE.
//
//   FORMAL PARAMETERS: keine
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: boolean: Ist ein s-a Test oder Rising/Falling init moeglich
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
boolean line_C::must_be_tested(value_t val_e)
{
ctest_pattern_b = FALSE;
if (test_done_b == TRUE)	// Test bereits generiert
	{
	hilf_C.fuege_ein(this,val_e);
	return(FALSE);
	}//if
if (val_e == VAL_X)
	if (test_complex_gate(this,TRUE) == TRUE)
		{
		ctest_pattern_b = TRUE;
		init_pattern_b = TRUE;
		return(TRUE);
		}//if
	else
		{
		if (tested_blocks_C.element_in_liste((*vorgaenger_p)[0]) == FALSE)
			tested_blocks_C.fuege_ein((*vorgaenger_p)[0]);
		if (phase3_b == TRUE)
			hilf_C.fuege_ein(this,VAL_X);
		return(FALSE);
		}//else
if (val_e == VAL_1)
   if ((s_a_1_untestable_b == FALSE) && (s_a_1_aborted_b == FALSE) && (s_a_1_test_b == FALSE))
	{ // s-a-1 Test kann durchgefuehrt werden
	good_value_e = VAL_0;
	bad_value_e = VAL_1;
	return(TRUE);
	}//if
   else if ((s_a_1_aborted_b == FALSE) && ((only_stuck_at_test_b == TRUE) || (only_iddq_test_b == TRUE)))
	{
	hilf_C.fuege_ein(this,val_e);
	return(FALSE);
	}//else if
if (val_e == VAL_0)
   if ((s_a_0_untestable_b == FALSE) && (s_a_0_aborted_b == FALSE) && (s_a_0_test_b == FALSE))
	{ // s-a-0 Test kann durchgefuehrt werden
	good_value_e = VAL_1;
	bad_value_e = VAL_0;
	return(TRUE);
	}//if
   else if ((s_a_0_aborted_b == FALSE) && ((only_stuck_at_test_b == TRUE) || (only_iddq_test_b == TRUE)))
	{
	hilf_C.fuege_ein(this,val_e);
	return(FALSE);
	}//else if
if ((val_e == VAL_U0) && (s_a_1_untestable_b == FALSE) && (s_a_1_aborted_b == FALSE))
   if ((falling_test_b == FALSE) && (falling_untestable_b == FALSE) && (falling_aborted_b == FALSE))
	{ // Initialisierung fuer Falling Test kann durchgefuehrt werden
	good_value_e = VAL_1;
	bad_value_e = VAL_0;
	init_pattern_b = TRUE;
	return(TRUE);
	}//if
if ((val_e == VAL_U1) && (s_a_0_untestable_b == FALSE) && (s_a_0_aborted_b == FALSE))
   if ((rising_test_b == FALSE) && (rising_untestable_b == FALSE) && (rising_aborted_b == FALSE))
	{ // Initialisierung fuer Rising Test kann durchgefuehrt werden
	good_value_e = VAL_0;
	bad_value_e = VAL_1;
	init_pattern_b = TRUE;
	return(TRUE);
	}//if
return(FALSE);
}//must_be_tested

//*********************************************************************
//
//   FUNCTION-NAME: should_be_tested
//
//   CLASS: line_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Es wird festgestellt, ob der naechste Rising/Falling
//	Test mit dem aktuellen Init-Muster durchgefuehrt werden kann. Dann ist der
//	Wert der Funktion TRUE.
//
//   FORMAL PARAMETERS: keine
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: boolean: Naechster Rising/Falling Test moeglich
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
boolean line_C::should_be_tested(value_t val_e)
{
ctest_pattern_b = FALSE;
if (test_done_b == TRUE)	// Test bereits generiert
	return(FALSE);
if ((val_e == VAL_X) && (prime_input_b == FALSE) && ((*vorgaenger_p)[0]->gib_type() == BLOCK))
	if (test_complex_gate(this,FALSE) == TRUE)
		{
		ctest_pattern_b = TRUE;
		return(TRUE);
		}//if
	else
		return(FALSE);
if ((val_e == VAL_0) && (s_a_0_untestable_b == FALSE) && (s_a_0_aborted_b == FALSE))
	{
	good_value_e = VAL_1;
	bad_value_e = VAL_0;
	return(TRUE);
	}//if
else if ((val_e == VAL_1) && (s_a_1_untestable_b == FALSE) && (s_a_1_aborted_b == FALSE))
	{
	good_value_e = VAL_0;
	bad_value_e = VAL_1;
	return(TRUE);
	}//if
else if ((val_e == VAL_U1) && (rising_untestable_b == FALSE) && (rising_aborted_b == FALSE))
   {
   if ((rising_test_b == FALSE) && ((init_value_e == VAL_0) || (init_value_e == VAL_D_BAR) ||
				    (iddq_stuck_on_test_b == TRUE)))
     {
     boolean test_gen_possible_b = TRUE;

     if (stuck_open_test_b == TRUE)
	{
	gate_C *follow_p = nachfolger_p->gib_only_gate();
	line_C *line_item_p;

			//nicht kontrollierende Werte muessen gefunden werden am Nachfolger
	switch ((*nachfolger_p)[0]->gib_type())
		{
		case AND:
		case NAND:
			{forall(line_item_p,*follow_p->gib_inputs())
				{
				if (line_item_p == this)
					continue;
				if (line_item_p->get_init_value() != VAL_1)
					test_gen_possible_b = FALSE;
				}}//forall
			break;
		case OR:
		case NOR:
			{forall(line_item_p,*follow_p->gib_inputs())
					{
				if (line_item_p == this)
					continue;
				if (line_item_p->get_init_value() != VAL_0)
					test_gen_possible_b = FALSE;
				}}//forall
			break;
		case EXOR:
			break;
		default:
			error("should_be_tested: Stuck-open test not possible for one gate");
		}//switch
	}//if
     if (test_gen_possible_b == TRUE)
	{
	good_value_e = VAL_1;
	bad_value_e = VAL_0;
	return(TRUE);
	}//if
     }//if
   }//elif
else if ((val_e == VAL_U0) && (falling_untestable_b == FALSE) && (falling_aborted_b == FALSE))
   {
   if ((falling_test_b == FALSE) && ((init_value_e == VAL_1) || (init_value_e == VAL_D) ||
				     (iddq_stuck_on_test_b == TRUE)))
     {
     boolean test_gen_possible_b = TRUE;

     if (stuck_open_test_b == TRUE)
	{
	gate_C *follow_p = nachfolger_p->gib_only_gate();
	line_C *line_item_p;

			//nicht kontrollierende Werte muessen gefunden werden am Nachfolger
	switch ((*nachfolger_p)[0]->gib_type())
		{
		case AND:
		case NAND:
			{forall(line_item_p,*follow_p->gib_inputs())
				{
				if (line_item_p == this)
					continue;
				if (line_item_p->get_init_value() != VAL_1)
					test_gen_possible_b = FALSE;
				}}//forall
			break;
		case OR:
		case NOR:
			{forall(line_item_p,*follow_p->gib_inputs())
				{
				if (line_item_p == this)
					continue;
				if (line_item_p->get_init_value() != VAL_0)
					test_gen_possible_b = FALSE;
				}}//forall
			break;
		case EXOR:
			break;
		default:
                        error("should_be_tested: Stuck-open test not possible for one gate");
		}//switch
	}//if
     if (test_gen_possible_b == TRUE)
	{
	good_value_e = VAL_0;
	bad_value_e = VAL_1;
	return(TRUE);
	}//if
     }//if
   }//elif
return(FALSE);
}//should_be_tested

//*********************************************************************
//
//   FUNCTION-NAME: mark_result
//
//   CLASS: line_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Der Test war aborted oder redundant. Die Funktion
//	schreibt das in die Datenstruktur
//
//   FORMAL PARAMETERS: fan_result_t result_e: ABORTED_FAULT oder REDUNDANT_FAULT
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: NONE
//
//   MODIFICATION-HISTORY: maerz 92 markieren von aequivalenten
//                         Redundanzen und aborted Faults mit
//                         mark_aequivalenzen() 
//			   dominate aborted Faults mit mark_dominanzen()	
//*********************************************************************
void line_C::mark_result(fan_result_t result_e,value_t ziel_val_e)
{
int reject_level_i = max_number_backtracks_i,null_i = 0,eins_i = 1,block_id_i,output_id_i;
gate_C *test_block_p;

if ((prime_input_b == FALSE) && ((*vorgaenger_p)[0]->gib_type() == BLOCK) && (ziel_val_e == VAL_X))
	{
	test_block_p = (*vorgaenger_p)[0];
//	block_id_i = test_block_p->get_name().gib_id();
	if (block_id_i == -1)
		{
		int instance_id_i,status_i;

//		test_block_p->get_name().gib_mixed_circuit_id(instance_id_i,block_id_i);
		select_instance(&instance_id_i,&status_i);
		if (status_i != 0)
			error("line_C::mark_result: select_instance terminated with error");
		}//if
//	output_id_i = name_sC.gib_id();
	reject_counter_i++;
	if ((result_e == ABORTED_FAULT) ||
	    ((hazard_occured_b == TRUE) && (hazard_test_ein_b == TRUE)))
		reject(&block_id_i,&output_id_i,&null_i);
	else
		reject(&block_id_i,&output_id_i,&eins_i);
	return;
	}//if
if ((hazard_occured_b == TRUE) && (hazard_test_ein_b == TRUE))
   {
   if (ziel_val_e == VAL_0)
	{
	s_a_0_aborted_b = TRUE;
	rising_aborted_b = TRUE;
	if ((init_value_e == VAL_0) && (rising_test_b == FALSE))
		rising_non_robust_b = TRUE;
#ifndef O_FS
	mark_all(this,Stuck_at_0);
#endif
	}
   else if (ziel_val_e == VAL_1)
	{
	s_a_1_aborted_b = TRUE;
	falling_aborted_b = TRUE;
	if ((init_value_e == VAL_1) && (falling_test_b == FALSE))
		falling_non_robust_b = TRUE;
#ifndef O_FS
	mark_all(this,Stuck_at_1);
#endif
	}//else
   else if (ziel_val_e == VAL_U1)
	{
	rising_aborted_b = TRUE;
	if ((init_value_e == VAL_0) && (rising_test_b == FALSE))
		rising_non_robust_b = TRUE;
	}//else
   else if (ziel_val_e == VAL_U0)
	{
	falling_aborted_b = TRUE;
	if ((init_value_e == VAL_1) && (falling_test_b == FALSE))
		falling_non_robust_b = TRUE;
	}//else
   else
	error("line_C::mark_result: Unknown Value");
   return;
   }//if
if (ziel_val_e == VAL_1)
	if (result_e == ABORTED_FAULT)
		{
		s_a_1_aborted_b = TRUE;
		falling_aborted_b = TRUE;
#ifndef O_FS
		mark_all(this,Stuck_at_1);
#endif
		}//if
	else
		{
		s_a_1_untestable_b = TRUE;
		falling_untestable_b = TRUE;
#ifndef O_FS
                mark_all(this,Stuck_at_1);
#endif
		hilf_C.fuege_ein(this,VAL_1);
		hilf_C.fuege_ein(this,VAL_U0);
		}//else
else if (ziel_val_e == VAL_0)
	if (result_e == ABORTED_FAULT)
		{
		s_a_0_aborted_b = TRUE;
		rising_aborted_b = TRUE;
#ifndef O_FS
                mark_all(this,Stuck_at_0);
#endif
		}//if
	else
		{
		rising_untestable_b = TRUE;
		s_a_0_untestable_b = TRUE;
#ifndef O_FS
                mark_all(this,Stuck_at_0);
#endif
		hilf_C.fuege_ein(this,VAL_0);
		hilf_C.fuege_ein(this,VAL_U1);
		}//else
else if (init_pattern_b == TRUE)
	if (result_e == ABORTED_FAULT)
		 {
		 falling_aborted_b = TRUE;
		 rising_aborted_b = TRUE;
		 }//if
	else
		{
		rising_untestable_b = TRUE;
		hilf_C.fuege_ein(this,VAL_U1);
		falling_untestable_b = TRUE;
		hilf_C.fuege_ein(this,VAL_U0);
		}//else
else if (ziel_val_e == VAL_U1)
        if (result_e == ABORTED_FAULT)
                {
                rising_aborted_b = TRUE;
                if (iddq_stuck_on_test_b == TRUE)
                        s_a_0_aborted_b = TRUE;
		}//if
        else
                {
                rising_untestable_b = TRUE;
                if (iddq_stuck_on_test_b == TRUE)
                        s_a_0_untestable_b = TRUE;
                hilf_C.fuege_ein(this,VAL_U1);
		}//else
else if (ziel_val_e == VAL_U0)
        if (result_e == ABORTED_FAULT)
                {
                falling_aborted_b = TRUE;
                if (iddq_stuck_on_test_b == TRUE)
                        s_a_1_aborted_b = TRUE;
		}//if
        else
                {
                falling_untestable_b = TRUE;
                if (iddq_stuck_on_test_b == TRUE)
                        s_a_1_untestable_b = TRUE;
                hilf_C.fuege_ein(this,VAL_U0);
		}//else
else
	error("line_C::mark_result: Unknown Value");
}//mark_result

//*********************************************************************
//
//   FUNCTION-NAME: search_next_path
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: Maerz 92
//
//   FUNCTIONAL-DESCRIPTION: Die Funktion ruft fuer den aktuellen Input die rekursive Funktion
//	search_next_path auf, um im Pfadverzoegerungsfehlermodell den naechsten zu testenden
//	Pfad zu ermitteln. Gegebenenfalls wird der aktuelle Input zurueckgegeben. Ist kein
//	weiterer Pfad mehr vorhanden, so wird NULL zurueckgegeben.
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: line_C*: Der entsprechende Input oder NULL
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void gate_C::search_next_path(line_C *line_p,boolean &path_set_b)
{
line_C *line_item_p;
value_t non_controll_val_e;

switch (gate_e)
	{
	case AND:
	case NAND:
		non_controll_val_e = VAL_1;
		break;
	case OR:
	case NOR:
		non_controll_val_e = VAL_0;
		break;
	case EXOR:
	case BLOCK:
	case FANOUT:
		non_controll_val_e = VAL_X;
		break;
	case D_LATCH:
	case D_FLIP_FLOP:
		error("gate_C::search_next_path: not implemented for sequential gates");
	}//switch
if (gate_e == FANOUT)
	line_item_p = (*output_p)[actual_branch_i];
else
	line_item_p = (*output_p)[0];
if ((line_item_p->search_next_path(path_set_b) == TRUE) && (path_set_b == TRUE))
	{
	selected_path_C.fuege_ein(this);
	if (non_controll_val_e != VAL_X)
		{forall(line_item_p,*input_p)
			if (line_item_p != line_p)
				{
				line_item_p->set_good_value(non_controll_val_e);
				if (line_item_p->gib_type() == BOUND_LINE)
					unjustified_bound_line_C.fuege_ein(line_item_p);
				}//if
		}
	return;
	}//if
if (gate_e != FANOUT)
	return;
actual_branch_i++;
if (output_p->no_elements() <= actual_branch_i)
	{
	actual_branch_i = 0;
	return;
	}
line_item_p = (*output_p)[actual_branch_i];
path_set_b = TRUE;
if (line_item_p->search_next_path(path_set_b) == TRUE)
        return;
else
	error("search_next_path: line_C::search_next_path is faulty");
}//search_next_path

//*********************************************************************
//
//   FUNCTION-NAME: search_next_path
//
//   CLASS: line_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: Maerz 92
//
//   FUNCTIONAL-DESCRIPTION: Die Funktion ruft search_next_path fuer das folgende Gatter auf,
//	wenn die aktuelle Line kein Primaerausgang ist.
//
//   FORMAL PARAMETERS: boolean io path_set_b: Eingabe: soll Pfad gesetzt werden
//					       Ausgabe: ist Pfad gesetzt worden
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: boolean: Pfad gefunden
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
boolean line_C::search_next_path(boolean &path_set_b)
{
if (prime_output_b == TRUE)
	return(TRUE);
nachfolger_p->gib_only_gate()->search_next_path(this,path_set_b);
if (prime_input_b == TRUE)
	return(path_set_b);
return(TRUE);
}//search_next_path

//*********************************************************************
//
//   FUNCTION-NAME: search_next_path
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: Maerz 92
//
//   FUNCTIONAL-DESCRIPTION: Die Funktion ruft fuer den aktuellen Input die rekursive Funktion
//	search_next_path auf, um im Pfadverzoegerungsfehlermodell den naechsten zu testenden
//	Pfad zu ermitteln. Gegebenenfalls wird der aktuelle Input zurueckgegeben. Ist kein
//	weiterer Pfad mehr vorhanden, so wird NULL zurueckgegeben.
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: line_C*: Der entsprechende Input oder NULL
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
line_C *search_next_path()
{
static int actual_input_i = -1;
line_C *line_item_p;
boolean set_path_b = FALSE;

selected_path_C.loesche_liste();
if (actual_input_i > -1)
   {
   line_item_p = input_lines_C[actual_input_i];
   if (line_item_p->search_next_path(set_path_b) == TRUE)
	{
	line_item_p->set_good_value(VAL_1);
	line_item_p->set_bad_value(VAL_0);
	return(line_item_p);
	}//if
   }//if
actual_input_i++;
set_path_b = TRUE;
if (input_lines_C.no_elements() <= actual_input_i)
	return(NULL);
line_item_p = input_lines_C[actual_input_i];
if (line_item_p->search_next_path(set_path_b) == TRUE)
	{
	line_item_p->set_good_value(VAL_1);
	line_item_p->set_bad_value(VAL_0);
	return(line_item_p);
	}//if
else
	error("search_next_path: line_C::search_next_path is faulty");
}//search_next_path
