//*********************************************************************
//
//   FILE-NAME: probability.c
//
//   CREATION-DATE: August 91
//
//   FILE-DESCRIPTION: Dieses File beinhaltet die Hauptroutinen fuer Berechnung
//	von Controllability und Observability
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
#include <math.h>
#include "extern.h"
#include "line_liste_C.h"
#include "line_C.h"
#include "gate_liste_C.h"
#include "gate_C.h"

	// Globale Variablen
extern double zero_contby_a[max_number_gate_inputs_i];
extern double one_contby_a[max_number_gate_inputs_i];
extern double durch_a[max_no_io_i];
extern double nachfolger_observaby_a[max_no_io_i];
extern double sum_1_contby_f;
extern double sum_0_contby_f;
extern int anzahl_outputs_i;
double prod_1_contby_f;
double prod_0_contby_f;
//*********************************************************************
//
//   FUNCTION-NAME: prob_compute_contby
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Controllability wird ermittelt. Fuer alle
//	Primaereingaenge wird prob_compute_contby aufgerufen
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void prob_compute_contby()
{
line_C *line_item_p;

forall(line_item_p,lines_C)
	line_item_p->reset_testaby();
{forall(line_item_p,input_lines_C)
	line_item_p->prob_compute_contby();}
}

//*********************************************************************
//
//   FUNCTION-NAME: prob_compute_observaby
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Observability wird ermittelt. Fuer alle
//	Primaerausgaenge wird compute_observaby aufgerufen
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void prob_compute_observaby()
{
line_C *line_item_p;

forall(line_item_p,output_lines_C)
	line_item_p->prob_compute_observaby();
{forall(line_item_p,lines_C)
	line_item_p->berechne_testaby_kehrwert();}
}

//*********************************************************************
//
//   FUNCTION-NAME: prob_compute_contby
//
//   CLASS: line_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Fuer alle Nachfolgenden Lines wird prob_compute_contby
//	aufgerufen.
//
//   FORMAL PARAMETERS: keine
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void gate_C::prob_compute_contby()
{
line_C *line_item_p;

forall(line_item_p,*output_p)
	line_item_p->prob_compute_contby();
}//prob_compute_contby

//*********************************************************************
//
//   FUNCTION-NAME: prob_compute_contby
//
//   CLASS: gate_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Fuer alle Nachfolgenden Gates wird prob_compute_contby
//	aufgerufen. Die Controllability wird nach der Formel von Bernhard
//	Krueger-Sprengel (ueberarbeitet von Thomas Lange) berechnet.
//	Fuer Einzelheiten siehe entsprechendes Paper. Falls die Controllability
//	bereits berechnet wurde, wird nichts gemacht.
//
//   FORMAL PARAMETERS: keine
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void line_C::prob_compute_contby()
{
line_C *line_item_p;
gate_C *gate_item_p;
double global_0_contby_f = 0.0;
double global_1_contby_f = 0.0;
double no_0_at_input_for_0_at_output_f;
double no_1_at_input_for_0_at_output_f;
cube_list_t *cube_array_p;
cube_list_t *cubes_array_p;
int i;
value_t mask_e;
boolean standard_b;
gate_t testability_gate_e;
belegung_t *beleg_p;

if (one_contby_f != 0.0)
	return;
	// Controllability wurde schon berechnet

if (prime_input_b == TRUE)
	{ // Eingaenge haben die Controllability 1
	zero_contby_f = 0.5;
	one_contby_f = 0.5;
	contby_f = 0.5;
		// Aufruf fuer die Nachfolger
	forall(gate_item_p,*nachfolger_p)
		gate_item_p->prob_compute_contby(); // Aufruf fuer das naechste Element
	return;
	}//if

i = 0;
gate_item_p = vorgaenger_p->gib_only_gate();
{forall(line_item_p,*gate_item_p->gib_inputs())
	{ // Die Controllability der Vorgaenger abholen
	zero_contby_a[i] = line_item_p->gib_zero_contby();
	one_contby_a[i] = line_item_p->gib_one_contby();
	i++;
	}}//forall

if (i != gate_item_p->no_inputs())
	error("prob_compute_contby: Problems with number of gate inputs");

for (i=0;i<gate_item_p->no_inputs();i++)
	if ((zero_contby_a[i] == 0.0) || (one_contby_a[i] == 0.0))
		{
		return; //Werte fuer die Vorgaenger nicht alle berechnet
		}//if

// Kommentar  siehe testability_ana.c  compute_contby()
testability_gate_e = gate_item_p->gib_type();
mask_e = VAL_X;
standard_b = TRUE;
if ( testability_gate_e == BLOCK )
   {
     cube_array_p = gate_item_p->gib_cubes(VAL_0);
     if ( cube_array_p->next_p == NULL )
        {
	  mask_e = cube_array_p->to_belegung_p->val_e;
	  for ( beleg_p = cube_array_p->to_belegung_p->next_p; beleg_p != NULL; beleg_p = beleg_p->next_p)
	      {
	        if ( beleg_p->val_e != mask_e )
		   {
		     standard_b = FALSE;
		     break;
		   }
	      }
	  if ( standard_b == TRUE)
	     if ( mask_e == VAL_0 ) testability_gate_e = OR;
	     else testability_gate_e = NAND;
	}
     cube_array_p = gate_item_p->gib_cubes(VAL_1);
     if ( cube_array_p->next_p == NULL )
	{
          mask_e = cube_array_p->to_belegung_p->val_e;
	  for ( beleg_p = cube_array_p->to_belegung_p->next_p; beleg_p != NULL; beleg_p = beleg_p->next_p)
	      {
	        if ( beleg_p->val_e != mask_e )
		   {
		     standard_b = FALSE;
                     break;
		   }
	      }
	  if ( standard_b == TRUE)
	     if ( mask_e == VAL_0 ) testability_gate_e = NOR;
	     else testability_gate_e = AND;
	}
      cube_array_p = gate_item_p->gib_cubes(VAL_0);
      cubes_array_p = gate_item_p->gib_cubes(VAL_1);
      if ( cube_array_p->next_p == NULL && cubes_array_p->next_p == NULL )
         if ( cube_array_p->to_belegung_p->next_p == NULL && cubes_array_p->to_belegung_p->next_p == NULL )
	    {
	      if ( cube_array_p->to_belegung_p->val_e == VAL_0 )
	         testability_gate_e = DRIVER;
	         else testability_gate_e = INVERTER;
	    }
}
	    
		  
switch (testability_gate_e)
	{	// Vorbelegung fuer alle Gatter
	case AND:
		one_contby_f = 1.0;
		for (i=0;i<gate_item_p->no_inputs();i++)
			one_contby_f = one_contby_f * one_contby_a[i];
		zero_contby_f = 1.0 - one_contby_f;
		break;
	case NOR:
		one_contby_f = 1.0;
		for (i=0;i<gate_item_p->no_inputs();i++)
			one_contby_f = one_contby_f * zero_contby_a[i];
		zero_contby_f = 1.0 - one_contby_f;
		break;
	case INVERTER:
		one_contby_f = zero_contby_a[0];
		zero_contby_f = 1.0 - one_contby_f;
		break;
	case FANOUT:
	case DRIVER:
		one_contby_f = one_contby_a[0];
		zero_contby_f = 1.0 - one_contby_f;
		break;
	case NAND:
		zero_contby_f = 1.0;
		for (i=0;i<gate_item_p->no_inputs();i++)
			zero_contby_f = zero_contby_f * one_contby_a[i];
		one_contby_f = 1.0 - zero_contby_f;
		break;
	case OR:
		zero_contby_f = 1.0;
		for (i=0;i<gate_item_p->no_inputs();i++)
			zero_contby_f = zero_contby_f * zero_contby_a[i];
		one_contby_f = 1.0 - zero_contby_f;
		break;
	case EXOR:
		zero_contby_f = zero_contby_a[0] * zero_contby_a[1] + one_contby_a[0] * one_contby_a[1];
		one_contby_f = 1.0 - zero_contby_f;
		break;
	case BLOCK:
		double anzahl_x_werte_f;
		double anzahl_1_werte_f,anzahl_0_werte_f;
		double null_wichtung_f,eins_wichtung_f;
		cube_list_t *cube_l_p;
		belegung_t *bel_p;
		
		anzahl_x_werte_f = 0.0;
		anzahl_1_werte_f = 0.0;
		anzahl_0_werte_f = 0.0;
		null_wichtung_f = 1.0;
		eins_wichtung_f = 1.0;
		for (i=0;i<gate_item_p->no_inputs();i++)
			eins_wichtung_f = eins_wichtung_f * one_contby_a[i];
		null_wichtung_f = 1.0 - eins_wichtung_f;
		no_0_at_input_for_0_at_output_f = 0.0;
		no_1_at_input_for_0_at_output_f = 0.0;
		cube_array_p = gate_item_p->gib_cubes(VAL_0);
		for (cube_l_p=cube_array_p;cube_l_p!=NULL;cube_l_p=cube_l_p->next_p)
		   {
		   for (bel_p=cube_l_p->to_belegung_p;bel_p!=NULL;bel_p=bel_p->next_p)
			if (bel_p->val_e == VAL_X)
				anzahl_x_werte_f++;
			else if (bel_p->val_e == VAL_1)
				anzahl_1_werte_f++;
			else if (bel_p->val_e == VAL_0)
				anzahl_0_werte_f++;
			else
				error("prob_compute_contby:  Unknown value in cubes");
		   no_0_at_input_for_0_at_output_f+=pow(2,anzahl_x_werte_f)*(anzahl_x_werte_f/2 + anzahl_0_werte_f);
		   no_1_at_input_for_0_at_output_f+=pow(2,anzahl_x_werte_f)*(anzahl_x_werte_f/2 + anzahl_1_werte_f);
		   }//for
		zero_contby_f = no_0_at_input_for_0_at_output_f * null_wichtung_f /
				(no_0_at_input_for_0_at_output_f + no_1_at_input_for_0_at_output_f);
		zero_contby_f += no_1_at_input_for_0_at_output_f * eins_wichtung_f /
				 (no_0_at_input_for_0_at_output_f + no_1_at_input_for_0_at_output_f);
		one_contby_f = 1 - zero_contby_f;
		break;
	case D_FLIP_FLOP:
	case D_LATCH:
	case UNDEFINED_GATE:
		error("prob_compute_contby: not implemented for sequential gates");
	}//switch

contby_f = 0.5 * (one_contby_f + zero_contby_f);

	// Aufruf fuer die Nachfolger
{forall(gate_item_p,*nachfolger_p)
	gate_item_p->prob_compute_contby(); // Aufruf fuer das naechste Element
}
}//prob_compute_contby

//*********************************************************************
//
//   FUNCTION-NAME: prob_compute_observaby
//
//   CLASS: gate_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Fuer alle Input Lines wird prob_compute_observaby aufgerufen.
//
//   FORMAL PARAMETERS: keine
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void gate_C::prob_compute_observaby()
{
for (int i=0;i<no_input_i;i++)
	(*input_p)[i]->prob_compute_observaby();
}//prob_compute_observaby

//*********************************************************************
//
//   FUNCTION-NAME: prob_compute_observaby
//
//   CLASS: line_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Fuer alle Vorgaenger Gates wird prob_compute_observaby
//	aufgerufen. Die Controllability wird nach der Formel von Bernhard
//	Krueger-Sprengel (ueberarbeitet von Thomas Lange) berechnet.
//	Fuer Einzelheiten siehe entsprechendes Paper. Falls die Observability bereits
//	berechnet wurde wird nichts gemacht.
//
//   FORMAL PARAMETERS: keine
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void line_C::prob_compute_observaby()
{
line_C *line_item_p;
gate_C *gate_item_p;
double local_observaby_f;
int j;
cube_list_t *cube_array_p;
cube_list_t *cubes_array_p;
value_t mask_e;
boolean standard_b;
gate_t testability_gate_e;
belegung_t *beleg_p;

if (observaby_f != 0.0)
	return;
	// Observability wurde schon berechnet
if (prime_output_b == TRUE)
	{ // Observability von Ausgaengen ist immer 1
	observaby_f = 1.0;
	testaby_f = contby_f * observaby_f;
		// Aufruf fuer den Vorgaenger der Line
	if (prime_input_b == FALSE)
		vorgaenger_p->gib_only_gate()->prob_compute_observaby(); //Aufruf fuer alle Vorgaenger
	return;
	}//if	Nachfolger ist Primaerausgang.
gate_item_p = nachfolger_p->gib_only_gate();
j = 0;
{forall(line_item_p,*gate_item_p->gib_outputs())
	nachfolger_observaby_a[j++] = line_item_p->gib_observaby();
}		// Die Observability Werte der Nachfolger werden geholt.
anzahl_outputs_i = j;

for (j=0;j<anzahl_outputs_i;j++)
	if (nachfolger_observaby_a[j] == 0.0)
		return;	//Observability fuer einen Nachfolger nicht berechnet.

prod_1_contby_f = 1.0;
prod_0_contby_f = 1.0;

gate_item_p = nachfolger_p->gib_only_gate();
{forall(line_item_p,*gate_item_p->gib_inputs())
	{ // Controllability Werte der Vorgaenger werden abgeholt
	if (line_item_p == this)
		continue;
	else
		{
		prod_1_contby_f = prod_1_contby_f * line_item_p->gib_one_contby();
		prod_0_contby_f = prod_0_contby_f * line_item_p->gib_zero_contby();
		}//else
	}}//forall

gate_item_p = nachfolger_p->gib_only_gate();

// Kommentar  siehe testability_ana.c  compute_contby()
testability_gate_e = gate_item_p->gib_type();
mask_e = VAL_X;
standard_b = TRUE;
if ( testability_gate_e == BLOCK )
   {
    cube_array_p = gate_item_p->gib_cubes(VAL_0);
    if ( cube_array_p->next_p == NULL )
       {
         mask_e = cube_array_p->to_belegung_p->val_e;
         for ( beleg_p = cube_array_p->to_belegung_p->next_p; beleg_p != NULL; beleg_p = beleg_p->next_p)
	     {
	       if ( beleg_p->val_e != mask_e )
		  {
		    standard_b = FALSE;
                    break;
		  }
	     }
	 if ( standard_b == TRUE)
	    if ( mask_e == VAL_0 ) testability_gate_e = OR;
	    else testability_gate_e = NAND;
       }
    cube_array_p = gate_item_p->gib_cubes(VAL_1);
    if ( cube_array_p->next_p == NULL )
       {
	mask_e = cube_array_p->to_belegung_p->val_e;
	for ( beleg_p = cube_array_p->to_belegung_p->next_p; beleg_p != NULL; beleg_p = beleg_p->next_p)
	    {
	      if ( beleg_p->val_e != mask_e )
	         {
                   standard_b = FALSE;
                   break;
		 }
	    }
	if ( standard_b == TRUE)
	   if ( mask_e == VAL_0 ) testability_gate_e = NOR;
	   else testability_gate_e = AND;
       }
     cube_array_p = gate_item_p->gib_cubes(VAL_0);
     cubes_array_p = gate_item_p->gib_cubes(VAL_1);
     if ( cube_array_p->next_p == NULL && cubes_array_p->next_p == NULL )
	if ( cube_array_p->to_belegung_p->next_p == NULL && cubes_array_p->to_belegung_p->next_p == NULL )
	   {
	     if ( cube_array_p->to_belegung_p->val_e == VAL_0 )
		testability_gate_e = DRIVER;
		else testability_gate_e = INVERTER;
	   }
}


switch (testability_gate_e)
	{ // Das Formelelement DURCH(...) wird berechnet
	case AND:
	case NAND:
		for(j=0;j<anzahl_outputs_i;j++)
			durch_a[j] = nachfolger_observaby_a[j] * prod_1_contby_f;
		break;
	case OR:
	case NOR:
		for(j=0;j<anzahl_outputs_i;j++)
			durch_a[j] = nachfolger_observaby_a[j] * prod_0_contby_f;
		break;
	case INVERTER:
	case DRIVER:
	case EXOR:
	case BLOCK:
		for(j=0;j<anzahl_outputs_i;j++)
			durch_a[j] = nachfolger_observaby_a[j];
		break;
	case FANOUT:
		for(j=0;j<anzahl_outputs_i;j++)
			durch_a[j] = nachfolger_observaby_a[j];
		break;
	case D_FLIP_FLOP:
	case D_LATCH:
	case UNDEFINED_GATE:
                error("prob_compute_observaby: not implemented for sequential gates");
	}//switch

	// Observability und Testability berechnen
local_observaby_f = durch_a[0];
for(j=0;j<anzahl_outputs_i;j++)
	if (durch_a[j] < local_observaby_f)
		local_observaby_f = durch_a[j];	// finde Minimum aller durch_a[j]

observaby_f = local_observaby_f;
testaby_f = contby_f * observaby_f;

	// Aufruf fuer den Vorgaenger der Line
if (prime_input_b == FALSE)
	vorgaenger_p->gib_only_gate()->prob_compute_observaby();
}//prob_compute_observaby

//*********************************************************************
//
//   FUNCTION-NAME: reko_compute_contby
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Controllability wird ermittelt. Fuer alle
//	Primaereingaenge wird reko_compute_contby aufgerufen
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void reko_compute_contby()
{
line_C *line_item_p;

forall(line_item_p,lines_C)
	line_item_p->reset_testaby();
{forall(line_item_p,input_lines_C)
	line_item_p->reko_compute_contby();}
}

//*********************************************************************
//
//   FUNCTION-NAME: reko_compute_observaby
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Observability wird ermittelt. Fuer alle
//	Primaerausgaenge wird compute_observaby aufgerufen
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void reko_compute_observaby()
{
line_C *line_item_p;

forall(line_item_p,output_lines_C)
	line_item_p->reko_compute_observaby();
}

//*********************************************************************
//
//   FUNCTION-NAME: reko_compute_contby
//
//   CLASS: line_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Fuer alle Nachfolgenden Lines wird reko_compute_contby
//	aufgerufen.
//
//   FORMAL PARAMETERS: keine
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void gate_C::reko_compute_contby()
{
line_C *line_item_p;

forall(line_item_p,*output_p)
	line_item_p->reko_compute_contby();
}//reko_compute_contby

//*********************************************************************
//
//   FUNCTION-NAME: reko_compute_contby
//
//   CLASS: gate_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Fuer alle Nachfolgenden Gates wird reko_compute_contby
//	aufgerufen. Die Controllability wird nach der Formel von Bernhard
//	Krueger-Sprengel (ueberarbeitet von Thomas Lange) berechnet.
//	Fuer Einzelheiten siehe entsprechendes Paper. Falls die Controllability
//	bereits berechnet wurde, wird nichts gemacht.
//
//   FORMAL PARAMETERS: keine
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void line_C::reko_compute_contby()
{
line_C *line_item_p;
gate_C *gate_item_p;
double global_0_contby_f = 0.0;
double global_1_contby_f = 0.0;
int i;

if (one_contby_f != 0.0)
	return;
	// Controllability wurde schon berechnet

if (prime_input_b == TRUE)
	{ // Eingaenge haben die Controllability 1
	zero_contby_f = 1.0;
	one_contby_f = 1.0;
	contby_f = 1.0;
		// Aufruf fuer die Nachfolger
	forall(gate_item_p,*nachfolger_p)
		gate_item_p->reko_compute_contby(); // Aufruf fuer das naechste Element
	return;
	}//if

i = 0;
gate_item_p = vorgaenger_p->gib_only_gate();
{forall(line_item_p,*gate_item_p->gib_inputs())
	{ // Die Controllability der Vorgaenger abholen
	zero_contby_a[i] = line_item_p->gib_zero_contby();
	one_contby_a[i] = line_item_p->gib_one_contby();
	i++;
	}}//forall

if (i != gate_item_p->no_inputs())
	error("reko_compute_contby: problems with number of gate inputs");

for (i=0;i<gate_item_p->no_inputs();i++)
	if ((zero_contby_a[i] == 0.0) || (one_contby_a[i] == 0.0))
		{
		return; //Werte fuer die Vorgaenger nicht alle berechnet
		}//if

for (i=0;i<gate_item_p->no_inputs();i++)
	zero_contby_f = (zero_contby_a[i] > zero_contby_f)?zero_contby_a[i]:zero_contby_f;

zero_contby_f = int(zero_contby_f);

if (gate_item_p->get_reconvergence() == TRUE)
	zero_contby_f++;

zero_contby_f += double(plane_no_i)/1000.0;
one_contby_f = zero_contby_f;
contby_f = 0.5 * (one_contby_f + zero_contby_f);

	// Aufruf fuer die Nachfolger
{forall(gate_item_p,*nachfolger_p)
	gate_item_p->reko_compute_contby(); // Aufruf fuer das naechste Element
}
}//reko_compute_contby

//*********************************************************************
//
//   FUNCTION-NAME: reko_compute_observaby
//
//   CLASS: gate_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Fuer alle Input Lines wird reko_compute_observaby aufgerufen.
//
//   FORMAL PARAMETERS: keine
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void gate_C::reko_compute_observaby()
{
for (int i=0;i<no_input_i;i++)
	(*input_p)[i]->reko_compute_observaby();
}//reko_compute_observaby

//*********************************************************************
//
//   FUNCTION-NAME: reko_compute_observaby
//
//   CLASS: line_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Fuer alle Vorgaenger Gates wird reko_compute_observaby
//	aufgerufen. Die Controllability wird nach der Formel von Bernhard
//	Krueger-Sprengel (ueberarbeitet von Thomas Lange) berechnet.
//	Fuer Einzelheiten siehe entsprechendes Paper. Falls die Observability bereits
//	berechnet wurde wird nichts gemacht.
//
//   FORMAL PARAMETERS: keine
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void line_C::reko_compute_observaby()
{
line_C *line_item_p;
gate_C *gate_item_p;
double local_observaby_f;
int j;

if (observaby_f != 0.0)
	return;
	// Observability wurde schon berechnet
if (prime_output_b == TRUE)
	{ // Observability von Ausgaengen ist immer 1
	observaby_f = 1.0;
	testaby_f = contby_f * observaby_f;
		// Aufruf fuer den Vorgaenger der Line
	if (prime_input_b == FALSE)
		vorgaenger_p->gib_only_gate()->reko_compute_observaby(); //Aufruf fuer alle Vorgaenger
	return;
	}//if	Nachfolger ist Primaerausgang.
gate_item_p = nachfolger_p->gib_only_gate();
j = 0;
forall(line_item_p,*gate_item_p->gib_outputs())
	nachfolger_observaby_a[j++] = line_item_p->gib_observaby();
		// Die Observability Werte der Nachfolger werden geholt.
anzahl_outputs_i = j;

for (j=0;j<anzahl_outputs_i;j++)
	if (nachfolger_observaby_a[j] == 0.0)
		return;	//Observability fuer einen Nachfolger nicht berechnet.

	// Observability und Testability berechnen
local_observaby_f = nachfolger_observaby_a[0];
for(j=0;j<anzahl_outputs_i;j++)
	if (nachfolger_observaby_a[j] < local_observaby_f)
		local_observaby_f = nachfolger_observaby_a[j];	// finde Minimum aller

observaby_f = local_observaby_f;
if (nachfolger_p->gib_only_gate()->get_reconvergence() == TRUE)
	observaby_f++;

observaby_f = int(observaby_f);
observaby_f += double(output_depth_i)/10000.0;
testaby_f = contby_f * observaby_f;

	// Aufruf fuer den Vorgaenger der Line
if (prime_input_b == FALSE)
	vorgaenger_p->gib_only_gate()->reko_compute_observaby();
}//reko_compute_observaby

//*********************************************************************
//
//   FUNCTION-NAME: level_compute_contby
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Controllability wird ermittelt. Fuer alle
//	Primaereingaenge wird level_compute_contby aufgerufen
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void level_compute_contby()
{
line_C *line_item_p;

forall(line_item_p,lines_C)
	line_item_p->reset_testaby();
{forall(line_item_p,input_lines_C)
	line_item_p->level_compute_contby();}
}

//*********************************************************************
//
//   FUNCTION-NAME: level_compute_observaby
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Observability wird ermittelt. Fuer alle
//	Primaerausgaenge wird compute_observaby aufgerufen
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void level_compute_observaby()
{
line_C *line_item_p;

forall(line_item_p,output_lines_C)
	line_item_p->level_compute_observaby();
}

//*********************************************************************
//
//   FUNCTION-NAME: level_compute_contby
//
//   CLASS: line_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Fuer alle Nachfolgenden Lines wird level_compute_contby
//	aufgerufen.
//
//   FORMAL PARAMETERS: keine
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void gate_C::level_compute_contby()
{
line_C *line_item_p;

forall(line_item_p,*output_p)
	line_item_p->level_compute_contby();
}//level_compute_contby

//*********************************************************************
//
//   FUNCTION-NAME: level_compute_contby
//
//   CLASS: gate_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Fuer alle Nachfolgenden Gates wird level_compute_contby
//	aufgerufen. Die Controllability wird nach der Formel von Bernhard
//	Krueger-Sprengel (ueberarbeitet von Thomas Lange) berechnet.
//	Fuer Einzelheiten siehe entsprechendes Paper. Falls die Controllability
//	bereits berechnet wurde, wird nichts gemacht.
//
//   FORMAL PARAMETERS: keine
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void line_C::level_compute_contby()
{
line_C *line_item_p;
gate_C *gate_item_p;
double global_0_contby_f = 0.0;
double global_1_contby_f = 0.0,hilf_f;
cube_list_t *cube_array_p;
cube_list_t *cubes_array_p;
int i;
value_t mask_e;
boolean standard_b;
gate_t testability_gate_e;
belegung_t *beleg_p;

if (one_contby_f != 0.0)
	return;
	// Controllability wurde schon berechnet

if (prime_input_b == TRUE)
	{ // Eingaenge haben die Controllability 1
	zero_contby_f = 1;
	one_contby_f = 1;
	contby_f = 1;
		// Aufruf fuer die Nachfolger
	forall(gate_item_p,*nachfolger_p)
		gate_item_p->level_compute_contby(); // Aufruf fuer das naechste Element
	return;
	}//if

i = 0;
gate_item_p = vorgaenger_p->gib_only_gate();
{forall(line_item_p,*gate_item_p->gib_inputs())
	{ // Die Controllability der Vorgaenger abholen
	zero_contby_a[i] = line_item_p->gib_zero_contby();
	one_contby_a[i] = line_item_p->gib_one_contby();
	i++;
	}}//forall

if (i != gate_item_p->no_inputs())
	error("level_compute_contby: problems with number of gate inputs");

for (i=0;i<gate_item_p->no_inputs();i++)
	if ((zero_contby_a[i] == 0.0) || (one_contby_a[i] == 0.0))
		{
		return; //Werte fuer die Vorgaenger nicht alle berechnet
		}//if

// Kommentar  siehe testability_ana.c  compute_contby()
testability_gate_e = gate_item_p->gib_type();
mask_e = VAL_X;
standard_b = TRUE;
if ( testability_gate_e == BLOCK )
   {
     cube_array_p = gate_item_p->gib_cubes(VAL_0);
     if ( cube_array_p->next_p == NULL )
        {
	  mask_e = cube_array_p->to_belegung_p->val_e;
	  for ( beleg_p = cube_array_p->to_belegung_p->next_p; beleg_p != NULL; beleg_p = beleg_p->next_p)
	      {
	        if ( beleg_p->val_e != mask_e )
		   {
		     standard_b = FALSE;
		     break;
		   }
	      }
	  if ( standard_b == TRUE)
	     if ( mask_e == VAL_0 ) testability_gate_e = OR;
	     else testability_gate_e = NAND;
	}
     cube_array_p = gate_item_p->gib_cubes(VAL_1);
     if ( cube_array_p->next_p == NULL )
	{
          mask_e = cube_array_p->to_belegung_p->val_e;
	  for ( beleg_p = cube_array_p->to_belegung_p->next_p; beleg_p != NULL; beleg_p = beleg_p->next_p)
	      {
	        if ( beleg_p->val_e != mask_e )
		   {
		     standard_b = FALSE;
                     break;
		   }
	      }
	  if ( standard_b == TRUE)
	     if ( mask_e == VAL_0 ) testability_gate_e = NOR;
	     else testability_gate_e = AND;
	}
      cube_array_p = gate_item_p->gib_cubes(VAL_0);
      cubes_array_p = gate_item_p->gib_cubes(VAL_1);
      if ( cube_array_p->next_p == NULL && cubes_array_p->next_p == NULL )
         if ( cube_array_p->to_belegung_p->next_p == NULL && cubes_array_p->to_belegung_p->next_p == NULL )
	    {
	      if ( cube_array_p->to_belegung_p->val_e == VAL_0 )
	         testability_gate_e = DRIVER;
	         else testability_gate_e = INVERTER;
	    }
}
	    
		  
switch (testability_gate_e)
	{	// Vorbelegung fuer alle Gatter
	case AND:
		zero_contby_f = max_number_gates_i;
		one_contby_f = 1.0;
		for (i=0;i<gate_item_p->no_inputs();i++)
			{
			if (one_contby_f < one_contby_a[i])
				one_contby_f = one_contby_a[i];
			if (zero_contby_f > zero_contby_a[i])
				zero_contby_f = zero_contby_f;
			}//for
		zero_contby_f++;
		one_contby_f++;
		break;
	case NOR:
		zero_contby_f = 1.0;
		one_contby_f = max_number_gates_i;
		for (i=0;i<gate_item_p->no_inputs();i++)
			{
			if (one_contby_f > one_contby_a[i])
				one_contby_f = one_contby_a[i];
			if (zero_contby_f < zero_contby_a[i])
				zero_contby_f = zero_contby_f;
			}//for
		hilf_f = zero_contby_f;
		zero_contby_f = one_contby_f + 1;
		one_contby_f = hilf_f + 1;
		break;
	case INVERTER:
		one_contby_f = zero_contby_a[0];
		zero_contby_f = one_contby_a[0];
		break;
	case FANOUT:
	case DRIVER:
		one_contby_f = one_contby_a[0];
		zero_contby_f = zero_contby_a[0];
		break;
	case NAND:
		zero_contby_f = max_number_gates_i;
		one_contby_f = 1.0;
		for (i=0;i<gate_item_p->no_inputs();i++)
			{
			if (one_contby_f < one_contby_a[i])
				one_contby_f = one_contby_a[i];
			if (zero_contby_f > zero_contby_a[i])
				zero_contby_f = zero_contby_f;
			}//for
		hilf_f = zero_contby_f;
		zero_contby_f = one_contby_f + 1;
		one_contby_f = hilf_f + 1;
		break;
	case OR:
		zero_contby_f = 1.0;
		one_contby_f = max_number_gates_i;
		for (i=0;i<gate_item_p->no_inputs();i++)
			{
			if (one_contby_f > one_contby_a[i])
				one_contby_f = one_contby_a[i];
			if (zero_contby_f < zero_contby_a[i])
				zero_contby_f = zero_contby_f;
			}//for
		hilf_f = zero_contby_f;
		zero_contby_f++;
		one_contby_f++;
		break;
	case EXOR:
		if (zero_contby_a[0] + zero_contby_a[1] < one_contby_a[0] + one_contby_a[1])
			zero_contby_f = zero_contby_a[0] + zero_contby_a[1] + 1;
		else
			zero_contby_f = one_contby_a[0] + one_contby_a[1] + 1;
		if (zero_contby_a[0] + one_contby_a[1] < one_contby_a[0] + zero_contby_a[1])
			one_contby_f = zero_contby_a[0] + one_contby_a[1] + 1;
		else
			one_contby_f = one_contby_a[0] + zero_contby_a[1];
		break;
	case BLOCK:
		cube_list_t *cube_l_p;
		belegung_t *bel_p;
		int i;

		zero_contby_f = (double)max_number_gates_i;
		cube_array_p = gate_item_p->gib_cubes(VAL_0);
		for (cube_l_p=cube_array_p;cube_l_p!=NULL;cube_l_p=cube_l_p->next_p)
		   {
		   hilf_f = 0.0;
		   i = 0;
		   for (bel_p=cube_l_p->to_belegung_p;bel_p!=NULL;bel_p=bel_p->next_p)
			{
			if (bel_p->val_e == VAL_X)
				continue;
			else if (bel_p->val_e == VAL_1)
				hilf_f += one_contby_a[i];
			else if (bel_p->val_e == VAL_0)
				hilf_f += zero_contby_a[i];
			else
				error("level_compute_contby:  Unknown value in cubes");
			i++;
			}//for
		   if (hilf_f < zero_contby_f)
			zero_contby_f = hilf_f;
		   }//for
		one_contby_f = (double)max_number_gates_i;
		cube_array_p = gate_item_p->gib_cubes(VAL_1);
		for (cube_l_p=cube_array_p;cube_l_p!=NULL;cube_l_p=cube_l_p->next_p)
		   {
		   hilf_f = 0.0;
		   i = 0;
		   for (bel_p=cube_l_p->to_belegung_p;bel_p!=NULL;bel_p=bel_p->next_p)
			{
			if (bel_p->val_e == VAL_X)
				continue;
			else if (bel_p->val_e == VAL_1)
				hilf_f += one_contby_a[i];
			else if (bel_p->val_e == VAL_0)
				hilf_f += one_contby_a[i];
			else
				error("level_compute_contby:  Unknown value in cubes");
			i++;
			}//for
		   if (hilf_f < one_contby_f)
			one_contby_f = hilf_f;
		   }//for
		break;
	case D_FLIP_FLOP:
	case D_LATCH:
	case UNDEFINED_GATE:
		error("level_compute_contby: not implemented for sequential gates");
	}//switch

contby_f = 0.5 * (one_contby_f + zero_contby_f);

	// Aufruf fuer die Nachfolger
{forall(gate_item_p,*nachfolger_p)
	gate_item_p->level_compute_contby(); // Aufruf fuer das naechste Element
}
}//level_compute_contby

//*********************************************************************
//
//   FUNCTION-NAME: level_compute_observaby
//
//   CLASS: gate_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Fuer alle Input Lines wird level_compute_observaby aufgerufen.
//
//   FORMAL PARAMETERS: keine
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void gate_C::level_compute_observaby()
{
for (int i=0;i<no_input_i;i++)
	(*input_p)[i]->level_compute_observaby();
}//level_compute_observaby

//*********************************************************************
//
//   FUNCTION-NAME: level_compute_observaby
//
//   CLASS: line_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Fuer alle Vorgaenger Gates wird level_compute_observaby
//	aufgerufen. Die Controllability wird nach der Formel von Bernhard
//	Krueger-Sprengel (ueberarbeitet von Thomas Lange) berechnet.
//	Fuer Einzelheiten siehe entsprechendes Paper. Falls die Observability bereits
//	berechnet wurde wird nichts gemacht.
//
//   FORMAL PARAMETERS: keine
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void line_C::level_compute_observaby()
{
line_C *line_item_p;
gate_C *gate_item_p;
double local_observaby_f;
int j;
gate_t testability_gate_e;

if (observaby_f != 0.0)
	return;
	// Observability wurde schon berechnet
if (prime_output_b == TRUE)
	{ // Observability von Ausgaengen ist immer 1
	observaby_f = 1.0;
	testaby_f = contby_f * observaby_f;
		// Aufruf fuer den Vorgaenger der Line
	if (prime_input_b == FALSE)
		vorgaenger_p->gib_only_gate()->level_compute_observaby(); //Aufruf fuer alle Vorgaenger
	return;
	}//if	Nachfolger ist Primaerausgang.
gate_item_p = nachfolger_p->gib_only_gate();
j = 0;
{forall(line_item_p,*gate_item_p->gib_outputs())
	nachfolger_observaby_a[j++] = line_item_p->gib_observaby();
}		// Die Observability Werte der Nachfolger werden geholt.
anzahl_outputs_i = j;

for (j=0;j<anzahl_outputs_i;j++)
	if (nachfolger_observaby_a[j] == 0.0)
		return;	//Observability fuer einen Nachfolger nicht berechnet.

gate_item_p = nachfolger_p->gib_only_gate();
// Kommentar  siehe testability_ana.c  compute_contby()
testability_gate_e = gate_item_p->gib_type();
switch (testability_gate_e)
	{ // Das Formelelement DURCH(...) wird berechnet
	case AND:
	case NAND:
	case OR:
	case NOR:
	case EXOR:
	case BLOCK:
		local_observaby_f = nachfolger_observaby_a[0] + 1;
		break;
	case INVERTER:
	case DRIVER:
		local_observaby_f = nachfolger_observaby_a[0];
		break;
	case FANOUT:
		local_observaby_f = max_number_gates_i;
		for(j=0;j<anzahl_outputs_i;j++)
			if (local_observaby_f > nachfolger_observaby_a[j])
				local_observaby_f = nachfolger_observaby_a[j];
		break;
	case D_FLIP_FLOP:
	case D_LATCH:
	case UNDEFINED_GATE:
		error("level_compute_observaby: not implemented for sequential gates");
	}//switch

	// Observability und Testability berechnen
observaby_f = local_observaby_f;
testaby_f = contby_f * observaby_f;

	// Aufruf fuer den Vorgaenger der Line
if (prime_input_b == FALSE)
	vorgaenger_p->gib_only_gate()->level_compute_observaby();
}//level_compute_observaby

