// ***************************************************************
//
//   FILE NAME: read_edif.c
//
//   FILE DESCRIPTION:
//
//   CREATION DATE:
//
//   AUTHOR: Uwe Glaeser
//                                     
//   MODIFICATION HISTORY: None
//
//   All rights reserved.
//
// ***************************************************************
#include "data2.h"
#include "stringliste_C.h"
#include "line_liste_C.h"
#include "gate_liste_C.h"
#include "line_C.h"
#include "gate_C.h"

extern line_liste_C lines_C,input_lines_C,output_lines_C,pseudo_inputs_C,pseudo_outputs_C;
extern gate_liste_C gates_C;
extern char *schaltungsname_s;
extern stat_t statistic_u;
boolean read_edif_called_b = FALSE;
extern line_C *search(string_C*,line_tree_u*);
extern gate_C *search(string_C*,gate_tree_u*);

line_C *search(char *s,line_tree_u *lt_p)
{
string_C s_sC(s);

return(search(&s_sC,lt_p));
}//search

gate_C *search(char *s,gate_tree_u *lt_p)
{
string_C s_sC(s);

return(search(&s_sC,lt_p));
}//search

// *********************************************************************
//
//   FUNCTION-NAME: read_edif
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: November '95
//
//   FUNCTIONAL-DESCRIPTION: Reads the Torino EDIF Interface and copies to
//	internal Data structure.
//
//   PARAMETERS: char *name_s: name of the circuit
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
// *********************************************************************
void read_edif(char *name_s)
{
int i,j,po_index = 0;
line_C *line_item_p;
gate_C *gate_item_p;
char *s;
gate_t g_e;
//line_liste_C io_lines_C;
stringliste_C namelist_C;

//io_lines_C.init(max_no_io_i);
read_edif_called_b = TRUE;
//create(name_s);
strcpy(schaltungsname_s,string_C(schaltungsname_s).basename();

lines_C.init(max_number_lines_i);
input_lines_C.init(n_pi+n_ff);
output_lines_C.init(n_po+n_ff);
pseudo_inputs_C.init(n_ff);
pseudo_outputs_C.init(n_ff);
gates_C.init(2 * n_descr);
po_array = new int[n_po];

for (i=0;i<n_descr;i++)
   { // Gates and Signals
   if (descr[i].attr != PO)
      {
      statistic_u.number_lines_i++;
      s = new char[strlen(descr[i].to_name) + 1];
      strcpy(s,descr[i].to_name);
      line_item_p = new line_C(s);
      lines_C.fuege_ein(line_item_p);
      }//if
   else
      po_array[po_index++] = i;
   if (descr[i].attr == INTERNAL)
      {
      statistic_u.number_gates_i++;
      s = new char[strlen(descr[i].name) + 1];
      strcpy(s,descr[i].name);
      gate_item_p = new gate_C(s);
      gates_C.fuege_ein(gate_item_p);
      }//if
   }//for

for (i=0;i<n_pi;i++)
   { // Input Signals of Circuit
   statistic_u.number_inputs_i++;
   line_item_p = search(descr[pi_array[i]].to_name,milef_line_tree_root_p);
   if (line_item_p == NULL)
      {
      cout << descr[pi_array[i]].to_name;
      error("Line not found in definition");
      }//if
   line_item_p->set_primary_input();
   input_lines_C.fuege_ein(line_item_p);
   }//for

for (i=0;i<n_po;i++)
   { // Output Signals of Circuit
   statistic_u.number_outputs_i++;
   line_item_p = search(descr[descr[po_array[i]].from[0]].to_name,milef_line_tree_root_p);
   if (line_item_p == NULL)
      {
      cout << descr[descr[po_array[i]].from[0]].to_name;
      error("Line not found in definition");
      }//if
   line_item_p->set_primary_output();
   output_lines_C.fuege_ein(line_item_p);
   }//for

for (i=0;i<n_descr;i++)
   { // Gates and Signals
   if (descr[i].attr != INTERNAL)
      continue;
   switch (descr[i].type)
      {
      case E_AND:
         g_e = AND;
	 statistic_u.number_and_gates_i++;
         break;
      case E_NAND:
         g_e = NAND;
	 statistic_u.number_nand_gates_i++;
         break;
      case E_OR:
         g_e = OR;
	 statistic_u.number_or_gates_i++;
         break;
      case E_NOR:
         g_e = NOR;
	 statistic_u.number_nor_gates_i++;
         break;
      case E_BUF:
         g_e = DRIVER;
	 statistic_u.number_drivers_i++;
         break;
      case E_NOT:
         g_e = INVERTER;
	 statistic_u.number_inverters_i++;
         break;
      case E_EXOR:
         g_e = EXOR;
	 statistic_u.number_exor_gates_i++;
         break;
//      case E_EXNOR:
//         g_e = EXNOR;
//	 statistic_u.number_exnor_gates_i++;
//         break;
      case E_FF:
         g_e = D_FLIP_FLOP;
	 statistic_u.number_flipflops_i++;
         break;
      default:
	 error("read_edif: Gate type not found");
      }//switch
   gate_item_p = search(descr[i].name,milef_gate_tree_root_p);
   if (gate_item_p == NULL)
      {
      cout << descr[i].name;
      error("Gate not found in definition");
      }//if
   if (g_e == D_FLIP_FLOP)
      gate_item_p->belege_gate(g_e, 1, 1);
   else
      gate_item_p->belege_gate(g_e, descr[i].fanin, 1);
   for (j=0;j<descr[i].fanin;j++)
      {
      s = descr[descr[i].from[j]].to_name;
//      if (g_e == D_FLIP_FLOP)
//	 cout << "Gattername: " << gate_item_p->get_name() << "   Portindex: " << j << "   Portname: " << s << "\n";
      if ((strcmp(s,"CK") == 0) || (strcmp(s,"RESET") == 0))
	 continue;
      namelist_C.add_element(descr[descr[i].from[j]].to_name);
//      line_item_p = search(descr[descr[i].from[j]].to_name,milef_line_tree_root_p);
//      if (line_item_p == NULL)
//	 {
//	 cout << descr[descr[i].from[j]].to_name;
//	 error("Line not found in definition");
//	 }//if
//      io_lines_C.fuege_ein(line_item_p);
      }//for
   gate_item_p->connect_input(/*io_lines_C*/ namelist_C);
//   io_lines_C.loesche_liste();
   namelist_C.add_element(descr[i].to_name);
//   line_item_p = search(descr[i].to_name,milef_line_tree_root_p);
//   if (line_item_p == NULL)
//      {
//      cout << descr[i].to_name;
//      error("Line not found in definition");
//      }//if
//   io_lines_C.fuege_ein(line_item_p);
   gate_item_p->connect_output(/*io_lines_C*/ namelist_C);
//   io_lines_C.loesche_liste();
   }//for

delete[] po_array;
}//read_edif

// *********************************************************************
//
//   FUNCTION-NAME: find_ppis_and_ppos
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: November '95
//
//   FUNCTIONAL-DESCRIPTION: Looks up ppis and ppos in the Torino EDIF
//	interface
//
//   PARAMETERS: NONE
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
// *********************************************************************
void find_ppis_and_ppos()
{
int i;
line_C *line_item_p;

for (i=0;i<n_ff;i++)
   { // Pseudo Input- and Pseudo Output Signals of Circuit
   line_item_p = search(descr[ppi_array[i]].to_name,milef_line_tree_root_p);
   if (line_item_p == NULL)
      {
      cout << descr[ppi_array[i]].to_name;
      error("Line not found in definition");
      }//if
   line_item_p->set_pseudo_input();
   pseudo_inputs_C.fuege_ein(line_item_p);
   line_item_p = (*(*line_item_p->vorgaenger())[0]->gib_inputs())[0];
   line_item_p->set_pseudo_output();
   pseudo_outputs_C.fuege_ein(line_item_p);
   }//for
}//find_ppis_and_ppos
