//*********************************************************************
//
//   FILE-NAME: testaby_ana.c
//
//   CREATION-DATE: August 91
//
//   FILE-DESCRIPTION: Dieses File beinhaltet die Hauptroutinen fuer Berechnung
//	von Controllability und Observability
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
#define CONTROLLING_VALUES 1
#include <math.h>
#include "extern.h"
#include "line_liste_C.h"
#include "line_C.h"
#include "gate_liste_C.h"
#include "gate_C.h"

#if DEBUG
line_liste_C lines_debug_C;
gate_liste_C gates_debug_C;
#endif

	// Globale Variablen
double zero_contby_a[max_number_gate_inputs_i];
double one_contby_a[max_number_gate_inputs_i];
double durch_a[max_no_io_i];
double nachfolger_observaby_a[max_no_io_i];
double sum_1_contby_f;
double sum_0_contby_f;
double sum_min_contby_f;
int anzahl_outputs_i;
//*********************************************************************
//
//   FUNCTION-NAME: compute_contby
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Controllability wird ermittelt. Fuer alle
//	Primaereingaenge wird compute_contby aufgerufen
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void compute_contby()
{
line_C *line_item_p;

forall(line_item_p,lines_C)
	line_item_p->reset_testaby();
{forall(line_item_p,input_lines_C)
	line_item_p->compute_contby();}
}

//*********************************************************************
//
//   FUNCTION-NAME: compute_observaby
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Observability wird ermittelt. Fuer alle
//	Primaerausgaenge wird compute_observaby aufgerufen
//
//   FORMAL PARAMETERS: void
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void compute_observaby()
{
line_C *line_item_p;

forall(line_item_p,output_lines_C)
	line_item_p->compute_observaby();
}

//*********************************************************************
//
//   FUNCTION-NAME: compute_contby
//
//   CLASS: line_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Fuer alle Nachfolgenden Lines wird compute_contby
//	aufgerufen.
//
//   FORMAL PARAMETERS: keine
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void gate_C::compute_contby()
{
line_C *line_item_p;

#if DEBUG
if (gates_debug_C.element_in_liste(this) == TRUE)
	error("compute_contby: Feedback in Circuit");
gates_debug_C.fuege_ein(this);
#endif
forall(line_item_p,*output_p)
	line_item_p->compute_contby();
#if DEBUG
gates_debug_C.fuege_aus(this);
#endif
}//compute_contby

//*********************************************************************
//
//   FUNCTION-NAME: compute_contby
//
//   CLASS: gate_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Fuer alle Nachfolgenden Gates wird compute_contby
//	aufgerufen. Die Controllability wird nach der Formel von Bernhard
//	Krueger-Sprengel (ueberarbeitet von Thomas Lange) berechnet.
//	Fuer Einzelheiten siehe entsprechendes Paper. Falls die Controllability
//	bereits berechnet wurde, wird nichts gemacht.
//
//   FORMAL PARAMETERS: keine
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void line_C::compute_contby()
{
line_C *line_item_p;
gate_C *gate_item_p;
double local_1_contby_f;
double local_0_contby_f;
double global_0_contby_f = 0.0;
double global_1_contby_f = 0.0;
double no_combinations_f;
double no_0_combinations_f;
double no_1_combinations_f;
double no_0_at_input_for_0_at_output_f;
double no_1_at_input_for_0_at_output_f;
double no_0_at_input_for_1_at_output_f;
double no_1_at_input_for_1_at_output_f;
cube_list_t *cube_array_p;
cube_list_t *cubes_array_p;
int i;
value_t mask_e;
boolean standard_b;
gate_t testability_gate_e;
belegung_t *beleg_p;

if (one_contby_f != 0.0)
	return;
	// Controllability wurde schon berechnet

#if DEBUG
if (lines_debug_C.element_in_liste(this) == TRUE)
	error("compute_contby: Feedback in Circuit");
lines_debug_C.fuege_ein(this);
#endif

if (prime_input_b == TRUE)
	{ // Eingaenge haben die Controllability 1
	zero_contby_f = 1.0;
	one_contby_f = 1.0;
	contby_f = 1.0;
		// Aufruf fuer die Nachfolger
	forall(gate_item_p,*nachfolger_p)
		gate_item_p->compute_contby(); // Aufruf fuer das naechste Element
#if DEBUG
	lines_debug_C.fuege_aus(this);
#endif
	return;
	}//if

i = 0;
gate_item_p = vorgaenger_p->gib_only_gate();
{forall(line_item_p,*gate_item_p->gib_inputs())
	{ // Die Controllability der Vorgaenger abholen
	zero_contby_a[i] = line_item_p->gib_zero_contby();
	one_contby_a[i] = line_item_p->gib_one_contby();
	i++;
	}}//forall

if (i != gate_item_p->no_inputs())
	error("compute_contby: Problems with number inputs of one gate");

for (i=0;i<gate_item_p->no_inputs();i++)
	if ((zero_contby_a[i] == 0.0) || (one_contby_a[i] == 0.0))
		{
#if DEBUG
		lines_debug_C.fuege_aus(this);
#endif
		return; //Werte fuer die Vorgaenger nicht alle berechnet
		}//if

no_combinations_f = (double) pow(2,gate_item_p->no_inputs());


// Falls Komplexgatter (Switch-Level) die gleiche Funktion wie eines der Standardgatter AND, NAND, OR, NOR, DRIVER
// oder INVERTER realisieren, soll dies bei der Berechnung der Testbarkeitsanalysewerte beruecksichtigt werden.
// Diese logische Funktion wird der Variablen testability_gate_e zur korrekten Berechnung der Testbarkeitsanalyse-
// werte zugewiesen!

testability_gate_e = gate_item_p->gib_type();
mask_e = VAL_X;
standard_b = TRUE;
if (testability_gate_e == BLOCK )
   {
     // Es gibt nur eine Kube, die den Wert 0 am Ausgang liefert und
     //   - jede Stelle der Kube hat den Wert 0   ==> logische Funktion OR
     //   - jede Stelle der Kube hat den Wert 1   ==> logische Funktioon NAND
     cube_array_p = gate_item_p ->gib_cubes(VAL_0);
     if ( cube_array_p->next_p  == NULL )
        {
	  mask_e = cube_array_p->to_belegung_p->val_e;
	  for ( beleg_p = cube_array_p->to_belegung_p->next_p; beleg_p != NULL; beleg_p = beleg_p->next_p )
	      {
	        if ( beleg_p->val_e != mask_e )
		   {
		     standard_b = FALSE;
		     break;
		   }
	      }
	  if ( standard_b == TRUE )
	     if ( mask_e == VAL_0 ) testability_gate_e = OR;
	     else testability_gate_e = NAND;
	}
     // Es gibt nur eine Kube, die den Wert 1 am Ausgang liefert und
     //   - jede Stelle der Kube hat den Wert 0   ==> logische Funktion NOR
     //   - jede Stelle der Kube hat den Wert 1   ==> logische Funktioon AND
     cube_array_p = gate_item_p ->gib_cubes(VAL_1);
     if ( cube_array_p->next_p  == NULL )
        {
          mask_e = cube_array_p->to_belegung_p->val_e;
	  for ( beleg_p = cube_array_p->to_belegung_p->next_p; beleg_p != NULL; beleg_p = beleg_p->next_p )
              {
                if ( beleg_p->val_e != mask_e )
		   {
		     standard_b = FALSE;
		     break;
		    }
	      }
	   if ( standard_b == TRUE )
	      if ( mask_e == VAL_0 ) testability_gate_e = NOR;
	      else testability_gate_e = AND;
        }
      // Es gibt genau eine 0- und genau eine 1-Kube, die jeweils genau eine Stelle haben  und
      // die Stelle der 0-Kube hat den Wert 0 ==> logische Funktion DRIVER
      // die Stelle der 0-Kube hat den Wert 1 ==> logische Funktion INVERTER
       cube_array_p = gate_item_p ->gib_cubes(VAL_0);
       cubes_array_p = gate_item_p ->gib_cubes(VAL_1);
       if( cube_array_p->next_p  == NULL && cubes_array_p->next_p  == NULL )
              if( cube_array_p->to_belegung_p->next_p  == NULL && cubes_array_p->to_belegung_p->next_p  == NULL )
        	     {  
	               if ( cube_array_p->to_belegung_p->val_e == VAL_0 )
		            testability_gate_e = DRIVER;
                       else testability_gate_e = INVERTER;
	             }
   }

switch (testability_gate_e)
	{	// Vorbelegung fuer alle Gatter
	case AND:
		no_1_combinations_f = 1.0;
		no_0_combinations_f = no_combinations_f - 1.0;
		no_0_at_input_for_0_at_output_f = no_combinations_f * gate_item_p->no_inputs() / 2.0;
		no_1_at_input_for_0_at_output_f = no_combinations_f * gate_item_p->no_inputs() / 2.0 - gate_item_p->no_inputs();
		no_0_at_input_for_1_at_output_f = 0.0;
		no_1_at_input_for_1_at_output_f = (double) gate_item_p->no_inputs();
		break;
	case NOR:
		no_1_combinations_f = 1.0;
		no_0_combinations_f = no_combinations_f - 1.0;
		no_0_at_input_for_0_at_output_f = no_combinations_f * gate_item_p->no_inputs() / 2.0 - gate_item_p->no_inputs();
		no_1_at_input_for_0_at_output_f = no_combinations_f * gate_item_p->no_inputs() / 2.0;
		no_0_at_input_for_1_at_output_f = (double) gate_item_p->no_inputs();
		no_1_at_input_for_1_at_output_f = 0.0;
		break;
	case INVERTER:
		no_1_combinations_f = 1.0;
		no_0_combinations_f = 1.0;
		no_0_at_input_for_0_at_output_f = 1.0;
		no_1_at_input_for_0_at_output_f = 1.0;
		no_0_at_input_for_1_at_output_f = 1.0;
		no_1_at_input_for_1_at_output_f = 1.0;
		break;
	case DRIVER:
		no_1_combinations_f = 1.0;
		no_0_combinations_f = 1.0;
		no_0_at_input_for_0_at_output_f = 1.0;
		no_1_at_input_for_0_at_output_f = 1.0;
		no_0_at_input_for_1_at_output_f = 1.0;
		no_1_at_input_for_1_at_output_f = 1.0;
		break;
	case NAND:
		no_0_combinations_f = 1.0;
		no_1_combinations_f = no_combinations_f - 1.0;
		no_0_at_input_for_0_at_output_f = 0.0;
		no_1_at_input_for_0_at_output_f = (double) gate_item_p->no_inputs();
		no_0_at_input_for_1_at_output_f = no_combinations_f * gate_item_p->no_inputs() / 2.0;
		no_1_at_input_for_1_at_output_f = no_combinations_f * gate_item_p->no_inputs() / 2.0 - gate_item_p->no_inputs();
		break;
	case OR:
		no_0_combinations_f = 1.0;
		no_1_combinations_f = no_combinations_f - 1.0;
		no_0_at_input_for_0_at_output_f = (double) gate_item_p->no_inputs();
		no_1_at_input_for_0_at_output_f = 0.0;
		no_0_at_input_for_1_at_output_f = no_combinations_f * gate_item_p->no_inputs() / 2.0 - gate_item_p->no_inputs();
		no_1_at_input_for_1_at_output_f = no_combinations_f * gate_item_p->no_inputs() / 2.0;
		break;
	case EXOR:
		no_0_combinations_f = 2.0;
		no_1_combinations_f = 2.0;
		no_0_at_input_for_0_at_output_f = 2.0;
		no_1_at_input_for_0_at_output_f = 2.0;
		no_0_at_input_for_1_at_output_f = 2.0;
		no_1_at_input_for_1_at_output_f = 2.0;
		break;
	case BLOCK:
		double anzahl_x_werte_f;
		double anzahl_1_werte_f,anzahl_0_werte_f;
		cube_list_t *cube_l_p;
		belegung_t *bel_p;
		
		anzahl_x_werte_f = 0.0;
		anzahl_1_werte_f = 0.0;
		anzahl_0_werte_f = 0.0;
		no_0_combinations_f = 0.0;
		no_1_combinations_f = 0.0;
		no_0_at_input_for_0_at_output_f = 0.0;
		no_1_at_input_for_0_at_output_f = 0.0;
		no_0_at_input_for_1_at_output_f = 0.0;
		no_1_at_input_for_1_at_output_f = 0.0;
		cube_array_p = gate_item_p->gib_cubes(VAL_0);
		for (cube_l_p=cube_array_p;cube_l_p!=NULL;cube_l_p=cube_l_p->next_p)
		   {
		   for (bel_p=cube_l_p->to_belegung_p;bel_p!=NULL;bel_p=bel_p->next_p)
			if (bel_p->val_e == VAL_X)
				anzahl_x_werte_f++;
			else if (bel_p->val_e == VAL_1)
				anzahl_1_werte_f++;
			else if (bel_p->val_e == VAL_0)
				anzahl_0_werte_f++;
			else
				error("compute_contby:  Unknown value in cubes");
		   no_0_combinations_f+=pow(2,anzahl_x_werte_f);
		   no_0_at_input_for_0_at_output_f+=pow(2,anzahl_x_werte_f)*(anzahl_x_werte_f/2 + anzahl_0_werte_f);
		   no_1_at_input_for_0_at_output_f+=pow(2,anzahl_x_werte_f)*(anzahl_x_werte_f/2 + anzahl_1_werte_f);
		   }//for
		cube_array_p = gate_item_p->gib_cubes(VAL_1);
		for (cube_l_p=cube_array_p;cube_l_p!=NULL;cube_l_p=cube_l_p->next_p)
		   {
		   for (bel_p=cube_l_p->to_belegung_p;bel_p!=NULL;bel_p=bel_p->next_p)
			if (bel_p->val_e == VAL_X)
				anzahl_x_werte_f++;
			else if (bel_p->val_e == VAL_1)
				anzahl_1_werte_f++;
			else if (bel_p->val_e == VAL_0)
				anzahl_0_werte_f++;
			else
				error("compute_contby:  Unknown value in cubes");
		   no_1_combinations_f+=pow(2,int(anzahl_x_werte_f));
		   no_0_at_input_for_1_at_output_f+=pow(2,anzahl_x_werte_f)*(anzahl_x_werte_f/2 + anzahl_0_werte_f);
		   no_1_at_input_for_1_at_output_f+=pow(2,anzahl_x_werte_f)*(anzahl_x_werte_f/2 + anzahl_1_werte_f);
		   }//for
		break;
	case FANOUT:
		break;
	case D_FLIP_FLOP:
	case D_LATCH:
	case UNDEFINED_GATE:
		error("contby: For sequential gates not implemented");
	}//switch
switch (testability_gate_e)
	{	// Vorbelegung fuer alle Gatter
#if CONTROLLING_VALUES
	case AND:
		// Lokale Controllability ermitteln
	   local_1_contby_f = 1.0 + lambda_f * no_combinations_f / no_1_combinations_f;

		// Globale Controllability berechnen
	   for (i=0;i<gate_item_p->no_inputs();i++)
		{
		double formel_f = 0.0;

		if ((i==0) || (zero_contby_a[i] < zero_contby_f))
			zero_contby_f = zero_contby_a[i];
		
		formel_f = no_0_at_input_for_1_at_output_f * zero_contby_a[i];
		formel_f = formel_f + no_1_at_input_for_1_at_output_f * one_contby_a[i];
		formel_f = formel_f / no_1_combinations_f;
		global_1_contby_f = global_1_contby_f + ((1.0 / gate_item_p->no_inputs()) * formel_f);
		}//for

			// Daraus ergibt sich die Gesamtcontrollability
	   one_contby_f = global_1_contby_f * local_1_contby_f;
	   break;
	case NOR:
		// Lokale Controllability ermitteln
	   local_1_contby_f = 1.0 + lambda_f * no_combinations_f / no_1_combinations_f;

		// Globale Controllability berechnen
	   for (i=0;i<gate_item_p->no_inputs();i++)
		{
		double formel_f = 0.0;

		if ((i==0) || (one_contby_a[i] < zero_contby_f))
			zero_contby_f = one_contby_a[i];

		formel_f = no_0_at_input_for_1_at_output_f * zero_contby_a[i];
		formel_f = formel_f + no_1_at_input_for_1_at_output_f * one_contby_a[i];
		formel_f = formel_f / no_1_combinations_f;
		global_1_contby_f = global_1_contby_f + ((1.0 / gate_item_p->no_inputs()) * formel_f);
		}//for

			// Daraus ergibt sich die Gesamtcontrollability
	   one_contby_f = global_1_contby_f * local_1_contby_f;
	   break;
	case NAND:
		// Lokale Controllability ermitteln
	   local_0_contby_f = 1.0 + lambda_f * no_combinations_f / no_0_combinations_f;

		// Globale Controllability berechnen
	   for (i=0;i<gate_item_p->no_inputs();i++)
		{
		double formel_f = 0.0;

		formel_f = no_0_at_input_for_0_at_output_f * zero_contby_a[i];
		formel_f = formel_f + no_1_at_input_for_0_at_output_f * one_contby_a[i];
		formel_f = formel_f / no_0_combinations_f;
		global_0_contby_f = global_0_contby_f + ((1.0 / gate_item_p->no_inputs()) * formel_f);

		if ((i==0) || (zero_contby_a[i] < one_contby_f))
			one_contby_f = zero_contby_a[i];
		}//for

			// Daraus ergibt sich die Gesamtcontrollability
	   zero_contby_f = global_0_contby_f * local_0_contby_f;
	   break;
	case OR:
		// Lokale Controllability ermitteln
	   local_0_contby_f = 1.0 + lambda_f * no_combinations_f / no_0_combinations_f;

		// Globale Controllability berechnen
	   for (i=0;i<gate_item_p->no_inputs();i++)
		{
		double formel_f = 0.0;

		formel_f = no_0_at_input_for_0_at_output_f * zero_contby_a[i];
		formel_f = formel_f + no_1_at_input_for_0_at_output_f * one_contby_a[i];
		formel_f = formel_f / no_0_combinations_f;
		global_0_contby_f = global_0_contby_f + ((1.0 / gate_item_p->no_inputs()) * formel_f);

		if ((i==0) || (one_contby_a[i] < one_contby_f))
			one_contby_f = one_contby_a[i];
		}//for

			// Daraus ergibt sich die Gesamtcontrollability
	   zero_contby_f = global_0_contby_f * local_0_contby_f;
	   break;
#else
	case AND:
	case NAND:
	case OR:
	case NOR:
#endif
	case INVERTER:
	case DRIVER:
	case EXOR:
	case BLOCK:
		// Lokale Controllability ermitteln
	   local_1_contby_f = 1.0 + lambda_f * no_combinations_f / no_1_combinations_f;
	   local_0_contby_f = 1.0 + lambda_f * no_combinations_f / no_0_combinations_f;

		// Globale Controllability berechnen
	   for (i=0;i<gate_item_p->no_inputs();i++)
		{
		double formel_f = 0.0;

		formel_f = no_0_at_input_for_0_at_output_f * zero_contby_a[i];
		formel_f = formel_f + no_1_at_input_for_0_at_output_f * one_contby_a[i];
		formel_f = formel_f / no_0_combinations_f;
		global_0_contby_f = global_0_contby_f + ((1.0 / gate_item_p->no_inputs()) * formel_f);

		formel_f = no_0_at_input_for_1_at_output_f * zero_contby_a[i];
		formel_f = formel_f + no_1_at_input_for_1_at_output_f * one_contby_a[i];
		formel_f = formel_f / no_1_combinations_f;
		global_1_contby_f = global_1_contby_f + ((1.0 / gate_item_p->no_inputs()) * formel_f);
		}//for

			// Daraus ergibt sich die Gesamtcontrollability
	   one_contby_f = global_1_contby_f * local_1_contby_f;
	   zero_contby_f = global_0_contby_f * local_0_contby_f;
	   break;
	case FANOUT:
		one_contby_f = gate_item_p->gib_inputs()->gib_only_line()->gib_one_contby();
		zero_contby_f = gate_item_p->gib_inputs()->gib_only_line()->gib_zero_contby();
		break;
	case D_FLIP_FLOP:
	case D_LATCH:
	case UNDEFINED_GATE:
                error("contby: For sequential gates not implemented");
	}//switch

contby_f = 0.5 * (one_contby_f + zero_contby_f);

	// Aufruf fuer die Nachfolger
{forall(gate_item_p,*nachfolger_p)
	gate_item_p->compute_contby(); // Aufruf fuer das naechste Element
}
#if DEBUG
lines_debug_C.fuege_aus(this);
#endif
}//compute_contby

//*********************************************************************
//
//   FUNCTION-NAME: compute_observaby
//
//   CLASS: gate_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Fuer alle Input Lines wird compute_observaby aufgerufen.
//
//   FORMAL PARAMETERS: keine
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void gate_C::compute_observaby()
{
for (int i=0;i<no_input_i;i++)
	(*input_p)[i]->compute_observaby();
}//compute_observaby

//*********************************************************************
//
//   FUNCTION-NAME: compute_observaby
//
//   CLASS: line_C
//
//   AUTHOR: Uwe Glaeser
//
//   CREATION-DATE: August 91
//
//   FUNCTIONAL-DESCRIPTION: Fuer alle Vorgaenger Gates wird compute_observaby
//	aufgerufen. Die Controllability wird nach der Formel von Bernhard
//	Krueger-Sprengel (ueberarbeitet von Thomas Lange) berechnet.
//	Fuer Einzelheiten siehe entsprechendes Paper. Falls die Observability bereits
//	berechnet wurde wird nichts gemacht.
//
//   FORMAL PARAMETERS: keine
//
//   IMPLICIT INPUT/OUTPUT: NONE
//
//   FUNCTION-VALUE: void
//
//   MODIFICATION-HISTORY:
//
//*********************************************************************
void line_C::compute_observaby()
{
line_C *line_item_p;
gate_C *gate_item_p;
double local_observaby_f;
int j;
cube_list_t *cube_array_p;
cube_list_t *cubes_array_p;
value_t mask_e;
boolean standard_b;
gate_t testability_gate_e;
belegung_t *beleg_p;

if (observaby_f != 0.0)
	return;
	// Observability wurde schon berechnet
if (prime_output_b == TRUE)
	{ // Observability von Ausgaengen ist immer 1
	observaby_f = 1.0;
	testaby_f = contby_f * observaby_f;
		// Aufruf fuer den Vorgaenger der Line
	if (prime_input_b == FALSE)
		vorgaenger_p->gib_only_gate()->compute_observaby(); //Aufruf fuer alle Vorgaenger
	return;
	}//if	Nachfolger ist Primaerausgang.
gate_item_p = nachfolger_p->gib_only_gate();
j = 0;
{forall(line_item_p,*gate_item_p->gib_outputs())
	nachfolger_observaby_a[j++] = line_item_p->gib_observaby();
		// Die Observability Werte der Nachfolger werden geholt.
}
anzahl_outputs_i = j;

for (j=0;j<anzahl_outputs_i;j++)
	if (nachfolger_observaby_a[j] == 0.0)
		return;	//Observability fuer einen Nachfolger nicht berechnet.

sum_1_contby_f = 0;
sum_0_contby_f = 0;
sum_min_contby_f = 0;
gate_item_p = nachfolger_p->gib_only_gate();
{forall(line_item_p,*gate_item_p->gib_inputs())
	{ // Controllability Werte der Vorgaenger werden abgeholt
	if (line_item_p == this)
		continue;
	else
		{
		sum_1_contby_f = sum_1_contby_f + line_item_p->gib_one_contby();
		sum_0_contby_f = sum_0_contby_f + line_item_p->gib_zero_contby();
		sum_min_contby_f = sum_min_contby_f + ((line_item_p->gib_one_contby() < line_item_p->gib_zero_contby())?
					line_item_p->gib_one_contby():line_item_p->gib_zero_contby());
		}//else
	}}//forall

gate_item_p = nachfolger_p->gib_only_gate();


// Falls Komplexgatter (Switch-Level) die gleiche Funktion wie eines der Standardgatter AND, NAND, OR, NOR, DRIVER
// oder INVERTER realisieren, soll dies bei der Berechnung der Testbarkeitsanalysewerte beruecksichtigt werden.
// Diese logische Funktion wird der Variablen testability_gate_e zur korrekten Berechnung der Testbarkeitsanalyse-
// werte zugewiesen!
testability_gate_e = gate_item_p->gib_type();
mask_e = VAL_X;
standard_b = TRUE;
if (testability_gate_e == BLOCK )
   {
     cube_array_p = gate_item_p ->gib_cubes(VAL_0);
     if ( cube_array_p->next_p  == NULL )
	{
	  mask_e = cube_array_p->to_belegung_p->val_e;
	  for ( beleg_p = cube_array_p->to_belegung_p->next_p; beleg_p != NULL; beleg_p = beleg_p->next_p )
	      {
		if ( beleg_p->val_e != mask_e )
		   {
		     standard_b = FALSE;
		     break;
		   }
	      }
	  if ( standard_b == TRUE )
		if ( mask_e == VAL_0 ) testability_gate_e = OR;
		else testability_gate_e = NAND;
         }
     cube_array_p = gate_item_p ->gib_cubes(VAL_1);
     if ( cube_array_p->next_p  == NULL )
        {
          mask_e = cube_array_p->to_belegung_p->val_e;
	  for ( beleg_p = cube_array_p->to_belegung_p->next_p; beleg_p != NULL; beleg_p = beleg_p->next_p )
	      {
	        if ( beleg_p->val_e != mask_e )
		   {
		     standard_b = FALSE;
		     break;
		    }
	      }
	  if ( standard_b == TRUE )
		if ( mask_e == VAL_0 ) testability_gate_e = NOR;
	        else testability_gate_e = AND;
	}

     cube_array_p = gate_item_p ->gib_cubes(VAL_0);
     cubes_array_p = gate_item_p ->gib_cubes(VAL_1);
     if( cube_array_p->next_p  == NULL && cubes_array_p->next_p  == NULL )
     if( cube_array_p->to_belegung_p->next_p  == NULL && cubes_array_p->to_belegung_p->next_p  == NULL )
       {
         if ( cube_array_p->to_belegung_p->val_e == VAL_0 )
	      testability_gate_e = DRIVER;
	 else testability_gate_e = INVERTER;
       }
   }

   
switch (testability_gate_e)
	{ // Das Formelelement DURCH(...) wird berechnet
	case AND:
	case NAND:
		for(j=0;j<anzahl_outputs_i;j++)
			durch_a[j] = nachfolger_observaby_a[j] + sum_1_contby_f;
		break;
	case OR:
	case NOR:
		for(j=0;j<anzahl_outputs_i;j++)
			durch_a[j] = nachfolger_observaby_a[j] + sum_0_contby_f;
		break;
	case INVERTER:
	case DRIVER:
	case EXOR:
	case BLOCK:
		for(j=0;j<anzahl_outputs_i;j++)
			durch_a[j] = nachfolger_observaby_a[j] + sum_min_contby_f;
		break;
	case FANOUT:
		for(j=0;j<anzahl_outputs_i;j++)
			durch_a[j] = nachfolger_observaby_a[j];
		break;
	case D_FLIP_FLOP:
	case D_LATCH:
	case UNDEFINED_GATE:
                error("compute_observaby: For sequential gates not implemented");
	}//switch

	// Observability und Testability berechnen
local_observaby_f = durch_a[0];
for(j=0;j<anzahl_outputs_i;j++)
	if (durch_a[j] < local_observaby_f)
		local_observaby_f = durch_a[j];	// finde Minimum aller durch_a[j]

observaby_f = local_observaby_f;
testaby_f = contby_f * observaby_f;

	// Aufruf fuer den Vorgaenger der Line
if (prime_input_b == FALSE)
	vorgaenger_p->gib_only_gate()->compute_observaby();
}//compute_observaby

